(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(** Status of properties.
    @since Boron-20100401 *)

open Cil_types
open Db_types
open Cil_datatype

type identified_complete = kernel_function * kinstr * string list
type identified_disjoint =  identified_complete
type identified_code_annotation = kernel_function * stmt * code_annotation
type identified_assigns =
    kernel_function
    * kinstr
    * funbehavior option
    * identified_term assigns list
type identified_decrease = kernel_function * kinstr * term variant
type identified_behavior = kernel_function * kinstr * funbehavior

type predicate_kind =
  | PKRequires of funbehavior
  | PKAssumes of funbehavior
  | PKEnsures of funbehavior * termination_kind
  | PKTerminates

let pretty_predicate_kind fmt = function
  | PKRequires _ -> Format.pp_print_string fmt "requires"
  | PKAssumes _ -> Format.pp_print_string fmt "assumes"
  | PKEnsures(_, tk)  ->
    Format.pp_print_string fmt
      (match tk with
      | Normal -> "ensures"
      | Exits -> "exits"
      | Breaks -> "breaks"
      | Continues -> "continues"
      | Returns -> "returns")
  | PKTerminates -> Format.pp_print_string fmt "terminates"

type identified_predicate =
    predicate_kind * kernel_function * kinstr * Cil_types.identified_predicate

type identified_spec = kernel_function * kinstr * funspec

type identified_property =
  | IPBlob of State.t (* an unidentified property *)
  | IPPredicate of identified_predicate
  | IPAxiom of string
  | IPCodeAnnot of identified_code_annotation
  | IPComplete of identified_complete
  | IPDisjoint of identified_disjoint
  | IPAssigns of identified_assigns
  | IPDecrease of identified_decrease
  | IPBehavior of identified_behavior

let get_kinstr = function
  | IPBlob _ -> Kglobal
  | IPPredicate (_,_,ki,_) -> ki
  | IPAxiom _ -> Kglobal
  | IPCodeAnnot (_,ki,_) -> Kstmt ki
  | IPComplete (_,ki,_) -> ki
  | IPDisjoint(_,ki,_) -> ki
  | IPAssigns(_,ki,_,_) -> ki
  | IPDecrease (_,ki,_) -> ki
  | IPBehavior(_,ki,_) -> ki

let get_kf = function
  | IPBlob _ -> None
  | IPPredicate (_,kf,_,_) -> Some kf
  | IPAxiom _ -> None
  | IPCodeAnnot (kf,_,_) -> Some kf
  | IPComplete (kf,_,_) -> Some kf
  | IPDisjoint(kf,_,_) -> Some kf
  | IPAssigns(kf,_,_,_) -> Some kf
  | IPDecrease (kf,_,_) -> Some kf
  | IPBehavior(kf,_,_) -> Some kf

let get_pk_behavior = function
  | PKRequires b -> Some b
  | PKAssumes b -> Some b
  | PKEnsures (b,_) -> Some b
  | PKTerminates -> None

let get_behavior = function
  | IPBlob _ -> None
  | IPPredicate (pk,_,_,_) -> get_pk_behavior pk
  | IPAxiom _ -> None
  | IPCodeAnnot (_,_,_) -> None
  | IPComplete (_,_,_) -> None
  | IPDisjoint(_,_,_) -> None
  | IPAssigns(_,_,b,_) -> b
  | IPDecrease (_,_,_) -> None
  | IPBehavior(_,_,b) -> Some b

include Datatype.Make_with_collections
    (struct

      include Datatype.Undefined
      type t = identified_property
      let name = "identified property"
      let reprs = [ IPBlob State.dummy ]
      let mem_project = Datatype.never_any_project

      let hash = function
	| IPBlob x -> Hashtbl.hash (0, State.hash x)
	| IPPredicate (_,_,_,x) -> Hashtbl.hash (1, x.ip_id)
	| IPAxiom x -> Hashtbl.hash (2, x)
	| IPCodeAnnot(_,_, ca) -> Hashtbl.hash (3, ca.annot_id)
	| IPComplete(f, ki, x) ->
	  Hashtbl.hash (4, Kernel_function.hash f, Kinstr.hash ki, x)
	| IPDisjoint(f, ki, x) ->
	  Hashtbl.hash (5, Kernel_function.hash f, Kinstr.hash ki, x)
	| IPAssigns(f, ki, b, _l) ->
	  Hashtbl.hash
	    (6,
	     Kernel_function.hash f,
	     Kinstr.hash ki,
	     Extlib.may_map ~dft:"" (fun b -> b.b_name) b)
	| IPDecrease(kf, ki, _v) ->
	  Hashtbl.hash (7, Kernel_function.hash kf, Kinstr.hash ki)
	| IPBehavior(kf, ki, b) ->
	  Hashtbl.hash (8, Kernel_function.hash kf, Kinstr.hash ki, b.b_name)

      let equal p1 p2 = match p1, p2 with
	| IPBlob s1, IPBlob s2 -> State.equal s1 s2
	| IPPredicate (_,_,_,s1), IPPredicate (_,_,_,s2) -> s1.ip_id = s2.ip_id
	| IPAxiom s1, IPAxiom s2 -> s1 = s2
	| IPCodeAnnot(_,_,ca1), IPCodeAnnot(_,_,ca2) ->
          ca1.annot_id = ca2.annot_id
	| IPComplete(f1, ki1, x1), IPComplete(f2, ki2, x2)
	| IPDisjoint(f1, ki1, x1), IPDisjoint(f2, ki2, x2) ->
	  Kernel_function.equal f1 f2 && Kinstr.equal ki1 ki2 && x1 = x2
	| IPAssigns (f1, ki1, b1, _), IPAssigns (f2, ki2, b2, _) ->
	  Kernel_function.equal f1 f2 && Kinstr.equal ki1 ki2
	  && (match b1, b2 with
	  | None, None -> true
	  | Some b1, Some b2 -> b1.b_name = b2.b_name
	  | Some _, None | None, Some _ -> false)
	| IPDecrease(f1, ki1, _), IPDecrease(f2, ki2, _) ->
	  Kernel_function.equal f1 f2 && Kinstr.equal ki1 ki2
	| IPBehavior(f1, ki1, b1), IPBehavior(f2, ki2, b2) ->
	  Kernel_function.equal f1 f2
	  && Kinstr.equal ki1 ki2
	  && b1.b_name = b2.b_name
	| _, _ -> false

      let compare x y = match x, y with
	| IPBlob s1, IPBlob s2 -> State.compare s1 s2
	| IPPredicate (_,_,_,s1), IPPredicate (_,_,_,s2) ->
          Datatype.Int.compare s1.ip_id s2.ip_id
	| IPAxiom s1, IPAxiom s2 -> String.compare s1 s2
	| IPCodeAnnot(_,_,ca1), IPCodeAnnot(_,_,ca2) ->
          Datatype.Int.compare ca1.annot_id ca2.annot_id
	| IPComplete(f1, ki1, x1), IPComplete(f2, ki2, x2)
	| IPDisjoint(f1, ki1, x1), IPDisjoint(f2, ki2, x2) ->
	  let n = Kernel_function.compare f1 f2 in
	  if n = 0 then
	    let n = Kinstr.compare ki1 ki2 in
	    if n = 0 then Extlib.compare_basic x1 x2 else n
	  else n
	| IPAssigns (f1, ki1, b1, _), IPAssigns (f2, ki2, b2, _) ->
	  let n = Kernel_function.compare f1 f2 in
	  if n = 0 then
	    let n = Kinstr.compare ki1 ki2 in
	    if n = 0 then
	      match b1, b2 with
	      | None, None -> 0
	      | Some b1, Some b2 -> Datatype.String.compare b1.b_name b2.b_name
	      | Some _, None -> -1
	      | None, Some _ -> 1
	    else n
	  else n
	| IPDecrease(f1, ki1, _), IPDecrease(f2, ki2, _) ->
	  let n = Kernel_function.compare f1 f2 in
	  if n = 0 then Kinstr.compare ki1 ki2 else n
	| IPBehavior(f1, ki1, b1), IPBehavior(f2, ki2, b2) ->
	  let n = Kernel_function.compare f1 f2 in
	  if n = 0 then
	    let n = Kinstr.compare ki1 ki2 in
	    if n = 0 then Datatype.String.compare b1.b_name b2.b_name else n
	  else n
	| x, y ->
	  let nb = function
	    | IPPredicate _ -> 1
	    | IPAssigns (_, _, _, _) -> 2
	    | IPDecrease _ -> 3
	    | IPAxiom _ -> 4
	    | IPCodeAnnot _ -> 5
	    | IPBehavior _ -> 6
	    | IPComplete (_, _, _) -> 7
	    | IPDisjoint (_, _, _) -> 8
	    | IPBlob _ -> 9
	  in
	  Datatype.Int.compare (nb x) (nb y)

      let pretty fmt = function
	| IPBlob s -> Format.fprintf fmt "%s" (State.get_name s)
	| IPPredicate (kind,_,_,p) ->
          Format.fprintf fmt "%a %a"
            pretty_predicate_kind kind Cil.d_identified_predicate p
	| IPAxiom s -> Format.fprintf fmt "axiom %s" s
	| IPCodeAnnot(_, _, a) -> Cil.d_code_annotation fmt a
	| IPComplete(_, _, l) ->
	  Format.fprintf fmt "complete %a"
	    (Pretty_utils.pp_list (fun fmt s ->  Format.fprintf fmt " %s" s))
	    l
	| IPDisjoint(_, _, l) ->
	  Format.fprintf fmt "disjoint %a"
	    (Pretty_utils.pp_list (fun fmt s ->  Format.fprintf fmt " %s" s))
	    l
	| IPAssigns(_, _, _, l) -> Cil.d_assigns fmt l
	| IPDecrease(_, _, _) -> Format.fprintf fmt "decrease: <TODO>"
	| IPBehavior(_, _, b) -> Format.fprintf fmt "behavior %s" b.b_name

     end)

(*
Local Variables:
compile-command: "make -C ../.."
End:
*)
