(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* -------------------------------------------------------------------------- *)
(* --- Heuristics for Axiom instanciation                                 --- *)
(* -------------------------------------------------------------------------- *)

open Cil_types
open Clabels

(* -------------------------------------------------------------------------- *)
(* --- Utilities                                                          --- *)
(* -------------------------------------------------------------------------- *)

let label_name = function
  | LogicLabel (None, l) -> l
  | _ -> Wp_parameters.fatal "Unexpected non-logical label"

let pp_labels fmt = function
  | [] -> Format.pp_print_string fmt "{}"
  | x::xs ->
      Format.fprintf fmt "@[{%s" x ;
      List.iter (fun x -> Format.fprintf fmt ",%s" x) xs ;
      Format.fprintf fmt "}@]"

let pp_sigma fmt = function
  | [] -> Format.pp_print_string fmt "{}"
  | (x,y)::s ->
      Format.fprintf fmt "@[{%s=%s" x y ;
      List.iter (fun (x,y) -> Format.fprintf fmt ",%s=%s" x y) s ;
      Format.fprintf fmt "}@]"

(* -------------------------------------------------------------------------- *)
(* --- Hint Global Table                                                  --- *)
(* -------------------------------------------------------------------------- *)

module Hints = Cil_datatype.Logic_var.Hashtbl

let globalhints = Hints.create 67

(* -------------------------------------------------------------------------- *)
(* --- Hint Application                                                   --- *)
(* -------------------------------------------------------------------------- *)

type hint = {
  h_name : string ;
  h_sigma : ( string * c_label ) list ; (* sorted by first element *)
}

let pp_sigma fmt = function
  | [] -> Format.pp_print_string fmt "{}"
  | (x,c)::s ->
      Format.fprintf fmt "@[{%s=%a" x Clabels.pretty c ;
      List.iter
	(fun (x,c) -> Format.fprintf fmt ",%s=%a" x Clabels.pretty c) s ;
      Format.fprintf fmt "}@]"

let pp_hint fmt h =
  Format.fprintf fmt "%s%a" h.h_name pp_sigma h.h_sigma

let apply def labels =
  try
    Wp_parameters.result ~level:3 ~current:true "Lookup hints for %a."
      !Ast_printer.d_logic_var def ;
    List.map
      (fun (h,s) ->
	 let sigma =
	   List.map
	     (fun (hdef,hused) ->
		hdef , Clabels.lookup labels hused
	     ) s
	 in 
	 let hint = { h_name = h ; h_sigma = sigma } in
	 Wp_parameters.result ~level:3 ~current:true
	   "For predicate %a : instanciate lemma %a"
	   !Ast_printer.d_logic_var def pp_hint hint ;
	 hint
      ) (Hints.find globalhints def)
  with Not_found -> []

(* -------------------------------------------------------------------------- *)
(* --- Hint Collector                                                     --- *)
(* -------------------------------------------------------------------------- *)

(* a hint (P,m) with (x,y) in m means :
   property P shall be instantiated with label x of its definition bound to
   the actual memory at label y *)

class hint_collector =
object(self)

  inherit Visitor.frama_c_inplace

  val mutable current = None

  method private hint f flabels =
    if flabels <> [] then
      match current with
	| None -> ()
	| Some a ->
	    let sigma =
	      List.map
		(fun (fdef,fuse) ->
		   let flabel = label_name fdef in
		   let ulabel = label_name fuse in
		   ulabel , flabel
		) flabels in
	    let hint = a,sigma in
	    begin
	      try
		let hints = Hints.find globalhints f.l_var_info in
		Hints.replace globalhints f.l_var_info (hint::hints)
	      with Not_found ->
		Hints.add globalhints f.l_var_info [hint]
	    end

  method vterm t =
    match t.term_node with
      | Tapp(f,labels,_) -> self#hint f labels ; Cil.DoChildren
      | _ -> Cil.DoChildren

  method vpredicate_named p =
    match p.content with
      | Papp(f,labels,_) -> self#hint f labels ; Cil.DoChildren
      | _ -> Cil.DoChildren

  method vannotation = function
    | Dlemma(f,_,_,_,_,_) as d ->
	current <- Some f ;
	Cil.ChangeDoChildrenPost( d , (fun d -> current <- None ; d) )
    | _ -> Cil.DoChildren

end

let pp_presigma fmt = function
  | [] -> Format.pp_print_string fmt "{}"
  | (x,y)::s ->
      Format.fprintf fmt "@[{%s=%s" x y ;
      List.iter
	(fun (x,y) -> Format.fprintf fmt ",%s=%s" x y) s ;
      Format.fprintf fmt "}@]"

let pp_prehint fmt (h,s) = Format.fprintf fmt "%s%a" h pp_presigma s

let compute () =
  if Wp_parameters.Axioms.get () then
    begin
      Visitor.visitFramacFile (new hint_collector) (Ast.get ()) ;
      if Wp_parameters.verbose_atleast 3 then
	Hints.iter
	  (fun f hs ->
	     Wp_parameters.result ~level:3 "Axiomatics hints for %a:@.%t" !Ast_printer.d_logic_var f
	       (fun fmt ->
		  List.iter (fun h -> Format.fprintf fmt " - %a@." pp_prehint h) hs
	       ) ;
	  ) globalhints ;
    end

(* -------------------------------------------------------------------------- *)
