(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(** Wp computation using the CFG *)

open Cil_types
open Cil_datatype
open Db_types

module Cfg (W : Mcfg.S) = struct

  (** Before storing something at a program point, we have to process the label
  * at that point. *)
  let do_labels wenv e obj =
    let do_lab o l =
      Wp_parameters.debug ~level:3 "[do_label] process %a@."
        Wp_error.pp_logic_label l;
      W.label wenv l o
    in
    let obj = do_lab obj Logic_const.here_label in
    let labels = Cil2cfg.get_edge_labels e in
    let obj = List.fold_left do_lab obj labels in
      obj

  let add_hyp wenv obj h =
    let h_id = WpAnnot.id_of_id_pred h in
    let p = WpAnnot.body_of_id_pred h in
      Wp_parameters.debug ~level:3 "add hyp (@[%s:@ %a@])@."
        (WpAnnot.prop_txt h_id) !Ast_printer.d_predicate_named p;
      W.add_hyp wenv h_id p obj

  let add_goal wenv obj g =
    let g_id = WpAnnot.id_of_id_pred g in
    let g_pred = WpAnnot.body_of_id_pred g in
        Wp_parameters.debug ~level:3 "add goal (@[%s:@ %a@])@."
          (WpAnnot.prop_txt g_id) !Ast_printer.d_predicate_named g_pred;
        W.add_goal wenv g_id g_pred obj

  let add_assigns_goal wenv id_assigns obj =
    let g_id = WpAnnot.id_of_id_assigns id_assigns in
    let a = WpAnnot.body_of_id_assigns id_assigns in
    Wp_parameters.debug ~level:3 "add assign goal (@[%s@])@." (WpAnnot.prop_txt g_id);
    W.add_assigns wenv g_id
      a.WpAnnot.a_label
      a.WpAnnot.a_kind
      a.WpAnnot.a_assigns
      obj

  (** Similar to [add_hyp] and [add_goal] except that [h] and [g] already
  * come from computations, so they are already represented in the memory model
  * (no translation needed) *)
  let add_inv_hyp wenv obj h = W.add_as_hyp wenv h obj
  let add_inv_goal wenv obj g = W.add_as_goal wenv g obj

  (** to store the results of computations :
  * we store a result for each edge, and also a list of proof obligations.
  *
  * Be careful that there are two modes of computation :
  * the first one ([Pass1]) is used to collect arbitrary invariants,
  * while the second (after [change_mode]) is the "normal" mode.
  * See {!R.set} for more details.
  * *)
  module R : sig
    type t
    val empty : keep:bool -> Cil2cfg.t -> t
    val change_mode : t -> unit
    val find : t -> Cil2cfg.edge -> W.t_prop
    val set :  WpAnnot.strategy -> W.t_env ->
               t -> Cil2cfg.edge -> W.t_prop -> W.t_prop
    val wp_loop :  
      Cil2cfg.t -> (* current cfg *)
      WpAnnot.strategy -> (* current strategy *)
      t -> (* current set of PO at each edge *)
      Cil2cfg.node -> (* Vloop *)
      Cil2cfg.edge -> (* edge through which the loop is currently entered *)
      (Cil2cfg.node -> W.t_prop) ->
      (* effects of a node. *)
      (W.t_prop -> W.t_prop) ->
      (* Effects of the statement contained in the Vloop node itself*)
      W.t_prop
    val add_oblig : t -> Cil_types.logic_label -> W.t_prop -> unit
  end =
  struct
    module HE = Cil2cfg.HE (struct type t = W.t_prop option end)
    module HI =
      Cil2cfg.HE (struct
                    type t = (W.t_prop list * W.t_prop list) (* hyp + goal *)
                  end)
    type t_mode = Pass1 | Pass2
    type t = {
      mutable mode : t_mode ;
      keep_res : bool ; (* whether to keep intermediate results or not *)
      cfg: Cil2cfg.t;
      tbl : HE.t ;
      inv : HI.t ;
      mutable obligs : (Cil_types.logic_label * W.t_prop list) list ;
    }

    let empty ~keep cfg =
      Wp_parameters.debug ~level:3 "[calculus] get invariants mode@.";
      { mode = Pass1; keep_res = keep; cfg = cfg;
        obligs = [] ; tbl = HE.create 97 ; inv = HI.create 97 ;}

    let get_obligs res label =
      let rec get l_obligs = match l_obligs with
        | [] -> []
        | (l, obligs)::_ when Logic_utils.is_same_logic_label l label -> obligs
        | _::tl -> get tl
      in get res.obligs

    let add_oblig res label obj =
      Wp_parameters.debug ~level:3
        "add proof obligation at label %a =@.  @[<hov2>  %a@]@."
        Wp_error.pp_logic_label label W.pretty obj;
      let rec add l_obligs = match l_obligs with
        | [] -> (label, [obj])::[]
        | (l, obligs)::tl when Logic_utils.is_same_logic_label l label ->
            (l, obj::obligs)::tl
        | o::tl -> o::(add tl)
      in
        res.obligs <- add (res.obligs)

    let memo_hyp_pass2 res obj edges =
      let add_hyp e =
        Wp_parameters.debug ~level:3
          "Memo hyp for Pass2 at %a=@.  @[<hov2>  %a@]@." 
          Cil2cfg.pp_edge e W.pretty obj;
        let (h,g) = try HI.find res.inv e with Not_found -> ([], []) in
         HI.replace res.inv e (obj::h,g)
      in List.iter add_hyp edges

    let memo_goal_pass2 res obj edges =
      let add_goal e =
        Wp_parameters.debug ~level:3
          "Memo goal for Pass2 at %a=@.  @[<hov2>  %a@]@." 
          Cil2cfg.pp_edge e W.pretty obj;
        let (h,g) = try HI.find res.inv e with Not_found -> ([], []) in
         HI.replace res.inv e (h,obj::g)
      in List.iter add_goal edges

    let find res e =
      let obj = HE.find res.tbl e in
        match obj with None ->
          Wp_parameters.warning "find edge annot twice (%a) ?"
            Cil2cfg.pp_edge e;
          raise Not_found
          | Some obj ->
              if (not res.keep_res)
              && (res.mode = Pass2)
              && (List.length
                    (Cil2cfg.pred_e res.cfg (Cil2cfg.edge_src e)) < 2) then
                begin
                  (* it should be used once only : can free it *)
                  HE.replace res.tbl e None;
                  Wp_parameters.debug ~level:3
                    "clear edge %a@." Cil2cfg.pp_edge e
                end;
              obj

    let change_mode res =
      Wp_parameters.debug ~level:3
        "[calculus] change to compute mode (clear wp table)@.";
      begin try
        let e_start = Cil2cfg.start_edge res.cfg in
        let start_goal = find res e_start in
          memo_goal_pass2 res start_goal [e_start]
      with Not_found -> ()
      end;
      res.mode <- Pass2; HE.clear res.tbl

    let collect_oblig wenv res e obj =
      let labels = Cil2cfg.get_edge_labels e in
      let add_oblig obj label =
        try
          Wp_parameters.debug ~level:3
            "get proof obligation at label %a@." Wp_error.pp_logic_label label;
          let obligs = get_obligs res label in
            List.fold_left (fun obj o -> add_inv_goal wenv o obj) obj obligs
        with Not_found -> obj
      in
      let obj = List.fold_left add_oblig obj labels in
        obj


    (** We have found some assigns hypothesis in the stategy :
    * it means that we skip the corresponding bloc, ie. we directly compute
    * the result before the block : (forall assigns. P),
    * and continue with empty. *)
    let use_assigns  wenv res e obj h_assigns = match h_assigns with
      | None -> obj
      | Some a ->
          let h_id = WpAnnot.id_of_id_assigns a in
          let a = WpAnnot.body_of_id_assigns a in
          let kind = a.WpAnnot.a_kind in
          let assigns = a.WpAnnot.a_assigns in
          let obj = do_labels wenv e obj in
          let obj = W.use_assigns wenv h_id kind assigns obj in
          let _ = add_oblig res a.WpAnnot.a_label obj in
            W.empty

    let get_memo res e = try HI.find res.inv e with Not_found -> [], []

    (** store the result p for the computation of the edge e.
    *
    * - In Pass1 mode :
      only Ainv from annotations are used. They are propagated to the next
      loop node and stored by [collect_inv] as Ahyp and Agoal to the loop edges.

    * - In Compute mode :
      if we have some hyps H about this edge, store H => p
      if we have some goal G about this edge, store G /\ p
      if we have annotation B to be used as both H and G, store B /\ B=>P
      We also have to add H and G from HI (invariants computed in Pass1 mode)
      So finaly, we build : [ H => [ B /\ (B => (G /\ P)) ] ]
    *)
    let set annots wenv res e obj =
      try let _ = HE.find res.tbl e in Wp_parameters.fatal "strange loop ?"
      with Not_found ->
        begin
          let e_annot = annots.WpAnnot.get_annots e in
          let h_prop = WpAnnot.get_hyp_only e_annot in
          let g_prop = WpAnnot.get_goal_only e_annot in
          let b_prop = WpAnnot.get_hyp_goal e_annot in
          let h_assigns = WpAnnot.get_asgn_hyp e_annot in
          let g_assigns = WpAnnot.get_asgn_goal e_annot in
          (* get_cut is ignored : see get_wp_edge *)
          let obj = collect_oblig wenv res e obj in
          let obj =
            if WpAnnot.new_loop_computation annots.WpAnnot.info then
              begin match res.mode with
                | Pass1 ->
                    let obj = List.fold_left (add_goal wenv) obj g_prop in
                    let obj = match g_assigns with None -> obj
                      | Some a -> add_assigns_goal wenv a obj
                    in
                    let obj = List.fold_left (add_hyp wenv) obj b_prop in
                    let obj = List.fold_left (add_goal wenv) obj b_prop in
                    let obj = List.fold_left (add_hyp wenv) obj h_prop in
                    let is_loop =
                      match Cil2cfg.node_type (Cil2cfg.edge_src e) with
                      | Cil2cfg.Vloop _ -> true
                        | _ -> false
                    in
                    let obj = 
                      if is_loop then obj
                      else use_assigns  wenv res e obj h_assigns 
                    in obj
                | Pass2 ->
                    let (hi, gi) = get_memo res e in
                      (*
                    let obj = List.fold_left (add_inv_goal wenv) obj gi in
                    let obj = List.fold_left (add_hyp wenv) obj b_prop in
                    let obj = List.fold_left (add_hyp wenv) obj h_prop in
            *)
                    let obj = match gi with
                      | [] ->
                          let obj = List.fold_left (add_hyp wenv) obj b_prop in
                          let obj = List.fold_left (add_hyp wenv) obj h_prop in
                            obj
                      | _ -> (* loop pred edges *)
                          (* don't add hyps : there are already in. *)
                          List.fold_left (add_inv_goal wenv) obj gi
                    in
                      assert (hi = []);
                      (*
                    let obj = List.fold_left (add_inv_hyp wenv) obj hi in
            *)
                    let obj = use_assigns  wenv res e obj  h_assigns in
                      obj
              end
            else (* old computation mode *)
              begin match res.mode with
                | Pass1 -> (* res.mode = Pass1 : use Invariants only *)
                    let i = WpAnnot.get_inv e_annot in
                      List.fold_left (add_goal wenv) obj i
                | Pass2 -> (* normal Pass2 mode *)
              let (hi, gi) = get_memo res e in
              let obj = List.fold_left (add_goal wenv) obj g_prop in
              let obj = match g_assigns with None -> obj
                | Some a -> add_assigns_goal wenv a obj
              in
              let obj =
                List.fold_left (add_inv_goal wenv) obj gi in
              let obj = List.fold_left (add_hyp wenv) obj b_prop in
              let obj = List.fold_left (add_goal wenv) obj b_prop in
              let obj =
                List.fold_left (add_hyp wenv) obj h_prop in
              let obj =
                List.fold_left (add_inv_hyp wenv) obj hi in
              let obj = use_assigns  wenv res e obj  h_assigns in
                obj
              end
          in
          let obj = do_labels wenv e obj in
            Wp_parameters.debug ~level:3
              "[set_wp_edge] %a@." Cil2cfg.pp_edge e;
            Wp_parameters.debug ~level:3 " = @[<hov2>  %a@]@." W.pretty obj;
            Format.print_flush ();
            HE.replace res.tbl e (Some obj);
            find res e (* this should give back obj, but also do more things *)
        end

    (** Compute the result for edge e which goes to the loop node [nloop].
    * Be very careful not to make an infinite loop by calling [get_loop_head]...
    * *)
    let wp_loop g annots res nloop e get_loop_head get_vloop_effects =
      if WpAnnot.new_loop_computation annots.WpAnnot.info then
        begin
          let has_inv () =
            let _loops, edges = Cil2cfg.loop_node_edges g nloop in
            (*if loops <> [] then
              (Wp_parameters.debug ~level:3 "Nested loop not handled yet"; (*TODO*)false)
            else*)
              let inv_edges = annots.WpAnnot.has_inv_prop in
              let has_inv =
                List.exists (fun e -> Cil2cfg.Eset.mem e edges) inv_edges
              in if has_inv then
                Wp_parameters.debug ~level:3 "Inv inside loop detected";
                 has_inv
          in match res.mode, Cil2cfg.is_back_edge e with
                 | Pass1, false ->
                     let obj_loop_head = get_loop_head nloop in
                     let _ = 
                       if has_inv () then
                         begin
                           let goal_edges = Cil2cfg.pred_e g nloop in
                             memo_goal_pass2 res obj_loop_head goal_edges;
                           (* let hyp_edges = Cil2cfg.succ_e g nloop in
                             memo_hyp_pass2 res obj_loop_head hyp_edges *)
                         end
                       else (* will be quantified *)
                         let loop_head_edges = Cil2cfg.succ_e g nloop in
                           memo_goal_pass2 res obj_loop_head loop_head_edges
                     in  W.empty
                 | Pass1, true -> 
                     (* back edge :
                          be careful not to use get_only_succ here
                                              (infinite loop) *)
                     W.empty
                 | Pass2, false -> get_loop_head nloop
                 | Pass2, true -> W.empty
      end
      else (* old computation mode *)
        if Cil2cfg.is_back_edge e then
          (* back edge: nothing to compute,
          * just propagate the invariant if any *)
          (* be careful not to use get_only_succ here (infinite loop) *)
          W.empty
        else (* edge going into the loop from outside *)
            let obj_loop_head = get_loop_head nloop in
              match res.mode with
                | Pass1 ->
    (* take the property [obj] at loop head
    * and put it as Goal to incoming edge and ad Hyp to the loop head edge.
    * We can then continue with [empty].
    * *)
                    let hyp_edges = Cil2cfg.succ_e g nloop in
                      memo_hyp_pass2 res obj_loop_head hyp_edges;
                    let goal_edges = Cil2cfg.pred_e g nloop in
                    let obj_loop_head = get_vloop_effects obj_loop_head in
                      memo_goal_pass2 res obj_loop_head goal_edges;
                    W.empty
                | Pass2 -> (* just propagate *) 
                  obj_loop_head


  end

  let wp_call wenv annots stmt res fct args pn pe =
    Wp_parameters.debug ~level:3 "[wp_call] %a@." !Ast_printer.d_exp fct;
    let get_pre = annots.WpAnnot.called_preconditions in
    let get_post = annots.WpAnnot.called_postconditions in
    let get_assigns = annots.WpAnnot.called_assigns in
    let get_exits = annots.WpAnnot.called_exit_postconditions in
    let pn =
      W.call_normal wenv ~get_pre ~get_post ~get_assigns stmt res fct args pn in
    let pe =
      W.call_exit wenv ~get_pre ~get_assigns ~get_exits stmt fct args pe in
    W.merge pn pe

  let wp_stmt wenv s obj = match s.skind with
  | Return (r, _) -> W.return wenv r obj
  | Instr i ->
      begin match i with
      | (Set (lv, e, _)) -> W.assign wenv lv e obj
      | (Call _) -> assert false
      | (Asm _) ->
          Wp_parameters.warning
	    "Unsupported inline assembler. Assuming no effects.@.";
          obj
      | Skip _ | Code_annot _ -> obj
      end
  | Break _ | Continue _ | Goto _ -> obj
  | Loop _-> (* this is not a real loop (exit before looping)
                just ignore it ! *) obj
  | If _ -> assert false
  | Switch _-> assert false
  | Block _-> assert false
  | UnspecifiedSequence _-> assert false
  | TryExcept _ | TryFinally _ -> assert false

  let wp_scope wenv vars scope obj =
    Wp_parameters.debug ~level:3 "[wp_scope] %s : %a@."
      (match scope with
	 | Mcfg.SC_Global -> "global"
	 | Mcfg.SC_Block_in -> "block in"
	 | Mcfg.SC_Block_out -> "block out"
	 | Mcfg.SC_Function_in -> "function in"
         | Mcfg.SC_Function_frame -> "function frame"
	 | Mcfg.SC_Function_out -> "function out" )
      (Pretty_utils.pp_list  ~sep:", " !Ast_printer.d_var) vars;
    W.scope wenv vars scope obj

  (** @return the WP stored for edge [e]. Compute it if it  is not already
  * there. Also handle the Acut annotations. *)
  let rec get_wp_edge ((_kf, _g, annots, he, wenv) as env) e =
    !Db.progress ();
    let v = Cil2cfg.edge_dst e in
    Wp_parameters.debug ~level:3
      "[get_wp_edge] get wp before %a@." Cil2cfg.pp_node v;
    try
      let res = R.find he e in
      Wp_parameters.debug ~level:3
        "[get_wp_edge] %a already computed@." Cil2cfg.pp_node v;
      res
    with Not_found ->
      (* Notice that other hyp and goal are handled in R.set as usual *)
      match WpAnnot.get_cut (annots.WpAnnot.get_annots e) with
	| [] ->
            let wp = compute_wp_edge env e in
            R.set annots wenv he e wp
	| cutp ->
            Wp_parameters.debug ~level:3
              "[get_wp_edge] cut at node %a@." Cil2cfg.pp_node v;
            let add_cut_goal (g,p) acc =
              if g then add_goal wenv acc p else acc
            in
            let edge_annot = List.fold_right add_cut_goal cutp W.empty in
            (* put cut properties are goals in e *)
            let edge_annot = R.set annots wenv he e edge_annot in
            let wp = compute_wp_edge env e in
            let oblig = List.fold_right
                          (fun (_,p) acc -> add_hyp wenv acc p) cutp wp
            in
            (* TODO : we could add hyp to the oblig if we have some in annots *)
              R.add_oblig he Logic_const.pre_label oblig;
              edge_annot

  and get_only_succ env g v = match Cil2cfg.succ_e g v with
    | [e'] -> get_wp_edge env e'
    | ls -> Wp_parameters.fatal "CFV node %a has %d successors instead of 1@."
              Cil2cfg.pp_node v (List.length ls)

  and compute_wp_edge ((kf, g, annots, he, wenv) as env) e =
    let v = Cil2cfg.edge_dst e in
    Wp_parameters.debug ~level:3
      "[compute_edge] before %a go...@." Cil2cfg.pp_node v;
    let old_loc = Cil.CurrentLoc.get () in
    let () = match Cil2cfg.node_stmt_opt v with
      | Some s -> Cil.CurrentLoc.set (Stmt.loc s)
      | None -> ()
    in
    let formals = Kernel_function.get_formals kf in
    let res = match Cil2cfg.node_type v with
      | Cil2cfg.Vstart ->
          Wp_parameters.fatal "No CFG edge can lead to Vstart"
      | Cil2cfg.VfctIn ->
          let obj = get_only_succ env g v in
	  let obj = wp_scope wenv formals Mcfg.SC_Function_in obj in
	  let obj = wp_scope wenv [] Mcfg.SC_Global obj in
	  obj
      | Cil2cfg.VblkIn (Cil2cfg.Bfct, b) ->
          let obj = get_only_succ env g v in
          let obj = wp_scope wenv b.blocals Mcfg.SC_Block_in obj in
          wp_scope wenv formals Mcfg.SC_Function_frame obj
      | Cil2cfg.VblkIn (_, b) ->
          let obj = get_only_succ env g v in
          wp_scope wenv b.blocals Mcfg.SC_Block_in obj
      | Cil2cfg.VblkOut (_, _b) ->
          let obj = get_only_succ env g v in
            obj (* cf. blocks_closed_by_edge below *)
      | Cil2cfg.Vstmt s ->
          let obj = get_only_succ env g v in
          wp_stmt wenv s obj
      | Cil2cfg.VcallIn (stmt, res, fct, args) ->
          begin
            let en, ee = Cil2cfg.get_call_out_edges g v in
            let objn = get_wp_edge env en in
            let obje = get_wp_edge env ee in
            wp_call wenv annots stmt res fct args objn obje
          end
      | Cil2cfg.VcallOut (_s, _res, _fct, _args) ->
          let obj = get_only_succ env g v in
          (* wp_call_post annots res fct args obj *) obj
      | Cil2cfg.VcallExit (_s, _res, _fct, _args) ->
          let obj = get_only_succ env g v in
          (* wp_call_exit annots res fct args obj *) obj
      | Cil2cfg.Vtest (true, _, c) ->
          let et, ef = Cil2cfg.get_test_edges g v in
          let t_obj = get_wp_edge env et in
          let f_obj = get_wp_edge env ef in
          W.test wenv c t_obj f_obj
      | Cil2cfg.Vtest (false, _, _) ->
          get_only_succ env g v
      | Cil2cfg.Vswitch (_, e) ->
          let cases, def_edge = Cil2cfg.get_switch_edges g v in
          let cases_obj = List.map (fun (c,e) -> c, get_wp_edge env e) cases in
          let def_obj = get_wp_edge env def_edge in
            W.switch wenv e cases_obj def_obj
      | Cil2cfg.Vloop (_, s) ->
            begin
              let get_loop_head = fun n -> get_only_succ env g n in
              let get_vloop_effect = fun p -> wp_stmt wenv s p in
                R.wp_loop g annots he v e get_loop_head get_vloop_effect
            end
      | Cil2cfg.VfctOut
      | Cil2cfg.Vexit ->
          let obj = get_only_succ env g v (* exitpost / postcondition *) in
          wp_scope wenv formals Mcfg.SC_Function_out obj
      | Cil2cfg.Vend ->
	  W.empty
(* LC : unused entry point...
          let obj = W.empty in
          wp_scope wenv formals Mcfg.SC_Function_after_POST obj
*)
    in
    let res =
      let blks = Cil2cfg.blocks_closed_by_edge g e in
      let free_locals res b = wp_scope wenv b.blocals Mcfg.SC_Block_out res in
        List.fold_left free_locals res blks
    in
      Wp_parameters.debug ~level:3
        "[compute_edge] before %a done@." Cil2cfg.pp_node v;
      Cil.CurrentLoc.set old_loc;
      res

  (** WP of global initialisations. *)
  let process_global_init wenv obj =
    let do_asgn lv e obj =
      Wp_parameters.debug ~level:3 "global init : %a <- %a@."
        !Ast_printer.d_lval lv !Ast_printer.d_exp e;
      W.assign wenv lv e obj
    in
    let rec do_init obj lv init = match init with
      | SingleInit e -> do_asgn lv e obj
      | CompoundInit (_, l) ->
          let doit obj (off, init) =
            do_init obj (Cil.addOffsetLval off lv) init
          in List.fold_left doit obj l
    in
    let do_glob_var var init obj = match init.init with None -> obj
        | Some init -> do_init obj (Var var, NoOffset) init
    in

    let do_glob glob obj = match glob with
      | GVar (var, initinfo, loc) ->
          let old_loc = Cil.CurrentLoc.get () in
            Cil.CurrentLoc.set loc;
          let obj = do_glob_var var initinfo obj in
            Cil.CurrentLoc.set old_loc;
            obj
      | _ -> obj
    in
    let ast = Ast.get () in
      List.fold_right do_glob ast.globals obj

  (** If [kf] is the 'main' function, process global initialisations
  * as if they were a kind of prelude of the 'main'. *)
  let do_main_init_if_needed wenv kf obj =
    try
      let main_kf, library = Globals.entry_point () in
        if (not library) && (Kernel_function.equal kf main_kf) then
          process_global_init wenv obj
        else obj
    with Globals.No_such_entry_point _ -> obj

  let get_weakest_precondition cfg ((kf, _g, annots, _he, wenv) as env) =
    Wp_parameters.debug ~level:3 "build final goals";
    let e_start = Cil2cfg.start_edge cfg in
    let obj = get_wp_edge env e_start in
    let obj = do_main_init_if_needed wenv kf obj in
    let obj = match annots.WpAnnot.strategy_kind with
      | WpAnnot.SKannots -> obj 
      | WpAnnot.SKfroms info -> 
          let pre = info.WpAnnot.get_pre () in
          let pre = WpAnnot.get_hyp_only pre in
          W.build_prop_of_from wenv pre obj 
    in
    let obj = W.close wenv obj in
      obj::[]

  let add_axiom (id, (name,labels,p)) = W.add_axiom id name labels p

  let compute cfg annots =
    Wp_parameters.feedback ~level:3
      "[wp-cfg] start computing with the strategy for %a"
          WpAnnot.pp_strategy_info annots.WpAnnot.info;
    if annots.WpAnnot.has_prop_goal || annots.WpAnnot.has_asgn_goal then
      begin
    let kf = Cil2cfg.cfg_kf cfg in
    let wenv = W.new_env kf in
    let keep = if Wp_parameters.Dot.get () then true else false in
    let he = R.empty ~keep cfg in
    let env = (kf, cfg, annots, he, wenv) in
    if annots.WpAnnot.has_inv_prop <> []
      || WpAnnot.new_loop_computation annots.WpAnnot.info
    then
      begin (* have to collect the Ainv properties *)
        (* this is only arbitrary inv. because loop inv are hyp+goal. *)
        Wp_parameters.debug ~level:3 "[wp-cfg] start Pass1";
        Cil2cfg.iter_edges (fun e -> ignore (get_wp_edge env e)) cfg ;
        Wp_parameters.debug ~level:3 "[wp-cfg] end of Pass1";
      end;
    List.iter add_axiom annots.WpAnnot.global_axioms ;
    R.change_mode he;
    let goals = get_weakest_precondition cfg env in
    let pp_cfg_edges_annot he fmt e =
      try W.pretty fmt (R.find he e)
      with Not_found -> Format.fprintf fmt "<released>"
    in
    let annot_cfg = pp_cfg_edges_annot he in
      Wp_parameters.feedback ~level:3 "[wp-cfg] computing done.";
    goals , annot_cfg
      end
    else
      begin
        Wp_parameters.feedback ~level:3
          "[wp-cfg] no goal in this strategy : ignore.";
        let annot_cfg fmt _e = Format.fprintf fmt "" in
          [], annot_cfg
      end

end
