(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* ------------------------------------------------------------------------ *)
(** Specific {!Formula.S} implementation for Caveat memory model implemented
* in {!Caveat_mem}                                                          *)
(* ------------------------------------------------------------------------ *)

open Cil_types
open Cil_datatype

let unsupported = Wp_error.unsupported

(* ------------------------------------------------------------------------ *)
(** {2 Types}                                                               *)
(* ------------------------------------------------------------------------ *)

(** Defines the data type for model Caveat. Everything is quite straitforward,
* except [D1var].
* Notice that the operations like [D1proj] and co take a [data] as argument
* instead of a [dterm] like in previous versions. This ensures that we only
* deel with known data expressions. That is why we need [D1var] then,
* even if it can look redondant with [Fol.Tvar].
* A [D1var] can represent a C variable when the option argument is [Some x];
* the label says at which program point it's value is taken.
* Nonetheless, it can be [None] to represent logic variables. Don't forget that
* some logic variables can be used as temporaries and can be replace later on by
* C data, so we need to keep the data specific representation. 
* Notice that the [mu] operation take an _optional_ expression :
* None means that we don't know the value, so it behaves like an [havoc].
**)
type data =
    | D1null
    | D1base of data
    | D1addr of Cil_types.varinfo * Formula.tau
             (** [D1addr(v,t)] address of [v] where [t] is the type of [v]
             * (not of its address. *)
    | D1var of  (Cil_types.varinfo *  Clabels.c_label) option * Fol.Var.t
    | D1depl of data * epath
    | D1shift of data * dterm
    | D1proj of data * epath
    | D1indir of data
    | D1mu of data * epath * dterm option (* no exp means havoc *)
    | D1muRange of data * dterm (* mu (tab, [+(i..j)]) *)
    | D1muI of data * dterm option * dterm option (* mu (d, [#(i), *] -> x) *)
    | D1muIrange of data * dterm (* mu (p, [#(i...j), * ]) *)
and epath =
    | Pfield of Cil_types.fieldinfo
    | Pidx of  dterm
and dterm = data Fol.d_term

type pred = dterm Fol.t_pred

type 'a term = dterm

(* ------------------------------------------------------------------------ *)

include Fol_gformula
include Fol_gformula.Langage
include Whylib.Create
(struct
   type 'a term = dterm
   type pred = dterm Fol.t_pred
   include Fol_gformula.Langage
 end)

(* ------------------------------------------------------------------------ *)
(** {2 Variables}                                                           *)
(* ------------------------------------------------------------------------ *)

type fvar = Fol.Var.t
type var = fvar

let eq_var = Fol.Var.equal

type pool = unit (* see Fol.Var.var_counter *)
let pool () = ()

let kind_of_var = Fol.Var.kind_of_var

let fresh _pool name kind =  match kind with
  | Formula.Model tau ->
      (* Problem with the variables created though here, because we don't have
      * there type : unable to build the [format] later on. *)
      Fol.Var.fresh_named_var name tau
  | Formula.Acsl (tau,t) ->
      Fol.Var.fresh_typed_var name t tau

let freshen _pool x = Fol.Var.fresh_var x

(** This variable has to be used for the value of [v] at [Here]. *)
let fvar_of_cvar v t = Fol.Var.mk_lvar_variable (Cil.cvar_to_lvar v) t

(** Get the wp variable of the C variable.
* With [create], create it if it doesn't exist.
* @raise Not_found when create=false and the variable doesn't exist.*)
let get_var ~create tbl_opt label v get_tau =
  try
    match tbl_opt with None -> raise Not_found
      | Some tbl -> Varinfo.Hashtbl.find tbl v
  with Not_found ->
    if create then
      let here = (label = Clabels.Here) in
      let t = get_tau v in
      let fv = fvar_of_cvar v t in
      let fv =
        if here then fv
        else freshen () (*pool m.env*) fv
      in
      let _ = match tbl_opt with
        | None -> if here then () else unsupported "get_var without tbl"
        | Some tbl -> Varinfo.Hashtbl.add tbl v fv
      in fv
     else raise Not_found

let fvar_of_cvar_at ~create tbl label v get_tau =
  get_var create (Some tbl) label v get_tau

let tau_of_var v = Fol.Var.var_type v
let name_of_var v = Fol.Var.var_name v
let basename_of_var v = Fol.Var.basename v


(*
let cvar_of_fvar fv = match Fol.Var.lvar_of_variable fv with
  | None -> None
  | Some lv -> match lv.lv_origin with None -> None | Some vi -> Some vi
*)

let data_of_lvar v = D1var (None, v)

let data_of_cvar_at tbl_opt label vi get_tau =
  let fv = get_var ~create:true tbl_opt label vi get_tau in
    D1var (Some (vi, label), fv)

let data_of_cvar_at_l tbl label vi get_tau =
  let fv = fvar_of_cvar_at ~create:true tbl label  vi get_tau in
    D1var (Some (vi, label), fv)

let term_of_var v =  Fol.e_var v

(** for Formula signature *)
let var v = Fol.e_var v

(* ------------------------------------------------------------------------ *)
(** {2 Comparison}                                                          *)
(* ------------------------------------------------------------------------ *)


let rec eq_data l1 l2 = match l1, l2 with
  | D1null, D1null -> true
  | D1base l1, D1base l2 -> eq_data l1 l2
  | D1addr (v1,_), D1addr (v2,_) -> v1.vid = v2.vid
  | D1var (_,v1), D1var (_,v2) -> Fol.Var.equal v1 v2
  | D1indir (l1), D1indir (l2) -> eq_data l1 l2
  | D1depl (l1, ch1), D1depl (l2, ch2)
  | D1proj (l1, ch1), D1proj (l2, ch2)
    -> eq_epath ch1 ch2 && eq_data l1 l2
  | D1shift (l1, t1), D1shift (l2, t2) ->
      eq_data l1 l2 && eq_term t1 t2
  | D1mu (l1, ch1, t1), D1mu (l2, ch2, t2) ->
      eq_data l1 l2 && eq_epath ch1 ch2 && eq_term_opt t1 t2
  | D1muRange (l1, ch1), D1muRange (l2, ch2) ->
      eq_data l1 l2 && eq_term ch1 ch2 
  | D1muI (l1, None, t1), D1muI (l2, None, t2) ->
      eq_data l1 l2 && eq_term_opt t1 t2
  | D1muI (l1, Some i1, t1), D1muI (l2, Some i2, t2) ->
      eq_data l1 l2 && eq_term i1 i2 && eq_term_opt t1 t2
  | D1muIrange (l1, i1), D1muIrange (l2, i2) ->
      eq_data l1 l2 && eq_term i1 i2 
  | _ -> false
and eq_epath e1 e2 = match e1, e2 with
  | Pfield f1, Pfield f2 ->
      f1.fcomp.ckey = f2.fcomp.ckey && f1.fname = f2.fname
  | Pidx i1, Pidx i2 -> eq_term i1 i2
  | _ -> false
and eq_term_opt t1 t2 = match t1, t2 with
  | None, None -> true
  | Some t1, Some t2 -> eq_term t1 t2
  | _, _ -> false
and eq_term t1 t2 = Fol.eq_terms eq_data t1 t2

(* ------------------------------------------------------------------------ *)
(** {2 Ctypes}                                                              *)
(* ------------------------------------------------------------------------ *)

let ctype_of_fvar v =  match Fol.Var.lvar_of_variable v with
  | None -> None
  | Some lv -> match lv.lv_type with
      | Ctype t -> Some t
      | _ -> None

(* ------------------------------------------------------------------------ *)
(** {2 Pretty Printers}                                                     *)
(* ------------------------------------------------------------------------ *)

open Format

let pp_var = Fol_pretty.pp_var

let rec pp_data fmt l =
  match l with
    | D1null -> Format.fprintf fmt "<null>"
    | D1base l -> Format.fprintf fmt "base(%a)" pp_data l
    | D1addr (v,_) -> Format.fprintf fmt "@@%s" v.vname (* (Fol.Var.var_name v) *)
    | D1var (Some (vi,l), fv) -> Format.fprintf fmt "%a(%s@@%a)"
        pp_var fv vi.vname Clabels.pretty l
    | D1var (_, v) -> Format.fprintf fmt "%a" pp_var v
    | D1depl (x, ch) ->
	Format.fprintf fmt "%a+%a" pp_data x pp_epath ch
    | D1proj (x, ch) ->
	Format.fprintf fmt "%a%a" pp_data x pp_epath ch
    | D1shift (p, i) ->
	Format.fprintf fmt "(%a#(%a))" pp_data p pp_term i
    | D1indir (p) -> Format.fprintf fmt "*(%a)"  pp_data p
    | D1mu (l, ch, x) -> Format.fprintf fmt "mu(%a, %a -> %a)"
        pp_data l pp_epath ch pp_term_opt x
    | D1muRange (l, ch) -> Format.fprintf fmt "muX(%a, %a -> _)"
        pp_data l pp_term ch
    | D1muI (l, None, x) -> Format.fprintf fmt "mu(%a, * -> %a)"
        pp_data l pp_term_opt x
    | D1muI (l, Some i, x) -> Format.fprintf fmt "mu(%a, #(%a),* -> %a)"
        pp_data l pp_term i pp_term_opt x
    | D1muIrange (l, i) -> Format.fprintf fmt "muX(%a, #(%a),* -> _)"
        pp_data l pp_term i 

and pp_epath fmt = function
  | Pfield f -> Format.fprintf fmt ".%s" f.fname
  | Pidx i -> Format.fprintf fmt "(%a)" pp_term i

and pp_term fmt e = Fol_pretty.fpretty_term pp_term pp_data fmt e

and pp_term_opt fmt e = match e with None -> Format.fprintf fmt "_"
  | Some e -> pp_term fmt e

module Data =
struct
  type d = data
  let pp_term = pp_term
  let pp_data = pp_data
end

include Fol_pretty.Create(Data)

(* ------------------------------------------------------------------------ *)
(** {2 Term smart constructors}                                             *)
(* ------------------------------------------------------------------------ *)

let null = D1null

let base l = D1base l (* TODO : simplify known cases *)

let addr v t = D1addr (v, t)

let shift_field loc f = D1depl (loc, Pfield f)

let shift_index loc i = D1depl (loc, Pidx i)

let shift_epath loc ch = match ch with
  | Pfield f -> shift_field loc f
  | Pidx i -> shift_index loc i

let rec mk_mu o ch x = match o with
  | D1mu (o', ch', _) when eq_epath ch ch' ->
      (* SMPrule mu (mu(o, ch, x'), ch -> x) --> mu(o, ch, x) *)
      mk_mu o' ch x
  | _ -> D1mu (o, ch, x) (* TODO : more SMP ? *)

let rec mk_mu_range o ch = match o with
  | D1muRange (o', ch') when eq_term ch ch' ->
      mk_mu_range o' ch
  | _ -> D1muRange (o, ch) (* TODO : SMP ? *)

let rec mk_mu_sh_indir p sh x = match p with 
  | D1muI (p', sh', _) when eq_term_opt sh sh' ->
      mk_mu_sh_indir p' sh x
  | _ -> D1muI (p, sh, x)

let rec mk_mu_range_indir p rg = match p with 
  | D1muIrange (p', rg') when eq_term rg rg' ->
      mk_mu_range_indir p' rg
  | _ -> D1muIrange (p, rg) (* TODO : SMP ? *)

let rec shift_pointer loc i =
  match loc with
    | D1depl (loc, Pidx j) ->
        (* SMPrule : (p+(i))#(j) ----> p+(i+j) *)
        shift_index loc (e_iop Formula.Iadd j i)
    | D1shift (loc, j) ->
        (* SMPrule : (p#(i))#(j) ----> p#(i+j) *)
        shift_pointer loc (e_iop Formula.Iadd j i)
    | D1muI (p, None, x) ->
        (* SMPrule : (mu(p, * -> x))#(i) --> mu(p#(i), [#(-i),*] -> x) *)
        D1muI (shift_pointer p i, Some (e_ineg i), x)
    | D1muI (p, Some j, x) ->
        (* SMPrule : (mu(p, [#(j),*] -> x))#(i) --> mu(p#(i), #(j-i),* -> x) *)
        D1muI (shift_pointer p i, Some (e_iop Formula.Isub j i), x)
    | _ -> D1shift (loc, i)

let rec proj loc ch = match loc with
  | D1mu (_, ch2, Some x) when eq_epath ch ch2 ->
      (* SMPrule mu (l, ch -> x).ch ----> x *)
      x
  | D1mu (l, ((Pfield _) as ch2), _) when not (eq_epath ch ch2) ->
      (* SMPrule mu (l, .a -> x).b -- a<>b --> l.b *)
      proj l ch
  (* D1muRange ??? *)
  | _ -> Fol.e_data (D1proj (loc, ch))

let data_of_term x = match x with
  | Fol.Tvar v -> data_of_lvar v (* useful for logic variables in data *)
  | Fol.Tdata d -> d
  | _ -> unsupported "location of abstract term: %a" pp_term x

let indir_at tbl label loc =
  let rec indir_rec loc = match loc with
    | D1addr (v,t) ->
        (* SMPrule *(@x) --> x *)
        let x = data_of_cvar_at tbl label v (fun _ -> t) in
          Fol.e_data x
    | D1mu _ | D1muRange _ ->
	Wp_parameters.fatal
          "Indirect access to (%a) which is not a pointer"
	  pp_data loc
    | D1muI (_, None, Some x) ->
        (* SMPrule *(mu (p, *, x)) --> x *)
        x
    | D1muI (p, Some (i), Some x) ->
        (* SMPrule *(mu(p, [#(i),*] -> q)) --> if (i=0) then x else *p *)
        e_cond (e_icmp Formula.Ceq i (e_int 0)) x (indir_rec p)
    | D1depl (p, ch) ->
        (* SMPrule *(p+ch) --> ( *p).ch *)
        proj (data_of_term (indir_rec p)) ch
    | _ ->
        Fol.e_data (D1indir (loc))
  in indir_rec loc

let indir_here loc = indir_at None Clabels.Here loc

let indir tbl label loc = indir_at (Some tbl) label loc

let term_of_data d = Fol.e_data d

(** {b Pre} the data should represent an object (as oposed to an address).
* @return the data that represent the address of the data [d].
* @raise Not_found when the data is based on a logic variable, because it
* cannot have any address.
* *)
let rec addr_of_data d = match d with
  | D1null | D1addr _ | D1base _ | D1depl _ | D1shift _
    -> unsupported "address of %a" pp_data d
  | D1var (None, _) -> (* logic variable *) raise Not_found
  | D1var (Some (vi, _), v) -> D1addr (vi, tau_of_var v)
  | D1indir p -> p
  | D1proj (s, ch) -> D1depl (addr_of_data s, ch)
  | D1mu (s, _, _) | D1muRange (s, _) -> addr_of_data s
  | D1muI _ | D1muIrange _ -> assert false (* should be eliminated before that*)

include Fol_gformula.Predicate

(* ------------------------------------------------------------------------ *)
(** {2 Propagation in data}                                                 *)
(* ------------------------------------------------------------------------ *)

let do_path do_exp ch = match ch with
  | Pfield _ -> ch
  | Pidx i -> Pidx (do_exp i)



(** propagate expression transformation [do_exp] in the data [d] *)
let rec do_exp_in_data do_exp d =
  let do_path ch = do_path do_exp ch in
  let do_exp_opt e = match e with None -> None 
    | Some e -> Some (do_exp e)
  in
  let do_data d = data_of_term (do_exp_in_data do_exp d) in
    match d with
      | D1null | D1addr _ -> Fol.e_data d
      | D1base p ->  Fol.e_data (D1base (do_data p))
      | D1var (_,v) ->
          let e = Fol.e_var v in
          let e' = do_exp e in
            if eq_term e e' then Fol.e_data d else e'
      | D1indir (e) -> indir_here (do_data e)
      | D1proj (e, ch) -> proj (do_data e) (do_path ch)
      | D1depl (e, Pfield f) -> Fol.e_data (shift_field (do_data e) f)
      | D1depl (e, Pidx i) -> Fol.e_data (shift_index (do_data e) (do_exp i))
      | D1shift (e, i) -> Fol.e_data (shift_pointer (do_data e)(do_exp i))
      | D1mu (e, ch, x) -> 
          Fol.e_data (D1mu (do_data e, do_path ch, do_exp_opt x))
      | D1muRange (e, x) -> Fol.e_data (D1muRange (do_data e, do_exp x))
      | D1muI (e, None, x) -> 
          Fol.e_data (D1muI (do_data e, None, do_exp_opt x))
      | D1muI (e, Some i, x) -> 
          Fol.e_data (D1muI (do_data e, Some (do_exp i), do_exp_opt x))
      | D1muIrange (e, x) -> Fol.e_data (D1muIrange  (do_data e, do_exp x))

(** propagate transformations [do_data] and [do_exp] in [d].
* Also apply [do_data] on the new [d].
*)
let rec do_data_rec do_data do_exp d =
  (* Format.printf "@[<v2>-> %a@," pp_data d; *)
  let do_path ch = do_path do_exp ch in
  let do_exp_opt e = match e with None -> None 
    | Some e -> Some (do_exp e)
  in
  let frec = do_data_rec do_data do_exp in
  let rec do_rec d = match d with (* poragate in sub-data and sub-exp *)
    | D1null | D1addr _ -> d
    | D1var _ -> d
    | D1base p -> D1base (frec p)
    | D1indir p -> data_of_term (indir_here (frec p))
    | D1proj (e, ch) -> (D1proj (frec e, do_path ch))
    | D1depl (e, ch) -> (D1depl (frec e, do_path ch))
    | D1shift (e, i) -> (D1shift (frec e, do_exp i))
    | D1mu (e, ch, x) -> (D1mu (frec e, do_path ch, (do_exp_opt x)))
    | D1muRange (e, x) -> (D1muRange (frec e, do_exp x))
    | D1muI (e, None, x) -> (D1muI (frec e, None, do_exp_opt x))
    | D1muI (e, Some i, x) -> (D1muI (frec e, Some (do_exp i), do_exp_opt x))
    | D1muIrange (e, x) -> (D1muIrange (frec e, do_exp x))
  in
  let d' = do_rec d in
  let d' = do_data d' in
  (* Format.printf "<- %a@]@," pp_data d; *)
    d'

(* ------------------------------------------------------------------------ *)
(** {2 Substitution}                                                        *)
(* ------------------------------------------------------------------------ *)

let subst_data_in_pred do_data p =
  Fol.change_data_in_pred (do_data_rec do_data) p

let mk_let ~fresh v exp p = match exp with
  | Fol.Tdata D1addr _
  | Fol.Tdata (D1var _)
    -> Fol.subst_in_pred do_exp_in_data v exp p
  | Fol.Tvar _v ->
      Fol.subst_in_pred do_exp_in_data v exp p
  | Fol.Tdata (D1muI _)
  | Fol.Tdata (D1muIrange _)
      (* TODO : build a let on x if it is not a single node *)
    -> Fol.subst_in_pred do_exp_in_data v exp p
  | _ -> Fol.let_pred do_exp_in_data ~fresh v exp p

let p_subst _alpha v exp p =
  mk_let ~fresh:true v exp p

let e_rename subst e =
  let rename_var v = match apply subst v with
    | None -> None
    | Some v' -> Some (Fol.e_var v')
  in
  let rename_data d = match d with
    | D1var (x, v) ->
        let d = match apply subst v with
          | None -> d
          | Some v' -> D1var (x, v')
        in d
    | _ -> d
  in
  let rec rename_data_rec d =
    let d' = do_data_rec rename_data rename_exp d in
      Fol.e_data d'
  and rename_exp e = Fol.change_in_exp rename_var rename_data_rec e
  in rename_exp e

(* ------------------------------------------------------------------------ *)
(** {2 Quantification}                                                      *)
(* ------------------------------------------------------------------------ *)

let p_forall vars p = Fol.p_forall_vars do_exp_in_data vars p

let p_exists vars p = Fol.p_exists_vars do_exp_in_data vars p

let rec data_has_var xs = function
  | D1null -> false
  | D1base d | D1indir d -> data_has_var xs d
  | D1addr _ -> false
  | D1var (_,x) -> List.exists (Fol.Var.equal x) xs
  | D1depl(d,e) | D1proj(d,e) -> data_has_var xs d || path_has_var xs e
  | D1shift(d,t) | D1muIrange(d,t) -> 
      data_has_var xs d || term_has_var xs t
  | D1muI(d,None,t) ->
      data_has_var xs d || term_opt_has_var xs t
  | D1muI(d,Some i,t) -> 
      data_has_var xs d || term_has_var xs i || term_opt_has_var xs t
  | D1mu(d,e,t) -> 
      data_has_var xs d || path_has_var xs e || term_opt_has_var xs t
  | D1muRange(d,t) -> data_has_var xs d || term_has_var xs t

and path_has_var xs = function
  | Pfield _ -> false
  | Pidx t -> term_has_var xs t

and term_has_var xs =
  Fol.e_has_var data_has_var xs

and term_opt_has_var xs t = match t with None -> false
      | Some t -> term_has_var xs t

let pred_has_var =
  Fol.p_has_var term_has_var

(*LC : sufficient approximation *)
let term_closed t = Fol.bad_term_closed [] t
let pred_closed t = Fol.bad_pred_closed [] t

type alpha = Fol.Var.t Fol.Vmap.t
let empty_alpha = Fol.Vmap.empty
let fold_alpha = Fol.Vmap.fold 

let data_alpha_cv alpha d = match d with
  | D1null -> alpha, d
  | _ -> (* TODO *) Wp_parameters.not_yet_implemented 
           "alpha conversion for data in caveat model"

let p_more_alpha_cv = Fol.pred_alpha_cv data_alpha_cv
let p_alpha_cv p = Fol.p_alpha_cv data_alpha_cv p
(* ------------------------------------------------------------------------ *)
(** {2 Visit/fold}                                                          *)
(* ------------------------------------------------------------------------ *)

let visit_path do_exp acc ch = match ch with
  | Pfield _ -> acc
  | Pidx i -> do_exp acc i

(** [do_sub_data] is a function applied to each sub-data of [d] (NOT on [d]).
* if not given, only visit sub-data to apply [do_exp] to expressions. *)
let rec visit_exp_in_data do_sub_data do_exp acc d =
  let do_path acc ch = visit_path do_exp acc ch in
  let do_exp_opt acc e = match e with None -> acc | Some e -> do_exp acc e in
  let frec acc d = visit_exp_in_data do_sub_data do_exp (do_sub_data acc d) d in
    match d with
      | D1null | D1addr _ | D1var _ -> acc
      | D1base e -> frec acc e
      | D1indir e -> frec acc e
      | D1proj (e, ch) -> let acc = frec acc e in do_path acc ch
      | D1depl (e, ch) -> let acc = frec acc e in do_path acc ch
      | D1shift (e, i) -> let acc = frec acc e in do_exp acc i
      | D1mu (e, ch, x) ->
          let acc = frec acc e in let acc = do_path acc ch in do_exp_opt acc x
      | D1muRange (e, x) -> let acc = frec acc e in do_exp acc x
      | D1muI (e, None, x) -> let acc = frec acc e in do_exp_opt acc x
      | D1muI (e, Some i, x) -> 
          let acc = frec acc e in let acc = do_exp acc i in do_exp_opt acc x
      | D1muIrange (e, x) -> let acc = frec acc e in do_exp acc x

(** use to collect things about data: call [do_data] on each data node of [e]
* and recursively inside the data. If [all], apply also [do_data] on sub-data,
* else only apply [do_data] on [d] in [Tdata d], ie. top data. *)
let visit_data_in_pred ~all do_data acc p =
  let do_var acc v = do_data (fun acc _ -> acc) acc (D1var (None, v)) in
  let do_top_data do_exp acc d =
    let do_sub_data acc d = if all then do_data do_exp acc d else acc in
      visit_exp_in_data do_sub_data do_exp (do_data do_exp acc d) d
  in
  Fol.fold_data_in_pred do_var do_top_data acc p

(* ------------------------------------------------------------------------ *)

let huge_pred = Fol.huge_pred
let huge_term = Fol.huge_term

