(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* -------------------------------------------------------------------------- *)
(** Memory Model for Caveat: more information about it in
 * {{:../../wp/Notes/m1.html}this document}
 * *)
(* -------------------------------------------------------------------------- *)

open Formula
open Cil_types
open Cil_datatype

(** Trace call from outside *)
let debug1 fmt = Wp_parameters.debug  fmt
let debug2 fmt = Wp_parameters.debug ~level:2 fmt
(** All the details *)
let debug3 fmt = Wp_parameters.debug ~level:3 fmt

let unsupported = Wp_error.unsupported

module Make(X:sig val with_guards:bool end)=
struct
  module WP = 
  struct
    
    (** Implementation of a specific {!Formula.S} for Caveat *)
    module F = Caveat_fol
    module D1 = Datalib.Create(F)
      
    (** Library functions from Hoare model *)
    module Shared = Hoare_lib.Create(D1)
    open Shared

    (** Library functions for  Caveat memory model implemented
     * in {!Caveat_mem}. Library functions are defined in
     * {{:../../../share/why/caveat.why} Caveat WHY file}.
     *)
    module CavLib = struct
      let indir (p:D1.pointer) (t:Ctypes.c_object) : D1.abstract =
	D1.e_app2 "indir" p (D1.format_of_object t)
	  
      let mu_indir (p:D1.pointer) (i:D1.integer) 
                   (t:Ctypes.c_object) (x:D1.abstract) : D1.abstract =
	D1.e_app4 "mu_indir" p i (D1.format_of_object t) x
	  
      (** similat to D1.set_range_index but for pointer *)
      let mu_range_indir (p:D1.pointer) (i:Formula.m_set F.term) : D1.abstract =
	D1.e_app2 "mu_range_indir" p i 
	  
      (** Used to specify the pointer data in which we can safely 
      * ignore [muI]. *)
      let ptr (p:D1.pointer) : D1.pointer = D1.e_app1 "ptr" p
	
    end

    module Model = struct
      
      module D = D1
      module A = Mint_natural.Create(D)
      module R = Mfloat_natural.Create(D)
	
      type loc = F.data
      let pp_loc = F.pp_data
	
      let pointer_of_loc p = (F.term_of_data p)
      let loc_of_pointer _ p = F.data_of_term p
	
      let cast_loc_to_int _tp _p _ti = unsupported "cast_loc_to_int"
      let cast_int_to_loc _ti _i _tp = unsupported "cast_int_to_loc"
	
    end

    let top_loc_term d = CavLib.ptr (Model.pointer_of_loc d)

    include  Datalib.Cvalues(Model)
    
    (*----------------------------------------------------------------------- *)
      
    type mem = {
      label : Clabels.c_label;
      variables : F.var Varinfo.Hashtbl.t;
    }
	
    type frame = {
      kf : Kernel_function.t;
      mutable states : (Clabels.c_label * mem) list;
      mutable result : F.var option;
      mutable exit_status : F.var option;
      return :  Cil_types.typ option ;
    }
	
    (** {2 Environement management} *)
	
    let global_frame () = {
      kf = Kernel_function.dummy ();
      states = [];
      result = None;
      exit_status = None ;
      return = None ;
    }
      
    let new_frame kf ~m_here ~m_pre ~m_post ropt =
      let may_cons acc lab x =
	match x with
	  | None -> acc
	  | Some x -> (lab,x)::acc
      in
      let states = may_cons
	(may_cons
	   (may_cons [] Clabels.Post m_post)
	   Clabels.Pre
	   m_pre)
	Clabels.Here
	m_here
      in
      {
	kf = kf;
	states = states;
	result = ropt ;
	exit_status = None;
	return = Some(Kernel_function.get_return_type kf);
      }
	
    (** gives the memory state for that label. Create it if it doesn't exist. *)
    let mem_at env clab =
      try List.assoc clab env.states
      with Not_found ->
	debug3 "[caveat:new_mem] for %a@." Clabels.pretty clab;
	let m =  { label = clab;
		   variables = Varinfo.Hashtbl.create 10;
		 } in
	env.states <- (clab,m)::env.states ;
	m
	  
    let result env = match env.result,env.return with
      | Some x, Some t -> x, t
      | Some x, None ->
	  let t = Kernel_function.get_return_type env.kf in
	  (x, t)
      | None,_ ->
	  let t = Kernel_function.get_return_type env.kf in
	  let x = D.fresh "result" (Mdata.Vacsl(Ctype t)) in
	  env.result <- Some x ; (x, t)
	    
    let exit_status env= match env.exit_status with
      | Some x -> x
      | None ->
	  let x = D.fresh "exit_status" (Mdata.Vmodel Formula.Integer) in
	  env.exit_status <- Some x ; x
	    
    let set_exit_status env x = {env with result = None; exit_status = Some x }
      
    let get_exit_status env = env.exit_status
      
    (** Get the wp variable of the C variable.
     * With [create], create it if it doesn't exist.
     * @raise Not_found when create=false and the variable doesn't exist.*)
    let get_var m v =
      F.fvar_of_cvar_at ~create:true m.variables m.label v
	(fun v -> D.tau_of_ctype v.vtype)
	
    (*------------------------------------------------------------------------*)
    (** {2 Build data and operations on locations}                            *)
    (*------------------------------------------------------------------------*)
	
    (*-----------------------------*)
    (** {3 Data for locations} *)
	
    let null = F.null
      
    let base_address _m l = F.base l
      
    let cvar _m v =
      if v.vglob && (get_array_length v.vtype <> None) then Globals.define v;
      F.addr v (D.tau_of_ctype v.vtype)
	
    let field loc f = F.shift_field loc f
      
    let index loc _te i = F.shift_index loc i
      
    let shift loc _te i = F.shift_pointer loc i
      
    let startof loc te = index loc te (F.e_int 0)
      
    (** see also [indir] in "Operations on pointers" *)
      
    let var_data m vi =
      let get_tau v = D.tau_of_ctype v.vtype in
      F.data_of_cvar_at_l m.variables m.label vi get_tau
	
    let var_data_term m vi = F.term_of_data (var_data m vi)
      
    (*-----------------------------*)
    (** {3 Operations on pointers} *)
      
    (** [d] should be a data that represent an adddress:
     * @return the data representing the content of [d] in [m]. *)
    let indir m d = F.indir m.variables m.label d
      
    (** We want to read [d] at [label], but in Caveat model, everything comes
     * from variables, so we cannot read at a label if the address is evaluated
     * at another one... Check that the data [d] is compatible with the label,
     * and @raise Unsupported otherwise. *)
    let rec check_indir_label label d = match d with
      | F.D1null | F.D1addr _ -> ()
      | F.D1var (None, _v) -> (* TODO: this might be incorrect !*) ()
          (* unsupported "indirect access at %a based on logical variable '%a' ?"
             Clabels.pretty label pp_var v *)
      | F.D1var (Some (_, l), _) ->
          if label = l then ()
          else unsupported "indirect access at %a of data at %a"
            Clabels.pretty label Clabels.pretty l
      | F.D1base e | F.D1indir e | F.D1proj (e, _) | F.D1depl (e, _)
      | F.D1shift (e, _)
      | F.D1mu (e, _, _) | F.D1muI (e, _, _) 
      | F.D1muRange (e, _) | F.D1muIrange (e, _)
	  -> check_indir_label label e
	  
    let load m cval loc =
      let _ = check_indir_label m.label loc in
      value_of_logic cval (indir m loc)
	
    let block_length _m l = Hlib.block_length (top_loc_term l)
      
    let disj p1 p2 = Hlib.separated (top_loc_term p1) (top_loc_term p2)
      
    let cast_loc_to_loc ty1 ty2 _l =
      unsupported "cast from (%a) to (%a)"
	!Ast_printer.d_type ty1 !Ast_printer.d_type ty2
	
    (*--------------------------------------------------------------------------*)
    (** {2 Anti-aliasing} *)
	
    (** @return true when there is no indirect access in d.
     * (Notice that there can be indirect access in subexpressions such as indexes)
     **)
    let rec direct_lval d = match d with
      | F.D1null | F.D1base _ 
      | F.D1mu _ | F.D1muRange _ | F.D1muI _ | F.D1muIrange _ ->
          unsupported "forbiden lvalue: %a" F.pp_data d
      | F.D1var (None, v) ->
          Wp_parameters.fatal "lvalue is a logic variable %a ?" F.pp_var v
      | F.D1addr _ -> true
      | F.D1var _ | F.D1indir _ | F.D1proj _ -> false
      | F.D1depl (d, _) | F.D1shift (d, _) -> direct_lval d
	  
    let rec logic_data d = match d with
      | F.D1null | F.D1addr _ -> false
      | F.D1var (None, _) -> true
      | F.D1var (Some _, _) -> false
      | F.D1indir p ->
          if logic_data p then
            unsupported "indirect access to logic data %a" F.pp_data d
          else false
      | F.D1base p | F.D1depl (p, _) | F.D1shift (p, _) |  F.D1proj (p, _)
      | F.D1mu (p,_,_) | F.D1muRange (p,_) 
      |  F.D1muI (p,_,_) | F.D1muIrange (p,_)->
          logic_data p
	    
    (** TODOopt should we have oredered lists ? *)
    let add_unique eq e acc = if List.exists (eq e) acc then acc else e::acc
      
    (** Collect the addresses of locations read by [p].
     * @param with_direct says if we have to include base addresses ([\@X]). *)
    let get_read_addr with_direct p =
      let add d acc =
	if logic_data d then
          (debug3 "[get_read_addr] ignore access to logic data %a" F.pp_data d;
           acc)
	else
	  if with_direct || not (direct_lval d) then add_unique F.eq_data d acc
	  else
	    (debug3 "[get_read_addr] ignore direct access to %a" F.pp_data d; acc)
      in
      let do_var acc (v_opt, fv) = match v_opt with
	| Some (vi, _l) ->
            (* TODO: maybe we should take only the variables at Here ? *)
            add (F.addr vi (F.tau_of_var fv)) acc
	| None -> debug3 "get_read_addr : ignore logic variable %a" F.pp_var fv;
            acc
      in

      let rec do_path do_exp acc ch = match ch with
	| F.Pfield _ -> acc
	| F.Pidx i -> do_exp acc i
	    
      and do_data do_exp acc d =
        let frec =  do_data do_exp in
        match d with
	  | F.D1null | F.D1addr _ -> acc
	  | F.D1base p -> frec acc p
	  | F.D1var (v, fv) -> do_var acc (v, fv)
	  | F.D1proj (s, ch) ->
              let acc =
		try
		  let p = F.addr_of_data d in
		  let acc = add p acc in
                  frec acc p
		with Not_found -> do_path do_exp (frec acc s) ch
              in acc
	  | F.D1depl (ps, F.Pfield _) -> frec acc ps
	  | F.D1depl (p, F.Pidx i) | F.D1shift (p, i)  ->
              let acc = do_exp acc i in frec acc p
	  | F.D1indir p  ->
              let acc = match frec [] p with
		| [] -> acc (* no modifiable var in p *)
		    (* TODO : check indir cannot be lowered *)
		| accp -> List.fold_left (fun acc d -> add d acc) acc accp
              in add p acc
	  | F.D1mu (s, F.Pfield _f, x)  ->
              let acc = match x with Some x -> do_exp acc x | None -> acc in
              frec acc s (* TODO what about f ? *)
	  | F.D1mu (t, F.Pidx i, x)  ->
              let acc = match x with Some x -> do_exp acc x | None -> acc in
              let acc = do_exp acc i in
              frec acc t (* TODO what about idx ? *)
	  | F.D1muRange (t, x)  ->
              let acc = do_exp acc x in
              frec acc t (* TODO what about idx ? *)
	  | F.D1muI _  | F.D1muIrange _ ->
              unsupported "[get_read_addr] of %a" F.pp_data d
    in
    let ptrs = F.visit_data_in_pred ~all:false do_data [] p in
    ptrs

    let prepare_guard lval p =
      try
	let ptrs = get_read_addr (not (direct_lval lval)) p  in
	debug3 "[get_read_addr] = %a"
          (Pretty_utils.pp_list  ~sep:", " F.pp_data) ptrs;
	let add_p g p =
          let keep_p = if F.eq_data lval p then false else true in
          if keep_p then F.p_and (disj lval p) g else g
	in List.fold_left add_p F.p_true ptrs
      with Wp_error.Error(model, msg) ->
	Wp_parameters.warning
          "Failed to compute guard in %s model because %s." model msg;
	D.dummy() (*TODO a valider avec Anne *)
	  
    (*--------------------------------------------------------------------------*)
    (** {2 WP calculus} *)
	  
    (** {3 Substitutions} *)
	  
    (** @return (v, mu) where [v] is the variable to substitute (base of lval),
     * and [mu] the expression of its new value after and assignment lval = exp;
     * Extended version of Hoare similar function [mk_mu]. See doc there. 
     *
     * [f_mu] can be used with a None expression.
     * *)
    let mk_mu m lval =
      let do_indir f_obj1 sh f_mu_of_mu2 =
	let f_obj2 = fun obj0 -> match sh with
          | None -> F.data_of_term (indir m (f_obj1 obj0)) 
          | Some i -> F.data_of_term (indir m (F.shift_pointer (f_obj1 obj0) i))
        in
	let f_mu_of_mu3 = fun obj0 mu3 ->
          let mu2 = F.term_of_data (F.mk_mu_sh_indir (f_obj1 obj0) sh mu3) in
          f_mu_of_mu2 obj0 (Some mu2)
	in f_obj2, f_mu_of_mu3
      in
      let rec addr_loc l = match l with
	| F.D1var (Some (vi,  Clabels.Here),_) -> cvar m vi
	| F.D1proj (l, ch) -> F.shift_epath (addr_loc l) ch
	| F.D1indir l -> l
	| _ -> unsupported "take address of unknown location: %a" pp_loc l
      in
      let rec mk_mu_rec lval = match lval with
	| F.D1null -> unsupported "assignment to NULL pointer !"
	| F.D1addr (v,_) -> v, (fun obj -> obj), (fun _obj mu -> mu)
	| F.D1depl (lv, d2) ->
            let v, f_obj1, f_mu_of_mu2 = mk_mu_rec lv in
            let f_obj2 = fun obj0 -> F.data_of_term (F.proj (f_obj1 obj0) d2) in
            let f_mu_of_mu3 = fun obj0 mu3 ->
              let mu2 = F.term_of_data (F.mk_mu (f_obj1 obj0) d2 mu3) in
              f_mu_of_mu2 obj0 (Some mu2)
            in
            v, f_obj2, f_mu_of_mu3
	| F.D1indir (lv) ->
            let v, f_obj1, f_mu_of_mu2 = mk_mu_rec lv in
            let f_obj2, f_mu_of_mu3 = do_indir f_obj1 None f_mu_of_mu2 in
            v, f_obj2, f_mu_of_mu3
	|  F.D1var (Some (v,  Clabels.Here),_) -> (* == D1indir (D1addr) *)
             let f_obj1 obj0 = obj0 in
             let f_mu_of_mu2 _o mu = mu in
             let f_obj2, f_mu_of_mu3 = do_indir f_obj1 None f_mu_of_mu2 in
             v, f_obj2, f_mu_of_mu3
	| F.D1proj (lv, ch) -> (* D1proj (x,ch) == D1indir (D1depl (&x,ch)) *)
            let lv_depl = F.shift_epath (addr_loc lv) ch in
            let v, f_obj1, f_mu_of_mu2 = mk_mu_rec lv_depl in
            let f_obj2, f_mu_of_mu3 = do_indir f_obj1 None f_mu_of_mu2 in
            v, f_obj2, f_mu_of_mu3
        | F.D1shift (lv, i) -> (* [lv] must be something like [*p] :
                                  process shift + indir in one step *)
            let p = addr_loc lv in
            let v, f_obj1, f_mu_of_mu2 = mk_mu_rec p in 
            let f_obj2, f_mu_of_mu3 = do_indir f_obj1 (Some i) f_mu_of_mu2 in
            v, f_obj2, f_mu_of_mu3
	| _ -> unsupported "assignment to unknown location: %a" pp_loc lval
      in mk_mu_rec lval
	   
    let get_var_and_mu m lval exp =
      debug3 "[get_var_and_mu] %a <- %a" pp_loc lval F.pp_term exp;
      let v, _, f_mu = mk_mu m lval in
      get_var m v, f_mu (var_data m v) (Some exp)
	
    let subst env lval exp p =
      let mem_here = mem_at env Clabels.Here in
      let var, mu = get_var_and_mu mem_here lval exp in
        match mu with None -> unsupported "substitution of unknown value"
          | Some mu ->
              debug2 "[subst_var_and_mu] %a := %a@." F.pp_var var F.pp_term mu;
              let p = D.subst var mu p in
                debug2 "= %a@." F.pp_pred p;
                p
	
    let subst_lval env _t lval v_exp p =
      debug1 "[subst_lval] %a <- %a@." pp_loc lval pp_value v_exp;
      let wp = subst env lval (logic_of_value v_exp) p in
      if X.with_guards
      then
        (* be carreful that the guard is computed on [p], NOT on wp *)
        let g = prepare_guard lval p in F.p_and g wp
      else wp

    let subst_result env v_exp p =
      match env.result with
	| None -> p
	| Some xres -> match v_exp with
            | Some v_exp ->
		debug1 "[subst_result] \\result <- %a@." pp_value v_exp ;
		D.subst xres (logic_of_value v_exp) p
          | None -> D.forall [xres] p
	      
    (** Substitute the variables that represent values at [label] by the
     * variables that represent values at [Here]. *)
    let update_at_label env label p =
      if label = Clabels.Here then p (* nothing to do (subst Here by Here !) *)
      else
	begin
          debug1 "[update_at_label] at %a" Clabels.pretty label;
          let mem_here = mem_at env Clabels.Here in
          let do_data d = match d with
            | F.D1var (Some (vi, l), _) when l = label -> var_data mem_here vi
            | _ -> d
          in let p = F.subst_data_in_pred do_data p in
          debug2 "[update_at_label] -> %a@." F.pp_pred p;
          p
	end
	  
    (** {3 Quantification} *)
	  
    (** Create names for address variables, field, structure types...
     * Needed so that they will be declared.
     * Be careful: not recursive, has to be used in a recursive function *)
    let create_data_name d = match d with
      | F.D1addr (v,_) -> D.Aindex.define v
      | F.D1depl (_, F.Pfield f)
      | F.D1proj (_, F.Pfield f)
      | F.D1mu (_, F.Pfield f, _) -> D.Findex.define f ; D.Tindex.define f.fcomp
      | _ -> ()
	  
    let quantify_at_label env label p =
      debug1 "[quantify_at_vars] at %a" Clabels.pretty label;
      let do_data _ acc d = match d with
	| F.D1var (Some (_vi, l), v_at_label) when l = label ->
            (* add the variable to the list of variable to quantify
             * if it is not already in. *)
            add_unique F.eq_var v_at_label acc
	| _ -> create_data_name d; acc
      in
      let vars = F.visit_data_in_pred ~all:true do_data [] p in
      let p = D.forall vars p in
      let p =
	try (* TODO: check why this is needed...
               All the vars should be in the data, shouldn't they ? *)
          let mem_lab = List.assoc label env.states in
          let dov _ v l = v::l in
          let l = Varinfo.Hashtbl.fold dov mem_lab.variables [] in
          D.forall l p
	with Not_found -> p
      in
      debug2 "[quantify_at_label] -> %a@." F.pp_pred p;
      p
	
    (*-----------------------------*)
    (** {3 declaration of local variables} *)
	
    (* process local scope : bool=true for allocation, false for free.
     * Add some hypotheses about the variable because we are going out of its
     * scope. The quantification could be done here, but it is done through
     * [quantify_at_vars] I guess. *)
    let local_scope _env _vars scope p =
      match scope with
	| Mcfg.SC_Block_in | Mcfg.SC_Function_in -> (* TODO *) p
	| _ -> p
	    
    (* --------------------------------------------------------------------- *)
    (** {3 Zones}                                                            *)
    (* --------------------------------------------------------------------- *)
	    
    let valid _mem = function
      | D.Aloc(_,loc) -> Hlib.valid_pointer (top_loc_term loc)
      | D.Arange(_,loc,rg) -> Hlib.valid_range (top_loc_term loc) (D.interval rg)
	  
    let separated z1 z2 =
      match z1,z2 with
	| D.Aloc(_,l1) , D.Aloc(_,l2) -> disj l1 l2
	| _ -> Wp_parameters.not_yet_implemented "separation of zones with range"
	    
    (* ---------------------------------------------------------------------- *)
    (** {3 Zones}                                                             *)
    (* ---------------------------------------------------------------------- *)
	    
    type m_dzone
    type dzone = m_dzone F.term
    let tau_of_dzone = Formula.ADT("zone",[])
      
    let dzone_assigned _z =
      Wp_parameters.not_yet_implemented "zone representation in Caveat"
	
    let dzone_subset _z1 _z2 =
      Wp_parameters.not_yet_implemented "zone subset in Caveat"
	
    let dzone_union _z1 _z2 =
      Wp_parameters.not_yet_implemented "zone union in Caveat"
	
    let dzone_empty () =
      Wp_parameters.not_yet_implemented "empty zone in Caveat"
	
    let dzone_of_var _m _x =
      Wp_parameters.not_yet_implemented "zone for allocated variable in Caveat"
	
    let effect_supported = false
      
    (* ---------------------------------------------------------------------- *)
    (** {3 Regions}                                                           *)
    (* ---------------------------------------------------------------------- *)
      
    type region
    let pp_region _ _ = ()
    let region_assigned _ = assert false
    let region_empty _ = assert false
    let region_union _ _ = assert false
    let region_included _ _ = assert false
    let region_separated _ _ = assert false
    let region_fingerprint _ _ = assert false

    let region_supported = false
      
    (* ---------------------------------------------------------------------- *)
    (* --- Havoc                                                          --- *)
    (* ---------------------------------------------------------------------- *)
    (** Try to find the data type needed to generated 'format' for [indir]
     * and [mu_indir] *)
    let rec ctype_of_data l = match l with
      | F.D1null | F.D1base _ -> Cil.voidPtrType
      | F.D1proj (_, F.Pfield f) -> f.ftype
      | F.D1proj (l, F.Pidx _) -> Cil.typeOf_array_elem (ctype_of_data l)
      | F.D1addr (vi,_) -> TPtr (vi.vtype, [])
      | F.D1var (Some (vi, _), _) -> vi.vtype
      | F.D1var (None,v) ->
          let t = match F.ctype_of_fvar v with
            | Some t -> t
            | None -> unsupported "unknown type of %a" F.pp_var v
          in t
      | F.D1depl (_, F.Pfield f) -> TPtr (f.ftype, [])
      | F.D1depl (l, F.Pidx _) ->
          let t_addr_array = ctype_of_data l in
          let t_array = Cil.typeOf_pointed t_addr_array in
          let t_elem = Cil.typeOf_array_elem t_array in
          TPtr (t_elem, [])
      | F.D1indir (l) -> Cil.typeOf_pointed (ctype_of_data l)
      | F.D1shift (l, _) -> ctype_of_data l
      | F.D1mu (l, _, _) -> ctype_of_data l
      | F.D1muRange (l, _) -> ctype_of_data l
      | F.D1muI (l, _, _) -> ctype_of_data l
      | F.D1muIrange (l, _) -> ctype_of_data l
	  
      
    let subst_havoc env assigned = match assigned with
      | D.Aloc(_t_elem,loc) ->
	  let tau = D.tau_of_object _t_elem in
	  let out = D.fresh "out" (Mdata.Vmodel tau) in
	  let exp = F.term_of_var out in
	  let m = mem_at env Clabels.Here in
          let x, _f_obj_n, f_mu = mk_mu m loc in
	  let newterm (sigma : D.substitution) : Formula.abstract F.term =
            let obj_0 = D.apply sigma (var_data_term m x) in
            let obj_0 = F.data_of_term obj_0 in
            let obj_0 = f_mu obj_0 (Some exp) in
            let exp = match obj_0 with None -> assert false
              | Some exp -> exp
            in
            debug3 "[subst_havoc:newterm] = %a@." F.pp_term exp;
              exp
	  in
          let x = get_var m x in
            [D.Fresh out ; D.Update (x,newterm)]
      | D.Arange (_t_elem, lval, rg) ->
            let m = mem_at env Clabels.Here in
            let t = ctype_of_data lval in
            let t = Ctypes.object_of t in
            let is_array = match Ctypes.object_of_pointed t with 
              | Ctypes.C_array _ -> true | _ -> false in
            let lval = if is_array then lval else F.addr_of_data lval in
            let x, f_obj_n, f_mu = mk_mu m lval in
            let upd_range x rg = 
              if is_array then F.mk_mu_range x (F.unwrap (D.interval rg))
              else F.mk_mu_range_indir x (F.unwrap (D.interval rg))
            in
            let newterm (sigma : D.substitution) : Formula.abstract F.term =
              let obj_0 = D.apply sigma (var_data_term m x) in
              let obj_0 = F.data_of_term obj_0 in
              let obj_n = f_obj_n obj_0 in
              let exp = upd_range obj_n rg in
              let exp = F.term_of_data exp in
              let exp = f_mu obj_0 (Some exp) in
                match exp with None -> assert false
                  | Some exp -> exp
            in
            let x = get_var m x in
              [D.Update(x,newterm)]
	    
    let assigns_goal _env _sloc _lto = unsupported "proof of assigns clauses"
      
    let assigns_supported = false
      
    (* ---------------------------------------------------------------------- *)
    (** {2 User Predicate environment}
     * This is not correct at the moment because closures are based on 
     * variables, and they should be based on data. For instance :
     * @predicate s_ok (Tstr s) = s.a < s.b;
     * should give : X1 < X2 with X1->s.a and X2->s.b
     * Then, when used in a property : @assert s_ok (s1);
     * it should give : s_ok (s1, s1.a, s1.b)
     *
     * TODO: this cannot be based on 'frame'. See with Loc how to change the
     * signatures of these [userdef_xxx] functions.
     **)
    (* ---------------------------------------------------------------------- *)
      
    type closure = Cil_types.varinfo * string

    let pp_closure fmt (vinfo,s) =
      Format.fprintf fmt "%a@@%s" !Ast_printer.d_var vinfo s
	
    let userdef_frame _pool =
      {
	kf = Kernel_function.dummy ();
	states = [];
	result = None;
	exit_status = None ;
	return = None ;
      }
	
    (* Nothings? TODO filter with respect to tused and pused *)
    let userdef_signature frame _tused _pused=
      let collect_sig label varinfo fvar (variables, closures) =
	fvar::variables, (varinfo, label)::closures
      in
      let do_lab_mem acc (clabel, mem) =
        match clabel with
          | Clabels.LabelParam label ->
              Varinfo.Hashtbl.fold (collect_sig label) mem.variables acc
          | _ -> Wp_parameters.fatal
              "Non-logic label in user-defined function or predicate"
      in List.fold_left do_lab_mem ([],[]) frame.states

    let userdef_closure mem (varinfo,_lparam) =
      var_data_term mem varinfo
	
    let label_at_closure (_,label) = label
      
  end (* of Module WP *)

  (* ------------------------------------------------------------------------ *)
  (** {2 Export to WHY}                                                       *)
  (* ------------------------------------------------------------------------ *)

  module Export = 
  struct
    
    open WP
    open WP.Shared
      
    let c_object_of_data ~indir d =
      let t = ctype_of_data d in
      let t = if indir then Cil.typeOf_pointed t else t in
      Ctypes.object_of  t
	
    let fterm_of_addr v = Hlib.addr (D.Aindex.get_ind v)
      
    let fterm_of_field f = (D.Findex.get_ind f)
      
    let fterm_of_field_access e f = D.acc_field  e f
      
    let fterm_of_mu_field s f e = D.upd_field s f e

    let fterm_of_havoc_index t idx = Hlib.havoc_index t idx
      
    let fterm_of_havoc_field s f =
      let fid = D.Findex.get_ind f in
        if f.fcomp.cstruct then Hlib.havoc_field s fid
        else Hlib.havoc_ufield s fid
      
    let fterm_of_mu_index t i v = D.upd_index t i v
      
    let fterm_of_havoc_range t i = D.e_app2 "set_range_index" t i
      
    let fterm_of_shift_field ps f =
      let tf = fterm_of_field f in
      if f.fcomp.cstruct
      then Hlib.shift_field ps tf
      else Hlib.shift_ufield ps tf
	
    let fterm_of_shift_index pt i = (Hlib.shift_index pt i)
      
    let fterm_of_shift_pointer p i = (Hlib.shift_pointer p i)
      
    let rec t_of_l (loc:F.data) = match loc with
      | F.D1null -> D.null
      | F.D1base x -> Hlib.base (t_of_l x)
      | F.D1addr (v,_) -> fterm_of_addr v
      | F.D1var (_,v) -> F.term_of_var v
      | F.D1indir l ->
          (CavLib.indir (t_of_l l) (c_object_of_data ~indir:false loc))
      | F.D1depl (l, F.Pidx i) ->
          fterm_of_shift_index (t_of_l l) (fterm_of_term i)
      | F.D1depl (l, F.Pfield f) -> fterm_of_shift_field (t_of_l l) f
      | F.D1shift (l,i) -> fterm_of_shift_pointer  (t_of_l l) (fterm_of_term i)
      | F.D1proj (x, F.Pfield f) -> fterm_of_field_access (t_of_l x) f
      | F.D1proj (x, F.Pidx i) ->  D.acc_index (t_of_l x) (fterm_of_term i)
      | F.D1mu (x, F.Pfield f, None) ->
          fterm_of_havoc_field (t_of_l x) f
      | F.D1mu (x, F.Pfield f, Some v) ->
          fterm_of_mu_field (t_of_l x) f (fterm_of_term v)
      | F.D1mu (x, F.Pidx i, None) ->
          fterm_of_havoc_index (t_of_l x) (fterm_of_term i)
      | F.D1mu (x, F.Pidx i, Some v) ->
          fterm_of_mu_index (t_of_l x) (fterm_of_term i) (fterm_of_term v)
      | F.D1muRange (x, i) ->
          fterm_of_havoc_range (t_of_l x) (fterm_of_term i)
      | F.D1muI (p,None, None) ->
          let i = F.e_int 0 in
          let idx = D.singleton (fterm_of_term i) in
          CavLib.mu_range_indir (t_of_l p) idx
      | F.D1muI (p,None, Some x) ->
          let format = c_object_of_data ~indir:true p in
          CavLib.mu_indir (t_of_l p) (F.e_int 0) format (fterm_of_term x)
      | F.D1muI (p,Some i, None) ->
          let idx = D.singleton (fterm_of_term i) in
          CavLib.mu_range_indir (t_of_l p) idx
      | F.D1muI (p,Some i, Some x) ->
          let format = c_object_of_data ~indir:true p in
          CavLib.mu_indir (t_of_l p) (fterm_of_term i) format (fterm_of_term x)
      | F.D1muIrange (p, x) ->
          CavLib.mu_range_indir (t_of_l p) (fterm_of_term x)
	    
    and fterm_of_term (t : F.dterm) : F.dterm =
      Fol.translate_data_in_term t_of_l t
	
    let uterm_of_data (d : F.data) : unit Fol.d_term =
      let hide_data (_ : F.data) : unit Fol.d_term = assert false in
      Fol.translate_data_in_term hide_data (t_of_l d)
	
    let fpred_of_pred (p : F.pred) : unit Fol.d_pred =
      Fol.translate_data_in_pred uterm_of_data p
	
    let fdecl_of_decl = Fol.translate_pred_in_decl fpred_of_pred
      
    module Create
      ( E : Mcfg.Export with type pred = unit Fol.d_pred )
      (* Mcfg.Export with type pred = F.pred *)
      =
    struct
      
      type pred = F.pred
	  
      let export_decl fmt d =
	try
	  (* translate all the [decl] before printing because it might fail ! *)
	  E.export_decl fmt (fdecl_of_decl d)
	with Wp_error.Error(model, msg) ->
	  Wp_parameters.warning
            "Failed to export from model %s to WHY because %s. Ignored:@.%a"
            model msg F.pp_decl d
	    
      let export_goal fmt g p =
	try
	  (* translate all the [decl] before printing because it might fail ! *)
	  E.export_goal fmt g (fpred_of_pred p)
	with Wp_error.Error(model, msg) ->
	  Wp_parameters.warning
            "Failed to export from model %s to WHY because %s. Ignored:@.%t"
            model msg (fun fmt -> F.pp_goal fmt g p) ;
	  E.export_goal fmt g Fol.Pfalse
	    
    end (* of Export.Create *)
      
  end (* of Export *)
    
  module Coq = Export.Create(Fol_coq)
  module Why = Export.Create(Fol_why)
  module Ergo91 = Export.Create(Fol_ergo.E91)
  module Ergo92 = Export.Create(Fol_ergo.E92)

  module Splitter =
  struct
    type pred = WP.F.pred
    let split _ p = Bag.elt p
  end

end (* of Caveat_mem.Make *)

module Safe = Make(struct let with_guards = true end)
module Unsafe = Make(struct let with_guards = false end)
