(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

open Cil_types
open Db_types

open Cfgpropid

class type computer =
object

  method add : Cil2cfg.t -> WpAnnot.strategy list -> unit
  method compute : Wpo.t list

end

module type Description =
sig
  
  val shared : string   (** Shared resource basename for the model *)
  val context : string  (** Basename for environment files ; unique per updater *)
  val updater : string  (** Unique name for the internal updater *)
  val name : string     (** Public name for user feedback *)

end

module Create
  (WpModel:Mwp.S)
  (Why:Mcfg.Export with type pred = WpModel.D.F.pred)
  (Coq:Mcfg.Export with type pred = WpModel.D.F.pred)
  (Ergo91:Mcfg.Export with type pred = WpModel.D.F.pred)
  (Ergo92:Mcfg.Export with type pred = WpModel.D.F.pred)
  (Splitter:Mcfg.Splitter with type pred = WpModel.D.F.pred)
  (Descr:Description)
  =
struct

  module Me = State_builder.Ref
    (Datatype.Unit)
    (struct
       let dependencies = [Ast.self]
       let name = "WP-" ^ Descr.updater
       let kind = `Correctness
       let default () = ()
     end)

  module F = WpModel.D.F
  module D = WpModel.D
  module CV =
    CfgWeakestPrecondition.Create
      (struct 
	 include WpModel 
	 let model = Descr.updater
       end)
  module PO = Cfgpropid.Create(CV)
  module WP = Calculus.Cfg(PO)

  module ST =
    Properties_status.Make_updater
      (struct
	 let name = Me.name
	 let emitter = Me.self
       end)

  let () = Wpo.register Descr.updater ST.set

  (* -------------------------------------------------------------------------- *)
  (* --- Proof Obligations : Export to WHY                                  --- *)
  (* -------------------------------------------------------------------------- *)

  type exportation = {
    env : string ;
    mutable goals : Wpo.t list ;
  }

  let assigns_method = PO.assigns_method

  (* ------------------------------------------------------------------------ *)
  (* --- Goal Registering                                                 --- *)
  (* ------------------------------------------------------------------------ *)

  let export_wpo export kf bhv wrn dep propid gpred =
    let gname = WpAnnot.prop_txt propid in
    let gid = Wpo.gid ~context:Descr.context ~kf ~bhv ~propid in
    (* --- HEADER --- *)
    Command.pp_to_file (Wpo.file_for_head ~gid)
      (fun fmt ->
	 Format.fprintf fmt "@[<v 0>Proof Obligation %s:@]@\n" gname ;
	 Format.fprintf fmt "Environment: %s@\n" export.env ;
	 List.iter (fun d -> Format.fprintf fmt "%a@\n" Wpo.pp_dependency d) dep ;
	 List.iter (fun w -> Format.fprintf fmt "%a@\n" Wpo.pp_warning w) wrn ;
      ) ;
    (* --- BODY --- *)
    Command.pp_to_file (Wpo.file_for_body ~gid)
      (fun fmt ->
	 Format.fprintf fmt "@[<v 2>Goal %s:@\n%a@]" gid F.pp_pred gpred ;
      ) ;
    (* --- WHY Others --- *)
    Command.pp_to_file (Wpo.file_for_goal ~gid Wpo.L_why)
      (fun fmt ->
	 Why.export_goal fmt gid gpred ;
      ) ;
    (* --- COQ --- *)
    Command.pp_to_file (Wpo.file_for_goal ~gid Wpo.L_coq)
      (fun fmt -> 
	 Coq.export_goal fmt gid gpred ;
      ) ;
    (* --- Alt-Ergo 91 --- *)
    Command.pp_to_file (Wpo.file_for_goal ~gid Wpo.L_ergo91)
      (fun fmt -> 
	 Ergo91.export_goal fmt gid gpred ; 
      ) ;
    (* --- Alt-Ergo 92 --- *)
    Command.pp_to_file (Wpo.file_for_goal ~gid Wpo.L_ergo92)
      (fun fmt -> 
	 Ergo92.export_goal fmt gid gpred ; 
      ) ;
    (* --- Warnings --- *)
    if wrn <> [] then
      begin
	let pp_warnings fmt ws =
	  let n = List.length ws in if n = 1
	  then Format.pp_print_string fmt "1 warning"
	  else Format.fprintf fmt "%d warnings" n
	in
	if List.exists (fun w -> w.Wpo.wrn_severe) wrn then
	  Wp_parameters.warning ~current:false
	    "Degenerated goal %s (%a)"
	    gid pp_warnings wrn
	else
	  Wp_parameters.warning ~current:false
	    "Stronger goal %s (%a)"
	    gid pp_warnings wrn ;
      end ;
    (* --- WPO --- *)
    let wpo = {
      Wpo.po_fun    = kf ;
      Wpo.po_bhv    = bhv ;
      Wpo.po_name   = gname ;
      Wpo.po_gid    = gid ;
      Wpo.po_model  = Descr.shared ;
      Wpo.po_env    = export.env ;
      Wpo.po_pid    = propid ;
      Wpo.po_dep    = dep ;
      Wpo.po_warn   = wrn ;
      Wpo.po_updater = Descr.updater ;
    } in
    Wpo.add wpo ;
    if F.is_true gpred then 
      Wpo.set_result wpo Wpo.WP Wpo.Valid ;
    export.goals <- wpo :: export.goals

  (* ------------------------------------------------------------------------ *)
  (* --- Goal Splitting                                                   --- *)
  (* ------------------------------------------------------------------------ *)

  let build_wpos export kf bhv wrn dep propid gpred =
    if Wp_parameters.Split.get () || 
      (WpAnnot.is_assigns propid && PO.assigns_method() = Mcfg.EffectAssigns)
    then
      let goals = Splitter.split (PO.assigns_method ()) gpred in
      if Bag.is_empty goals then
	export_wpo export kf bhv wrn dep propid F.p_true
      else
	WpAnnot.split (export_wpo export kf bhv wrn dep) propid goals
    else
      export_wpo export kf bhv wrn dep propid gpred
    
  let add_goal export kf bhv po =
    let gpred = CV.zip po.PO.g_prop in
    let wrn = ref [] in
    let dep = ref [] in
    PO.iter_description
      (fun w -> wrn := w :: !wrn)
      (fun d -> dep := d :: !dep)
      po.PO.g_descr ;
    build_wpos export kf bhv (List.rev !wrn) (List.rev !dep) po.PO.g_id gpred

  (* ------------------------------------------------------------------------ *)
  (* --- Proof Obilgation Generation                                      --- *)
  (* ------------------------------------------------------------------------ *)

  class computer =
  object

    val mutable wptasks = []
    val mutable exported = None

    method add cfg strategies =
      wptasks <- (cfg,strategies) :: wptasks

    method compute =
      begin

	exported <- None ;
	Wp_error.set_model Descr.name ;
	D.clear () ;
	let env = Wpo.new_env ~context:Descr.context in
	let export = { env=env ; goals=[] } in

	(* Generates Wpos and accumulate exported goals *)
	List.iter
	  (fun (cfg,strategies) ->
	     let kf = Cil2cfg.cfg_kf cfg in
	     let names = WpAnnot.missing_rte kf in
	     if names <> [] then 
	       Wp_parameters.warning ~current:false ~once:true
		 "Missing RTE guards" ;
	     List.iter
	       (fun strategy ->
		  !Db.progress ();
		  let bhv = strategy.WpAnnot.behavior_name in
		  let goals,annotations = WP.compute cfg strategy in
		  if Wp_parameters.Dot.get () then
		    ignore (Cil2cfg.dot_wp_res cfg Descr.shared annotations) ;
		  List.iter
		    (List.iter
		       (add_goal export kf bhv))
		    goals
	       ) strategies
	  ) wptasks ;

	if export.goals <> [] then
	  begin
	    
	    (* --- Env Description --- *)
	    Command.pp_to_file (Wpo.file_for_ctxt ~env)
	      (fun fmt -> D.iter_all (fun d -> F.pp_decl fmt d)) ;
	    
	    (* --- WHY Others --- *)
	    Command.pp_to_file (Wpo.file_for_env ~env Wpo.L_why)
	      (fun fmt -> D.iter_all (fun d -> Why.export_decl fmt d)) ;
	    
	    (* --- COQ --- *)
	    Command.pp_to_file (Wpo.file_for_env ~env Wpo.L_coq)
	      (fun fmt -> 
		 Format.fprintf fmt 
		   "Require Import Reals.@\n\
                    Require Import wp.@\n\
                    Require Import %s.@\n"
		   (Wpo.coq_for_model Descr.shared) ;
		 D.iter_all (fun d -> Coq.export_decl fmt d)) ;
	    
	    (* --- Alt-Ergo 91 --- *)
	    Command.pp_to_file (Wpo.file_for_env ~env Wpo.L_ergo91)
	      (fun fmt -> D.iter_all (fun d -> Ergo91.export_decl fmt d)) ;
	    
	    (* --- Alt-Ergo 92 --- *)
	    Command.pp_to_file (Wpo.file_for_env ~env Wpo.L_ergo92)
	      (fun fmt -> D.iter_all (fun d -> Ergo92.export_decl fmt d)) ;
	    
	  end
	else
	  Wpo.release_env env ;
	    
	(* --- Generated Goals --- *)
	export.goals

      end (* method compute *)

  end (* class computer *)

  let create () = (new computer :> computer)

end
