(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* -------------------------------------------------------------------------- *)
(* --- Generator of Preconditions                                         --- *)
(* -------------------------------------------------------------------------- *)

open Cil_types
open Wp_error

module Create
  (WpModel:
    sig 
      include Mwp.S 
      val model : string
    end)
  =
struct

  module F = WpModel.D.F
  module D = WpModel.D
  module E = Translate_expr.Create(WpModel)
  module L = Translate_prop.Create(WpModel)

  let predicate = Wp_error.protect_translation L.prop
  let expr = Wp_error.protect_translation E.expr
  let addr = Wp_error.protect_translation E.addr
  let cond = Wp_error.protect_translation E.prop
  let cast = Wp_error.protect_translation3 E.expr_cast

  type t_env = F.pool * L.env

  type t_prop = assigns_kind * property

  and property = {
    bindings : D.bindings ;
    instances : L.instance list ;
    property : F.pred ;
  }

  and assigns_kind =
    | NoAssigns
    | EffectAssigns of effect_assigns
    | RegionAssigns of region_assigns

  and effect_assigns = {
    a_pid : WpAnnot.prop_id ;
    a_label : Clabels.c_label ;
    a_effect : F.var ; (* accumulated zones for effects *)
    a_locals : F.var ; (* accumulated zones for local variables *)
  }

  and region_assigns = {
    r_pid : WpAnnot.prop_id ;
    r_vars : F.var list ; (* finger prints of the assigns *)
    r_global : WpModel.region ; (* global region *)
    r_label : Clabels.c_label ; (* reference labels *)
    mutable r_effect : int ; (* number of effects *)
    mutable r_locals : (varinfo list * F.var list * WpModel.region) list ; (* stack of locals *)
  }

  let empty = NoAssigns , {
    bindings = D.closed ;
    instances = [] ;
    property = F.p_true ;
  }

  let zip (_,omega) =
    let _ , q = L.instantiate omega.instances omega.property in
    D.close omega.bindings q

  let merge_assigns s1 s2 =
    match s1,s2 with
      | NoAssigns,NoAssigns -> NoAssigns
      | NoAssigns,a -> a
      | a,NoAssigns -> a
      | EffectAssigns a1 , EffectAssigns a2 when a1 == a2 -> s1
      | RegionAssigns r1 , RegionAssigns r2 when r1 == r2 -> s1
      | _ -> Wp_parameters.fatal "Merging different assigns goals"

  let merge_property f w1 w2 =
    D.capture (List.map L.as_hint w1.instances) w1.bindings ;
    D.capture (List.map L.as_hint w2.instances) w2.bindings ;
    {
      bindings=D.closed ;
      instances=L.merge w1.instances w2.instances ;
      property=f
	(D.close w1.bindings w1.property)
	(D.close w2.bindings w2.property) ;
    }

  let merge (s1,p1) (s2,p2) = 
    merge_assigns s1 s2 , merge_property F.p_and p1 p2

  let pretty fmt wp = F.pp_pred fmt (zip wp)

  let new_env kf =
    let frame =
      WpModel.new_frame
	kf ~m_pre:None ~m_post:None ~m_here:None
	None
    in
    F.pool() , L.env_body frame

  type closing =
    | Keep_opened
    | Apply_hints
    | Close_context

  type assigns =
    | Keep_assigns
    | Clear_assigns
    | Label_assigns of Clabels.c_label
    | Goal_assigns of assigns_kind ref

  (* --- Utilities --- *)

  let pp_vars fmt = function
    | [] -> Format.pp_print_string fmt "-"
    | x::xs -> 
	Format.fprintf fmt "@[<hov 2>%s" x.vname ;
	List.iter
	  (fun x -> Format.fprintf fmt ",@,%s" x.vname)
	  xs ;
	Format.fprintf fmt "@]"

  let region_assigns k = D.p_app1 "assigns" (F.e_int k)

  let region_included re r =
    r.r_effect <- succ r.r_effect ;
    let authorized =
      List.fold_left
	(fun re (_,_,rx) ->
	   WpModel.region_union re rx
	) r.r_global r.r_locals
    in
    let p_included = WpModel.region_included re authorized in
    F.p_iff (region_assigns r.r_effect) p_included

  (* ------------------------------------------------------------------------ *)
  (* ---  Context Management                                              --- *)
  (* ------------------------------------------------------------------------ *)

  let close_property where context q = function
    | Keep_opened ->
	let instances = L.pop_instances where in
	let bindings = D.pop where context in
	{
	  bindings=bindings ;
	  instances=instances ;
	  property=q ;
	}
    | Apply_hints ->
	let instances = L.pop_instances where in
	let instances,q = L.instantiate instances q in
	if instances = []
	then {
	  bindings=D.closed ;
	  instances=[];
	  property=D.flush where context q;
	} else {
	  bindings=D.pop where context ;
	  instances=instances ;
	  property=q ;
	}
    | Close_context ->
	let instances = L.pop_instances where in
	let _ , q = L.instantiate instances q in
	{
	  bindings=D.closed ;
	  instances=[] ;
	  property=D.flush where context q ;
	}
	  
  exception Failed

  let on_context (pool,env) (where:string)
      (akind,omega) closing assigns
      (f : L.env -> assigns_kind -> F.pred -> F.pred) : t_prop
      =
    let hints = List.map L.as_hint omega.instances in
    let context = D.push where pool omega.bindings hints in
    L.push_instances where omega.instances ;
    try
      let wp = ref F.p_false in
      begin
	try
	  let prop = 
	    try f env akind omega.property
	    with Failed -> F.p_false
	  in
	  let huge = 
	    let m = Wp_parameters.Huge.get () in
	    if m < 1 then 1 else if m > 29 then max_int else 1 lsl m
	  in
	  if F.huge_pred huge prop then
	    (raise(Wp_error.Error("WP","Huge property"))) ;
	  wp := prop ;
	with err ->
	  let (source,reason) = Wp_error.protect err in
	  Datalib.Collector.add_warning ~source ~reason ~severe:true
	    "Abort goal generation" ;
      end ;
      (* f must be computed before, because of lazy bindings for assign goals *)
      let asgns =
	match assigns , akind with
	  | Goal_assigns gref , _ -> !gref
	  | Keep_assigns , a -> a
	  | Clear_assigns , _ | _ , NoAssigns -> NoAssigns
	  | Label_assigns l , EffectAssigns a when a.a_label = l ->
	      let ze = WpModel.dzone_empty () in
	      wp := D.subst a.a_effect ze (D.subst a.a_locals ze !wp) ;
	      NoAssigns
	  | Label_assigns l , RegionAssigns r when r.r_label = l ->
	      let rec all_assigns p k = 
		if k = 0 then p else
		  let pk = region_assigns k in
		  all_assigns (F.p_and pk p) (pred k)
	      in
	      let p_goal = region_assigns 0 in
	      let p_gdef = F.p_iff p_goal (all_assigns F.p_true r.r_effect) in
	      wp := D.forall r.r_vars (F.p_implies p_gdef !wp) ;
	      NoAssigns
	  | Label_assigns _ , a -> a
      in
      asgns , close_property where context !wp closing 
    with err ->
      ignore (L.pop_instances where) ;
      D.kill where context ;
      raise err

  let label env l wp =
    let lab = Clabels.c_label l in
    on_context env "label" wp Keep_opened (Label_assigns lab)
      (fun env _assigns p ->
	 let frame = L.get_frame env in
	 if lab = Clabels.Here then p
	 else
	   let wp = WpModel.update_at_label frame lab p in
	   Clabels.label F.p_named lab wp)

  (* ------------------------------------------------------------------------ *)
  (* ---  Hypotheses and Goal Management                                  --- *)
  (* ------------------------------------------------------------------------ *)

  let merge_with f (a,p) (a',p') =
    merge_assigns a a' , merge_property f p p'

  let add_as_hyp _env = merge_with F.p_implies
  let add_as_goal _env = merge_with F.p_and

  let add_hyp env hid h wp =
    on_context env "add_hyp" wp Keep_opened Keep_assigns
      (fun env _assigns p ->
	 if F.is_true p then p else
	   match predicate env h with
	     | Result h ->
		 Datalib.Collector.add_depend hid "hypothesis";
		 F.p_implies h p
	     | Warning(source,reason) ->
		 Datalib.Collector.add_warning ~source ~reason
		   "Ignored hypothesis %a" WpAnnot.pp_id hid ; p)

  let add_goal env pid pn wp =
    on_context env "add_goal" wp Keep_opened Clear_assigns
      (fun env _assigns p ->
	 if F.is_false p then p else
	   match predicate env pn with
	     | Result g ->
                 F.p_and g p
	     | Warning(source,reason) ->
		 Datalib.Collector.add_warning ~severe:true ~source ~reason
                   "Goal %a can not be translated"
		   WpAnnot.pp_id pid ;
		 F.p_false)

  (* ------------------------------------------------------------------------ *)
  (* ---  Axiom Rule                                                      --- *)
  (* ------------------------------------------------------------------------ *)

  let add_axiom _id name labels pn =
    match
      Wp_error.protect_translation3 L.add_axiom name labels pn
    with
      | Result () -> ()
      | Warning(source,reason) ->
          Datalib.Collector.add_warning ~source ~reason
	    "Ignored user-defined axiom '%s'" name

  (* ------------------------------------------------------------------------ *)
  (* ---  Assignment Rule                                                 --- *)
  (* ------------------------------------------------------------------------ *)

  let assign env lv e wp =
    on_context env "assign" wp Keep_opened Keep_assigns
      (fun env assigns p ->
	 let frame = L.get_frame env in
	 let m = WpModel.mem_at frame Clabels.Here in
	 match addr m lv with
	   | Warning(source,reason) ->
	       Datalib.Collector.add_warning
		 ~severe:true ~source ~reason
		 "No translation for l-value '%a'" !Ast_printer.d_lval lv ;
	       F.p_false
	   | Result l ->
	       let te = Cil.typeOf e in
	       let obj = Ctypes.object_of te in
	       let wp =
		 match expr m e with
		   | Warning(source,reason) ->
		       Datalib.Collector.add_warning ~source ~reason
			 "Ignored r-value of assignment" ;
		       let modified = D.Aloc(obj,l) in
		       D.havoc_static (WpModel.subst_havoc frame modified) p
		   | Result v ->
		       let t = Ctypes.object_of (Cil.typeOf e) in
		       WpModel.subst_lval frame t l v p
	       in
	       begin
		 match assigns with
		   | NoAssigns -> wp
		   | EffectAssigns a ->
		       let zl = WpModel.dzone_assigned (D.Aloc(obj,l)) in
		       let zs = WpModel.dzone_union (F.var a.a_effect) zl in
		       D.subst a.a_effect zs wp
		   | RegionAssigns r ->
		       let rl = WpModel.region_assigned (D.Aloc(obj,l)) in
		       F.p_implies (region_included rl r) wp
	       end )

  (* ------------------------------------------------------------------------ *)
  (* ---  Return Rule                                                     --- *)
  (* ------------------------------------------------------------------------ *)

  let return env e wp =
    on_context env "return" wp Keep_opened Keep_assigns
      (fun env _assigns p ->
	 let frame = L.get_frame env in
	 match e with
	   | None -> p
	   | Some e ->
               begin
		 match expr (WpModel.mem_at frame Clabels.Here) e with
		   | Warning(source,reason) ->
                       Datalib.Collector.add_warning ~source ~reason
			 "Ignored returned value" ;
                       WpModel.subst_result frame None p
		   | Result v ->
		       let _,ty_to = WpModel.result frame in
		       let ty_from = Cil.typeOf e in
		       let r_cast = cast ty_to ty_from v in
		       begin
			 match r_cast with
			   | Warning(source,reason) ->
			       Datalib.Collector.add_warning ~source ~reason
				 "Ignored returned value (because of a cast)" ;
			       WpModel.subst_result frame None p
			   | Result v -> WpModel.subst_result frame (Some v) p
		       end
	       end)

  (* ------------------------------------------------------------------------ *)
  (* ---  Conditional Rule                                                --- *)
  (* ------------------------------------------------------------------------ *)

  let test env e wpt wpf =
    let pt = zip wpt in
    let pf = zip wpf in
    let wpe = merge_assigns (fst wpt) (fst wpf) , snd empty in
    on_context env "test" wpe Keep_opened Keep_assigns
      (fun env _assigns _true ->
	 match cond (WpModel.mem_at (L.get_frame env) Clabels.Here) e with
	   | Result b -> 
	       F.p_and
		 (F.p_named "Then" (F.p_implies b pt))
		 (F.p_named "Else" (F.p_implies (F.p_not b) pf))
	   | Warning(source,reason) ->
               Datalib.Collector.add_warning ~source ~reason
		 "Ignored condition of if-statement (%a)" !Ast_printer.d_exp e ;
	       F.p_and pt pf)

  (* ------------------------------------------------------------------------ *)
  (* ---  Switch Rule                                                     --- *)
  (* ------------------------------------------------------------------------ *)

  let case_of_exp m_here e =
    match Ctypes.get_int e with
      | Some k -> F.e_int64 k
      | None ->
	  match expr m_here e with
	    | Result( WpModel.V_int(_,term) ) -> term
	    | Result( _ ) -> 
		Datalib.Collector.add_warning
		  ~severe:true ~source:"wp" 
		  ~reason:"non-integer expression"
		  "Can not translate switch statement@[@ (%a)@]" 
		  !Ast_printer.d_exp e ;
		raise Failed
	    | Warning(source,reason) ->
		Datalib.Collector.add_warning
		  ~severe:true ~source ~reason
		  "Can not translate switch statement@[@ (%a)@]"
		  !Ast_printer.d_exp e ;
		raise Failed
		  
  let switch env e wp_cases wp_def =
    let cases = List.map (fun (es,wp) -> es , zip wp) wp_cases in
    let p_def = zip wp_def in
    let assigns =
      List.fold_left
	(fun ak (_es,wp) -> merge_assigns ak (fst wp)) 
	(fst wp_def) wp_cases in
    let wpe = assigns , { (snd empty) with property = p_def } in
    on_context env "swith" wpe Keep_opened Keep_assigns
      (fun env _assigns _true ->
	 let frame = L.get_frame env in
	 let m_here = WpModel.mem_at frame Clabels.Here in

         let typ_e = Cil.typeOf e in
	 let int_e = case_of_exp m_here e in
         let var_e = D.fresh "k" (Mdata.Vacsl(Ctype typ_e)) in
	 let val_e = F.var var_e in
	 
	 let (default,cases) =
	   List.fold_left
	     (fun (default,cases) (es_case,p_case) ->
		let ks = List.map (case_of_exp m_here) es_case in
		let hs = F.p_disj (List.map (F.p_eq val_e) ks) in
		let ds = F.p_conj (List.map (F.p_neq val_e) ks) in
		F.p_and ds default ,
		F.p_and (F.p_implies hs p_case) cases
	     ) (F.p_true , F.p_true) cases
	 in
	 let wp_switch = F.p_and cases (F.p_implies default p_def) in
	 D.subst var_e int_e wp_switch)

  (* ------------------------------------------------------------------------ *)
  (* ---  Scope Rule                                                      --- *)
  (* ------------------------------------------------------------------------ *)

  let scope env vars sc wp =
    on_context env "scope" wp Keep_opened Keep_assigns
      (fun env assigns p ->
	 let frame = L.get_frame env in
	 let wp = WpModel.local_scope frame vars sc p in
	 begin
	   match assigns , sc with
	     | EffectAssigns a , (Mcfg.SC_Block_in | Mcfg.SC_Function_frame) ->
                 let m = WpModel.mem_at frame Clabels.Here in
		 let zs =
		   List.fold_left
		     (fun zs x ->
			let te = Ctypes.object_of x.vtype in
			let ax = D.Aloc(te,WpModel.cvar m x) in
			let zx = WpModel.dzone_assigned ax in
			WpModel.dzone_union zs zx)
		     (F.var a.a_locals) vars
		 in
		 D.subst a.a_locals zs wp
		   
	     | RegionAssigns r , 
		 (Mcfg.SC_Block_out | Mcfg.SC_Function_out) when vars <> [] ->
		 begin
                   let m = WpModel.mem_at frame Clabels.Here in
		   let region =
		     List.fold_left
		       (fun r x ->
			  let lx = WpModel.cvar m x in
			  let tx = Ctypes.object_of x.vtype in
			  let rx = WpModel.region_assigned (D.Aloc(tx,lx)) in
			  WpModel.region_union r rx)
		       (WpModel.region_empty()) vars in
		   let pool = D.pool () in
		   let xs,bindings,local = WpModel.region_fingerprint pool region in
		   r.r_locals <- (vars,xs,local) :: r.r_locals ;
		   F.p_and bindings wp
		 end

	     | RegionAssigns r , 
		   (Mcfg.SC_Block_in | Mcfg.SC_Function_frame) when vars <> [] ->
		 begin
		   match r.r_locals with
		     | (xvars,xs,_) :: stack ->
			 if xvars != vars then
			   Wp_parameters.fatal ~current:true 
			     "Expected locals %a instead of %a" 
			     pp_vars vars pp_vars xvars ;
			 r.r_locals <- stack ;
			 D.forall xs wp
		     | [] ->
			 Wp_parameters.fatal ~current:true 
			   "No opened block for locals %a" pp_vars vars
		 end
		   
	     | _ -> wp
	 end)

  (* ------------------------------------------------------------------------ *)
  (* ---  Build property to prove the FROMs                               --- *)
  (* ------------------------------------------------------------------------ *)
      
      (* TODO: ask Loc if the parameters for [on_context] are correct 
      * because I don't really understand what it means... Anne. *)
  let build_prop_of_from wenv pre wp =
    on_context wenv "build_froms" wp Keep_opened Keep_assigns
      (fun env _assigns p -> 
         let alpha, p' = F.p_more_alpha_cv F.empty_alpha p in
         let p = F.p_implies p p' in
         let add_pre (alpha, p) (id, pre) = match predicate env pre with
           | Result pre ->
               let frame = L.get_frame env in
               let pre = WpModel.update_at_label frame Clabels.Pre pre in
               Datalib.Collector.add_depend id "hypothesis";
               let p = F.p_implies pre p in
               let alpha', pre' = F.p_more_alpha_cv alpha pre in
               let p = F.p_implies pre' p in
                 alpha', p
           | Warning(source,reason) ->
               Datalib.Collector.add_warning ~source ~reason
                 "Ignored hypothesis %a" WpAnnot.pp_id id ; 
               alpha, p
         in
         let alpha, p = List.fold_left add_pre (alpha, p) pre in
         let vars = F.fold_alpha (fun _v v' acc -> v'::acc) alpha [] in
           let p = F.p_forall vars p in
             p
      )

  (* ------------------------------------------------------------------------ *)
  (* ---  Closing Goal                                                    --- *)
  (* ------------------------------------------------------------------------ *)
      
  let close env wp =
    on_context env "close" wp Close_context Clear_assigns
      (fun env _assigns p ->
	 let frame = L.get_frame env in
	 WpModel.quantify_at_label frame Clabels.Here p)

  (* ------------------------------------------------------------------------ *)
  (* --- Normal Assigns clauses                                           --- *)
  (* ------------------------------------------------------------------------ *)

  let rec translate_assigned_targets env assigned =
    match assigned with
      | [] -> Result []
      | lv::others ->
	  match translate_assigned_targets env others with
	    | Warning(p,m) -> Warning(p,m)
	    | Result acc ->
		Wp_error.protect_translation
		  (fun e t -> L.assigned e t @ acc)
		  env lv

  let translate_assigned env assigned =
    let lvset = List.fold_left
      (fun lvset (loc,_from) ->
	match loc with
	| Nothing -> lvset
	| Location {Cil_types.it_content=lv} ->
	  Cil_datatype.Term.Set.add lv lvset)
      Cil_datatype.Term.Set.empty assigned
    in
    translate_assigned_targets env (Cil_datatype.Term.Set.elements lvset)

  (* None for Everything or Some loc assigned list *)
  let assigned_of_assigns env assigned =
    if assigned=[] then None (* Everything *)
    else
      match translate_assigned env assigned with
	| Result region -> Some region
	| Warning(source,reason) ->
	    Datalib.Collector.add_warning ~source ~reason
              "Can not translate assign hypothesis, assigns everything instead" ;
	    None

  (* Region for proving assign close *)
  let assigned_for_assigns_goal kind from env assigned =
    let env_assigned =
      match kind with
	| WpAnnot.StmtAssigns -> L.env_at env from
	| WpAnnot.LoopAssigns -> env
    in
    match translate_assigned env_assigned assigned with
      | Warning(wsrc, msg) -> raise (Wp_error.Error(wsrc,msg))
      | Result region -> region

  (* Apply region during wp *)
  let havoc_region hkind caller_frame region prop =
    match region with
      | None ->
	  (* Assigns Everythings *)
          WpModel.quantify_at_label caller_frame Clabels.Here prop
      | Some region ->
	  let hs =
	    List.concat
	      (List.map (WpModel.subst_havoc caller_frame) region) in
	  begin
	    match hkind with
	      | WpAnnot.StmtAssigns -> D.havoc_static hs prop
	      | WpAnnot.LoopAssigns -> D.havoc_inductive hs prop
	  end

  (* ------------------------------------------------------------------------ *)
  (* --- Assigns Method                                                   --- *)
  (* ------------------------------------------------------------------------ *)

  let assigns_method () =
    let mth = Wp_parameters.get_assigns_method () in
    match mth with
      | Mcfg.NoAssigns -> mth
      | Mcfg.EffectAssigns when WpModel.effect_supported -> mth
      | Mcfg.RegionAssigns when WpModel.region_supported -> mth
      | _ ->
	  if WpModel.assigns_supported 
	  then Mcfg.NormalAssigns
	  else Mcfg.NoAssigns

  (* ------------------------------------------------------------------------ *)
  (* --- Generate Observational Assigns Goal                              --- *)
  (* ------------------------------------------------------------------------ *)

  let add_normal_assigns env pid label kind assigned wp =
    on_context env "add_assigns" wp Keep_opened Clear_assigns
      (fun env _noassigns prop ->
	 if assigned = [] then prop (* everything assigned : nothing to do *)
	 else
	   try
	     let frame = L.get_frame env in
	     let from = Clabels.c_label label in
	     let asgned = assigned_for_assigns_goal kind from env assigned in
	     let goal = WpModel.assigns_goal
	       frame  (* frame of model *)
	       from   (* label from *)
	       asgned (* assigned region *)
	       Clabels.Here (* label to *)
	     in
	     F.p_and goal prop
	   with e ->
	     let (source,reason) = Wp_error.protect e in
             Datalib.Collector.add_warning
	       ~severe:true ~source ~reason
	       "Goal %a can not be translated"
	       WpAnnot.pp_id pid ;
             F.p_false)

  (* ------------------------------------------------------------------------ *)
  (* ---  Prove assigns clauses with WpModel.dzones                       --- *)
  (* ------------------------------------------------------------------------ *)

  let add_effect_assigns env pid label kind assigned wp =
    let from = Clabels.c_label label in
    let goal = ref NoAssigns in
    on_context env "add_assigns" wp Keep_opened (Goal_assigns goal)
      (fun env _noassigns prop ->
	 if assigned = [] then prop (* everything assigned : nothing to do *)
	 else
	   try
	     let asgns = assigned_for_assigns_goal kind from env assigned in
	     let ze = D.fresh "ze" (Mdata.Vmodel WpModel.tau_of_dzone) in
	     let zx = D.fresh "zx" (Mdata.Vmodel WpModel.tau_of_dzone) in
	     let zs = List.fold_left
	       (fun zs a ->
		  let zx = WpModel.dzone_assigned a in
		  WpModel.dzone_union zs zx)
	       (F.var zx) asgns
	     in
	     goal := EffectAssigns {
	       a_pid = pid ;
	       a_label = from ;
	       a_effect = ze ;
	       a_locals = zx ;
	     } ;
	     WpModel.dzone_subset (F.var ze) zs
	 with e ->
	   let (source,reason) = Wp_error.protect e in
	   Datalib.Collector.add_warning
	     ~severe:true ~source ~reason
	     "Goal %a can not be translated"
	     WpAnnot.pp_id pid ;
           F.p_false)

  (* ------------------------------------------------------------------------ *)
  (* ---  Prove assigns clauses with WpModel.region                       --- *)
  (* ------------------------------------------------------------------------ *)

  let add_region_assigns env pid label kind assigned wp =
    let from = Clabels.c_label label in
    let pool = fst env in
    let goal = ref NoAssigns in
    on_context env "add_assigns" wp Keep_opened (Goal_assigns goal)
      (fun env _noassigns prop ->
	 if assigned = [] then prop (* everything assigned : nothing to do *)
	 else
	   try
	     let asgned = assigned_for_assigns_goal kind from env assigned in
	     let region = List.fold_left
	       (fun re a ->
		  let rx = WpModel.region_assigned a in
		  WpModel.region_union re rx)
	       (WpModel.region_empty()) asgned
	     in
	     let xs,bindings,global = WpModel.region_fingerprint pool region in
	     goal := RegionAssigns {
	       r_pid = pid ;
	       r_label = from ;
	       r_global = global ;
	       r_vars = xs ;
	       r_effect = 0 ;
	       r_locals = [] ;
	     } ;
	     F.p_implies bindings (region_assigns 0)
	 with e ->
	   let (source,reason) = Wp_error.protect e in
	   Datalib.Collector.add_warning
	     ~severe:true ~source ~reason
	     "Goal %a can not be translated"
	     WpAnnot.pp_id pid ;
           F.p_false)

  (* ------------------------------------------------------------------------ *)
  (* --- Dispatch Assigns Goal against selected method                    --- *)
  (* ------------------------------------------------------------------------ *)

  let add_assigns env pid label kind assigned wp =
    match assigns_method () with
      | Mcfg.NoAssigns -> Wp_parameters.abort "Unsupported assigns with the model"
      | Mcfg.NormalAssigns -> add_normal_assigns env pid label kind assigned wp
      | Mcfg.EffectAssigns -> add_effect_assigns env pid label kind assigned wp
      | Mcfg.RegionAssigns -> add_region_assigns env pid label kind assigned wp

  (* Assigns Hypothesis *)

  let check_assigns goal region wp =
    match goal , region with
      | NoAssigns , _ -> wp
      | (EffectAssigns _ | RegionAssigns _) , None -> wp
	  (* TODO: Should be Assigns Everything !! *)
	  (* Seems to be a bug in CIL because Nothing leads
	     to no assigned, hence None. *)
	  (*
	    Datalib.Collector.add_warning
	    ~severe:true
	    ~reason:"Assigns everything during calculus"
	    "Can not prove the assign goal" ;
	    F.p_false
	  *)
      | EffectAssigns a , Some zones ->
	  let ze = List.fold_left
	    (fun zs a ->
	       let za = WpModel.dzone_assigned a in
	       WpModel.dzone_union zs za)
	    (F.var a.a_effect) zones
	  in
	  D.subst a.a_effect ze wp
      | RegionAssigns r , Some zones ->
	  let re = List.fold_left
	    (fun rs a ->
	       let ra = WpModel.region_assigned a in
	       WpModel.region_union rs ra)
	    (WpModel.region_empty()) zones
	  in
	  F.p_implies (region_included re r) wp

  let use_assigns env hid kind assigned wp =
    on_context env "use_assigns" wp Close_context Keep_assigns
      (fun env assignsgoal p ->
	 Datalib.Collector.add_depend hid "assigns hypothesis" ;
	 let asgnd = assigned_of_assigns env assigned in
	 let p0 = check_assigns assignsgoal asgnd p in
	 havoc_region kind (L.get_frame env) asgnd p0)

  (* ------------------------------------------------------------------------ *)
  (* --- CALL Rule                                                        --- *)
  (* ------------------------------------------------------------------------ *)

  let call caller_env ~get_pre ~get_post ~get_exits ~get_assigns
      stmt lv fct args c_assigns p_normal p_exits
      =
    let frame = L.get_frame caller_env in
    let called_kf = match fct.enode with
      | Lval (Var vinfo,NoOffset) -> Globals.Functions.get vinfo
      | _ ->
	  Datalib.Collector.add_warning
	    ~severe:true
	    ~reason:"call through function pointer not implemented yet"
	    "Can not treat function call" ;
	  raise Failed
    in
    
    let kf_pid = WpAnnot.pid_for_function called_kf WpModel.model in
    Datalib.Collector.add_depend kf_pid "calling" ;
    
    let p_normal, x_opt = match lv with
      | None -> p_normal, None
      | Some lv ->
	  let t_result = Kernel_function.get_return_type called_kf in
	  let x_result = D.fresh "result" (Mdata.Vacsl (Ctype t_result)) in
	  
	  let lv_t = Ctypes.object_of t_result in
	  let casted_result =
	    cast t_result (Cil.typeOfLval lv)
	      (WpModel.value_of_logic lv_t (F.var x_result))
	  in
	  let v_result = match casted_result with
	    | Warning(source,reason) ->
		Datalib.Collector.add_warning
		  ~severe:true ~source ~reason
		  "Can not cast the returned value" ;
		raise Failed
	    | Result res -> res
	  in
	  
	  let p =
	    begin
	      match addr (WpModel.mem_at frame Clabels.Here) lv with
		| Warning(source, reason) ->
		    Datalib.Collector.add_warning
		      ~severe:true ~source ~reason
		      "Can not assign the returned value, no translation for l-value" ;
		    raise Failed
		| Result loc ->
		    WpModel.subst_lval frame lv_t loc v_result p_normal
	    end
	  in
	  p, Some x_result
    in
    
    let pre_label = Clabels.CallAt stmt.sid in
    let m_pre = WpModel.mem_at frame pre_label in
    let m_post = WpModel.mem_at frame Clabels.Here in
    
    let called_pre =  get_pre called_kf in
    let called_assigns = get_assigns called_kf in
    let called_post = get_post called_kf in
    let called_exits = get_exits called_kf in
    
    let translate_arg e =
      match expr m_pre e with
	| Warning(source,reason) ->
	    Datalib.Collector.add_warning ~source ~reason
	      "Can not call function, no translation for parameter '%a'"
	      !Ast_printer.d_exp e ;
	    raise Failed
	| Result v -> v
    in
    let values = List.map translate_arg args in
    
    let do_prop env item p =
      match predicate env p with
	| Warning(source,reason) ->
	    Datalib.Collector.add_warning
	      ~source ~reason
	      "Ignored %s for function call" item ;
	    F.p_true
	| Result p -> p
    in
    
    let env_post = L.call_post called_kf values m_pre m_post x_opt
      
    in
    let do_post (_id, post) p =
      (* [LC] handled by the global call dependency
	 Datalib.Collector.add_depend id "called postcondition"; *)
      F.p_implies (do_prop env_post "post-condition" post) p
    in
    let pn0 = (* avoid dependencies when not necessary *)
      if F.is_true p_normal then p_normal else
	List.fold_right do_post called_post p_normal
    in
    let pn = match x_opt with None -> pn0 | Some r -> D.forall [r] pn0 in
    
    let x_status, env_exits =
      match WpModel.get_exit_status frame with
	| None ->
	    let x_status =
	      D.fresh "exit_status" (Mdata.Vmodel Formula.Integer)
	    in
	    x_status, L.call_exit env_post x_status
	| Some x -> x, L.call_exit env_post x
    in
    
    let do_exits (_id, e) p =
      (* [LC] handled by the function call consistency
         Datalib.Collector.add_depend id "called exits postcondition"; *)
      F.p_implies (do_prop env_exits "exit-condition" e) p
    in
    let pe = (* avoid dependencies when not necessary *)
      if F.is_true p_exits then p_exits else
	D.forall [x_status]
	  (List.fold_right do_exits called_exits p_exits) in
    
    let env_pre = L.call_pre called_kf values m_pre in
    let asgnd = assigned_of_assigns env_pre called_assigns in
    
    let p1n = havoc_region WpAnnot.StmtAssigns frame asgnd pn in
    let p1e = havoc_region WpAnnot.StmtAssigns frame asgnd pe in
    
    let do_precond (_id, p) =
      (* [LC] handled by the function call consistency
         Datalib.Collector.add_depend id "called precondition" *)
      do_prop env_pre "pre-condition" p
    in
    let preconditions = List.map do_precond called_pre in
    let precondition = F.p_conj preconditions in
    
    let p2n = F.p_implies precondition p1n in
    let p2e = F.p_implies precondition p1e in
    
    let p3n = check_assigns c_assigns asgnd p2n in
    let p3e = check_assigns c_assigns asgnd p2e in
    
    let condlabel = "Requires_" ^ Kernel_function.get_name called_kf in
    let p4n = F.p_and (F.p_named condlabel precondition) p3n in
    let p4e = F.p_and (F.p_named condlabel precondition) p3e in
    
    let pn = WpModel.update_at_label frame pre_label p4n in
    let pe = WpModel.update_at_label frame pre_label p4e in
    (pn,pe)
      
  let call_normal caller_env ~get_pre ~get_post ~get_assigns
      stmt lv fct args wp_normal
      =
    on_context caller_env  "call_normal" wp_normal Close_context Keep_assigns
      (fun caller_env p_assigns p_normal ->
	 let get_exits _kf = [] in
	 let p_exit = F.p_true in
	 fst ( call caller_env ~get_pre ~get_post ~get_exits ~get_assigns
		 stmt lv fct args p_assigns p_normal p_exit ))

  let call_exit caller_env ~get_pre ~get_exits ~get_assigns
      stmt fct args wp_exits
      =
    on_context caller_env "call_exit" wp_exits Close_context Keep_assigns
      (fun caller_env p_assigns p_exits ->
	 let lv = None in
	 let get_post _kf = [] in
	 let p_normal = F.p_true in
	 snd ( call caller_env ~get_pre ~get_post ~get_exits ~get_assigns
		 stmt lv fct args p_assigns p_normal p_exits ))

(* --- End of Weakest Precondition Rules --- *)

end
