(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

open Cil_types

(** abstract type of a cfg *)
type t 

(** abstract type of the cfg nodes *)
type node
val pp_node : Format.formatter -> node -> unit

(** abstract type of the cfg edges *)
type edge
val pp_edge : Format.formatter -> edge -> unit

(** set of edges *)
module Eset : Set.S with type elt = edge

(** node and edges relations *)
val edge_src : edge -> node 
val edge_dst : edge -> node 
val pred_e : t -> node -> edge list
val succ_e : t -> node -> edge list

(** iterators *)
val fold_nodes : (node -> 'a -> 'a) -> t -> 'a -> 'a
val iter_nodes : (node -> unit) -> t -> unit
val iter_edges : (edge -> unit) -> t -> unit

(** Be careful that only Bstmt are real Block statements *)
type block_type = private
  | Bstmt of stmt | Bthen of stmt | Belse of stmt | Bloop of stmt | Bfct

type node_type = private
  | Vstart | Vend | Vexit 
  | VfctIn | VfctOut 
  | VblkIn of block_type * block
  | VblkOut of block_type * block
  | Vstmt of stmt
  | VcallIn of stmt * lval option * exp * exp list 
  | VcallOut of stmt * lval option * exp * exp list 
  | VcallExit of stmt * lval option * exp * exp list
  | Vtest of bool * stmt * exp
  | Vswitch of stmt * exp
  | Vloop of bool * stmt (* boolean is is_natural *)

val node_type : node -> node_type
val pp_node_type : Format.formatter -> node_type -> unit

val node_stmt_opt : node -> stmt option
val start_stmt_of_node : node -> stmt option

(** similar to [succ_e g v] 
* but tests the branch to return (then-edge, else-edge) 
  * @raise Invalid_argument if the node is not a test.
* *)
val get_test_edges : t -> node -> edge * edge

(** similar to [succ_e g v]
but give the switch cases and the default edge *)
val get_switch_edges : t -> node -> (exp list * edge) list * edge

(** similar to [succ_e g v]
but gives the edge to VcallOut first and the edge to VcallExit second. *)
val get_call_out_edges : t -> node -> edge * edge

val blocks_closed_by_edge : t -> edge -> block list

val is_back_edge : edge -> bool

(** @return a label that represent the first point of a loop body. *)
val loop_head_label : stmt -> logic_label

(** @return the list of the loop nodes in the cfg. *)
val loop_nodes : t ->  node list

(** @raise Invalid_argument when the node is not a loop node.
* @raise Log.FeatureRequest for non natural loops.
* @return the list a loop nodes inside the given loop and the edges
* of the body. *)
val loop_node_edges : t -> node -> node list * Eset.t

(** @return the (normalized) labels at the program point of the edge. *)
val get_edge_labels : edge ->  Cil_types.logic_label list

(** @return None when the edge leads to the end of the function. *)
val get_edge_next_stmt : t -> edge -> stmt option

(** @raise Kernel_function.No_Definition when [kf] is a declaration.
  * @raise Log.FeatureRequest for non natural loops and 'exception'
  * @return the graph and the list of unreachable nodes.
  * *)
val get : Kernel_function.t -> t * node_type list

(** get the starting edges *)
val start_edge : t -> edge 

(** wether an exit edge exists or not *)
val has_exit : t -> bool

(** Find the node that follows the input node statement.
* The statement postcondition can then be stored to the edges before that node.
* @raise Not_found if the node is not a complex not (test, loop, ...),
* or when the node after has been removed (unreachable) *)
val node_after : t -> node -> node

(** Find the edges [e] that goes to the [Vexit] node inside the statement
* begining at node [n] *)
val get_exit_edges : t -> node -> edge list

(** Find the edges [e] of the statement node [n] postcondition
* and the set of edges that are inside the statement ([e] excluded).
* For instance, for a single statement node, [e] is [succ_e n],
* and the set is empty. For a test node, [e] are the last edges of the 2
* branches, and the set contains all the edges between [n] and the [e] edges.
* *)
val get_internal_edges : t -> node -> edge list * Eset.t

val cfg_kf : t -> Kernel_function.t

(** signature of a mapping table from cfg edges to some information. *)
module type HEsig = 
sig
  type ti
  type t
  val create : int -> t
  val find : t -> edge -> ti
  val find_all : t -> edge -> ti list
  val add : t -> edge -> ti -> unit
  val replace : t -> edge -> ti -> unit
  val clear : t -> unit
end

module HE (I : sig type t end) : HEsig with type ti = I.t

(** type of functions to print things related to edges *)
type pp_edge_fun = Format.formatter -> edge -> unit

(** Output the graph in a dot file and return its name **)
val dot_cfg : t -> string

(** [dot_annots cfg bhv_name pp_annots] *)
val dot_annots : t -> string option -> pp_edge_fun -> string

(** [dot_wp_res cfg model pp_wp_res] *)
val dot_wp_res : t -> string -> pp_edge_fun -> string

val mk_logic_label : Cil_types.stmt -> Cil_types.logic_label
  (** create a virtual label to a statement (it can have no label) *)
