(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

module WpLog = Wp_parameters
open Format
open Ctypes
open Cil_types
open Formula
open Mdata

let mk_access_ufield = "access_ufield"
let mk_update_ufield = "update_ufield"

let mk_empty      = "empty"
let mk_singleton  = "singleton"
let mk_union      = "union"
let mk_inter      = "inter"
let mk_remove     = "remove"

let mk_range      = "range"
let mk_range_inf  = "range_inf"
let mk_range_sup  = "range_sup"
let mk_integers   = "integers_set"
let mk_radd       = "plus_int"
let mk_rmult      = "mult_int"
let mk_rneg       = "neg_int"

let get_range_index = "get_range_index"
let set_range_index = "set_range_index"

let farray_access = "access"
let farray_update = "update"

let mk_iformat i = Pretty_utils.sfprintf "%a_format" Ctypes.pp_int i
let mk_imodulo = "as_int"
let mk_iguard = "is_in_format"

let mk_fformat f = Pretty_utils.sfprintf "%a_format" Ctypes.pp_float f
let mk_round = "as_float"
let mk_fguard = "is_in_format"


(* pointer arithmetics *)

let mk_equal_pointer_bool = "equal_pointer"
let mk_lt_pointer_bool = "lt_pointer"
let mk_le_pointer_bool = "le_pointer"
let mk_minus_pointer = "minus_pointer"

let mk_lt_pointer = "lt_pointer_rel"
let mk_le_pointer = "le_pointer_rel"

let mk_is_null = "is_null"
let mk_null = "null"


let mk_int_format     = "int_format"
let mk_real_format    = "real_format"
let mk_pointer_format = "pointer_format"
let mk_record_format   = "record_format"
let mk_urecord_format   = "urecord_format"
let mk_array_format  = "array_format"

let mk_eq_record = "eq_struct"
let mk_eq_urecord = "eq_union"
let mk_eq_array = "eq_array"
let mk_eq_record_bool = "eq_struct_bool"
let mk_eq_urecord_bool = "eq_union_bool"
let mk_eq_array_bool = "eq_array_bool"

let identifier x =
  let range a c b = a <= c && c <= b in
  let buffer = Buffer.create 80 in
  for i=0 to String.length x - 1 do
    let c = x.[i] in
    if range 'a' c 'z'
      || range 'A' c 'Z'
      || (i > 0 && range '0' c '9')
      || c = '_'
    then
      Buffer.add_char buffer c ;
  done ;
  Buffer.contents buffer

(* Does not need to be projectified but could be. *)
let reserved_prefix_tbl = Hashtbl.create 17
let register_prefix s =
  Hashtbl.replace reserved_prefix_tbl s ()
let has_reserved_prefix name =
  try
    let index = String.index name '_' in
    let prefix = String.sub name 0 index in
    Hashtbl.mem reserved_prefix_tbl prefix
  with Not_found -> false

module Tau =
struct

  (* --------------------------------------------------------------------- *)
  (* --- Pure Type Conversions                                         --- *)
  (* --------------------------------------------------------------------- *)

  let tau_of_logic_ctype c =
    match Ctypes.object_of c with
      | C_int _ -> Integer
      | C_float _ -> Real
      | C_pointer _ -> Pointer
      | C_comp c -> Record c
      | C_array a -> Array a

  let rec pp_tau fmt = function
    | Integer -> pp_print_string fmt "int"
    | Real -> pp_print_string fmt "real"
    | Boolean -> pp_print_string fmt "bool"
    | Pointer -> pp_print_string fmt "pointer"
    | Record c -> pp_print_string fmt (if c.cstruct then "record" else "union")
    | Array a ->
	Format.fprintf fmt "%a array" pp_tau
	  (tau_of_logic_ctype a.arr_element)
    | Formula.Set te ->
	if Wp_parameters.verbose_atleast 2
	then Format.fprintf fmt "%a set" pp_tau te
	else pp_print_string fmt "set"
    | ADT(s,[]) -> pp_print_string fmt s
    | ADT(s,[t]) -> Format.fprintf fmt "%a %s" pp_tau t s
    | ADT(s,t::ts) ->
	Format.fprintf fmt "@[(%a" pp_tau t ;
	List.iter (fun t -> Format.fprintf fmt ",@,%a" pp_tau t) ts ;
	Format.fprintf fmt ") %s@]" s

end

(* ------------------------------------------------------------------------ *)
(* --- Goals information collection                                     --- *)
(* ------------------------------------------------------------------------ *)

module Collector =
struct

  type t = {
    mutable c_warning : Wpo.warning list;
    mutable c_depends : Wpo.dependency list;
  }

  let stack = ref []

  let push () =
    let c = { c_warning = []; c_depends = [] } in
      stack := c::(!stack);
      c

  let pop c = match !stack with
    | top::tl when top == c ->
        stack := tl; (List.rev c.c_warning, List.rev c.c_depends)
    | _ -> Wp_parameters.fatal "Datalib.Collector: inconsistent stack"

  let add_warning ?(severe=false) ?source ~reason effect =
    let f _ =
      let msg = Format.flush_str_formatter () in
      let wrn = {
	Wpo.wrn_loc = Log.get_current_source () ;
	Wpo.wrn_severe = severe ;
	Wpo.wrn_source = (match source with None -> "wp" | Some r -> r) ;
	Wpo.wrn_reason = reason ;
	Wpo.wrn_effect = msg ;
      } in
      match !stack with
        | top::_ -> top.c_warning <- wrn::top.c_warning
        | _ -> Wp_parameters.fatal "Datalib.Collector: empty stack"
    in Format.kfprintf f Format.str_formatter effect

  let add_depend prop_id fmt =
    let f _ =
      let msg = Format.flush_str_formatter () in
      let dep = { Wpo.dep_pid = prop_id ; Wpo.dep_descr = msg } in
      match !stack with
        | top::_ -> top.c_depends <- dep::top.c_depends
        | _ -> Wp_parameters.fatal "Datalib.Collector: empty stack"
    in Format.kfprintf f Format.str_formatter fmt

end

(* -------------------------------------------------------------------------- *)
(* --- Creates the Data Library                                           --- *)
(* -------------------------------------------------------------------------- *)

module Create (F:Formula.S) =
struct

  module F = F

  type abstract = Formula.abstract F.term
  type format = Formula.m_format F.term
  type record = Formula.m_array F.term
  type urecord  = Formula.m_array F.term
  type array = Formula.m_array F.term
  type set = Formula.m_set F.term
  type name = Formula.integer F.term
  type pointer = m_pointer F.term
  type integer = Formula.integer F.term
  type real = Formula.real F.term
  type boolean = Formula.boolean F.term

  type interval = {
    inf : integer option ;
    sup : integer option ;
  }

  let pp_interval fmt rg =
    let pp_opt fmt = function
      | None -> ()
      | Some d -> F.pp_term fmt d
    in
    Format.fprintf fmt "[%a..%a]" pp_opt rg.inf pp_opt rg.sup

  type 'a assigned =
    | Aloc of c_object * 'a
    | Arange of c_object * 'a * interval

  (* ------------------------------------------------------------------------ *)
  (* --- Calling Logic Functions and Predicates                           --- *)
  (* ------------------------------------------------------------------------ *)

  let p_app0 f = F.p_call f []
  let p_app1 f a = F.p_call f [F.wrap a]
  let p_app2 f a b = F.p_call f [F.wrap a;F.wrap b]
  let p_app3 f a b c = F.p_call f [F.wrap a;F.wrap b;F.wrap c]
  let p_app4 f a b c d = F.p_call f [F.wrap a;F.wrap b;F.wrap c;F.wrap d]
  let p_app5 f a b c d e = F.p_call f [F.wrap a; F.wrap b; F.wrap c; F.wrap d;
                                       F.wrap e]

  let e_app0 f = F.unwrap (F.e_call f [])
  let e_app1 f a = F.unwrap (F.e_call f [F.wrap a])
  let e_app2 f a b = F.unwrap (F.e_call f [F.wrap a;F.wrap b])
  let e_app3 f a b c = F.unwrap (F.e_call f [F.wrap a;F.wrap b;F.wrap c])
  let e_app4 f a b c d =
    F.unwrap (F.e_call f [F.wrap a;F.wrap b;F.wrap c;F.wrap d])
  let e_app5 f a b c d e =
    F.unwrap (F.e_call f [F.wrap a;F.wrap b;F.wrap c;F.wrap d; F.wrap e])

  (* ------------------------------------------------------------------------ *)
  (* --- Logic Integer Cast                                               --- *)
  (* ------------------------------------------------------------------------ *)

  let modulo ti e = e_app2 mk_imodulo (e_app0 (mk_iformat ti)) e
  let guard ti e = p_app2 mk_iguard (e_app0 (mk_iformat ti)) e
  let i_convert tfrom tto e =
    if Ctypes.sub_c_int tfrom tto then e else modulo tto e

  (* ------------------------------------------------------------------------ *)
  (* --- Logic Real Cast                                                  --- *)
  (* ------------------------------------------------------------------------ *)

  let round tf e = e_app2 mk_round (e_app0 (mk_fformat tf)) e
  let f_guard tf e = p_app2 mk_fguard (e_app0 (mk_fformat tf)) e
  let f_convert tfrom tto e =
    if Ctypes.sub_c_float tfrom tto then e else round tto e

  (* ------------------------------------------------------------------------ *)
  (* --- Declaration Management                                           --- *)
  (* ------------------------------------------------------------------------ *)

  let part_of_item = function
    | Formula.Type _ -> 1
    | Formula.Name _ -> 2 | Formula.Cons _ -> 3
    | Formula.Function _ | Formula.Predicate _ -> 4
    | Formula.Axiom _ -> 5

  module Dset = Set.Make
    (struct
       type t = F.pred declaration
       let compare d1 d2 =
	 let p = part_of_item d1.d_item - part_of_item d2.d_item in
	 if p = 0 then String.compare d1.d_name d2.d_name else p
     end)

  let gindex = Hashtbl.create 731
  let gsection = Array.create 4 Dset.empty
  let gfresh = Hashtbl.create 131

  let fresh_name prefix basename =
    let x = identifier basename in
    let m = if x="" then prefix else Printf.sprintf "%s_%s" prefix x in
      try
	let p = Hashtbl.find gfresh m in
	incr p ; Printf.sprintf "%s_%d" m !p
      with Not_found ->
	Hashtbl.add gfresh m (ref 0) ; m

  let s_index = function
    | S_name -> 0
    | S_typedef -> 1
    | S_userdef -> 2
    | S_goals -> 3

  let gclear = ref []

  let clear () =
    begin
      Hashtbl.clear gfresh ;
      Hashtbl.clear gindex ;
      for i=0 to 3 do gsection.(i) <- Dset.empty done ;
      List.iter (fun f -> f ()) !gclear ;
    end
  let on_clear f = gclear := !gclear @ [f]

  let locked = ref false

  let add_declaration d =
    try
      if !locked then Wp_parameters.fatal
	"Locked datalib (when declaring %t)" d.d_title ;
      let old = Hashtbl.find gindex d.d_name in
      Wp_parameters.not_yet_implemented
	"Overloaded operator %t and %t"
	 old.d_title d.d_title
    with Not_found ->
      begin
	Hashtbl.add gindex d.d_name d ;
	let s = s_index d.d_section in
	gsection.(s) <- Dset.add d gsection.(s) ;
      end

  let iter s f =
    try
      locked := true ;
      Dset.iter f gsection.(s_index s) ;
      locked := false ;
    with e ->
      locked := false ; raise e

  let iter_all f =
    begin
      iter S_name f ;
      iter S_typedef f ;
      iter S_userdef f ;
      iter S_goals f ;
    end

  (* -------------------------------------------------------------------------- *)
  (* --- Dummy                                                              --- *)
  (* -------------------------------------------------------------------------- *)

  let gdummy = ref 0

  let () = on_clear (fun () -> gdummy :=0)

  let dummy () = incr gdummy; p_app1 "dummy" (F.e_int !gdummy)

  (* -------------------------------------------------------------------------- *)
  (* --- Declaration Functors                                               --- *)
  (* -------------------------------------------------------------------------- *)

  module type Identifiable =
  sig
    type t
    module H : Hashtbl.S
    val index : t -> H.key
    val prefix : string
    val basename : t -> string
    val location : t -> Log.source option
    val pp_title : Format.formatter -> t -> unit
    val pp_descr : Format.formatter -> t -> unit
  end

  module type Registry =
  sig
    type t
    val define : t -> unit
    val get_definition : t -> F.pred declaration
    val on_definition : (t -> F.pred declaration -> unit) -> unit
  end

  module type Declarator =
  sig
    include Identifiable
    val clear : unit -> unit
    val section : section
    val declare : t -> string -> F.pred item
  end

  module Register
    (D : Declarator) :
    (Registry with type t = D.t) =
  struct
    let () = register_prefix D.prefix

    type t = D.t
    let index : F.pred declaration D.H.t = D.H.create 131
    let demons : (t -> F.pred declaration -> unit) list ref = ref []

    let () = on_clear (fun () -> D.clear () ; D.H.clear index)

    let get_definition x =
      let k = D.index x in
      try D.H.find index k
      with Not_found ->
	let name = fresh_name D.prefix (D.basename x) in
	let item = D.declare x name in
	let d = {
	  d_name = name ;
	  d_section = D.section ;
	  d_source = D.location x ;
	  d_item = item ;
	  d_title = (fun fmt -> D.pp_title fmt x) ;
	  d_descr = (fun fmt -> D.pp_descr fmt x) ;
	} in
	add_declaration d ;
	D.H.add index k d ;
	List.iter (fun f -> f x d) !demons ;
	d

    let define x = ignore (get_definition x)
    let on_definition f = demons := !demons @ [f]

  end

  (* -------------------------------------------------------------------------- *)
  (* --- Index Functor                                                       --- *)
  (* -------------------------------------------------------------------------- *)

  let gindexref = ref 0

  module type Indexed =
  sig
    include Registry
    val get_ind : t -> integer
  end

  module Dindex
    (I : Identifiable) :
    (Indexed with type t = I.t) =
  struct
    let hindex = I.H.create 131
    include Register
      (struct
	 include I
	 let clear () = I.H.clear hindex
	 let section = S_name
	 let declare _ _ = incr gindexref ;
            Formula.Cons (!gindexref)
        end)
    let get_ind x =
      let k = I.index x in
      try I.H.find hindex k
      with Not_found ->
	let d = get_definition x in
	let t : integer = F.unwrap (F.e_call d.d_name []) in
	I.H.add hindex k t ; t
  end

  (* -------------------------------------------------------------------------- *)
  (* --- Built-in Identifiables                                             --- *)
  (* -------------------------------------------------------------------------- *)
  let rec pp_dim fmt = function
    | TArray (typ_elt,lo,_,_) ->
	(match lo with
	   | Some lo ->
	       Format.fprintf fmt "[%a]" !Ast_printer.d_exp lo
	   | None ->
	       Format.fprintf fmt "[]"); pp_dim fmt typ_elt
    | _ -> ()


  let rec pp_ctype dim fmt = function
    | TInt(ikind,_) -> Format.fprintf fmt "%a" Cil.d_ikind ikind
    | TFloat(fkind,_) -> Format.fprintf fmt "%a" Cil.d_fkind fkind
    | TPtr(typ,_) -> Format.fprintf fmt "%a*" (pp_ctype dim) typ
    | TFun _ as t ->  Format.fprintf fmt "%a*" (pp_ctype dim) t
    | TEnum (e,_) -> Format.fprintf fmt "enum %s " e.ename
    | TComp (comp,_,_) ->
	Format.fprintf fmt
          "%s %s"
          (if comp.cstruct then "struct" else "union")
          comp.cname

    | TArray (typ_elt,_,_,_) as t ->
	pp_ctype false fmt typ_elt ;
	if dim then pp_dim fmt t;

    | TBuiltin_va_list _ -> pp_print_string fmt "builtin type"
    | TVoid _ -> pp_print_string fmt "void"
    | TNamed (t,_)  -> Format.fprintf fmt "%s" t.tname

  module Varinfo : Identifiable with type t = varinfo =
  struct
    type t = varinfo
    let prefix = "X"
    let index x = x
    let basename x = x.vname
    let location x = Some(Cil.source x.vdecl)
    module H = Cil_datatype.Varinfo.Hashtbl
    let pp_title fmt x =
      if x.vglob
      then Format.fprintf fmt "Global '%a'" !Ast_printer.d_var x
      else Format.fprintf fmt "Local '%a'" !Ast_printer.d_var x
    let pp_descr fmt x =
      Format.fprintf fmt "%a %a ;"
	(pp_ctype true) x.vtype !Ast_printer.d_var x
  end

  module Varaddr : Identifiable with type t = varinfo =
  struct
    include Varinfo
    let prefix = "A"
    let pp_title fmt x = Format.fprintf fmt "Address of '%a'" pp_title x
  end

  module Fieldinfo : Identifiable with type t = fieldinfo =
  struct
    type t = fieldinfo
    let prefix = "F"
    let index f = f
    let basename f = f.fname
    let location f = Some(Cil.source f.floc)
    module H = Cil_datatype.Fieldinfo.Hashtbl
    let pp_title fmt f = Format.fprintf fmt "Field '%s'" f.fname
    let pp_descr fmt f =
      Format.fprintf fmt "@[<hov 0>@[<hov 2>%s %s {@ ... ;@ %a %s ;@ ...@]@ }@]"
	(if f.fcomp.cstruct then "struct" else "union")
	f.fcomp.cname
	!Ast_printer.d_type f.ftype f.fname
  end

  module Compinfo : Identifiable with type t = compinfo =
  struct
    type t = compinfo
    let prefix = "C"
    let index c = c
    let basename c = c.cname
    let location c =
      match c.cfields with
	| f :: _ ->
	    let s = Cil.source f.floc in
	    if s.Log.src_file = "" then None else Some s
	| [] -> None
    module H = Cil_datatype.Compinfo.Hashtbl
    let pp_title fmt c =
      Format.fprintf fmt "%s '%s'"
	(if c.cstruct then "Struct" else "Union") c.cname
    let pp_descr fmt c =
      Format.fprintf fmt "typedef %s %s { ... }"
	(if c.cstruct then "struct" else "union") c.cname
  end

  module Arrayinfo : Identifiable with type t = arrayinfo =
  struct
    type t = arrayinfo
    let prefix = "A"
    module H = Hashtbl.Make(AinfoComparable)
    let index a = a
    let location _ = None
    let basename a = Ctypes.basename (C_array a)
    let pp_title fmt a = Ctypes.pretty fmt (C_array a)
    let pp_descr fmt _ = Format.fprintf fmt "Logic array"
  end

  module LTypeinfo : Identifiable with type t = Cil_types.logic_type =
  struct
    type t = Cil_types.logic_type
    let prefix = "AT"
    let index c = c
    let basename ty =
      let rec typ_basename ty =
	match object_of ty with
	  | C_int i -> Pretty_utils.sfprintf "%a" Ctypes.pp_int i
	  | C_float f -> Pretty_utils.sfprintf "%a" Ctypes.pp_float f
	  | C_pointer _ -> "pointer"
	  | C_comp c -> (if c.cstruct then "struct_" else "union_")^
	      c.cname
	  | C_array arr -> typ_basename arr.Ctypes.arr_element^"_array"
      in
      match ty with
      | Ctype c -> "is_"^typ_basename c
      | Linteger | Lreal  -> ""
      | Lvar x -> x
      | Ltype (lt,_) -> lt.lt_name
      | Larrow _ ->""

    let location _ = None
    module H = Cil_datatype.Logic_type.Hashtbl
    let pp_title fmt _ =
      Format.fprintf fmt "Acsl type"
    let pp_descr fmt _ =
      Format.fprintf fmt "Declaration"
  end

  module Logicvar : Identifiable with type t = logic_var =
  struct
    type t = logic_var
    let prefix = "D"
    let index x = x
    let basename x = x.lv_name
    let location _ = None
    module H = Cil_datatype.Logic_var.Hashtbl
    let pp_title fmt x = Format.fprintf fmt "Logic variable '%s'" x.lv_name
    let pp_descr fmt _x = Format.fprintf fmt "Declaration"
  end

  (* ----------------------------------------------------------------------- *)
  (* --- User Type Registry                                              --- *)
  (* ----------------------------------------------------------------------- *)

  module LTinfo =
  struct
    type t = logic_type_info
    let compare t1 t2 = String.compare t1.lt_name t2.lt_name
    let hash v = Hashtbl.hash v.lt_name
    let equal t1 t2 = t1.lt_name = t2.lt_name
  end

  module LTinfoId : Identifiable with type t = logic_type_info  =
  struct
    type t = logic_type_info
    let prefix = "T"
    let index t = t
    let basename c = c.lt_name
    let location _c = None
    module H = Hashtbl.Make(LTinfo)
    let pp_title fmt x =
      Format.fprintf fmt "Logic type '%s'" x.lt_name
    let pp_descr fmt _x =
      Format.fprintf fmt "Declaration"
  end


  module ADTDecl = Register
    (struct
       include LTinfoId
       let declare t _ =
	 Formula.Type (List.length t.lt_params)
       let section = S_typedef
       let clear () = ()
     end)

  (* -------------------------------------------------------------------------- *)
  (* --- Tau of Logic Type                                                  --- *)
  (* -------------------------------------------------------------------------- *)

  let rec tau_of_object = function
    | C_int _ -> Formula.Integer
    | C_float _ -> Formula.Real
    | C_pointer _ -> Formula.Pointer
    | C_comp c -> Formula.Record c
    | C_array t -> Formula.Array t

  let tau_of_ctype t = tau_of_object (Ctypes.object_of t)

  let rec tau_of_logic_type = function
    | Ctype c -> Tau.tau_of_logic_ctype c
    | Linteger -> Integer
    | Lreal -> Real
    | Ltype( d , [] ) when d.lt_name = Utf8_logic.boolean -> Boolean
    | Ltype( {lt_name="set"} , [t] ) -> Set (tau_of_logic_type t)
    | Ltype( lt , ts) ->
	let d = (ADTDecl.get_definition lt).d_name in
	ADT (d,List.map tau_of_logic_type ts)
    | Lvar _ -> Wp_parameters.not_yet_implemented "logic type variables"
    | Larrow _ -> Wp_parameters.not_yet_implemented "type of logic function"

  (* -------------------------------------------------------------------------- *)
  (* --- Built-in Names                                                     --- *)
  (* -------------------------------------------------------------------------- *)

  module Findex = Dindex(Fieldinfo)
  module Xindex = Dindex(Varinfo)
  module Tindex = Dindex(Compinfo)
  module Aindex = Dindex(Varaddr)
  module LTindex = Dindex(LTypeinfo)

  let pp_kind fmt = function
     | Kcint _ -> Format.fprintf fmt "Kcint"
     | Kint -> Format.fprintf fmt "Kint"
     | Kreal-> Format.fprintf fmt "Kreal"
     | Kptr _-> Format.fprintf fmt "Kptr"
     | Kset _-> Format.fprintf fmt "Kset"
     | Kstruct _-> Format.fprintf fmt "Kstruct"
     | Karray _-> Format.fprintf fmt "Karray"
     | Kadt (s,_) -> Format.fprintf fmt "Kadt %s" s
     | Kbool ->  Format.fprintf fmt "Kbool"

  (* ------------------------------------------------------------------------ *)
  (* --- Set Interface                                                    --- *)
  (* ------------------------------------------------------------------------ *)

  let empty = e_app0 mk_empty
  let singleton e = e_app1 mk_singleton e

  let union a b = e_app2 mk_union a b

  let unions l =
    let rec unions_aux l s =
      match l with
	| [] -> s
	| a::m -> unions_aux m (union a s)
    in
    match l with
      | [] -> empty
      | a::m -> unions_aux m a

  let inter a b = e_app2 mk_inter a b
  let remove a b = e_app2 mk_remove a b
  let add_set s1 s2 = e_app2 mk_radd s1 s2

  let mult_set s1 s2 = e_app2 mk_rmult s1 s2
  let neg_set s1 = e_app1 mk_rneg s1

  (* ------------------------------------------------------------------------ *)
  (* --- Integer Range Inteface                                           --- *)
  (* ------------------------------------------------------------------------ *)

  let loc_range = e_app2 mk_range
  let range_inf = e_app1 mk_range_inf
  let range_sup = e_app1 mk_range_sup
  let integers  = e_app0 mk_integers

  let rec set_of_list = function
    | [] -> empty
    | a::m -> union (singleton a) (set_of_list m)

  let interval r =
    match r.inf,r.sup with
      | None,None -> integers
      | Some i, None -> range_inf i
      | None, Some j -> range_sup j
      | Some i, Some j -> loc_range i j

  let get_range_index t ri = e_app2 get_range_index t (interval ri)
  let set_range_index t ri = e_app2 set_range_index t (interval ri)

  (* ------------------------------------------------------------------------ *)
  (* --- Pointer Arithmetics                                              --- *)
  (* ------------------------------------------------------------------------ *)

  let lt_pointer = p_app2 mk_lt_pointer
  let le_pointer = p_app2 mk_le_pointer

  let minus_pointer = e_app2 mk_minus_pointer

  let equal_pointer_bool = e_app2 mk_equal_pointer_bool
  let lt_pointer_bool = e_app2 mk_lt_pointer_bool
  let le_pointer_bool = e_app2 mk_le_pointer_bool

  let null = e_app0 mk_null
  let is_null = e_app1 mk_is_null

  (* ------------------------------------------------------------------------ *)
  (* --- Formats                                                          --- *)
  (* ------------------------------------------------------------------------ *)

  let int_format :     format = e_app0 mk_int_format
  let real_format :    format = e_app0 mk_real_format
  let pointer_format : format = e_app0 mk_pointer_format
  let record_format :  format = e_app0 mk_record_format
  let urecord_format : format = e_app0 mk_urecord_format
  let array_format (typ_elt: format) : format
      = e_app1 mk_array_format typ_elt

  (* ------------------------------------------------------------------------ *)
  (* --- Acls Types Guards Generations                                    --- *)
  (* ------------------------------------------------------------------------ *)

  let access = e_app2 farray_access

  let has_type_int i e = guard i e

  let rec has_type e = function
    | Ctype ty  ->
	begin
	  match object_of ty with
	    | C_int i -> has_type_int i e
	    | C_array _ -> F.p_true
	    | C_comp _ -> F.p_true
	    | C_float _| C_pointer _ -> F.p_true
	end
    | (Ltype _ | Linteger | Lreal | Lvar _ | Larrow _ ) -> F.p_true

  (* ------------------------------------------------------------------------ *)
  (* --- Constrained Terms                                                --- *)
  (* ------------------------------------------------------------------------ *)

  class type hint =
  object
    method has_var : F.var list -> bool
    method capture : F.var list -> unit
    method subst : F.var -> Formula.abstract F.term -> unit
    method pretty : Format.formatter -> unit
  end

  type bindings = binding list and binding =
    | Forall of F.var list
    | Exists of F.var list
    | Any of F.var * F.pred
    | Let of F.var * Formula.abstract F.term

  let pp_binding fmt = function
    | Forall xs -> Format.fprintf fmt "Forall @[%a@]" (Pretty_utils.pp_list F.pp_var) xs
    | Exists xs -> Format.fprintf fmt "Exists @[%a@]" (Pretty_utils.pp_list F.pp_var) xs
    | Any(x,p) -> Format.fprintf fmt "Any %a s.t. @[%a@]" F.pp_var x F.pp_pred p
    | Let(x,t) -> Format.fprintf fmt "Let %a = @[%a@]" F.pp_var x F.pp_term t

  type context = {
    pool : F.pool ;
    hints : hint list ;
    mutable bindings : bindings ;
  }

  let closed : bindings = []
  let context = ref []

  let occur_check y xs =
    if List.exists (F.eq_var y) xs then
      Wp_parameters.fatal
	"Quantification of constrained variable"

  let capture hints xps =
    if hints <> [] && xps <> [] then
      List.iter
	(function
	   | Forall xs | Exists xs ->
	       List.iter (fun h -> h#capture xs) hints
	   | Let(x,t) -> List.iter (fun h -> h#subst x t) hints
	   | Any(y,_) -> List.iter (fun h -> h#capture [y]) hints
	) xps

  let pp_vkind fmt = function
    | Mdata.Vmodel t -> Tau.pp_tau fmt t
    | Mdata.Vacsl ty -> !Ast_printer.d_logic_type fmt ty

  let guards_with f xs p =
    let do_with f p x =
      match F.kind_of_var x with
	| Model _ -> p
	| Acsl (_,Ctype ty) ->
	  begin
	    match object_of ty with
	    | C_int i -> f (guard i (F.var x)) p
	    | C_float _ -> p
	    | C_pointer _ -> p
	    | _ -> p
	  end
	| Acsl _ -> p
    in
    List.fold_left (do_with f) xs p

  let sub y xs = List.filter (fun x -> not (F.eq_var x y)) xs

  let some_alpha pool x = Some(F.freshen pool x)

  let rec apply_bindings alpha bindings p =
    match bindings with
      | [] -> p
      | b::bindings ->
	  let p = apply_bindings alpha bindings p in
	  match b with
	    | Forall xs -> F.p_forall xs p
	    | Exists xs -> F.p_exists xs p
	    | Any(x,q) -> F.p_forall [x] (F.p_implies q p)
	    | Let(x,t) -> F.p_subst alpha x t p

  let alpha x =
    match !context with
      | [] -> None
      | c::_ -> Some(F.freshen c.pool x)

  let do_subst x v p =
    if F.pred_has_var [x] p 
    then F.p_subst alpha x v p
    else p

  let do_forall xs p =
    if F.pred_has_var xs p
    then
      let ys = List.filter (fun y -> F.pred_has_var [y] p) xs in
      if ys = [] then p
      else
	F.p_forall ys (guards_with F.p_implies p ys)
    else p

  let do_exists xs p =
    if F.pred_has_var xs p
    then
      let ys = List.filter (fun y -> F.pred_has_var [y] p) xs in
      if ys = [] then p
      else F.p_exists ys (guards_with F.p_and p ys)
    else p

  let rec has_var xs bindings hints =
    xs <> [] &&
      match bindings with
	| [] ->
	    List.exists (fun h -> h#has_var xs) hints
	| b::others ->
	    match b with
	      | Forall ys | Exists ys ->
		  let xs = List.fold_right sub ys xs in
		  has_var xs others hints
	      | Any(y,p) ->
		  F.pred_has_var xs p || has_var (sub y xs) others hints
	      | Let(y,t) ->
		  F.term_has_var xs t || has_var (sub y xs) others hints

  let forall xs p =
    match !context with
      | [] -> do_forall xs p
      | c :: _ ->
	  if has_var xs c.bindings c.hints
	  then ( c.bindings <- Forall xs :: c.bindings ; p )
	  else do_forall xs p

  let exists xs p =
    match !context with
      | [] -> do_exists xs p
      | c :: _ ->
	  if has_var xs c.bindings c.hints
	  then ( c.bindings <- Exists xs :: c.bindings ; p )
	  else do_exists xs p

  let fresh id vk =
    match !context with
	| [] ->
	    Wp_parameters.fatal "Bad context (for %S:%a)" id pp_vkind vk
	| c::_ ->
	    F.fresh c.pool id
	      (match vk with
		 | Mdata.Vmodel t -> Model t
		 | Mdata.Vacsl t -> Acsl (tau_of_logic_type t,t)
	      )

  let pool () =
    match !context with
      | [] -> Wp_parameters.fatal "Bad context (no pool available)"
      | c::_ -> c.pool

  let vkind_of_var x =
    match F.kind_of_var x with
      | Model t -> Mdata.Vmodel t
      | Acsl(_,c) -> Mdata.Vacsl c

  let term_such_that tau phi =
    match !context with
      | [] -> Wp_parameters.fatal "No context opened for constraints"
      | c::_ ->
	  let x = F.fresh c.pool "X" (Model tau) in
	  let t = F.var x in
	  c.bindings <- Any(x,phi t) :: c.bindings ; t

  let subst_in_bindings c x v p =
    if has_var [x] c.bindings c.hints
    then ( c.bindings <- Let(x,v) :: c.bindings ; p )
    else ( do_subst x v p )

  let subst x v p =
    match !context with
      | [] -> do_subst x v p
      | c::_ -> subst_in_bindings c x (F.wrap v) p

  let close bindings p =
    apply_bindings (fun _ -> None) bindings p

  let has_vars xs p =
    F.pred_has_var xs p ||
      ( match !context with
	  | [] -> false
	  | c::_ -> has_var xs c.bindings c.hints )

  (* -------------------------------------------------------------------------- *)
  (* --- Context Management                                                 --- *)
  (* -------------------------------------------------------------------------- *)

  let push where pool bindings hints =
    Wp_parameters.debug ~level:3 "PUSH %d: %S@." (List.length !context) where ;
    let c = { pool=pool ; bindings=bindings ; hints=hints } in
    context := c :: !context ; c

  let pop where c0 =
    Wp_parameters.debug ~level:3 "POPK %d: %S@." (pred (List.length !context)) where ;
    match !context with
      | [] -> Wp_parameters.fatal "No context for constrained term"
      | c::stack ->
	  if not (c0 == c)
	  then Wp_parameters.fatal "Context mismatch for constrained term" ;
	  context := stack ;
	  c.bindings

  let flush where c0 p =
    Wp_parameters.debug ~level:3 "FLUSH %d %S@." (pred (List.length !context)) where ;
    match !context with
      | [] -> Wp_parameters.fatal "No context for constrained term"
      | c::stack ->
	  if not (c0 == c)
	  then Wp_parameters.fatal "Context mismatch for constrained term" ;
	  context := stack ;
	  apply_bindings (some_alpha c.pool) c.bindings p

  let kill where c = ignore (pop where c)

  (* ------------------------------------------------------------------------ *)
  (* --- Generalized Substitutions                                        --- *)
  (* ------------------------------------------------------------------------ *)

  type substitution = (F.var * F.var) list

  let pp_sigma fmt s =
    begin
      Format.fprintf fmt "[" ;
      List.iter (fun (x,x') -> Format.fprintf fmt " %a:=%a" F.pp_var x F.pp_var x') s ;
      Format.fprintf fmt " ]" ;
    end

  let pp_bindings fmt xts =
    List.iter (fun (x,t) -> Format.fprintf fmt "%a:= %a@\n" F.pp_var x F.pp_term t) xts

  type havoc =
    | Fresh of F.var
    | Update of F.var * (substitution -> Formula.abstract F.term)

  let freshen x =
    match !context with
      | [] -> Wp_parameters.fatal "no context opened for havoc"
      | c::_ -> F.freshen c.pool x

  let apply = F.e_rename

  let rec domain xs d = function
    | [] -> List.rev xs,List.rev d
    | h::hs ->
	match h with
	  | Fresh x -> domain (x::xs) d hs
	  | Update(x,_) ->
	      let d' = if List.exists (F.eq_var x) d then d else (x::d) in
	      domain xs d' hs

  let reverse = List.map (fun (x,y) -> (y,x))

  let freshen_sigma y =
    List.map (fun u -> if F.eq_var (fst u) y then y,freshen y else u)

  let rec compute_bindings bindings sigmaR sigma = function
    | [] -> bindings , sigma
    | Fresh _ :: havocs -> compute_bindings bindings sigmaR sigma havocs
    | Update(x,phi) :: havocs ->
	let sigma' = freshen_sigma x sigma in
	let x' = snd (List.find (fun (y,_) -> F.eq_var x y) sigma') in
	let t' = apply sigmaR (phi sigma) in
	compute_bindings ((x',t')::bindings) sigmaR sigma' havocs


  let rec apply_bindings xts p =
    match xts with
      | [] -> p
      | (x,t)::xts -> apply_bindings xts (subst x t p)

  let rec rename s p =
    match s with
      | [] -> p
      | (x,x')::s -> subst x (F.var x') (rename s p)

  let rec fixpoint sf sn p =
    match sf,sn with
      | (_,x')::sf , (_,y)::sn ->
	F.p_implies
	  (F.p_eq (F.var x') (F.var y))
	  (fixpoint sf sn p)
      | [] , [] -> p
      | _ -> Wp_parameters.fatal "inconsistent domains in fixpoint substitutions"

  let havoc_inductive hs p =
    let xs,d = domain [] [] hs in
    let sigma_0 = List.map (fun x -> x,freshen x) d in
    let sigma_F = List.map (fun x -> x,freshen x) d in
    let bindings,sigma_N = compute_bindings [] sigma_F sigma_0 hs in
     forall ( xs @ List.map snd sigma_F )
      (rename (reverse sigma_0)
	 (apply_bindings bindings
	    (fixpoint sigma_F sigma_N
	       (rename sigma_F p))))

  let havoc_static hs p =
    let xs,d = domain [] [] hs in
    let sigma_0 = List.map (fun x -> x,freshen x) d in
    let bindings,sigma_N = compute_bindings [] sigma_0 sigma_0 hs in
    forall xs
      (rename (reverse sigma_0)
	 (apply_bindings bindings
	    (rename sigma_N p)))

  (* -------------------------------------------------------------------------- *)
  (* --- Decode and Encode of Ctypes                                        --- *)
  (* -------------------------------------------------------------------------- *)

  let tau_of_object = function
    | C_int _ -> Integer
    | C_float _ -> Real
    | C_pointer _ -> Pointer
    | C_comp c -> Record c
    | C_array a -> Array a

  (* -------------------------------------------------------------------------- *)
  (* --- Formats                                                            --- *)
  (* -------------------------------------------------------------------------- *)

  let rec format_of_object = function
    | C_int _ -> int_format
    | C_float _ -> real_format
    | C_pointer _ -> pointer_format
    | C_comp c ->
        if c.cstruct
        then record_format
        else urecord_format
    | C_array arr ->
        array_format
          (format_of_object (Ctypes.object_of arr.arr_element))

  (* ------------------------------------------------------------------------ *)
  (* --- Records, Unions and Arrays as First Class Values                 --- *)
  (* ------------------------------------------------------------------------ *)

  let acc_field s f =
    let fmt = format_of_object (Ctypes.object_of f.ftype) in
    let fid = Findex.get_ind f in
    let acc = if f.fcomp.cstruct then farray_access else mk_access_ufield in
    e_app2 "decode" fmt (e_app2 acc s fid)

  let upd_field s f v  =
    let fmt = format_of_object (Ctypes.object_of f.ftype) in
    let fid = Findex.get_ind f in
    let upd = if f.fcomp.cstruct then farray_update else mk_update_ufield in
    e_app3 upd s fid (e_app2 "encode" fmt v)

  let acc_index = e_app2 farray_access
  let upd_index = e_app3 farray_update

  let encode fmt v = e_app2 "encode" fmt v
  let decode fmt x = e_app2 "decode" fmt x

  (* ------------------------------------------------------------------------ *)
  (* --- Comparison of Record and Arrays                                  --- *)
  (* ------------------------------------------------------------------------ *)

  let equal_rec = ref (fun _ _ _ -> assert false)

  let eq_record_definition tcomp eq_name =
    let pool = F.pool() in
    let xa = F.fresh pool "a" (Model (Record tcomp)) in
    let xb = F.fresh pool "b" (Model (Record tcomp)) in
    let ra = F.var xa in
    let rb = F.var xb in
    let p_comp =
      List.fold_left
	(fun p field ->
	   let tf = Ctypes.object_of field.ftype in
	   let va = acc_field ra field in
	   let vb = acc_field rb field in
	   let ef = !equal_rec tf va vb in
	   if tcomp.cstruct
	   then F.p_and p ef
	   else F.p_or p ef
	) (if tcomp.cstruct
	   then F.p_true
	   else F.p_false) tcomp.cfields
    in
    F.p_forall [xa;xb]
      (F.p_iff (p_app2 eq_name ra rb) p_comp)

  let eq_array_definition ta eq_name =
    let pool = F.pool() in
    let te = Ctypes.object_of ta.arr_element in
    let vmodel = Model(Array ta) in
    let xa = F.fresh pool "a" vmodel in
    let xb = F.fresh pool "b" vmodel in
    let ra = F.var xa in
    let rb = F.var xb in
    let xi = F.fresh pool "i" (Model Integer) in
    let vi = F.var xi in
    let i_pos = F.p_icmp Cleq (F.e_int 0) vi in
    let i_max =
      match ta.arr_flat with
	| None -> F.p_true
	| Some f -> F.p_icmp Clt vi (F.e_icst (Int64.to_string f.arr_size))
    in
    let i_range = F.p_and i_pos i_max in
    let a_comp =
      F.p_forall [xi]
	(F.p_implies
	   i_range
	   (!equal_rec te (acc_index ra vi) (acc_index rb vi))) in
    F.p_forall [xa;xb]
      (F.p_iff (p_app2 eq_name ra rb) a_comp)

  module RecEqName = Register
    (struct
       include Compinfo
       let declare tcomp _eqname =
	 Formula.Predicate [Record tcomp;Record tcomp]
       let prefix = "Eqrec"
       let section = S_typedef
       let clear () = ()
       let pp_title fmt x =
         Format.fprintf fmt "Equality for %a"
	   Compinfo.pp_title x
     end)

  module RecEqDef = Register
    (struct
       include Compinfo
       let declare tcomp _axname =
	 let eq_name = RecEqName.get_definition tcomp in
         Formula.Axiom(eq_record_definition tcomp eq_name.d_name)
       let prefix = "EqrecDef"
       let section = S_typedef
       let clear () = ()
       let pp_title fmt x =
         Format.fprintf fmt "Equality for %a"
	   Compinfo.pp_title x
       let pp_descr fmt _ =
	 Format.pp_print_string fmt "Axiomatic Definition"
     end)

  module ArrEqName = Register
    (struct
       include Arrayinfo
       let declare arr _eqname =
	 Formula.Predicate [Array arr;Array arr]
       let prefix = "Eqarr"
       let section = S_typedef
       let clear () = ()
       let pp_title fmt x =
         Format.fprintf fmt "Equality for %a" Arrayinfo.pp_title x
     end)

  module ArrEqDef = Register
    (struct
       include Arrayinfo
       let declare arr _axname =
	 let eq_name = ArrEqName.get_definition arr in
	 Formula.Axiom(eq_array_definition arr eq_name.d_name)
       let prefix = "EqarrDef"
       let section = S_typedef
       let clear () = ()
       let pp_title fmt x =
         Format.fprintf fmt "Equality for %a"
	   Arrayinfo.pp_title x
       let pp_descr fmt _ =
	 Format.pp_print_string fmt "Axiomatic Definition"
     end)

  let eq_record comp (ta:record) (tb:record) =
    if comp.cfields <>[] then RecEqDef.define comp ;
    let eq = RecEqName.get_definition comp in
    p_app2 eq.d_name ta tb

  let eq_array arr (ta:array) (tb:array) =
    ArrEqDef.define arr ;
    let eq = ArrEqName.get_definition arr in
    p_app2 eq.d_name ta tb

  let equal te a b =
    match te with
      | C_int _ | C_float _ | C_pointer _ -> F.p_eq a b
      | C_comp comp -> eq_record comp (F.unwrap a) (F.unwrap b)
      | C_array arr -> eq_array arr (F.unwrap a) (F.unwrap b)

  let () = equal_rec := equal

end

(* -------------------------------------------------------------------------- *)
(* --- Interpret C-runtime values                                         --- *)
(* -------------------------------------------------------------------------- *)

module Cvalues(M : Mvalues.Model) =
struct

  module A = M.A
  module R = M.R
  module D = M.D
  module F = M.D.F
  type loc = M.loc
  type m_cell
  type cell = D.abstract

  let loc_of_pointer = M.loc_of_pointer
  let pointer_of_loc = M.pointer_of_loc

  let cast_int_to_loc = M.cast_int_to_loc
  let cast_loc_to_int = M.cast_loc_to_int

  type value =
    | V_int of Ctypes.c_int * D.integer
    | V_float of Ctypes.c_float * D.real
    | V_pointer of Cil_types.typ * M.loc
    | V_record of compinfo * D.record
    | V_union of compinfo * D.urecord
    | V_array of arrayinfo * D.array

  let rec logic_of_value = function
    | V_int(_,t) -> F.wrap t
    | V_float(_,t) -> F.wrap t
    | V_pointer(_,loc) -> F.wrap (M.pointer_of_loc loc)
    | V_record(_,t) -> F.wrap t
    | V_union(_,t) -> F.wrap t
    | V_array(_,t) -> F.wrap t


  let rec value_of_logic c_obj t =
    match c_obj with
      | C_int i -> V_int(i,F.unwrap t)
      | C_float f -> V_float(f,F.unwrap t)
      | C_pointer typ ->
	  V_pointer(typ,M.loc_of_pointer (object_of typ) (F.unwrap t))
      | C_comp cinfo ->
	  if cinfo.cstruct
	  then V_record(cinfo,F.unwrap t)
	  else V_union(cinfo,M.D.F.unwrap t)
      | C_array ainfo -> V_array(ainfo,F.unwrap t)

  let pp_loc = M.pp_loc
  let pp_value fmt = function
    | V_pointer(_,loc) -> M.pp_loc fmt loc
    | v -> F.pp_term fmt (logic_of_value v)

end

(*
Local Variables:
compile-command: "make -C ../.."
End:
*)
