(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

open Cil_datatype

(* ------------------------------------------------------------------------ *)
(** {2 Constants} *)

type constant =
  | ConstInt of string
  | ConstBool of bool
  | ConstUnit
  | ConstFloat of string

let c_bool b = ConstBool b
let c_int_of_str s = ConstInt s
let c_int i = ConstInt (string_of_int i)
let c_float_of_str s = ConstFloat s
let c_float f = ConstFloat (string_of_float f)

(* ------------------------------------------------------------------------ *)
(** {2 Variables} *)

module Var = struct

  type t =
    | VarL of Cil_types.logic_var * Formula.tau
    | VarT of string * int option * Formula.tau * Cil_types.logic_type option

  let mk_lvar_variable lv t = VarL (lv, t)

  let lvar_of_variable v = match v with VarL (lv, _) -> Some lv | _ -> None

  let var_type v = match v with 
    | VarL (_, t) | VarT (_, _, t,_) -> t 

  let var_name = function
    | VarT (v, None, _,_) -> v
    | VarT (v, Some id, _,_) -> Printf.sprintf "%s_%d" v id
    | VarL (v, _) ->
	let name = v.Cil_types.lv_name in
	match v.Cil_types.lv_origin with
	  | Some _ -> name
          | _ -> Printf.sprintf "%s__%d" name v.Cil_types.lv_id

  let basename = function
    | VarT(x,_,_,_) -> x
    | VarL(v,_) ->
	match v.Cil_types.lv_origin with
	  | Some x -> x.Cil_types.vorig_name
	  | None -> v.Cil_types.lv_name

  let kind_of_var = function
    | VarL (lv,t) -> Formula.Acsl(t,lv.Cil_types.lv_type)
    | VarT (_,_,t,Some ty) -> Formula.Acsl(t,ty)
    | VarT (_,_,t,None)->  Formula.Model t

  let var_counter = ref 0
  let ident_named_var name k t = VarT(name,Some k,t,None)
  let fresh_named_var name t =
    incr var_counter;
    VarT (name, Some !var_counter, t,None)

  let fresh_typed_var name t tau =
    let lv = Cil_const.make_logic_var name t in
    VarL (lv, tau)

  let fresh_var v = match v with
    | VarT (name, _, t,_) ->
	fresh_named_var name t
    | VarL (lv, t) ->
        fresh_typed_var lv.Cil_types.lv_name lv.Cil_types.lv_type t

  let equal v1 v2 = match v1, v2 with
    | VarT(x,c,_,_) , VarT(y,d,_,_) -> c=d && x=y
    | VarL (v1, _), VarL (v2, _) -> Logic_var.equal v1 v2
    | _ -> false

  let compare v1 v2 = match v1, v2 with
    | VarT _, VarL _ -> 1
    | VarL _, VarT _ -> -1
    | VarT (v1, id1, _,_), VarT (v2, id2, _,_) ->
      let cmp = String.compare v1 v2 in
      if cmp <> 0 then cmp
      else
	begin
	  match id1 , id2 with
	    | None , None -> 0
	    | Some _ , None -> 1
	    | None , Some _ -> (-1)
	    | Some id1 , Some id2 -> Pervasives.compare id1 id2
	end
    | VarL (v1, _), VarL (v2, _) -> Logic_var.compare v1 v2

end

module Vset = Set.Make(Var)
module Vmap = Map.Make(Var)


(* ------------------------------------------------------------------------ *)
(** {2 Terms} *)
(* ------------------------------------------------------------------------ *)

type 'd d_term =
  | Tconst of constant
  | Tvar of Var.t
  | Tdata of 'd
  | Tapp of string * 'd d_term list
  | Tif of 'd d_term * 'd d_term * 'd d_term
  | Tlet of Var.t * 'd d_term * 'd d_term

(* -----------------------------------*)
(** {3 Term comparison} *)

(** compare terms without knowing how to compare data : better than something
* for smart constructors (see [eq_terms]) *)
let rec bad_eq_terms a b =
    match a , b with
      | Tconst k1 , Tconst k2 -> k1 = k2 (* strings,int and bools *)
      | Tvar x , Tvar y -> Var.equal x y
      | Tapp(f,xs) , Tapp(g,ys) when f=g ->
	  (try List.for_all2 bad_eq_terms xs ys
	   with _ -> false)
      | Tif(x,y,z) , Tif(x',y',z') ->
	  bad_eq_terms x x' && bad_eq_terms y y' && bad_eq_terms z z'
      | _ -> a==b (* at least ! *)

let eq_terms eq_data e1 e2 =
  let rec eq e1 e2 = match e1, e2 with
    | Tconst c1, Tconst c2 -> c1 = c2
    | Tconst _, _ -> false
    | Tvar v1, Tvar v2 -> Var.equal v1 v2
    | Tvar _, _ -> false
    | Tdata d1, Tdata d2 -> eq_data d1 d2
    | Tdata _, _ -> false
    | Tapp (f1, args1), Tapp (f2, args2) ->
        f1 = f2 && List.for_all2 eq args1 args2
    | Tapp _, _ -> false
    | Tif (c1, t1, e1), Tif (c2, t2, e2) ->
        (eq c1 c2) && (eq t1 t2) && (eq e1 e2)
    | Tif _, _ -> false
    | Tlet (x,v,t),Tlet(x',v',t') ->
	(Var.equal x x') && (eq v v') && (eq t t')
    | Tlet _, _ -> false
  in eq e1 e2

(* -----------------------------------*)
(** {3 Term visitors} *)

let rec e_has_var d_has_var xs e =
  let frec = e_has_var d_has_var xs in
  match e with
  | Tconst _ -> false
  | Tvar y -> List.exists (Var.equal y) xs
  | Tdata a -> d_has_var xs a
  | Tapp(_,ts) -> List.exists (frec) ts
  | Tif(a,b,c) -> frec a || frec b || frec c
  | Tlet(x,v,t) ->
      frec v ||
	(let xs = List.filter (fun y -> not (Var.equal x y)) xs in
	 xs <> [] && e_has_var d_has_var xs t)

let bad_term_has_var t = e_has_var (fun _ _ -> true) t

let rec e_closed f xs = function
  | Tconst _ -> true
  | Tvar y -> List.exists (Var.equal y) xs
  | Tdata d -> f xs d
  | Tapp(_,ts) -> List.for_all (e_closed f xs) ts
  | Tif(a,b,c) -> e_closed f xs a && e_closed f xs b && e_closed f xs c
  | Tlet(x,v,t) -> e_closed f xs v && e_closed f (x::xs) t
  
let bad_term_closed xs e = e_closed (fun _ _ -> false) xs e

let rec fold_data_in_exp do_var do_data acc exp =
  let frec = fold_data_in_exp do_var do_data in
  match exp with
  | Tconst _ -> acc
  | Tvar v -> do_var acc v
  | Tdata d -> do_data acc d
  | Tapp (_n,tl) ->  List.fold_left frec acc tl
  | Tif (t1,t2,t3) -> frec (frec (frec acc t1) t2) t3
  | Tlet (_x,v,t) -> frec (frec acc v) t
	  
(* -----------------------------------*)
(** {3 Term smart constructors} *)

let e_true = Tconst (c_bool true)
let e_false = Tconst (c_bool false)
let e_int i = Tconst (c_int i)
let e_int64 i = Tconst (ConstInt (Int64.to_string i))
let e_float f = Tconst (c_float f)
let e_cnst c = Tconst c
let e_var v = Tvar v
let e_data d = Tdata d

let e_if c t f = match c with
  | Tconst (ConstBool true) -> t
  | Tconst (ConstBool false) -> f
  | _ -> if bad_eq_terms t f then t else Tif (c, t, f)

let i_compute zop a b =
  let z_a = Big_int.big_int_of_string a in
  let z_b = Big_int.big_int_of_string b in
  Tconst(ConstInt(Big_int.string_of_big_int (zop z_a z_b)))

let simpl() = Wp_parameters.Simpl.get()

let e_app f args = 
  if not (simpl()) then Tapp(f,args)
  else
    match f , args with
      | "add_int" , [ Tconst (ConstInt a) ; Tconst (ConstInt b) ] ->
	  i_compute Big_int.add_big_int a b
      | "sub_int" , [ Tconst (ConstInt a) ; Tconst (ConstInt b) ] ->
	  i_compute Big_int.sub_big_int a b
      | "mul_int" , [ Tconst (ConstInt a) ; Tconst (ConstInt b) ] ->
	  i_compute Big_int.mult_big_int a b
      | "add_int" , [ Tconst (ConstInt "0") ; x ] -> x
      | "add_int" , [ x ; Tconst (ConstInt "0") ] -> x
      | "sub_int" , [ x ; Tconst (ConstInt "0") ] -> x
      | "mul_int" , [ Tconst (ConstInt "1") ; x ] -> x
      | "mul_int" , [ x ; Tconst (ConstInt "1") ] -> x
      | "mul_int" , [ (Tconst (ConstInt "0")) as z ; _ ] -> z
      | "mul_int" , [ _ ; (Tconst (ConstInt "0")) as z ] -> z
      | "add_int" , [ b ; Tapp("sub_int",[a;c]) ] when bad_eq_terms b c -> a ;
      | "add_int" , [ Tapp("sub_int",[a;b]) ; c ] when bad_eq_terms b c -> a
      | "sub_int" , [ Tapp("add_int",[a;b]) ; c ] when bad_eq_terms b c -> a
      | "encode" , [ Tapp("decode" , [ a ; fmt ]) ; fmt' ] when bad_eq_terms fmt fmt' -> a
      | "decode" , [ Tapp("encode" , [ a ; fmt ]) ; fmt' ] when bad_eq_terms fmt fmt' -> a
      | "access" , [ Tapp("update" , [ _ ; k ; v ]) ; k' ] when bad_eq_terms k k' -> v
      | _ -> Tapp (f, args)

let e_let x exp t = if bad_term_has_var [x] t then Tlet (x, exp, t) else t

let case_of = function
  | Tconst(ConstInt s) -> Some s
  | _ -> None

(* -----------------------------------*)
(** {3 Term transformation} *)

(** Apply [do_var] and [do_data] in term subexpressions.
* Notice that [do_data] is in charge of propagating
* recursively in its own subexpressions if any, and that it returns a new
* expression instead of a new data because there can be data transformation. *)
let rec change_in_exp do_var do_data exp =
  let frec = change_in_exp do_var do_data in
  match exp with
  | Tconst c -> e_cnst c
  | Tvar v -> (match do_var v with Some e -> e | None -> e_var v)
  | Tdata d -> do_data d
  | Tapp (n,tl) -> e_app n (List.map (frec) tl)
  | Tif (t1,t2,t3) -> e_if (frec t1) (frec t2) (frec t3)
  | Tlet (x,v,t) -> e_let x (frec v) (frec t)

(** Similar to [change_in_exp] but only transforms data into data.
* [do_data_rec] is  supposed to process the data [d], to propagate recursively
* in the subdata, and to call [do_exp] in the subexpressions.
* *)
let rec change_data_in_exp do_data_rec exp =
  let do_exp = change_data_in_exp do_data_rec in
  let do_data d = do_data_rec do_exp d in
  match exp with
  | Tconst c -> e_cnst c
  | Tvar v -> e_var v
  | Tdata d -> e_data (do_data d)
  | Tapp (n,tl) -> e_app n (List.map (do_exp) tl)
  | Tif (t1,t2,t3) -> e_if (do_exp t1) (do_exp t2) (do_exp t3)
  | Tlet (x,v,t) -> e_let x (do_exp v) (do_exp t)

(** This one doesn't know how to substitute in [data],
* so it builds a let if needed.
* [alpha] is used
**)
let rec term_replace alpha x exp t =
  let frec = term_replace alpha x exp in
    match t with
      | Tconst _ -> t
      | Tvar x0 -> if Var.equal x0 x then exp else t
      | Tdata _ -> Tlet(x,exp,t)
      | Tapp (f, ts) -> e_app f (List.map (frec) ts)
      | Tif (a, b, c) -> e_if (frec a) (frec b) (frec c)
      | Tlet (x0, a, b) ->
          if bad_term_has_var [x0] exp then
            match alpha x with
              | None -> Tlet(x,exp,t)
              | Some y ->
                  let by = term_replace alpha x0 (Tvar y) b in
                  Tlet(y,frec a,frec by)
          else
            let b' = if Var.equal x0 x then b else frec b in
            Tlet(x0,frec a,b')

let alpha_bound_var alpha v = 
  let old = try Some (Vmap.find v alpha) with Not_found -> None in
  old, Vmap.add v v alpha

let alpha_unbound alpha v old = match old with
  | Some v' -> Vmap.add v v' alpha
  | None -> Vmap.remove v alpha

let apply_alpha alpha v = 
  try alpha, Vmap.find v alpha
  with Not_found -> 
      let v' = Var.fresh_var v in
      let alpha = Vmap.add v v' alpha in
        alpha, v'

let rec term_alpha_cv do_data alpha t =
  let do_term alpha t = term_alpha_cv do_data alpha t in
  let rec do_terms alpha l = terms_alpha_cv do_data alpha l in
    match t with
      | Tconst _ -> alpha, t
      | Tvar v -> 
          let alpha, v = apply_alpha alpha v in alpha, Tvar v
      | Tdata d -> let alpha, d = do_data alpha d in alpha, Tdata d
      | Tapp (f, ts) -> 
          let alpha, ts = do_terms alpha ts in
            alpha, Tapp(f, ts)
      | Tif (a, b, c) -> 
          let alpha, a = do_term alpha a in
          let alpha, b = do_term alpha b in
          let alpha, c = do_term alpha c in
          alpha, Tif (a, b, c)
      | Tlet (x, a, b) ->
          let alpha, a = do_term alpha a in (* a doesn't see x *)
          let old_x, alpha = alpha_bound_var alpha x in
          let alpha, b = do_term alpha b in
          let alpha = alpha_unbound alpha x old_x in (* restore old_x *)
            alpha, Tlet (x, a, b)
and terms_alpha_cv do_data alpha l = match l with [] -> alpha, []
    | t::l -> 
        let alpha, t = term_alpha_cv do_data alpha t in
        let alpha, l = terms_alpha_cv do_data alpha l in
          alpha, t::l

(* ------------------------------------------------------------------------ *)
(** {2 Predicates} *)
(* ------------------------------------------------------------------------ *)

type 't t_pred =
  | Papp of string * 't list
  | Ptrue
  | Pfalse
  | Pimplies of 't t_pred * 't t_pred
  | Pif of 't * 't t_pred * 't t_pred
  | Pand of 't t_pred * 't t_pred
  | Por of 't t_pred * 't t_pred
  | Piff of 't t_pred * 't t_pred
  | Pnot of 't t_pred
  | Pforall of Var.t * 't t_pred
  | Pexists of Var.t * 't t_pred
  | Plet of Var.t * 't * 't t_pred
  | Pnamed of string * 't t_pred

type 'd d_pred = ('d d_term) t_pred

(* -----------------------------------*)
(** {3 Predicates smart constructors} *)

let i_compare zop a b =
  if zop (Big_int.big_int_of_string a) (Big_int.big_int_of_string b)
  then Ptrue else Pfalse

let p_app name args = 
  if not (simpl()) then Papp(name,args) 
  else
    match name , args with
      | "eq", [ a ; b ] when bad_eq_terms a b -> Ptrue
      | "neq", [ a ; b ] when bad_eq_terms a b -> Pfalse
      | "le_int" , [a;b] when bad_eq_terms a b -> Ptrue
      | "lt_int" , [a;b] when bad_eq_terms a b -> Pfalse
      | "eq" , [ Tconst(ConstInt a) ; Tconst(ConstInt b) ] ->
	  i_compare Big_int.eq_big_int a b
      | "neq" , [ Tconst(ConstInt a) ; Tconst(ConstInt b) ] -> 
	  i_compare (fun za zb -> not (Big_int.eq_big_int za zb)) a b
      | "le_int" , [ Tconst(ConstInt a) ; Tconst(ConstInt b) ] -> 
	  i_compare Big_int.le_big_int a b
      | "lt_int" , [ Tconst(ConstInt a) ; Tconst(ConstInt b) ] -> 
	  i_compare Big_int.lt_big_int a b
      | _ -> Papp (name,args)


let rec val_of = function Pnamed (_,p) -> val_of p | p -> p 
let rec cut p = function Pnamed (a,q) -> Pnamed(a,cut p q) | _ -> p

let p_not p = match val_of p with
  | Ptrue -> cut Pfalse p
  | Pfalse -> cut Ptrue p
  | Papp( "eq" , w ) -> p_app "neq" w
  | Papp( "neq" , w ) -> p_app "eq" w
  | Papp( "le_int" , [a;b] ) -> p_app "lt_int" [b;a]
  | Papp( "lt_int" , [a;b] ) -> p_app "le_int" [b;a]
  | _ -> Pnot p

let p_and p1 p2 = match val_of p1, val_of p2 with
  | Ptrue, _ -> p2
  | _, Ptrue  -> p1
  | Pfalse,_-> cut Pfalse p1
  | _,Pfalse -> cut Pfalse p2
  | _ -> Pand (p1, p2)

let p_or p1 p2 = match val_of p1, val_of p2 with
  | Ptrue, _ -> cut Ptrue p1
  | _ , Ptrue -> cut Ptrue p2
  | Pfalse ,_ -> p2 
  | _ ,Pfalse -> p1
  | _ -> Por (p1,p2)

let p_xor p1 p2 = match val_of p1, val_of p2 with
  | Ptrue , Ptrue -> cut (cut Pfalse p2) p1
  | Ptrue ,_ -> cut Ptrue p1
  | _,Ptrue -> cut Ptrue p2
  | Pfalse , _ -> p2 
  | _ , Pfalse -> p1
  | _ -> Pnot(Piff(p1,p2))

let p_implies p1 p2 = 
  match val_of p1, val_of p2 with
    | Ptrue, _ -> p2
    | Pfalse, _ -> cut Ptrue p1
    | _, Ptrue -> cut Ptrue p2
    | _ -> Pimplies (p1, p2)

let rec p_conj = function
  | [] -> Ptrue
  | [p] -> p
  | p::ps -> p_and p (p_conj ps)

let rec p_disj = function
  | [] -> Pfalse
  | [p] -> p
  | p::ps -> p_or p (p_disj ps)

let p_if c p1 p2 = match c, val_of p1, val_of p2 with
  | (_,Ptrue, Ptrue ) -> cut (cut Ptrue p2) p1
  | (_,Pfalse, Pfalse ) -> cut (cut Pfalse p2) p1
  | (t,_ , _) -> Pif (t,p1,p2)

let p_iff p1 p2 = 
  match val_of p1,val_of p2 with
    | Ptrue ,_ -> p2
    | _ ,Ptrue -> p1
    | Pfalse, _ -> p2
    | _ , Pfalse -> p_not p1
    | _ -> Piff (p1,p2)

let p_eq e1 e2 =
  if bad_eq_terms e1 e2 then Ptrue else p_app "eq" [e1; e2]
let p_neq e1 e2 =
  if bad_eq_terms e1 e2 then Pfalse else p_app "neq" [e1; e2]

let p_forall x p =
  match val_of p with
    | Ptrue | Pfalse | Papp ("dummy",_) -> p
    | _ -> Pforall(x,p)

let p_exists x p =
  match val_of p with
    | Ptrue | Pfalse | Papp ("dummy",_) -> p
    | _ -> Pexists(x,p)

let p_let x v p =
  match val_of p with
    | Ptrue | Pfalse| Papp ("dummy",_) -> p
    | _ -> Plet(x,v,p)

let p_named name p = Pnamed(name,p)

(* -----------------------------------*)
(** {3 Predicates visitors} *)

let rec p_has_var f xs = function
  | Papp(_,ts) -> List.exists (f xs) ts
  | Ptrue | Pfalse -> false
  | Pimplies(p,q) | Pand(p,q) | Por(p,q) | Piff(p,q) ->
      p_has_var f xs p || p_has_var f xs q
  | Pif(t,p,q) ->
      f xs t || p_has_var f xs p || p_has_var f xs q
  | Pnot p | Pnamed(_,p) -> p_has_var f xs p
  | Pforall(x,p) | Pexists(x,p) ->
      let xs = List.filter (fun y -> not (Var.equal x y)) xs in
      xs <> [] && p_has_var f xs p
  | Plet(x,a,p) ->
      f xs a ||
	(let xs = List.filter (fun y -> not (Var.equal x y)) xs in
	 xs <> [] && p_has_var f xs p)

let bad_pred_has_var p = p_has_var bad_term_has_var p

let rec p_closed f xs = function
  | Papp(_,ts) -> List.for_all (f xs) ts
  | Ptrue | Pfalse -> false
  | Pimplies(p,q) | Pand(p,q) | Por(p,q) | Piff(p,q) ->
      p_closed f xs p && p_closed f xs q
  | Pif(t,p,q) ->
      f xs t && p_closed f xs p && p_closed f xs q
  | Pnot p | Pnamed(_,p) -> p_closed f xs p
  | Pforall(x,p) | Pexists(x,p) -> p_closed f (x::xs) p
  | Plet(x,a,p) -> f xs a && p_closed f (x::xs) p

let bad_pred_closed xs p = p_closed bad_term_closed xs p

(** TODO: this function is probably not exactly the good one
* since bounded variables might escape. *)
let rec fold_exp_in_pred do_exp acc p =
  let frec = fold_exp_in_pred do_exp in
  match p with
  | Ptrue | Pfalse  -> acc
  | Pif (t,p1,p2) ->
      let acc = do_exp acc t in
      let acc = frec acc p1 in
      let acc = frec acc p2 in
        acc
  | Pnot p -> frec acc p
  | Pforall (_v,p') | Pexists (_v,p') ->
        fold_exp_in_pred do_exp acc p'
  | Plet (_x,e,p) ->
      let acc = do_exp acc e in
      fold_exp_in_pred do_exp acc p
  | Pnamed (_n,p) -> frec acc p
  | Pimplies (p1,p2) | Pand (p1,p2) | Por (p1,p2) | Piff (p1,p2)
    -> frec (frec acc p1) p2
  | Papp (_n,t) -> List.fold_left do_exp acc t

let fold_data_in_pred do_var do_data_rec acc p =
  let rec do_exp acc e = fold_data_in_exp do_var do_data acc e
  and do_data acc d = do_data_rec do_exp acc d
  in fold_exp_in_pred do_exp acc p

(* -----------------------------------*)
(** {3 Predicates transformation} *)

(* ------------------------------------------------------------------------ *)
(* ---  Propagation of transformations                                  --- *)
(* ------------------------------------------------------------------------ *)


(** apply [do_exp] on each sub expression of the predicate.
* [quantif_do_exp] is called to change [do_exp] if needed
* when we go under a quantification.
* This version makes possible to have a different flavor of term,
* ie. it can be used for translation.
* TODOopt: we could have another optimized version if the types of terms
* are the same in order to avoid building new terms when there is no
* modification.
* *)
let rec change_exp_in_pred (do_exp:'a d_term -> 'b d_term)
      (quantif_do_exp: ('a d_term -> 'b d_term) -> Var.t ->
                       ('a d_term -> 'b d_term)) p =
  let subst_pred = change_exp_in_pred do_exp quantif_do_exp in
  match p with
  | Ptrue -> Ptrue
  | Pfalse -> Pfalse
  | Pif (t,p1,p2) ->
      Pif (do_exp t, subst_pred p1, subst_pred p2)
  | Pnot p -> p_not (subst_pred p)
  | Pforall (v,p') ->
      let f = quantif_do_exp do_exp v in
        Pforall (v,change_exp_in_pred f quantif_do_exp p')
  | Pexists (v,p') ->
      let f = quantif_do_exp do_exp v in
        Pexists (v,change_exp_in_pred f quantif_do_exp p')
  | Plet (x,v,p) ->
      let f =  quantif_do_exp do_exp x in
        Plet (x, do_exp v,change_exp_in_pred f quantif_do_exp p)
  | Pnamed (n,p) -> Pnamed (n,subst_pred p)
  | Pimplies (p1,p2) -> p_implies (subst_pred p1) (subst_pred p2)
  | Pand (p1,p2) -> p_and (subst_pred p1) (subst_pred p2)
  | Por (p1,p2) -> p_or (subst_pred p1) (subst_pred p2)
  | Piff (p1,p2) -> p_iff (subst_pred p1) (subst_pred p2)
  | Papp (n,t) -> p_app n (List.map do_exp t)

let no_quantif_do_exp do_exp qqvar =
  let var_term = e_var qqvar in
    match do_exp var_term with
      | Tvar v when Var.equal qqvar v -> do_exp
      | _ -> assert false

let change_data_in_pred do_data_rec p =
  change_exp_in_pred (change_data_in_exp do_data_rec) no_quantif_do_exp p

(* ------------------------------------------------------------------------ *)
(* ---  Alpha-conversion                                                --- *)
(* ------------------------------------------------------------------------ *)

let rec pred_alpha_cv do_data alpha p = 
  let do_pred alpha p = pred_alpha_cv do_data alpha p in
  let rec do_preds alpha l = match l with [] -> alpha, l
    | p::l ->
        let alpha, p = do_pred alpha p in
        let alpha, l = do_preds alpha l in
          alpha, p::l
  in
  match p with
  | Ptrue | Pfalse -> alpha, p
  | Pif (t,p1,p2) ->
      let alpha, t = term_alpha_cv do_data alpha t in
      let alpha, p1 = do_pred alpha p1 in
      let alpha, p2 = do_pred alpha p2 in
        alpha, Pif (t,p1,p2)
  | Pnot p -> let alpha, p = do_pred alpha p in alpha, Pnot p
  | Pforall (x,p) ->
      let old_x, alpha = alpha_bound_var alpha x in
      let alpha, p = do_pred alpha p in
      let alpha = alpha_unbound alpha x old_x in
        alpha, Pforall (x,p)
  | Pexists (x,p) ->
      let old_x, alpha = alpha_bound_var alpha x in
      let alpha, p = do_pred alpha p in
      let alpha = alpha_unbound alpha x old_x in
        alpha, Pexists (x,p)
  | Plet (x,t,p) ->
      let alpha, t = term_alpha_cv do_data alpha t in
      let old_x, alpha = alpha_bound_var alpha x in
      let alpha, p = do_pred alpha p in
      let alpha = alpha_unbound alpha x old_x in
        alpha, Plet (x,t,p)
  | Pnamed (n,p) -> let alpha, p = do_pred alpha p in alpha, Pnamed (n,p)
  | Pimplies (p1,p2) -> 
      let alpha, p1 = do_pred alpha p1 in
      let alpha, p2 = do_pred alpha p2 in
        alpha, Pimplies (p1,p2)
  | Pand (p1,p2) -> 
      let alpha, p1 = do_pred alpha p1 in
      let alpha, p2 = do_pred alpha p2 in
        alpha, Pand (p1,p2)
  | Por (p1,p2) -> 
      let alpha, p1 = do_pred alpha p1 in
      let alpha, p2 = do_pred alpha p2 in
        alpha, Por (p1,p2)
  | Piff (p1,p2) -> 
      let alpha, p1 = do_pred alpha p1 in
      let alpha, p2 = do_pred alpha p2 in
        alpha, Piff (p1,p2)
  | Papp (n,lt) -> 
      let alpha, lt = terms_alpha_cv do_data alpha lt in alpha, Papp (n,lt)


let p_alpha_cv do_data p = 
  let alpha, p = pred_alpha_cv do_data Vmap.empty p in
  let vars = Vmap.fold (fun _v v' acc -> v'::acc) alpha [] in
    vars, p

(* ------------------------------------------------------------------------ *)
(* ---  Translation (data type can be modified)                         --- *)
(* ------------------------------------------------------------------------ *)

let translate_data_in_term
    (do_data:'a -> 'b d_term) (exp:'a d_term) : 'b d_term =
  change_in_exp (fun _v -> None) do_data exp

let translate_data_in_pred
    (do_data: 'a -> 'b d_term) (p: 'a d_pred) : 'b d_pred =
  let do_exp e = translate_data_in_term do_data e in
  let quantif_do_exp f_exp _ = f_exp in
    change_exp_in_pred do_exp quantif_do_exp p

let translate_pred_in_item
    (do_pred: 'p -> 'd d_pred)
    (d: 'p Formula.item) :
    ('d d_pred) Formula.item =
  match d with
    | Formula.Name k -> Formula.Name k
    | Formula.Cons k -> Formula.Cons k
    | Formula.Type n -> Formula.Type n
    | Formula.Function (tl, t) -> Formula.Function (tl, t)
    | Formula.Predicate tl -> Formula.Predicate tl
    | Formula.Axiom p -> Formula.Axiom (do_pred p)

let translate_pred_in_decl
    (do_pred: 'p -> 'd d_pred)
    (d: 'p Formula.declaration) :
    ('d d_pred) Formula.declaration = 
  { 
    Formula.d_section = d.Formula.d_section ;
    Formula.d_name = d.Formula.d_name ;
    Formula.d_title = d.Formula.d_title ;
    Formula.d_descr = d.Formula.d_descr ;
    Formula.d_source = d.Formula.d_source ;
    Formula.d_item = translate_pred_in_item do_pred d.Formula.d_item ;
  }

let change_exp_in_pred do_exp = change_exp_in_pred do_exp no_quantif_do_exp

(* ------------------------------------------------------------------------ *)
(* ---  Variable substitutions                                              *)
(* ------------------------------------------------------------------------ *)

(** Similar to [change_vars_in_exp] but on predicates.
* Notice that we assume (and check) that [var_subst] only works on free
* variables (and that they are different from bounded ones).
*)
let subst_vars_in_pred prop_in_data var_subst p =
  let rec do_exp exp = change_in_exp var_subst do_data exp
  and do_data d = prop_in_data do_exp d
  in change_exp_in_pred do_exp p

(** Specialized version of [subst_vars_in_pred] to substitute one variable [v]
* by and expression [exp] in a predicate [p]. *)
let subst_in_pred prop_in_data x exp p =
  let var_subst v = if Var.equal v x then Some exp else None in
    subst_vars_in_pred prop_in_data var_subst p

       (*
let nb_var_in_pred prop_in_data v p =
  let nb_occ = ref 0 in
  let do_var () var = if Var.equal v var then nb_occ := !nb_occ + 1 in
  let rec do_exp () e = fold_data_in_exp do_var do_data () e
  and do_data () d = ignore (prop_in_data (fun e -> do_exp () e; e) d) in
  let _ = fold_exp_in_pred do_exp () p in
    !nb_occ
*)

(** Build a predicate equivalent to [let v = e in p] but may use the
* substitution in some simple cases (like substitute a variable by another
* variable for instance).
* [fresh] is only meaningfull when the [let] is actually built: it tells if we
* have to build a new variable for [v].
*)
let let_pred prop_in_data ~fresh v e p =
  if p = Ptrue then Ptrue
  else begin
    match e with
    | Tconst _ | Tvar _ -> subst_in_pred prop_in_data v e p
    | _ ->
        (* let nb_occ = nb_var_in_pred prop_in_data v p in
          (* TODOopt : do only one visit *)
          if nb_occ = 0 then p
          (* else if nb_occ = 1 then subst_in_pred prop_in_data v e p *)
          else *)
        if fresh then
          let v' = Var.fresh_var v in
          let p = subst_in_pred prop_in_data v (e_var v') p in
            p_let v' e p
        else p_let v e p
  end

let rec pred_replace alpha x exp p =
  let frec = pred_replace alpha x exp in
    match p with
      | Papp(f,ts) -> p_app f (List.map (term_replace alpha x exp) ts)
      | (Ptrue | Pfalse) as p -> p
      | Pimplies(p,q) -> p_implies (frec p) (frec q)
      | Pif(a,p,q) -> p_if (term_replace alpha x exp a) (frec p) (frec q)
      | Pand(p,q) -> p_and (frec p) (frec q)
      | Por(p,q)  -> p_or (frec p) (frec q)
      | Piff(p,q) -> p_iff (frec p) (frec q)
      | Pnot(p)   -> p_not (frec p)
      | Pforall(x0,p) as p0 ->
          begin
            if Var.equal x x0 then p0
            else if bad_term_has_var [x0] exp then
                match alpha x with
                  | None -> Plet(x,exp,p0)
                  | Some y ->
                      let py = pred_replace alpha x0 (Tvar y) p in
                      let py' = frec py in
                        Pforall( y , py' )
            else Pforall(x0, frec p)
          end
      | Pexists(x0,p) as p0 ->
          begin
            if Var.equal x x0 then p0
            else if bad_term_has_var [x0] exp then
                match alpha x with
                  | None -> Plet(x,exp,p0)
                  | Some y ->
                      let py = pred_replace alpha x0 (Tvar y) p in
                      let py' = frec py in
                        Pexists( y , py' )
            else Pexists(x0,frec p)
          end
      | Plet(x0,t,p) as p0 ->
          begin
            if bad_term_has_var [x0] exp then
              match alpha x with
                | None -> Plet(x,exp,p0)
                | Some y ->
                    let t' = term_replace alpha x exp t in
                    let py = pred_replace alpha x0 (Tvar y) p in
                    let py' = frec py in
                      Plet( y , t' , py' )
            else
              let t' = term_replace alpha x exp t in
              let p' = if Var.equal x0 x then p else frec p in
                Plet(x0,t',p')
          end
      | Pnamed(a,p) -> Pnamed(a,frec p)

(* ------------------------------------------------------------------------ *)
(* ---  Quantification                                                      *)
(* ------------------------------------------------------------------------ *)

let fresh_vars_in_pred prop_in_data vars p =
  let do_var (vars, p) v =
    let v' = Var.fresh_var v in
    let p = subst_in_pred prop_in_data v (e_var v') p in
      v'::vars, p
  in List.fold_left do_var ([], p) vars

let p_forall_vars (prop_in_data: ('a d_term -> 'a d_term) -> 'a -> 'a d_term)
                   vars (p: 'a d_pred) : 'a d_pred =
    let vars, p = fresh_vars_in_pred prop_in_data vars p in
      List.fold_left (fun p v -> p_forall v p) p vars

let p_exists_vars (prop_in_data: ('a d_term -> 'a d_term) -> 'a -> 'a d_term)
                    vars (p: 'a d_pred) : 'a d_pred =
    let vars, p = fresh_vars_in_pred prop_in_data vars p in
      List.fold_left (fun p v -> p_exists v p) p vars

(* ------------------------------------------------------------------------ *)
(* ---  Huge                                                                *)
(* ------------------------------------------------------------------------ *)

exception Huge

let rec check_term m t =
  if m < 0 then raise Huge ;
  match t with
    | Tconst _ | Tvar _ | Tdata _ -> pred m
    | Tapp(_,ts) -> List.fold_left check_term (pred m) ts
    | Tif(a,b,c) -> check_term (check_term (check_term (pred m) a) b) c
    | Tlet(_,a,b) -> check_term (check_term (pred m) a) b
 
let rec check_pred m p =
  if m < 0 then raise Huge ;
  match p with
    | Ptrue | Pfalse -> pred m
    | Papp(_,ts) -> List.fold_left check_term (pred m) ts
    | Pimplies(a,b) | Pand(a,b) | Por(a,b) | Piff(a,b) -> check_pred (check_pred (pred m) a) b
    | Pnamed(_,p) | Pnot p | Pforall(_,p) | Pexists(_,p) -> check_pred (pred m) p
    | Pif(a,p,q) -> check_pred (check_pred (check_term (pred m) a) p) q
    | Plet(_,t,p) -> check_pred (check_term (pred m) t) p
  
let huge_term m t = try ignore (check_term m t) ; false with Huge -> true
let huge_pred m p = try ignore (check_pred m p) ; false with Huge -> true


