(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

type constant = private
  | ConstInt of string
  | ConstBool of bool
  | ConstUnit
  | ConstFloat of string

val c_bool : bool -> constant
val c_int_of_str : string -> constant
val c_int : int -> constant
val c_float : float -> constant
val c_float_of_str : string -> constant

(** Fol Variable definition*)
module Var : sig

  type t = 
    | VarL of Cil_types.logic_var * Formula.tau
    | VarT of string * int option * Formula.tau * Cil_types.logic_type option

  val ident_named_var : string -> int -> Formula.tau -> t
  val fresh_named_var : string -> Formula.tau -> t
  val fresh_typed_var : string -> Cil_types.logic_type -> Formula.tau -> t
  val fresh_var : t -> t
  val mk_lvar_variable : Cil_types.logic_var -> Formula.tau -> t
  val var_type : t -> Formula.tau
  val var_name : t -> string
  val basename : t -> string
  val kind_of_var : t -> Formula.kind
  val lvar_of_variable : t -> Cil_types.logic_var option
  val equal : t -> t -> bool
  val compare : t -> t -> int

end

module Vset : Set.S with type elt = Var.t
module Vmap : Map.S with type key = Var.t

(*----------------------------------------------------------------------------*)
(** Terms *)

type 'd d_term = private
  | Tconst of constant
  | Tvar of Var.t
  | Tdata of 'd
  | Tapp of string * 'd d_term list
  | Tif of 'd d_term * 'd d_term * 'd d_term 
  | Tlet of Var.t * 'd d_term * 'd d_term 

val e_true : 'd d_term
val e_false : 'd d_term
val e_int : int -> 'd d_term
val e_int64 : int64 -> 'd d_term
val e_float : float -> 'd d_term
val e_cnst : constant -> 'd d_term
val e_var : Var.t -> 'd d_term
val e_data : 'd -> 'd d_term
val e_if : 'd d_term -> 'd d_term -> 'd d_term -> 'd d_term
val e_app : string -> 'd d_term list -> 'd d_term
val e_let : Var.t -> 'd d_term -> 'd d_term -> 'd d_term

(** in [bad_eq_terms], we don't know how to compare data. *)
val bad_eq_terms : 'd d_term -> 'd d_term -> bool
val eq_terms : ('d -> 'd -> bool) -> 'd d_term -> 'd d_term -> bool

val term_replace : 
  (Var.t -> Var.t option) -> Var.t -> 'd d_term -> 'd d_term -> 'd d_term

val e_has_var : (Var.t list -> 'd -> bool) -> Var.t list -> 'd d_term -> bool
val e_closed : (Var.t list -> 'd -> bool) -> Var.t list -> 'd d_term -> bool
val bad_term_closed : Var.t list -> 'd d_term -> bool

val change_in_exp : (Var.t -> 'a d_term option) -> ('b -> 'a d_term) -> 
  'b d_term -> 'a d_term

(** [fold_data_in_exp do_var do_data acc exp] *)
val fold_data_in_exp : ('a -> Var.t -> 'a) -> ('a -> 'b -> 'a) -> 
  'a -> 'b d_term -> 'a  

(*----------------------------------------------------------------------------*)
(** Predicates *)

type 't t_pred =
  | Papp of string * 't list
  | Ptrue
  | Pfalse
  | Pimplies of 't t_pred * 't t_pred
  | Pif of 't * 't t_pred * 't t_pred
  | Pand of 't t_pred * 't t_pred
  | Por of 't t_pred * 't t_pred
  | Piff of 't t_pred * 't t_pred
  | Pnot of 't t_pred
  | Pforall of Var.t * 't t_pred
  | Pexists of Var.t * 't t_pred
  | Plet of Var.t * 't * 't t_pred
  | Pnamed of string * 't t_pred

type 'd d_pred = ('d d_term) t_pred

val p_and :  't t_pred -> 't t_pred -> 't t_pred
val p_app: string -> 'd d_term list -> 'd d_pred
val p_or : 't t_pred -> 't t_pred -> 't t_pred
val p_xor : 't t_pred -> 't t_pred -> 't t_pred
val p_iff : 'd d_pred -> 'd d_pred -> 'd d_pred
val p_not : 'd d_pred -> 'd d_pred
val p_if :  'd d_term -> 'd d_pred -> 'd d_pred -> 'd d_pred
val p_implies: 'd d_pred -> 'd d_pred -> 'd d_pred
val p_conj : 'd d_pred list -> 'd d_pred
val p_disj : 'd d_pred list -> 'd d_pred

val p_eq : 'd d_term -> 'd d_term -> 'd d_pred
val p_neq : 'd d_term -> 'd d_term -> 'd d_pred

val p_forall : Var.t -> 'd d_pred -> 'd d_pred
val p_exists : Var.t -> 'd d_pred -> 'd d_pred
val p_let : Var.t -> 'd d_term -> 'd d_pred -> 'd d_pred
val p_named : string -> 'd d_pred -> 'd d_pred

val pred_replace : 
  (Var.t -> Var.t option) -> Var.t -> 'd d_term -> 'd d_pred -> 'd d_pred

val p_has_var : (Var.t list -> 't -> bool) -> Var.t list -> 't t_pred -> bool

val p_closed : (Var.t list -> 't -> bool) -> Var.t list -> 't t_pred -> bool
val bad_pred_closed : Var.t list -> 'd d_pred -> bool

val subst_in_pred : (('a d_term -> 'b d_term) -> 'a -> 'b d_term) ->
  Var.t -> 'b d_term -> 'a d_term t_pred -> 'b d_term t_pred

val let_pred : (('a d_term -> 'a d_term) -> 'a -> 'a d_term) ->
  fresh:bool -> Var.t -> 'a d_term -> 'a d_term t_pred -> 'a d_term t_pred

val p_forall_vars : (('a d_term -> 'a d_term) -> 'a -> 'a d_term) ->
                   Var.t list ->  'a d_pred -> 'a d_pred

val p_exists_vars : (('a d_term -> 'a d_term) -> 'a -> 'a d_term) ->
                   Var.t list ->  'a d_pred -> 'a d_pred

(** [fold_exp_in_pred do_exp acc p] *)
val fold_exp_in_pred : ('a -> 'b -> 'a) -> 'a -> 'b t_pred -> 'a   

val fold_data_in_pred : ('a -> Var.t -> 'a) ->
  (('a -> 'd d_term -> 'a) -> 'a -> 'd -> 'a) ->
  'a -> 'd d_pred -> 'a

(** [change_data_in_pred do_data_rec p] where [do_data_rec do_exp d] *)
val change_data_in_pred : (('a d_term -> 'b d_term) -> 'a -> 'b ) ->
  'a d_term t_pred -> 'b d_term t_pred  

(** [pred_alpha_c data_alpha alpha p] -> [alpha', p']*)
val pred_alpha_cv :
    (Var.t Vmap.t -> 'd -> Var.t Vmap.t * 'd) -> 
     Var.t Vmap.t  -> 'd d_pred -> Var.t Vmap.t * 'd d_pred

val p_alpha_cv : 
    (Var.t Vmap.t -> 'd -> Var.t Vmap.t * 'd) -> 
    'd d_pred -> Var.t list * 'd d_pred

(*----------------------------------------------------------------------------*)
(** Translations *)

(** [translate_data_in_term do_data exp] *)
val translate_data_in_term : ('a -> 'b d_term) -> 'a d_term -> 'b d_term

(** [translate_data_in_pred do_data pred] *)
val translate_data_in_pred : ('a -> 'b d_term) -> 'a d_pred -> 'b d_pred

(** [translate_pred_in_item do_pred item] *)
val translate_pred_in_item : ('p -> 'd d_pred) -> 
  'p Formula.item -> ('d d_pred) Formula.item

(** [translate_pred_in_decl do_pred item] *)
val translate_pred_in_decl : ('p -> 'd d_pred) -> 
  'p Formula.declaration -> ('d d_pred) Formula.declaration

(*----------------------------------------------------------------------------*)
val huge_term : int -> 'd d_term -> bool
val huge_pred : int -> 'd d_pred -> bool
(*----------------------------------------------------------------------------*)

