(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* ------------------------------------------------------------------------ *)
(* ---  Remove let with unversally quantified variables                     *)
(* ------------------------------------------------------------------------ *)

open Fol

(* II. Alternative 
   logic closure, each let-bound variable x becomes a universally
   quantified variable, constrained by an equality to the bound term t : 
   let x = t in ... becomes forall x, x = t -> ...
   
*)      

type 'd env = { mutable foralls : Vset.t ; mutable eqs : 'd d_term t_pred}

let new_name x sigma cpt =
  incr cpt;
  let sx = Var.basename x^(string_of_int !cpt) in 
  let nx = Var.fresh_named_var sx (Var.var_type x) in
  nx,(Vmap.add x nx sigma) 

let rec cct sigma env cpt = function
  | Tlet(x,t1,t2) -> 
      let t1 = cct sigma env cpt t1 in
      let nx ,sigma = new_name x sigma cpt in 
      env.foralls <- Vset.add x env.foralls; 
      env.eqs <- p_and env.eqs (p_eq (e_var nx) t1) ;
      cct sigma env cpt t2 
  | Tapp(f,tl) ->
      e_app f  (List.map (cct sigma env cpt ) tl)
  | Tif (c,ta,tb) ->
      e_if(cct sigma env cpt c)(cct sigma env cpt ta)(cct sigma env cpt tb)
  | Tdata _ -> Wp_parameters.fatal "Normalization of Fol(data)"
  | Tconst _ as c -> c
  | Tvar x as t ->
      begin
	try e_var (Vmap.find x sigma)
	with Not_found -> 
	  Wp_parameters.failure ~once:true "Suspiscous normalization of %s" 
	    (Fol.Var.var_name x) ; t
      end

let flush env p = 
 Vset.fold (fun x p -> p_forall x p ) env.foralls (p_implies env.eqs p)

let init_env () = {foralls = Vset.empty;eqs = Ptrue}  

let rec ccpred sigma cpt p = 
  let env = init_env() in 
  let p =  ccp sigma env cpt p in 
  flush env p 

and ccp sigma env cpt = function 
  | Pimplies (p1,p2) -> 
      let e1 = init_env () in
      let h1 = ccp sigma e1 cpt p1 in
      if Vset.is_empty e1.foralls then
	( env.eqs <- p_and env.eqs (p_and e1.eqs h1) ; ccp sigma env cpt p2 )
      else
	p_implies (flush e1 h1) (ccpred sigma cpt p2)

  | Pif (t,p1,p2) ->
      p_if (cct sigma env cpt t) (ccpred sigma cpt p1)(ccpred sigma cpt p2)
  | Pand (p1,p2) -> p_and(ccpred sigma cpt p1)(ccpred sigma cpt p2)
  | Por (p1,p2) -> p_or(ccpred sigma cpt p1)(ccpred sigma cpt p2)
  | Piff (p1,p2) -> p_iff (ccpred sigma cpt p1)(ccpred sigma cpt p2)
  | Pnot p -> p_not (ccpred sigma cpt p)
  | Papp (f, tl) -> Papp(f,List.map (cct sigma env cpt) tl)

  | Pexists(x,p) -> 
      let nx,sigma = new_name x sigma cpt in 
      Pexists(nx,ccpred sigma cpt p)

  | Pforall (x,p) -> 
      let nx,sigma = new_name x sigma cpt in 
      env.foralls <- Vset.add nx env.foralls; 
      ccp sigma env cpt p 

  | Plet (x,t,p) ->
      let t1 = cct sigma env cpt t in 
      let nx,sigma =  new_name x sigma cpt in
      env.foralls <- Vset.add nx env.foralls;
      env.eqs <- p_and env.eqs (p_eq (e_var nx) t1) ;
      ccp sigma env cpt p

  | Pnamed(s,p) -> Pnamed(s,ccp sigma env cpt p)
  | Pfalse -> Pfalse
  | Ptrue -> Ptrue
  
let cc_pred p= ccpred Vmap.empty (ref 0) p   
