(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* -------------------------------------------------------------------------- *)
(* --- Alt-ergo Export                                                    --- *)
(* -------------------------------------------------------------------------- *)

open Format
open Cilutil
open Formula
open Fol
open Fol_norm

let constant fmt = function
  | ConstInt n -> pp_print_string fmt n
  | ConstBool b -> pp_print_string fmt (if b then "true" else "false")
  | ConstUnit -> pp_print_string fmt "void"
  | ConstFloat f -> pp_print_string fmt f

let pp_list pp fmt = function
  | [] -> ()
  | x::xs ->
      pp fmt x ;
      List.iter (fun x -> fprintf fmt ",@,%a" pp x) xs

let pp_flow fmt nil op pp = function
    | [] -> 
	pp_print_string fmt nil
    | x::xs ->
	fprintf fmt "@[<hov 1>(%a" pp x ;
	List.iter (fun x -> fprintf fmt "@,%s%a" op pp x) xs ;
	fprintf fmt ")@]"

let pp_block fmt tab op pp ps =
  match ps with
    | [] -> ()
    | p::ps ->
	fprintf fmt "%s %a" tab pp p ;
	List.iter (fun p -> fprintf fmt "@\n%s %a" op pp p) ps

let pp_var fmt v =
  let name = Datalib.identifier (Var.var_name v) in
  fprintf fmt "%s" name

let rec collect_assoc op xs = function
  | [] -> List.rev xs
  | Tapp(id,ts) :: others when id = op -> 
      collect_assoc op (collect_assoc op xs ts) others
  | t::others ->
      collect_assoc op (t::xs) others


let array_flag = ref false
let array () = !array_flag

let fpp_term term pp_data fmt t =
  match t with
    | Tconst c -> constant fmt c
    | Tvar v -> pp_var fmt v
    | Tdata d -> pp_data fmt d
    | Tapp (id, []) -> pp_print_string fmt id
    | Tapp ("ite",[c;a;b]) | Tif (c,a,b) ->
	fprintf fmt "(@[<v 0>if %a@ then %a@ else %a@])"
	  term c term a term b
    | Tapp ("add_int", ts) ->
	let xs = collect_assoc "add_int" [] ts in
	pp_flow fmt "0" "+" term xs
    | Tapp ("sub_int", [a;b]) ->
	fprintf fmt "@[<hov 1>(%a@,-%a)@]" term a term b
    | Tapp ("mul_int", ts) ->
	let xs = collect_assoc "mul_int" [] ts in
	pp_flow fmt "1" "*" term xs
    | Tapp("access",[a;k]) when array() ->
	Format.fprintf fmt "%a[%a]" term a term k
    | Tapp("update",[a;k;b]) when array () ->
	Format.fprintf fmt "@[<hv 2>%a[%a@,<-%a]@]" 
	  term a term k term b
    | Tapp (id, t::ts) ->
	fprintf fmt "@[<hov 2>%s(@,%a" id term t ;
	List.iter (fun t -> fprintf fmt ",@,%a" term t) ts ;
	fprintf fmt ")@]"
    | Tlet (_x,_v,_t) -> 
	Wp_parameters.fatal 
	  "[Fol_ergo] Works only on free let construction term\n"

let rec collect_or ps = function
  | Por(a,b) -> collect_or (collect_or ps b) a
  | p -> p :: ps

let rec collect_and ps = function
  | Pand(a,b) -> collect_and (collect_and ps b) a
  | p -> p :: ps

let rec collect_imply ps = function
  | Pimplies(a,b) -> collect_and (collect_imply ps b) a
  | p -> p :: ps

let rec collect_iff ps = function
  | Piff(a,b) -> collect_iff (collect_iff ps b) a
  | p -> p :: ps


type 'a pp = Format.formatter -> 'a -> unit
type 'a pp_env = {
  pp_type : Formula.tau pp ;
  pp_term : 'a Fol.d_term pp ;
  pp_pred : 'a Fol.d_pred pp ;
}

open Fol_cc

let pp_args pp_tau fmt = function 
  | [] -> () 
  | [x] -> Format.fprintf fmt "(%a@,:@,%a)@,:%a@,=@" 
      pp_var x pp_tau (Var.var_type x) pp_tau (Var.var_type x)
  | x::m -> 
      Format.fprintf fmt "(%a@,:@,%a@," 
       pp_var x pp_tau (Var.var_type x);
      List.iter (fun x -> Format.fprintf fmt ",@,%a@,:@,%a@,"
		   pp_var x pp_tau (Var.var_type x) ) m;
      Format.fprintf fmt ")@,:%a@,=@\n" pp_tau (Var.var_type x)

let fpp_f_let pp_tau pp_term fmt fl = 
  Format.fprintf fmt "@[<hov 2>function %s@,%a%a@]"
    fl.name (pp_args pp_tau) fl.param  
    pp_term fl.body 

let fpp_lf_let pp_tau pp_term fmt = function
  | [] -> ()
  | x::xs ->
      let pp = fpp_f_let pp_tau pp_term in 
      pp fmt x ;
      List.iter (fun x -> fprintf fmt "%a@\n" pp x ) xs


let rec epp_pred_vbox env fmt p =
  match p with
    | Pand _ -> pp_block fmt "   " "and" env.pp_pred (collect_and [] p)
    | Por _ -> pp_block fmt "  " "or" env.pp_pred (collect_or [] p)
    | Pimplies _ -> pp_block fmt "  " "->" env.pp_pred (collect_imply [] p)
    | Piff _ -> pp_block fmt "   " "<->" env.pp_pred (collect_iff [] p)
    | Pforall(x,p) ->
	fprintf fmt "forall %a:%a.@\n" pp_var x env.pp_type (Var.var_type x) ;
	epp_pred_vbox env fmt p
    | Pexists(x,p) ->
	fprintf fmt "exists %a:%a.@\n" pp_var x env.pp_type (Var.var_type x) ;
	epp_pred_vbox env fmt p 
    | Pif(t,p,q) ->
	fprintf fmt "@[<hov 0>if @[<hov 2>%a@]@ then@]@\n  %a@\nelse@\n  %a"
	  env.pp_term t env.pp_pred p env.pp_pred q
    | (Ptrue | Pfalse | Papp _ | Pnot _ | Pnamed _) ->
	env.pp_pred fmt p
    | Plet(_x,_t,_p) -> Wp_parameters.fatal 
	  "[Fol_ergo] Works only on free let construction predicate\n"

let rec epp_pred_atom env fmt p =
  match p with
    | Pand _ | Por _ | Pimplies _ | Piff _ | Pif _
    | Pforall _ | Pexists _ | Plet _ ->
	fprintf fmt "@[<v 1>(%a)@]" (epp_pred_vbox env) p
    | Pnot p ->
	fprintf fmt "@[<hov 2>(not@ %a)@]" (epp_pred_atom env) p
    | Pnamed(tag,p) ->
	fprintf fmt "@[<hov 0>%s:@,%a@]" tag (epp_pred_atom env) p
    | Ptrue -> pp_print_string fmt "true"
    | Pfalse -> pp_print_string fmt "false"
    | Papp(id,[]) -> pp_print_string fmt id
    | Papp (("eq" | "eq_int"), [t1; t2]) ->
	fprintf fmt "@[<hov 1>(%a@ =@ %a)@]" env.pp_term t1 env.pp_term t2
    | Papp (("neq" | "neq_int"), [t1; t2]) ->
	fprintf fmt "@[<hov 1>(%a@ <>@ %a)@]" env.pp_term t1 env.pp_term t2
    | Papp ("lt_int", [t1; t2]) ->
	fprintf fmt "@[<hov 1>(%a@ <@ %a)@]" env.pp_term t1 env.pp_term t2
    | Papp ("le_int", [t1; t2]) ->
	fprintf fmt "@[<hov 1>(%a@ <=@ %a)@]" env.pp_term t1 env.pp_term t2
    | Papp(id,t::ts) ->
	fprintf fmt "@[<hov 2>%s(@,%a" id env.pp_term t ;
	List.iter (fun t -> fprintf fmt ",@ %a" env.pp_term t) ts ;
	fprintf fmt ")@]"

let fpp_pred predicate pp_term pp_type fmt p =
  match p with
    | Ptrue -> fprintf fmt "true"
    | Pfalse -> fprintf fmt "false"
    | Papp (id, [])-> fprintf fmt "%s" id
    | Papp ("eq", [t1; t2]) -> fprintf fmt "(%a =@ %a)" pp_term t1 pp_term t2
    | Papp ("neq", [t1; t2]) -> fprintf fmt "(%a <>@ %a)" pp_term t1 pp_term t2
    | Papp (id, l) -> fprintf fmt "@[%s(%a)@]" id (pp_list pp_term) l
    | Pimplies (a, b) -> fprintf fmt "(@[%a ->@ %a@])" predicate a predicate b
    | Piff (a, b) -> fprintf fmt "(@[%a <->@ %a@])" predicate a predicate b
    | Pand (a, b) -> fprintf fmt "(@[%a and@ %a@])" predicate a predicate b
    | Por (a, b) -> fprintf fmt "(@[%a or@ %a@])" predicate a predicate b
    | Pnot a -> fprintf fmt "(not %a)" predicate a
    | Pif (a, b, c) ->
	fprintf fmt "(@[if %a then@ %a else@ %a@])"
	  pp_term a predicate b predicate c
    | Pforall (v,p) ->
	fprintf fmt "@[<hov 0>(forall %a:%a.@ %a@])"
	  pp_var v pp_type (Var.var_type v) predicate p
    | Pexists (v,p) ->
	fprintf fmt "@[<hov 0>(exists %a:%a.@ %a@])"
	  pp_var v pp_type (Var.var_type v) predicate p
    | Pnamed (n, p) ->
	fprintf fmt "@[%s: %a@]" n predicate p
    | Plet (_x,_v,_p) ->
	Wp_parameters.fatal 
	  "[Fol_ergo] Works only on free let construction predicate\n" 

let fpp_item predicate pp_tau pp_term fmt x =
  function
    | Formula.Name k ->
	fprintf fmt "function %s (): name = mk_name(%d)@\n" x k
    | Formula.Cons k ->
	fprintf fmt "function %s (): int = %d@\n" x k
    | Formula.Function ([], t) ->
	fprintf fmt "logic %s: %a@\n" x pp_tau t
    | Formula.Function (tl, t) ->
	fprintf fmt "logic %s: @[<hov 0>%a -> %a@]@\n" x 
	  (pp_list pp_tau) tl pp_tau t
    | Formula.Predicate [] ->
	fprintf fmt "logic %s: prop@\n" x
    | Formula.Predicate tl ->
	fprintf fmt "logic %s: @[<hov 0>%a -> prop@]@\n" x 
	  (pp_list pp_tau) tl
    | Formula.Axiom p ->
	begin
	 match Fol_norm.ergo_normalize p with 
	   | Pred p' -> fprintf fmt "@[<hv 2>axiom %s:@ %a@]@\n" x predicate p'
	   | Conv (defs,p') ->
	       fpp_lf_let pp_tau pp_term fmt defs  ;
	       fprintf fmt "@[<hv 2>axiom %s:@ %a@]@\n" x predicate p'
	end
    | Formula.Type 0 ->
	fprintf fmt "type %s@\n" x
    | Formula.Type 1 ->
	fprintf fmt "type 'a %s@\n" x
    | Formula.Type n ->
	fprintf fmt "@[<hov 2>type ('a" ;
	for k=2 to n do 
	  fprintf fmt ",%c" (char_of_int (int_of_char 'a'+k-1)) 
	done ;
	fprintf fmt ") %s@]@\n" x

let fpp_header fmt d =
  begin
    d.d_title fmt ;
    Format.pp_print_newline fmt () ;
    ( match d.d_source with
	| Some { Log.src_file=f ; Log.src_line=k } ->
	    Format.fprintf fmt "%s:%d: " f k
	| None -> () ) ;
    d.d_descr fmt ;
  end

let fpp_decl predicate pp_tau pp_term fmt d =
  begin
    fpp_header fmt d ;
    Format.pp_print_newline fmt () ;
    fpp_item predicate pp_tau pp_term fmt d.d_name d.d_item ;
    Format.pp_print_newline fmt () ;
  end

let fpp_goal predicate fmt x p =
  fprintf fmt "@[<hv 2>goal %s:@ %a@]@." x predicate p


type pred = unit Fol.d_pred
    
let export_data _ _ = ()

let rec export_tau fmt = function
  | Integer -> pp_print_string fmt "int"
  | Real -> pp_print_string fmt "real"
  | Boolean -> pp_print_string fmt "bool"
  | Pointer -> pp_print_string fmt "pointer"
  | Record _ -> pp_print_string fmt 
      (if array() then "data farray" else "data why__farray")
  | Array arr ->
      Format.fprintf fmt "%a %s"
	export_tau 
	(Datalib.Tau.tau_of_logic_ctype arr.Ctypes.arr_element)
	(if array() then "farray" else "why__farray")
  | Set te ->
      Format.fprintf fmt "%a set" export_tau te
  | ADT("farray",[t]) -> 
      Format.fprintf fmt "%a %s" export_tau t 
	(if array() then "farray" else "why__farray")
  | ADT(s,[]) -> pp_print_string fmt s
  | ADT(s,[t]) -> Format.fprintf fmt "%a %s" export_tau t s
  | ADT(s,t::ts) ->
      Format.fprintf fmt "@[(%a" export_tau t ;
      List.iter (fun t -> Format.fprintf fmt ",@,%a" export_tau t) ts ;
      Format.fprintf fmt ") %s@]" s
	
let rec export_term fmt t = fpp_term export_term export_data fmt t
  
let rec pp_pred_atom fmt p =
  epp_pred_atom {
    pp_type = export_tau ;
    pp_term = export_term ;
    pp_pred = pp_pred_atom ;
  } fmt p
    
let export_pred fmt p =
  epp_pred_vbox {
    pp_type = export_tau ;
    pp_term = export_term ;
    pp_pred = pp_pred_atom ;
  } fmt p
    
let export_item fmt name item = fpp_item export_pred export_tau fmt name item
  
let export_decl fmt d =
  Pretty_utils.pp_trail fpp_header fmt d ;
  Format.pp_print_newline fmt () ;
  fpp_item export_pred export_tau export_term fmt d.d_name d.d_item
    
let export_goal fmt x g=
  match Fol_norm.ergo_normalize g with 
    | Pred p' ->  fpp_goal export_pred fmt x p'
    | Conv (defs,p') ->
	fpp_lf_let export_tau export_term fmt defs  ;
	fpp_goal export_pred fmt x p'
    

module E91 =
struct

  type pred = unit Fol.d_pred

  let export_tau fmt t =
    array_flag := false ;
    export_tau fmt t

  let export_decl fmt d =
    array_flag := false ;
    export_decl fmt d
  
  let export_goal fmt g p =
    array_flag := false ;
    export_goal fmt g p

end

module E92 =
struct

  type pred = unit Fol.d_pred

  let export_tau fmt t =
    array_flag := true ;
    export_tau fmt t

  let export_decl fmt d =
    array_flag := true ;
    export_decl fmt d
  
  let export_goal fmt g p =
    array_flag := true ;
    export_goal fmt g p

end
