(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

open Fol

module Langage =
struct

  let e_true = Fol.e_true
  let e_false = Fol.e_false

  let e_int k   = Fol.e_int k
  let e_int64 k = Fol.e_int64 k
  let e_float k = Fol.e_float k

  let e_icst z = Fol.e_cnst (Fol.c_int_of_str z)
  let e_rcst z =  Fol.e_cnst (Fol.c_float_of_str z)

  let wrap t = t
  let unwrap t = t

  let e_call f ts = Fol.e_app f ts
  let p_call f xs = Fol.p_app f xs

end

module Pool =
struct

  type pool = (string,int option) Hashtbl.t

  type var = Fol.Var.t

  let pool () = Hashtbl.create 7

  let fresh_var pool basename =
    let new_contents = try
      let counter = Hashtbl.find pool basename in
      match counter with
      | None -> Some 0
      | Some i -> Some (succ i)
    with Not_found -> None
    in
    Hashtbl.replace pool basename new_contents;
    new_contents

  let fresh pool x kind =
    let x = if Datalib.has_reserved_prefix (x^"_") then
      let new_name = "G_"^x in
      if Datalib.has_reserved_prefix new_name then
        Wp_parameters.fatal
          "Reserved prefix for '%s' is clashing. Frama-C WP plugin cannot reserve the prefix 'G'"
          new_name
      else new_name
    else x
    in
      let vx =
      match kind with
      | Formula.Model t -> Fol.Var.VarT(x,fresh_var pool x,t,None)
      | Formula.Acsl(t,ty) -> Fol.Var.VarT(x,fresh_var pool x,t,Some ty)
    in vx

  let var v = Fol.e_var v

  let tau_of_var = Fol.Var.var_type
  let name_of_var = Fol.Var.var_name
  let basename_of_var = Fol.Var.basename
  let kind_of_var = Fol.Var.kind_of_var

  let freshen pool v = fresh pool (Fol.Var.basename v) (Fol.Var.kind_of_var v)

end


module Predicate =
struct

  let p_true = Fol.Ptrue
  let p_false = Fol.Pfalse

  let p_not a   = Fol.p_not a
  let p_bool a  = Fol.p_eq Langage.e_true a

  let equal_terms e1 e2 = Fol.bad_eq_terms e1 e2

  let p_and a b = Fol.p_and a b
  let p_or  a b = Fol.p_or a b
  let p_xor  a b = Fol.p_xor a b
  let p_eq a b = Fol.p_eq a b
  let p_neq a b = Fol.p_neq a b
  let p_implies a b = Fol.p_implies a b
  let p_iff a b = Fol.p_iff a b
  let p_conj l = Fol.p_conj l
  let p_disj l = Fol.p_disj l
  let p_cond b pt pf = Fol.p_if (Langage.wrap b) pt pf
  let rec is_true = function 
    | Fol.Ptrue -> true 
    | Fol.Pnamed(_,p) -> is_true p
    | _ -> false
  let rec is_false = function 
    | Fol.Pfalse -> true 
    | Fol.Pnamed(_,p) -> is_false p
    | _ -> false

  let p_named = Fol.p_named

  let eq_var = Fol.Var.equal

  let p_forall xs p =
    List.fold_right Fol.p_forall xs p

  let p_exists xs p =
    List.fold_right Fol.p_exists xs p

  let p_subst alpha x v p =
    match v with
      | Fol.Tvar y when (Fol.Var.equal x y) -> p
      | (Fol.Tconst _ | Fol.Tvar _ | Fol.Tapp(_,[])) ->
	  Fol.pred_replace alpha x v p
      | _ -> Fol.p_let x v p

  let e_subst alpha x v t =
    match v with
      | Fol.Tvar y when (Fol.Var.equal x y) -> t
      | (Fol.Tconst _ | Fol.Tvar _ | Fol.Tapp(_,[])) ->
          Fol.term_replace alpha x v t
      | _ -> Fol.e_let x v t

  let rec apply alpha x = match alpha with
    | [] -> None
    | (y,y')::s ->
        if Fol.Var.equal x y then Some y' else apply s x

  let rec e_rename s t =
    match t with
      | Tconst _ -> t
      | Tvar x ->
	  ( match apply s x with
	      | None -> t
	      | Some y -> Fol.e_var y )
      | Tdata _ as t ->
	  List.fold_left
	    (fun t (x,x') -> e_let x (Fol.e_var x') t)
	    t s
      | Tapp(f,ts) -> e_app f (List.map (e_rename s) ts)
      | Tif(a,b,c) -> e_if (e_rename s a) (e_rename s b) (e_rename s c)
      | Tlet(y,a,b) ->
	  let a' = e_rename s a in
	  let s' = (y,y)::s in (* defensive ! *)
            e_let y a' (e_rename s' b)

end
