(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* ---------------------------------------------------------------------- *)
(* --- Output Utilities                                               --- *)
(* ---------------------------------------------------------------------- *)

open Format
open Cilutil
open Formula
open Fol

let constant fmt = function
  | ConstInt n -> pp_print_string fmt n
  | ConstBool b -> pp_print_string fmt (if b then "true" else "false")
  | ConstUnit -> pp_print_string fmt "void"
  | ConstFloat f -> pp_print_string fmt f

let pp_list pp fmt = function
  | [] -> ()
  | x::xs ->
      pp fmt x ;
      List.iter (fun x -> fprintf fmt ",@,%a" pp x) xs

let pp_flow fmt nil op pp = function
    | [] -> 
	pp_print_string fmt nil
    | x::xs ->
	fprintf fmt "@[<hov 1>(%a" pp x ;
	List.iter (fun x -> fprintf fmt "@,%s%a" op pp x) xs ;
	fprintf fmt ")@]"

let pp_block fmt tab op pp ps =
  match ps with
    | [] -> ()
    | p::ps ->
	fprintf fmt "%s %a" tab pp p ;
	List.iter (fun p -> fprintf fmt "@\n%s %a" op pp p) ps

let pp_var fmt v =
  let name = Datalib.identifier (Var.var_name v) in
  fprintf fmt "%s" name

let rec collect_assoc op xs = function
  | [] -> List.rev xs
  | Tapp(id,ts) :: others when id = op -> 
      collect_assoc op (collect_assoc op xs ts) others
  | t::others ->
      collect_assoc op (t::xs) others

let fpp_term term pp_data fmt t =
  match t with
    | Tconst c -> constant fmt c
    | Tvar v -> pp_var fmt v
    | Tdata d -> pp_data fmt d
    | Tapp (id, []) -> pp_print_string fmt id
    | Tapp ("ite",[c;a;b]) | Tif (c,a,b) ->
	fprintf fmt "(@[<v 0>if %a@ then %a@ else %a@])"
	  term c term a term b
    | Tapp ("add_int", ts) ->
	let xs = collect_assoc "add_int" [] ts in
	pp_flow fmt "0" "+" term xs
    | Tapp ("sub_int", [a;b]) ->
	fprintf fmt "@[<hov 1>(%a@,-%a)@]" term a term b
    | Tapp ("mul_int", ts) ->
	let xs = collect_assoc "mul_int" [] ts in
	pp_flow fmt "1" "*" term xs
    | Tapp (id, t::ts) ->
	fprintf fmt "@[<hov 2>%s(@,%a" id term t ;
	List.iter (fun t -> fprintf fmt ",@,%a" term t) ts ;
	fprintf fmt ")@]"
    | Tlet (x,v,t) ->
	fprintf fmt "(@[<v 0>let %a@ = %a@ in %a@])"
	  pp_var x term v term t

let fpretty_term term pp_data fmt t = 
  if Wp_parameters.verbose_atleast 2 
  then fpp_term term pp_data fmt t
  else match t with

    | Tapp(("encode"|"decode"),[_;v]) ->
	Format.fprintf fmt "{%a}" term v

    | Tapp("access",[a;k]) ->
	Format.fprintf fmt "%a[%a]" term a term k

    | Tapp("update",[a;k;b]) ->
	Format.fprintf fmt "@[<hv 2>%a[%a@,->%a]@]" 
	  term a term k term b

    | t -> 
	fpp_term term pp_data fmt t

let rec collect_or ps = function
  | Por(a,b) -> collect_or (collect_or ps b) a
  | p -> p :: ps

let rec collect_and ps = function
  | Pand(a,b) -> collect_and (collect_and ps b) a
  | p -> p :: ps

let rec collect_imply ps = function
  | Pimplies(a,b) -> collect_and (collect_imply ps b) a
  | p -> p :: ps

let rec collect_iff ps = function
  | Piff(a,b) -> collect_iff (collect_iff ps b) a
  | p -> p :: ps

type 'a pp = Format.formatter -> 'a -> unit
type 'a pp_env = {
  pp_type : Formula.tau pp ;
  pp_term : 'a Fol.d_term pp ;
  pp_pred : 'a Fol.d_pred pp ;
}

let rec epp_pred_vbox env fmt p =
  match p with
    | Pand _ -> pp_block fmt "   " "and" env.pp_pred (collect_and [] p)
    | Por _ -> pp_block fmt "  " "or" env.pp_pred (collect_or [] p)
    | Pimplies _ -> pp_block fmt "  " "->" env.pp_pred (collect_imply [] p)
    | Piff _ -> pp_block fmt "   " "<->" env.pp_pred (collect_iff [] p)
    | Pforall(x,p) ->
	fprintf fmt "forall %a:%a.@\n" pp_var x env.pp_type (Var.var_type x) ;
	epp_pred_vbox env fmt p
    | Pexists(x,p) ->
	fprintf fmt "exists %a:%a.@\n" pp_var x env.pp_type (Var.var_type x) ;
	epp_pred_vbox env fmt p
    | Plet(x,t,p) ->
	fprintf fmt "@[<hov 2>let %a=@ %a@ in@]@\n" pp_var x env.pp_term t ;
	epp_pred_vbox env fmt p
    | Pif(t,p,q) ->
	fprintf fmt "@[<hov 0>if @[<hov 2>%a@]@ then@]@\n  %a@\nelse@\n  %a"
	  env.pp_term t env.pp_pred p env.pp_pred q
    | (Ptrue | Pfalse | Papp _ | Pnot _ | Pnamed _) ->
	env.pp_pred fmt p

let rec epp_pred_atom env fmt p =
  match p with
    | Pand _ | Por _ | Pimplies _ | Piff _ | Pif _
    | Pforall _ | Pexists _ | Plet _ ->
	fprintf fmt "@[<v 1>(%a)@]" (epp_pred_vbox env) p
    | Pnot p ->
	fprintf fmt "@[<hov 2>(not@ %a)@]" (epp_pred_atom env) p
    | Ptrue -> pp_print_string fmt "true"
    | Pfalse -> pp_print_string fmt "false"
    | Papp(id,[]) -> pp_print_string fmt id
    | Papp (("eq" | "eq_int"), [t1; t2]) ->
	fprintf fmt "@[<hov 1>(%a@ =@ %a)@]" env.pp_term t1 env.pp_term t2
    | Papp (("neq" | "neq_int"), [t1; t2]) ->
	fprintf fmt "@[<hov 1>(%a@ <>@ %a)@]" env.pp_term t1 env.pp_term t2
    | Papp ("lt_int", [t1; t2]) ->
	fprintf fmt "@[<hov 1>(%a@ <@ %a)@]" env.pp_term t1 env.pp_term t2
    | Papp ("le_int", [t1; t2]) ->
	fprintf fmt "@[<hov 1>(%a@ <=@ %a)@]" env.pp_term t1 env.pp_term t2
    | Papp(id,t::ts) ->
	fprintf fmt "@[<hov 2>%s(@,%a" id env.pp_term t ;
	List.iter (fun t -> fprintf fmt ",@ %a" env.pp_term t) ts ;
	fprintf fmt ")@]"
    | Pnamed(tag,p) ->
	fprintf fmt "@[<hov 0>@[<hov 0>%s:%a@]" tag (epp_pred_named env) p

and epp_pred_named env fmt = function
  | Pnamed(tag,p) ->
      fprintf fmt "@,%s:" tag ;
      epp_pred_named env fmt p
  | p ->
      fprintf fmt "@]@,%a" (epp_pred_atom env) p

let fpp_pred predicate pp_term pp_type fmt p =
  match p with
    | Ptrue -> fprintf fmt "true"
    | Pfalse -> fprintf fmt "false"
    | Papp (id, [])-> fprintf fmt "%s" id
    | Papp ("eq", [t1; t2]) -> fprintf fmt "(%a =@ %a)" pp_term t1 pp_term t2
    | Papp ("neq", [t1; t2]) -> fprintf fmt "(%a <>@ %a)" pp_term t1 pp_term t2
    | Papp (id, l) -> fprintf fmt "@[%s(%a)@]" id (pp_list pp_term) l
    | Pimplies (a, b) -> fprintf fmt "(@[%a ->@ %a@])" predicate a predicate b
    | Piff (a, b) -> fprintf fmt "(@[%a <->@ %a@])" predicate a predicate b
    | Pand (a, b) -> fprintf fmt "(@[%a and@ %a@])" predicate a predicate b
    | Por (a, b) -> fprintf fmt "(@[%a or@ %a@])" predicate a predicate b
    | Pnot a -> fprintf fmt "(not %a)" predicate a
    | Pif (a, b, c) ->
	fprintf fmt "(@[if %a then@ %a else@ %a@])"
	  pp_term a predicate b predicate c
    | Pforall (v,p) ->
	fprintf fmt "@[<hov 0>(forall %a:%a.@ %a@])"
	  pp_var v pp_type (Var.var_type v) predicate p
    | Pexists (v,p) ->
	fprintf fmt "@[<hov 0>(exists %a:%a.@ %a@])"
	  pp_var v pp_type (Var.var_type v) predicate p
    | Plet (x,v,p) ->
	fprintf fmt "@[<hov 0>(let %a =@[<hov 2>@ %a@ in@]@ %a@])"
	  pp_var x pp_term v predicate p
    | Pnamed (n, p) ->
	fprintf fmt "@[%s: %a@]" n predicate p

let fpp_item predicate pp_tau fmt x =
  function
    | Formula.Name k ->
	fprintf fmt "function () %s: name = mk_name(%d)@\n" x k
    | Formula.Cons k ->
	fprintf fmt "function %s () : int = %d@\n" x k
    | Formula.Function ([], t) ->
	fprintf fmt "logic %s: %a@\n" x pp_tau t
    | Formula.Function (tl, t) ->
	fprintf fmt "logic %s: @[<hov 0>%a -> %a@]@\n" x (pp_list pp_tau) tl pp_tau t
    | Formula.Predicate [] ->
	fprintf fmt "logic %s: prop@\n" x
    | Formula.Predicate tl ->
	fprintf fmt "logic %s: @[<hov 0>%a -> prop@]@\n" x (pp_list pp_tau) tl
    | Formula.Axiom p ->
	fprintf fmt "@[<hv 2>axiom %s:@ %a@]@\n" x predicate p
    | Formula.Type 0 ->
	fprintf fmt "type %s@\n" x
    | Formula.Type 1 ->
	fprintf fmt "type 'a %s@\n" x
    | Formula.Type n ->
	fprintf fmt "@[<hov 2>type ('a" ;
	for k=2 to n do fprintf fmt ",%c" (char_of_int (int_of_char 'a'+k-1)) done ;
	fprintf fmt ") %s@]@\n" x

let fpp_header fmt d =
  begin
    d.d_title fmt ;
    Format.pp_print_newline fmt () ;
    ( match d.d_source with
	| Some { Log.src_file=f ; Log.src_line=k } ->
	    Format.fprintf fmt "%s:%d: " f k
	| None -> () ) ;
    d.d_descr fmt ;
  end

let fpp_decl predicate pp_tau fmt d =
  begin
    fpp_header fmt d ;
    Format.pp_print_newline fmt () ;
    fpp_item predicate pp_tau fmt d.d_name d.d_item ;
    Format.pp_print_newline fmt () ;
  end

let fpp_goal predicate fmt x p =
  fprintf fmt "@[<hv 2>goal %s:@ %a@]@." x predicate p

(* -------------------------------------------------------------------------- *)
(* --- Pretty Print                                                       --- *)
(* -------------------------------------------------------------------------- *)

module Create
  (T : sig
     type d
     val pp_term : formatter -> d d_term -> unit
   end) =
struct

  let pp_tau = Datalib.Tau.pp_tau
  let pp_term = T.pp_term

  let rec pp_pred_atom fmt p =
    epp_pred_atom {
      pp_type = pp_tau ;
      pp_term = T.pp_term ;
      pp_pred = pp_pred_atom ;
    } fmt p

  let pp_pred_vbox fmt p =
    epp_pred_vbox {
      pp_type = pp_tau ;
      pp_term = T.pp_term ;
      pp_pred = pp_pred_atom ;
    } fmt p

  let rec pp_pred_debug fmt p =
    fpp_pred pp_pred_debug T.pp_term pp_tau fmt p

  let pp_pred fmt p =
    if Wp_parameters.debug_atleast 1 then
      pp_pred_debug fmt p
    else
      pp_pred_vbox fmt p

  let pp_decl fmt d = fpp_decl pp_pred pp_tau fmt d
  let pp_goal fmt x g = fpp_goal pp_pred fmt x g

end
