(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* ------------------------------------------------------------------------ *)
(* ---  FOL Implementation with Qed                                     --- *)
(* ------------------------------------------------------------------------ *)

open Qed
open Qed.Logic
open Ctypes
open Formula

(* ------------------------------------------------------------------------ *)
(* ---  Qed Types                                                       --- *)
(* ------------------------------------------------------------------------ *)

let tau_of = function Model t | Acsl(t,_) -> t

module Tau : Logic.Type with type tau = Formula.kind 
  =
struct

  type tau = Formula.kind
  type data = string

  let prop = Model Formula.Boolean
  let int = Model Formula.Integer
  let real = Model Formula.Real
  let data ts s = Model (Formula.ADT(s,List.map tau_of ts))
  let rec t_sort = function
    | Formula.Integer -> Logic.Int
    | Formula.Real -> Logic.Real
    | Formula.Boolean -> Logic.Prop
    | Formula.Set t -> Logic.Data([t_sort t],"set")
    | Formula.Record _ -> Logic.Data([Logic.Data([],"data")],"farray")
    | Formula.Array arr ->
	let t = Datalib.Tau.tau_of_logic_ctype arr.arr_element in
	Logic.Data([t_sort t],"farray")
    | Formula.Pointer _ -> Logic.Data([],"pointer")
    | Formula.ADT(s,ts) ->
	Logic.Data(List.map t_sort ts,s)
	  
  let sort k = t_sort (tau_of k)

  let basename k = match tau_of k with
    | Formula.Integer -> "x"
    | Formula.Real -> "z"
    | Formula.Boolean -> "p"
    | Formula.Set _ -> "S"
    | Formula.Record _ -> "R"
    | Formula.Array _ -> "A"
    | Formula.Pointer _ -> "P"
    | Formula.ADT(s,_) -> s

  let name k = match tau_of k with
    | Formula.Integer -> "int"
    | Formula.Real -> "real"
    | Formula.Boolean -> "prop"
    | Formula.Set _ -> "set"
    | Formula.Record _ -> "record"
    | Formula.Array _ -> "array"
    | Formula.Pointer _ -> "pointer"
    | Formula.ADT(s,_) -> s

  type t = kind
  let hash t = Hashtbl.hash (basename t)
  let pretty fmt k = Datalib.Tau.pp_tau fmt (tau_of k)

  let equal t1 t2 = sort t1 = sort t2
  let compare t1 t2 = Pervasives.compare (sort t1) (sort t2)

end

(* ------------------------------------------------------------------------ *)
(* ---  Qed Functions and Predicates                                    --- *)
(* ------------------------------------------------------------------------ *)

let categories : (string,Logic.category) Hashtbl.t = Hashtbl.create 31
let set_category = Hashtbl.replace categories

module Fun : Logic.Function with type f = string
  =
struct

  type f = string
  type t = string

  let id _ = assert false
  let category f =
    try Hashtbl.find categories f
    with Not_found -> Logic.Function

  let name f = f
  let pretty = Format.pp_print_string
  let hash = Hashtbl.hash
  let compare = String.compare
  let equal = (=)

end

(* ------------------------------------------------------------------------ *)
(* ---  Vernacular                                                      --- *)
(* ------------------------------------------------------------------------ *)

module T = Qed.Term.Make(Tau)(Fun)

module V =
struct

  type 'a term = T.term
  type pred = T.term

  let e_true = T.e_true
  let e_false = T.e_false
  let e_int = T.e_int

  let e_call = T.e_fun
  let p_call f es = set_category f Logic.Property ; T.e_fun f es
  let wrap t = t
  let unwrap t = t

end

module F =
struct

  include V

  let unop f a = T.e_fun f [a]
  let binop f a b = T.e_fun f [a;b]

  let e_float f = T.e_r (R.of_string (string_of_float f))
  let e_icst s = T.e_z (Z.of_string s)
  let e_rcst s = T.e_r (R.of_string s)
  let e_int64 k = T.e_z (Z.of_string (Int64.to_string k))
  let e_ineg = T.e_opp
  let e_rneg = T.e_opp
  let e_iop = function
    | Iadd -> T.e_add
    | Isub -> T.e_sub
    | Imul -> T.e_mul
    | Idiv -> T.e_div
    | Imod -> T.e_mod
  let e_rop = function
    | Radd -> T.e_add
    | Rsub -> T.e_sub
    | Rmul -> T.e_mul
    | Rdiv -> binop "div_real"
  let e_cmp = function
    | Ceq -> T.e_eq
    | Cneq -> T.e_neq
    | Clt -> T.e_lt
    | Cleq -> T.e_leq
  let e_icmp = e_cmp
  let e_rcmp = e_cmp
  let p_icmp = e_cmp
  let p_rcmp = e_cmp
  let e_bnot   = unop "bnot"
  let e_band   = binop "band"
  let e_bor    = binop "bor"
  let e_bxor   = binop "bxor"
  let e_lshift = binop "lshift"
  let e_rshift = binop "rshift"

  let integer_of_real = unop "truncate_real_to_int"
  let real_of_integer = unop "real_of_int"

  let e_bool c = T.e_fun "ite" [c;T.e_int 1;T.e_int 0]
  let e_cond c a b = T.e_fun "ite" [c;a;b]
  let e_not = T.e_not
  let e_and a b = T.e_and [a;b]
  let e_or a b = T.e_or [a;b]

  let p_cond c a b = T.e_and [T.e_imply [c] a ; T.e_imply [T.e_not c] b]
  let p_true = T.e_true
  let p_false = T.e_false
  let p_bool x = x
  let p_not = T.e_not
  let p_and a b = T.e_and [a;b]
  let p_or a b = T.e_or [a;b]
  let p_xor = T.e_neq
  let p_implies a b = T.e_imply [a] b
  let p_iff = T.e_eq
  let p_eq = T.e_eq
  let p_neq = T.e_neq
  let p_conj = T.e_and

  let is_true e = e == T.e_true
  let is_false e = e == T.e_false

  type var = T.var
  type pool = T.pool
      
  let pool = T.pool
  let fresh pool _ k = T.fresh pool k
  let freshen pool x = T.fresh pool (T.type_of_var x)
  let var = T.e_var
  let eq_var x y = (T.e_var x) == (T.e_var y)
  let name_of_var = T.name_of_var
  let kind_of_var = T.type_of_var
  let tau_of_var x = tau_of (T.type_of_var x)
  let term_has_var = T.has_vars
  let pred_has_var = T.has_vars

  let p_forall = T.e_forall
  let p_exists = T.e_exists

  let primitive = function
    | True | False | Kint _ | Kreal _ | Bvar _  | Fun(_,[]) -> true
    | _ -> false
	
  let subst x v t = 
    if not (T.is_closed t) then
      Wp_parameters.fatal ~current:true 
	"Qed: substitution required with non-closed term" ;
    let rec apply s tx v e =
      if primitive (T.repr e) then e else
	if e == tx then v else
	  try T.Tmap.find e !s
	  with Not_found ->
	    let e' = T.e_map ~closure:true (apply s tx v) e in
	    s := T.Tmap.add e e' !s ; e'
    in apply (ref T.Tmap.empty) (T.e_var x) v t
	 
  let e_rename xys t =
    let txys = List.map (fun (x,y) -> T.e_var x,T.e_var y) xys in
    let rec apply s txys e =
      if primitive (T.repr e) then e else
	try List.assq e txys
	with Not_found ->
	  try T.Tmap.find e !s
	  with Not_found ->
	    let e' = T.e_map ~closure:true (apply s txys) e in
	    s := T.Tmap.add e e' !s ; e'
    in apply (ref T.Tmap.empty) txys t

  let e_subst _ = subst
  let p_subst _ = subst

  let equal_terms = T.equal
    
  (* ------------------------------------------------------------------------ *)
  (* ---  Pretty Printers                                                 --- *)
  (* ------------------------------------------------------------------------ *)

  let pp_var fmt x = T.pretty fmt (T.e_var x)
  let pp_term = T.pretty
  let pp_pred = T.pretty
  let pp_decl = Fol_gformula.fpp_decl pp_pred Datalib.Tau.pp_tau
  let pp_goal = Fol_gformula.fpp_goal pp_pred

end

module E =
struct

  type pred = T.term

  module S = Shared.Make(T)

  let pp_pred fmt p = Qed.Why_output.pp_prop fmt (S.output_prop p)
  let pp_tau fmt t = Qed.Why_output.pp_tau fmt (S.output_type (Model t))

  let export_goal = Fol_gformula.fpp_goal pp_pred
  let export_decl = Fol_gformula.fpp_decl pp_pred pp_tau
    
end
