(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* ------------------------------------------------------------------------ *)
(* ---  Splitter for FOL                                                --- *)
(* ------------------------------------------------------------------------ *)

open Fol

module Env = Map.Make(Fol.Var)

let is_zone = function
  | Tapp(s,_) -> List.mem s
      [ "zunion" ; "zempty" ; "zrange" ; 
	"zrange_of_addr" ; "zrange_of_addr_range" ]
  | _ -> false

let rec e_unfold env = function
  | Tvar x as t -> (try Env.find x env with Not_found -> t)
  | Tlet(x,a,b) ->
      let a0 = e_unfold env a in
      if is_zone a0 
      then e_unfold (Env.add x a0 env) b
      else e_let x a0 (e_unfold (Env.remove x env) b)
  | Tapp(f,xs) -> e_app f (List.map (e_unfold env) xs)
  | Tif(c,a,b) -> e_if (e_unfold env c) (e_unfold env a) (e_unfold env b)
  | Tconst _ as t -> t
  | Tdata _ -> assert false

let rec p_unfold env = function
  | Plet(x,t,p) ->
      let t0 = e_unfold env t in
      if is_zone t0
      then p_unfold (Env.add x t0 env) p
      else p_let x t0 (p_unfold (Env.remove x env) p)
  | Pnamed(a,p) -> Pnamed(a,p_unfold env p)
  | (Ptrue | Pfalse) as p -> p
  | Papp(f,es) -> p_app f (List.map (e_unfold env) es)
  | Pimplies(p,q) -> p_implies (p_unfold env p) (p_unfold env q)
  | Pif(c,p,q) -> p_if (e_unfold env c) (p_unfold env p) (p_unfold env q)
  | Pand(p,q) -> p_and (p_unfold env p) (p_unfold env q)
  | Por(p,q) -> p_and (p_unfold env p) (p_unfold env q)
  | Piff(p,q) -> p_iff (p_unfold env p) (p_unfold env q)
  | Pnot p -> p_not (p_unfold env p)
  | Pforall(x,p) -> p_forall x (p_unfold env p)
  | Pexists(x,p) -> p_exists x (p_unfold env p)

(* ------------------------------------------------------------------------ *)
(* --- Case Split                                                       --- *)
(* ------------------------------------------------------------------------ *)
      
let rec flatten xs = function
  | Tapp( "zunion" , es ) -> List.fold_left flatten xs es
  | Tapp( "zempty" , [] ) -> xs
  | t -> t::xs

let rec zunion = function
  | [z] -> z
  | [] -> e_app "zempty" []
  | z::zs -> e_app "zunion" [ z ; zunion zs ]

let included a b =
  let zas = flatten [] a in
  let zbs = flatten [] b in
  let zb = zunion zbs in
  List.fold_left
    (fun acc z ->
       if List.exists (Fol.bad_eq_terms z) zbs
       then acc
       else Bag.add (p_app "included" [z;zb]) acc)
    Bag.empty zas

let rec dispatch = function
  | Ptrue -> Bag.empty
  | Papp( "included" , [a;b] ) -> included a b
  | (Pfalse|Por _|Piff _|Papp _) as p -> Bag.elt p
  | Pand(p,q) -> Bag.concat (dispatch p) (dispatch q)
  | Pimplies(h,p) -> Bag.map (fun p -> p_implies h p) (dispatch p)
  | Pif(c,p,q) -> 
      Bag.concat
	(Bag.map (fun p -> p_implies (p_eq c e_true) p) (dispatch p))
	(Bag.map (fun q -> p_implies (p_eq c e_false) q) (dispatch q))
  | Pnamed(a,p) ->
      Bag.map (fun p -> Pnamed(a,p)) (dispatch p)
  | Pnot p ->
      Bag.map p_not (dispatch p)
  | Pforall(x,p) ->
      Bag.map (p_forall x) (dispatch p)
  | Pexists(x,p) ->
      Bag.map (p_exists x) (dispatch p)
  | Plet(x,t,p) ->
      Bag.map (p_let x t) (dispatch p)

(* ------------------------------------------------------------------------ *)
(* --- Splitter Interface                                               --- *)
(* ------------------------------------------------------------------------ *)

type pred = unit Fol.d_pred

let split meth p =
  match meth with
    | Mcfg.EffectAssigns -> dispatch (p_unfold Env.empty p)
    | _ -> dispatch p

  
