(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* -------------------------------------------------------------------------- *)
(** Logic Formulae                                                            *)
(* -------------------------------------------------------------------------- *)

(** {2 Pure types} *)

type boolean  (** The set of two elements [{true,false}]. *)
type integer  (** Natural signed integers ({b Z}). *)
type real     (** Real numbers ({b R}). *)
type abstract (** Abstract Data Types (user-defined WHY-types). *)

type m_name    (** type name [data_lib.why] *)
type m_format  (** type 'a format [data_lib.why] *)
type m_pointer (** type pointer [data_lib.why] *)
type m_array   (** type 'a farray [array.why] *)
type m_record  (** type record [data_lib.why] *)
type m_urecord (** type urecord [data_lib.why] *)
type m_set     (** type 'a set [data_lib.why] *)
type m_zone    (** type of elementary regions *)

(** {2 Logic types }*)

type tau = 
  | Integer 
  | Real 
  | Boolean
  | Pointer
  | Set     of tau 
  | Record  of Cil_types.compinfo
  | Array   of Ctypes.arrayinfo
  | ADT     of string * (tau list)

type kind = 
  | Model of tau 
  | Acsl of tau * Cil_types.logic_type

(** {2 Arithmetics Operators} *)

type integer_op = Iadd | Isub | Imul | Idiv | Imod 
type real_op = Radd | Rsub | Rmul | Rdiv
type cmp_op = Ceq | Cneq | Clt | Cleq

(** {2 Declarations} *)

type section =
  | S_name
  | S_typedef
  | S_userdef
  | S_goals

type 'a item =
  | Type of int
  | Name of int
  | Cons of int
  | Function of tau list * tau
  | Predicate of tau list
  | Axiom of 'a

type 'a declaration = 
    {
      d_section : section ;
      d_name    : string ;
      d_title   : (Format.formatter -> unit) ;
      d_descr   : (Format.formatter -> unit) ;
      d_source  : Log.source option ;
      d_item    : 'a item ;
    }

(** {2 Signature for logic formulae} *)

module type Vernacular =
sig

  type 'a term
  type pred

  val e_int  : int -> integer term
  val e_call : string -> abstract term list -> abstract term
  val p_call : string -> abstract term list -> pred
  val wrap   : 'a term -> abstract term
  val unwrap : abstract term -> 'a term

end

module type S =
sig

  include Vernacular

  (** {2 Terms} *)

  val e_true    : boolean term
  val e_false   : boolean term

  val e_int     : int -> integer term
  val e_float   : float -> real term
  val e_icst    : string -> integer term
  val e_rcst    : string -> real term
  val e_int64   : int64 -> integer term

  (** {2 Arithmetics} *)

  val e_ineg : integer term -> integer term
  val e_rneg : real term -> real term
  val e_iop  : integer_op -> integer term -> integer term -> integer term
  val e_rop  : real_op -> real term -> real term -> real term
  val e_icmp : cmp_op -> integer term -> integer term -> boolean term
  val e_rcmp : cmp_op -> real term -> real term -> boolean term
  val p_icmp : cmp_op -> integer term -> integer term -> pred
  val p_rcmp : cmp_op -> real term -> real term -> pred
    
  val e_bnot   : integer term -> integer term
  val e_band   : integer term -> integer term -> integer term
  val e_bor    : integer term -> integer term -> integer term
  val e_bxor   : integer term -> integer term -> integer term
  val e_lshift : integer term -> integer term -> integer term
  val e_rshift : integer term -> integer term -> integer term

  val integer_of_real : real term -> integer term
  val real_of_integer : integer term -> real term

  (** {2 Booleans} *)

  val e_bool : boolean term -> integer term
  val e_not  : boolean term -> boolean term
  val e_and  : boolean term -> boolean term -> boolean term
  val e_or   : boolean term -> boolean term -> boolean term
    
  (** {2 Conditional} *)

  val e_cond : boolean term -> 'a term -> 'a term -> 'a term
  val p_cond : boolean term -> pred -> pred -> pred

  (** {2 Predicates} *)

  val p_true    : pred
  val p_false   : pred
  val p_bool    : boolean term -> pred
  val p_and     : pred -> pred -> pred
  val p_or      : pred -> pred -> pred
  val p_xor     : pred -> pred -> pred
  val p_not     : pred -> pred
  val p_implies : pred -> pred -> pred
  val p_iff     : pred ->pred -> pred
  val p_eq      : 'a term -> 'a term -> pred  
  val p_neq     : 'a term -> 'a term -> pred  
  val p_conj    : pred list -> pred
  val p_disj    : pred list -> pred
  val p_named   : string -> pred -> pred

  (** {2 Utilities} *)
  
  val is_true : pred -> bool 
  val is_false : pred -> bool
  val huge_term : int -> 'a term -> bool
  val huge_pred : int -> pred -> bool

  (** {2 Variables} 

	Pools are used to generate fresh free variables.  Do not mix
	non-closed terms from different pools.  
  *)

  type var
  type pool
  
  val pool : unit -> pool
  val fresh : pool -> string -> kind ->  var
  val freshen : pool -> var -> var

  val var : var -> 'a term
  val eq_var : var -> var -> bool
  val name_of_var : var -> string
  val tau_of_var : var -> tau
  val kind_of_var : var  -> kind

  val term_has_var : var list -> 'a term -> bool
  val pred_has_var : var list -> pred -> bool

  val term_closed : 'a term -> bool
  val pred_closed : pred -> bool

  val p_forall : var list -> pred -> pred
  val p_exists : var list -> pred -> pred
  val p_subst  : (var -> var option) -> var -> 'a term -> pred -> pred
  val e_subst  : (var -> var option) -> var -> 'a term -> 'b term -> 'b term

  (** Requires domain to be disjoint from co-domain *)
  val e_rename : (var * var) list -> 'a term -> 'a term

  (** Returns true when the two terms are syntactically equals *)
  val equal_terms : 'a term -> 'a term -> bool

  (** {3 Alpha conversion} *)

  (** Maping from old var to new var *)
  type alpha
  (** Empty mapping *)
  val empty_alpha : alpha
  val fold_alpha : (var -> var -> 'a -> 'a) -> alpha -> 'a -> 'a

  (** [alpha', p' = p_more_alpha_cv alpha p] build p' from p by renaming
  * all the variable v into v' according to the mapping alpha.
  * Add new mappings for the variables that are not already mapped. *)
  val p_more_alpha_cv : alpha -> pred -> alpha * pred

  (** easier to use when doing a simple alpha conversion.
  * @return the new predicate and the newly created variables. *)
  val p_alpha_cv : pred -> var list * pred

  (** {2 Pretty printers} *)

  val pp_var :
    Format.formatter -> var -> unit
  val pp_term : 
    Format.formatter -> 'a term -> unit
  val pp_pred : 
    Format.formatter -> pred -> unit
  val pp_decl :
    Format.formatter -> pred declaration -> unit
  val pp_goal :
    Format.formatter -> string -> pred -> unit

end
