(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* -------------------------------------------------------------------------- *)
(* --- Functional Variable Memory Model Functor                           --- *)
(* -------------------------------------------------------------------------- *)

module WpLog = Wp_parameters
open Cil_types
open Cil_datatype
open Formula
open Ctypes
open Clabels
open Mdata

module Create
  (M:Mwp.S)
  =
struct

  module Lookup =
  struct

    module F = M.D.F
    module D = M.D

    type poffset =
      | Ofield of  fieldinfo
      | Oindex of  D.integer * Ctypes.c_object

    type path = M.mem * varinfo * poffset list * Ctypes.c_object

    let get_offset p = function
      | Ofield f -> D.acc_field (D.F.unwrap p) f
      | Oindex (i,_) -> D.acc_index (D.F.unwrap p) i

    let access v poff = List.fold_left get_offset v poff

    let mloc_of_path m x poff =
      List.fold_left
	(fun loc offset ->
	  match offset with
	    | Ofield f -> M.field loc f
	    | Oindex(k,te) -> M.index loc te k
	) (M.cvar m x) poff

    let pp_path fmt (_m,vinfo,loff,_) =
      Format.fprintf fmt "PATH:%a of %d" !Ast_printer.d_var vinfo
	(List.length loff)

  end

  module Model =
  struct

    open Lookup

    module F = M.D.F
    module A = M.A
    module R = M.R
    module D = M.D

    type loc = Path of path | Mloc of M.loc

    let loc_of_pointer ty p = Mloc (M.loc_of_pointer ty p)
    let pointer_of_loc = function
      | Path _ -> Wp_parameters.fatal "Pointer of functional variable"
      | Mloc l -> M.pointer_of_loc l

    let pp_loc fmt = function
      | Mloc l -> M.pp_loc fmt l
      | Path p -> pp_path fmt p

    let cast_loc_to_int ty l ty2 =
      match l with
	| Mloc l -> M.cast_loc_to_int ty l ty2
	| _ -> Wp_parameters.fatal "int of functional variable"

    let cast_int_to_loc ty i ty2 = Mloc (M.cast_int_to_loc ty i ty2)

  end

  open Model

  include Datalib.Cvalues(Model)

  (* ------------------------------------------------------------------------ *)
  (* ---  Env and Memory                                                  --- *)
  (* ------------------------------------------------------------------------ *)

  type env = (Cil_types.varinfo * (F.var * Ctypes.c_object)) list ref

  type mem = { mem : M.mem ; env : env }

  type fun_frame = (Clabels.c_label * env) list ref

  let new_fun_frame opt_here opt_pre opt_post =
    let may_cons acc lab x =
      match x with
	| None -> acc
	| Some x -> (lab,x)::acc
    in
    let frame =
      (may_cons
	 (may_cons
	    (may_cons [] Clabels.Post opt_post)
	    Clabels.Pre opt_pre)
	 Clabels.Here opt_here)
    in
    ref frame

  let create () = ref []

  let env_at frame lab =
    try List.assoc lab !frame
    with Not_found ->
      let env_lab = create () in
      frame := (lab,env_lab)::!frame ; env_lab

  type frame =
      {
	mframe : M.frame;
	fframe : fun_frame;
      }
	
  let opt_mem = function None -> None | Some m -> Some m.mem
  let opt_env = function None -> None | Some m -> Some m.env

  let new_frame kf ~m_here ~m_pre ~m_post ret =
    {
      mframe = M.new_frame kf
	(opt_mem m_here) (opt_mem m_pre) (opt_mem m_post) ret;
      fframe = new_fun_frame
	(opt_env m_here) (opt_env m_pre) (opt_env m_post);
    }

  let mem_at frame label =
    {
      mem = M.mem_at frame.mframe label ;
      env = env_at frame.fframe label;
    }

  (* ------------------------------------------------------------------------ *)
  (* --- Locations                                                        --- *)
  (* ------------------------------------------------------------------------ *)

  let get_funvar env vinfo =
    try
      let cell =
	List.find (fun (v,_) -> Varinfo.equal v vinfo) !env
      in
      fst (snd cell)
    with Not_found ->
      let tx = object_of vinfo.vtype in
      let x =
	D.fresh vinfo.vname
	  (Vacsl (Ctype vinfo.vtype))
      in
      env := (vinfo,(x,tx))::!env ; x

  let cvar m vinfo =
     if Vaddr.is_funvar vinfo
     then Path (m.mem, vinfo,[],object_of vinfo.vtype)
     else Mloc (M.cvar m.mem vinfo)

  let shift l ty i =
    match l with
      | Mloc l -> Mloc (M.shift l ty i)
      | Path _ -> Wp_parameters.fatal "Shift of a functional variable"

  let index l ty i =
    match l with
      | Mloc l -> Mloc (M.index l ty i)
      | Path (m,x,off,_obj) -> Path(m,x, off@[Lookup.Oindex (i,ty)], ty)

  let field l finfo =
    match l with
      | Mloc l ->
	  Mloc (M.field l finfo)
      | Path (m,x,off,_obj) ->
	  let ty = object_of finfo.ftype in
	  Path(m,x, off@[Lookup.Ofield finfo],ty )

  let startof l ty =
    match l with
      | Mloc l -> Mloc (M.startof l ty)
      | Path _ -> Wp_parameters.fatal "startof a functional variable"

  (* ------------------------------------------------------------------------ *)
  (* --- Pointers                                                         --- *)
  (* ------------------------------------------------------------------------ *)

  let cast_loc_to_loc t1 t2 = function
    | Mloc l -> Mloc (M.cast_loc_to_loc t1 t2 l)
    | Path _ ->
	Wp_parameters.not_yet_implemented
	  "Cast of over a hoare-variable (try -wp-nofunvar)"

  (* ------------------------------------------------------------------------ *)
  (* --- Load                                                             --- *)
  (* ------------------------------------------------------------------------ *)

  let fun_load e (_m,x,poff,obj) =
    let xv = F.var(get_funvar e x) in
    let vload =
      match poff with
	| [] -> xv
	| _ -> Lookup.access xv poff
    in value_of_logic obj vload

  let loc_of_path m p =
   match fun_load m.env p with
     | V_pointer (_,Mloc l) -> l
     | _ -> assert false

  let array_of_path env p =
    match fun_load env p with
      | V_array (_,t) -> t
      | _ -> assert false

  let value_of_mvalue = function
    | M.V_int (i,t) -> V_int (i,t)
    | M.V_float (f,t) -> V_float (f,t)
    | M.V_array (a,t) -> V_array(a,t)
    | M.V_record (c,t) -> V_record (c,t)
    | M.V_union (c,t) -> V_union (c,t)
    | M.V_pointer (te,l) -> V_pointer (te,Mloc l)


  let mvalue_of_value = function
    | V_int (i,t) -> M.V_int (i,t)
    | V_float (f,t) -> M.V_float (f,t)
    | V_array (a,t) -> M.V_array(a,t)
    | V_record (c,t) -> M.V_record (c,t)
    | V_union (c,t) -> M.V_union (c,t)
    | V_pointer (te,Mloc l) -> M.V_pointer (te, l)
    | V_pointer (_,Path _) -> assert false

  let load m cv l =
    match l with
      | Mloc l -> value_of_mvalue (M.load m.mem cv l)
      | Path p -> fun_load m.env p

  (* ------------------------------------------------------------------------ *)
  (* --- Zones                                                            --- *)
  (* ------------------------------------------------------------------------ *)
(*
  type zone =
    | FunPath of Lookup.path
    | FunRange of Lookup.path * D.interval
    | MemZone of M.zone
*)

  let mloc_of_loc = function
    | Path (m,x,p,_) -> Lookup.mloc_of_path m x p
    | Mloc loc -> loc

  let massigned = function
    | D.Aloc(te,l) -> D.Aloc(te,mloc_of_loc l)
    | D.Arange(te,l,rg) -> D.Arange(te,mloc_of_loc l,rg)
	  
  type m_dzone = M.m_dzone
  type dzone = M.dzone
  let tau_of_dzone = M.tau_of_dzone

  let dzone_assigned z = M.dzone_assigned (massigned z)
  let dzone_subset = M.dzone_subset
  let dzone_union = M.dzone_union
  let dzone_empty = M.dzone_empty

  let effect_supported = M.effect_supported

  (* ------------------------------------------------------------------------ *)
  (* --- Region                                                           --- *)
  (* ------------------------------------------------------------------------ *)

  module Z = Region.Create(D)
    (struct
       module F = F
       type addr = D.integer
       let pp_addr fmt a = Format.fprintf fmt "Lptr %a" F.pp_term a
       let closed = F.term_closed
       let equal = F.p_eq
       let fresh pool =
	 let p = F.fresh pool "X" (Model Integer) in
	 p , F.var p
     end)

  type region = Z.region * M.region
	  
  let pp_region fmt (f,m) =
    Z.pp_region fmt f ; M.pp_region fmt m

  let rec funvar_offset ~atomic base path offset range =
    match offset , range with
      | [] , None -> Z.path base path
      | [] , Some(te,rg) -> Z.path base (Z.p_interval ~atomic path te rg)
      | Lookup.Ofield f :: offset , _ ->
	  funvar_offset ~atomic base (Z.p_field ~atomic path f) offset range
      | Lookup.Oindex(k,te) :: offset , _ ->
	  funvar_offset ~atomic base (Z.p_index ~atomic path te ~offset:k) offset range

  let funvar_path x offset range =
    let base = Z.cvar (D.Xindex.get_ind x) x in
    let te = Ctypes.object_of x.vtype in
    let atomic = Z.is_atomic te in
    match offset,range with
      | [],None -> Z.block base
      | _ -> funvar_offset ~atomic base (Z.p_block ~atomic te) offset range

  let region_assigned = function
    | D.Aloc(_,Path(_,x,offset,_)) -> 
	funvar_path x offset None , M.region_empty ()
    | D.Arange(te,Path(_,x,offset,_),rg) -> 
	funvar_path x offset (Some(te,rg)) , M.region_empty ()
    | D.Aloc(te,Mloc l) -> 
	Z.empty , M.region_assigned(D.Aloc(te,l))
    | D.Arange(te,Mloc l,rg) -> 
	Z.empty , M.region_assigned(D.Arange(te,l,rg))
    
  let region_empty () = 
    Z.empty , M.region_empty ()
  let region_union (fr1,mr1) (fr2,mr2) = 
    Z.union fr1 fr2 , M.region_union mr1 mr2
  let region_included (fr1,mr1) (fr2,mr2) =
    F.p_and (Z.included fr1 fr2) (M.region_included mr1 mr2)
    
  let region_separated (fr1,mr1) (fr2,mr2) =
    F.p_and (Z.separated fr1 fr2) (M.region_separated mr1 mr2)

  let region_fingerprint pool (fr,mr) =
    let xs,fc,fr = Z.fingerprint pool fr in
    let ys,mc,mr = M.region_fingerprint pool mr in
    xs @ ys , F.p_and fc mc , (fr,mr)
    
  let region_supported = M.region_supported

  (* ------------------------------------------------------------------------ *)
  (* --- Result and Exit status                                           --- *)
  (* ------------------------------------------------------------------------ *)

  let result frame = M.result frame.mframe
  let exit_status frame = M.exit_status frame.mframe
  let get_exit_status frame = M.get_exit_status frame.mframe
  let set_exit_status frame x =
    {
      fframe = frame.fframe ;
      mframe = M.set_exit_status frame.mframe x
    }

  (* ------------------------------------------------------------------------ *)
  (* ---  Pointers                                                        --- *)
  (* ------------------------------------------------------------------------ *)

  let loc_of_funloc = function
    | Mloc l -> l
    | Path(m,x,p,_) -> Lookup.mloc_of_path m x p

  let base_address m l = Mloc (M.base_address m.mem (loc_of_funloc l))
  let block_length m l = M.block_length m.mem (loc_of_funloc l)

  let valid m z = M.valid m.mem (massigned z)

  let separated z1 z2 = M.separated (massigned z1) (massigned z2)

  (* ------------------------------------------------------------------------ *)
  (* ---  Functional Closure                                              --- *)
  (* ------------------------------------------------------------------------ *)

  type fun_closure = Cil_types.varinfo * string

  type closure =
    | Fclos of fun_closure
    | Mclos of M.closure

  let pp_closure fmt = function
    | Mclos cl -> M.pp_closure fmt cl
    | Fclos (vinfo,s) -> Format.fprintf fmt "%a@@%s" !Ast_printer.d_var vinfo s

  let collect_signature label varinfo fvar (variables,closures) =
    fvar :: variables , (Fclos (varinfo,label))::closures

  let userdef_fun_signature frame =
    List.fold_left
      (fun s (clabel,env) ->
	 match clabel with
	   | Clabels.LabelParam label ->
	       List.fold_left
		 (fun s (v,(x,_))  -> collect_signature label v x s)
		 s !env
	   | _ ->
               Wp_parameters.fatal
                 "Non-logic label in user-defined function or predicate"
      ) ([],[]) !frame

  let userdef_fun_closure env (vinfo,_) =
    F.var(get_funvar env vinfo)

  let userdef_frame pool =
    {
      fframe = ref [] ;
      mframe = M.userdef_frame pool ;
    }

  let userdef_signature frame tused pused =
    let fp,fclos = userdef_fun_signature frame.fframe in
    let mp,mclos = M.userdef_signature frame.mframe tused pused in
    fp@mp , fclos@(List.map (fun c -> Mclos c) mclos)

  let label_at_closure = function
    | Fclos(_,label) -> label
    | Mclos mc -> M.label_at_closure mc

  let userdef_closure env = function
    | Fclos  fc -> userdef_fun_closure env.env fc
    | Mclos  mc -> M.userdef_closure env.mem mc

  (* ------------------------------------------------------------------------ *)
  (* ---  Labels & Quantification                                         --- *)
  (* ------------------------------------------------------------------------ *)

  let fupdate_at_label frame lab p =
    let env_here = env_at frame Clabels.Here in
    let env_lab = env_at frame lab in
    List.fold_left
      (fun p (vinfo,(x_var,_)) ->
	 let x_here = get_funvar env_here vinfo in
	 D.subst x_var (F.var x_here) p
      ) p !env_lab

  let update_at_label frame label p =
    M.update_at_label frame.mframe label
      (fupdate_at_label frame.fframe label p)

  let close_env env p =
    D.forall
      (List.map (fun (_,(f,_)) -> f) !env) p

  let fquantify_at_label frame lab p =
    close_env (env_at frame lab) p

  let quantify_at_label frame label p =
    M.quantify_at_label frame.mframe label
      (fquantify_at_label frame.fframe label p)

  (* ------------------------------------------------------------------------ *)
  (* ---  Assignments                                                     --- *)
  (* ------------------------------------------------------------------------ *)

  let rec update phi current poffset =
    match poffset with
      | [] -> phi current
      | poff::m ->
	  let v = Lookup.get_offset current poff in
	  let r = update phi v m in
	  begin
	    match poff with
	      | Lookup.Ofield f ->
		  F.wrap (D.upd_field (F.unwrap current) f r)
	      | Lookup.Oindex (i,_obj) ->
		  F.wrap (D.upd_index (F.unwrap current) i r)
	  end

  let store frame (_m,r,off,_) v p =
    let e = env_at frame Clabels.Here in
    let x = get_funvar e r in
    let v' = update (fun _ -> logic_of_value v) (F.var x) off in
    D.subst x v' p

  let subst_lval frame obj loc v p =
    match loc with
      | Mloc l -> M.subst_lval frame.mframe obj l (mvalue_of_value v) p
      | Path ph -> store frame.fframe ph v p

  let subst_result frame v_opt p =
    let v = match v_opt with
      | Some v -> Some (mvalue_of_value v) | None -> None
    in
    M.subst_result frame.mframe v p

  (* ------------------------------------------------------------------------ *)
  (* ---  Zone Havoc                                                      --- *)
  (* ------------------------------------------------------------------------ *)

  let subst_havoc (frame:frame) = function
    | D.Aloc(_,Path(_,x,[],_)) ->
	let x = get_funvar (env_at frame.fframe Clabels.Here) x in
	let v = D.fresh "v" (D.vkind_of_var x) in
	[D.Fresh v;D.Update(x,fun _ -> F.var v)]
    | D.Aloc(_,Path(_,x,path,ty)) ->
	let x = get_funvar (env_at frame.fframe Clabels.Here) x in
	let v = D.fresh "v" (Vmodel (D.tau_of_object ty)) in
	let newterm (sigma : D.substitution) : abstract F.term =
	  F.wrap ( update (fun _ -> F.var v) (D.apply sigma (F.var x)) path )
	in
	[D.Fresh v;D.Update(x,newterm)]
    | D.Arange(_,Path(_,x,path,_),rg) ->
	let x = get_funvar (env_at frame.fframe Clabels.Here) x in
	let upd_range rg = fun array ->
	  F.wrap (D.set_range_index (F.unwrap array) rg)
	in
	let newterm (sigma : D.substitution) : abstract F.term =
	  F.wrap (update (upd_range rg) (D.apply sigma (F.var x)) path)
	in
	[D.Update(x,newterm)]
    | D.Aloc(te,Mloc l) -> M.subst_havoc frame.mframe (D.Aloc(te,l))
    | D.Arange(te,Mloc l,rg) -> M.subst_havoc frame.mframe (D.Arange(te,l,rg))

  let assigns_goal frame lab1 reg lab2 =
    (* Not very usefull !! *) 
    let r_mem =
      List.map
	(function
	   | D.Aloc(_,Path _) | D.Arange(_,Path _,_) ->
	       Wp_parameters.fatal "Proof of assigns-clause with hoare-region"
	   | D.Aloc(te,Mloc l) -> D.Aloc(te,l)
	   | D.Arange(te,Mloc l,rg) -> D.Arange(te,l,rg))
	reg
    in
    M.assigns_goal frame.mframe lab1 r_mem lab2

  let assigns_supported = false

  (* ------------------------------------------------------------------------ *)
  (* ---  Local Scope                                                     --- *)
  (* ------------------------------------------------------------------------ *)

  let local_scope frame lx b p =
    let mframe = frame.mframe in
    let mvars = List.filter (fun x -> not (Vaddr.is_funvar x)) lx in
    M.local_scope mframe mvars b p

end
