(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* -------------------------------------------------------------------------- *)
(** Library functions for Hoare memory model implemented
* in {!Hoare_mem}. Library functions are defined in
* {{:../../../share/why/hoare.why} Hoare WHY file}.
*
* Notice that it is also use in Caveat model and that is why it is built as a
* functor.
*)
(* -------------------------------------------------------------------------- *)

open Cil_types

module Create (D:Mdata.S)
  =
struct

  module Hlib = struct
  let addr : D.name -> D.pointer = D.e_app1 "m0_addr"

  let shift_pointer : D.pointer -> Formula.integer D.F.term -> D.pointer = 
    D.e_app2 ("shift_pointer")

  let shift_field : D.pointer-> D.name -> D.pointer =
    D.e_app2 ("shift_field")

  let shift_ufield : D.pointer -> D.name -> D.pointer =
    D.e_app2 ("shift_ufield")

  let shift_index : (D.pointer -> Formula.integer D.F.term -> D.pointer) =
    D.e_app2 ("shift_index")

  let havoc_index : (D.array -> Formula.integer D.F.term -> D.array) =
    D.e_app2 "havoc_index"

  let havoc_field : (D.record -> D.name -> D.record) =
    D.e_app2 "havoc_index"

  let havoc_ufield : (D.urecord -> D.name -> D.urecord) =
    D.e_app2 "havoc_ufield"

  let base : D.pointer -> D.pointer = D.e_app1 ("m0_base")

  let block_length : D.pointer -> Formula.integer D.F.term =
    D.e_app1 ("m0_block_length")

  let valid_pointer : D.pointer -> D.F.pred = D.p_app1 ("m0_valid_pointer") 

  let valid_range : D.pointer -> D.set -> D.F.pred =
    D.p_app2 ("m0_valid_range")

  let separated : D.pointer -> D.pointer -> D.F.pred = D.p_app2 ("m0_separated")

  let array_length : D.pointer -> Formula.integer D.F.term =
    D.e_app1 ("m0_array_length")
  end

  (** build [array_length t = sz] when we know how to interpet the size sz *)
  let get_array_length ctype =
    match ctype with
      | TArray (_t, Some {enode = Const sz}, _, _) ->
          begin
            match Ast_info.possible_value_of_integral_const sz with
              | Some  sz_int64 ->
                  let sz = Int64.to_string sz_int64 in
                  let sz = D.F.e_icst sz in
                    Some sz
              | _ -> None
          end
      | _ -> None

  let mk_v_array_length term sz =
    let array_length = Hlib.array_length term in
    let eq =  D.F.p_eq array_length sz in
      eq

  module Globals = D.Register
    (struct
       include D.Varinfo

     (** Hyp about the global arrays size: [array_length (addr (v)) = sz] *)
       let declare v _ =
         let sz = match get_array_length v.vtype with
           | None -> Wp_parameters.fatal "should be check before"
           | Some sz -> sz
         in
         let p = Hlib.addr (D.Xindex.get_ind v) in
         Formula.Axiom (mk_v_array_length p sz)

       let section = Formula.S_typedef
       let prefix = "Sizeof"
       let basename x = x.vname
       let clear () = ()
       let pp_descr fmt _x =
         Format.fprintf fmt "Array length"
     end)

end

(* -------------------------------------------------------------------------- *)
