(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* -------------------------------------------------------------------------- *)
(** Memory Model for Hoare.                                                   *)
(* -------------------------------------------------------------------------- *)

open Cil_types
open Cil_datatype

module Create
  (D:Mdata.S)
  (A:Mint.S   with module D = D)
  (R:Mfloat.S with module D = D)
  =
struct

  let unsupported = Wp_error.unsupported

  module Shared = Hoare_lib.Create (D)
  open Shared

  module F = D.F

  module Model = struct
    module D = D
    module A = A
    module R = R

    type epath =
      | Pfield of Cil_types.fieldinfo
      | Pidx of D.integer

    type loc =
      | L0addr of Cil_types.varinfo
      | L0depl of loc * epath
      | L0ptr of D.pointer (* in order to be able to have pointers *)
      | L0shift of loc * D.integer

    let rec pointer_of_loc loc : D.pointer =
      match loc with
	| L0addr v -> Hlib.addr (D.Xindex.get_ind v)
	| L0depl (l, Pidx i) -> Hlib.shift_index (pointer_of_loc l) i
	| L0depl (l, Pfield f) ->
            let lt = pointer_of_loc l in
            let tf = D.Findex.get_ind f in
            if f.fcomp.cstruct
            then Hlib.shift_field lt tf else Hlib.shift_ufield lt tf
	| L0ptr p -> p
	| L0shift (l,i) -> Hlib.shift_pointer (pointer_of_loc l) i

    let loc_of_pointer _ p = L0ptr p

    let rec pp_loc fmt l = match l with
      | L0addr v -> Format.fprintf fmt "@@%s" v.vname
      | L0depl (x, ch) ->
          Format.fprintf fmt "%a+%a" pp_loc x pp_epath ch
      | L0shift (p, i) ->
          Format.fprintf fmt "(%a+%a)" pp_loc p F.pp_term i
      | L0ptr p -> Format.fprintf fmt "%a" F.pp_term p
    and pp_epath fmt ch = match ch with
      | Pfield f -> Format.fprintf fmt ".%s" f.fname
      | Pidx i -> Format.fprintf fmt "(%a)" F.pp_term i

    let cast_loc_to_int _tp _p _ti = unsupported "cast from pointer to int"
    let cast_int_to_loc _ti _i _tp = unsupported "cast from int to pointer"

  end

  let cast_loc_to_loc ty1 ty2 _l =
    unsupported "cast from pointer (%a) to pointer (%a)"
      !Ast_printer.d_type ty1 !Ast_printer.d_type ty2

  include Datalib.Cvalues(Model)

  open Model

  let cvar _m vi =
    if vi.vglob && (get_array_length vi.vtype <> None) then Globals.define vi ;
    L0addr vi

  let field loc f = L0depl (loc, Pfield f)

  let index loc _te i = L0depl (loc, Pidx i)

  let shift loc _te i =
    match loc with
      | L0depl (loc, Pidx j) -> L0depl (loc,Pidx (F.e_iop Formula.Iadd j i))
      | L0shift (loc, j) -> L0shift (loc,F.e_iop Formula.Iadd j i)
      | _ -> L0shift(loc,i)

  let startof loc te = index loc te (F.e_int 0)

  let field_access e f = D.acc_field (F.unwrap e) f

  type mem = { variables : F.var Varinfo.Hashtbl.t; }

  type frame = {
    kf : Kernel_function.t ;
    mutable states : (Clabels.c_label * mem) list;
    mutable result : F.var option;
    mutable exit_status : F.var option;
    mutable ret_type :  Cil_types.typ option ;
  }

  let new_frame kf ~m_here ~m_pre ~m_post ropt =
    let may_cons acc lab x = match x with
      | None -> acc
      | Some x -> (lab,x)::acc
    in
    let states = may_cons []     Clabels.Post m_post in
    let states = may_cons states Clabels.Pre m_pre in
    let states = may_cons states Clabels.Here m_here in
    {
      kf = kf;
      states = states;
      result = ropt ;
      exit_status = None;
      ret_type = Some (Kernel_function.get_return_type kf);
    }

  let new_mem () = { variables = Varinfo.Hashtbl.create 10; }

  (** gives the memory state for that label. Create it if it doesn't exist. *)
  let mem_at env clab =
    try List.assoc clab env.states
    with Not_found ->
      let m = new_mem () in env.states <- (clab,m)::env.states ; m

  (** Get the wp variable of the C variable.
  * With [create], create it if it doesn't exist.
  * @raise Not_found when create=false and the variable doesn't exist.*)
  let get_var ~create m lv =
    try Varinfo.Hashtbl.find m.variables lv
    with Not_found ->
      if create then
      (let v = D.fresh lv.vname (Mdata.Vacsl(Ctype lv.vtype)) in
         Varinfo.Hashtbl.add m.variables lv v;
         v)
      else raise Not_found

  let result env =
    let t = match env.ret_type with Some t -> t
      | None ->
          let t = Kernel_function.get_return_type env.kf in
            env.ret_type <- Some t ; t
    in
    let x = match env.result with Some x -> x
      | None ->
          let x = D.fresh "result" (Mdata.Vacsl (Ctype t)) in
            env.result <- Some x ; x
    in (x, t)

  let exit_status env= match env.exit_status with
    | Some x -> x
    | None ->
        let x =
	  D.fresh "exit_status" (Mdata.Vmodel Formula.Integer)
	in
          env.exit_status <- Some x ; x

  let set_exit_status env x =
    {
      kf = env.kf;
      states = env.states;
      result = None;
      exit_status = Some x ;
      ret_type = None;
    }

  let get_exit_status env = env.exit_status

  let rec term_load m = function
    | L0addr v -> F.var (get_var ~create:true m v)
    | L0depl (loc, Pfield f) -> field_access (term_load m loc) f
    | L0depl (loc, Pidx i) -> D.acc_index (F.unwrap (term_load m loc)) i
    | (L0ptr _ | L0shift _) as loc ->
	unsupported "indirect access to %a" pp_loc loc

  let load m cobj loc = value_of_logic cobj (term_load m loc)

  let rec base_address m l  = match l with
    | L0addr _ -> l
    | L0depl (l, _) -> base_address m l
    | L0ptr _ | L0shift _ -> L0ptr (Hlib.base (pointer_of_loc l))

  let block_length _mem l = Hlib.block_length (pointer_of_loc l)

  (* --- WP Calculus *)

  type wp = F.pred

  let update_at_label env l p =
    try
      let mem_lab = List.assoc l env.states in
      let mem_here = mem_at env Clabels.Here in
      let dov vi v_l p =
        let v_here = get_var ~create:true mem_here vi in
        D.subst v_l (F.var v_here) p
      in  Varinfo.Hashtbl.fold dov mem_lab.variables p
    with  Not_found -> p

  let quantify_at_label env l p =
    try
      let mem_lab = List.assoc l env.states in
        let dov _ v l = v::l in
        let l = Varinfo.Hashtbl.fold dov mem_lab.variables [] in
        D.forall l p
    with Not_found -> p

  (** compute m_0 where :  mu_i = mu (obj_(i-1), d_i -> mu_(i+1))
  * and at the last step (n), mu_(n+1) = exp.
  * obj_i is the expression of lval at step (i).
  *
  * For instance, if lval is x.d1.d2.d3,
  * obj_0 = x      ; mu_0 = mu_1
  * obj_1 = x.d1   ; mu_1 = mu (obj_0, d1 -> mu_2)
  * etc. So finally, we should get :
  * mu_0 = mu (x, d1 -> mu (x.d1, d2 -> mu (x.d1.d2, d3 -> exp)))
  *
  * The recursive call at step (i) returns :
  * - the base variable v,
  * - a function that computes obj_i from obj_0
  * - and a function that computes mu_0 from mu_(i+1) and obj_0.
  *
  * obj_0 is not always (F.var v) because of range (see havoc functions).
  *)
  let mk_mu m lval =
    let rec mk_mu_rec lval = match lval with
      | L0addr v -> get_var ~create:true m v ,
                    (fun obj -> obj), (fun _obj mu -> mu)
      | L0depl (lv, d2) ->
          let v, f_obj_1, f_mu_of_mu_2 = mk_mu_rec lv in
          let f_obj_2 = fun obj_0 -> match d2 with
            | Pfield f -> field_access (f_obj_1 obj_0) f
            | Pidx i -> D.acc_index (F.unwrap (f_obj_1 obj_0)) i
          in
          let f_mu_2 = fun obj_0 mu_3 -> match d2 with
            | Pfield f ->
                F.wrap (D.upd_field  (F.unwrap (f_obj_1 obj_0)) f mu_3)
            | Pidx i ->
                F.wrap (D.upd_index (F.unwrap (f_obj_1 obj_0)) i mu_3)
          in
          let f_mu_of_mu_3 = fun o mu3 -> f_mu_of_mu_2 o (f_mu_2 o mu3) in
            v, f_obj_2, f_mu_of_mu_3
      | lval -> unsupported "indirect assignment to %a" pp_loc lval
    in mk_mu_rec lval

  (** @return (v, mu) where [v] is the variable to substitute (base of
   * lval), and [mu] the expression of its new value after and
   * assignment lval = exp; *)
  let get_var_and_mu m lval exp =
    let v, _, f_mu = mk_mu m lval in
      v, f_mu (F.var v) exp

  let subst env lval exp p =
    let mem_here = mem_at env Clabels.Here in
    let v, mu = get_var_and_mu mem_here lval exp in
      Wp_parameters.debug ~level: 3 "hoare:subst %a <-- %a@."
        F.pp_term (F.var v) F.pp_term mu;
    D.subst v mu p

  let subst_lval env _t lval v_exp p =
    subst env lval (logic_of_value v_exp) p

  let subst_result env v_exp p =
    match env.result with
      | None -> p
      | Some xres -> match v_exp with
          | Some v_exp -> D.subst xres (logic_of_value v_exp) p
          | None -> D.forall [xres] p

  (*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*)
  let add_array_length_hyp _m var p =
    match get_array_length var.vtype with
      | None -> p
      | Some sz ->
          let h = mk_v_array_length (pointer_of_loc (L0addr var)) sz in
          let p = F.p_implies h p in
          p

  (**
   * Add some hypotheses about the variable because we are going out of its
   * scope. The quantification could be done here, but it is done through
   * [quantify_at_vars] I guess. *)
  let add_var_decl_hyp m p v =
    let p = (* add hyp about the variable *)
      try
        (* This might be dangerous because it makes possible to prove the
        * validity of &x even after exiting x scope... :
        let h = Hlib.valid_pointer (pointer_of_loc (L0addr v)) in
        (F.p_implies  h p )
        *) p
      with Not_found -> (* var_addr doesn't exist : don't need it ! *) p
    in
    let p = add_array_length_hyp m v p in
      p

  (** process local scope : bool=true for allocation, false for free *)
  let local_scope env vars scope p =
    match scope with
      | Mcfg.SC_Block_in | Mcfg.SC_Function_in ->
          let mem_here = mem_at env Clabels.Here in
          List.fold_left (add_var_decl_hyp mem_here) p vars
      | _ -> p

  (* ----------------------------------------------------------------------- *)
  (* --- Setof Locations                                                 --- *)
  (* ----------------------------------------------------------------------- *)

  let valid _mem = function
    | D.Aloc(_,loc) -> Hlib.valid_pointer (pointer_of_loc loc)
    | D.Arange(_,loc,rg) -> Hlib.valid_range (pointer_of_loc loc) (D.interval rg)

  let separated z1 z2 =
    match z1,z2 with
      | D.Aloc(_,l1) , D.Aloc(_,l2) ->
	  Hlib.separated (pointer_of_loc l1) (pointer_of_loc l2)
      | _ -> Wp_parameters.not_yet_implemented "separated of regions in Hoare"

  (* ----------------------------------------------------------------------- *)
  (* --- Zone Assigns                                                    --- *)
  (* ----------------------------------------------------------------------- *)

  type m_dzone
  type dzone = m_dzone F.term
  let tau_of_dzone = Formula.ADT("zone",[])

  let dzone_assigned _ =
    Wp_parameters.not_yet_implemented "zone representation in Hoare"

  let dzone_subset _z1 _z2 =
    Wp_parameters.not_yet_implemented "zone subset in Hoare"

  let dzone_union _z1 _z2 =
    Wp_parameters.not_yet_implemented "zone union in Hoare"

  let dzone_empty () =
    Wp_parameters.not_yet_implemented "empty zone in Hoare"

  let effect_supported = false

  (* ----------------------------------------------------------------------- *)
  (* --- Region Assigns                                                  --- *)
  (* ----------------------------------------------------------------------- *)

  type region
  let pp_region _ _ = ()
  let region_assigned _ = assert false
  let region_empty _ = assert false
  let region_union _ _ = assert false
  let region_included _ _ = assert false
  let region_separated _ _ = assert false
  let region_fingerprint _ _ = assert false

  let region_supported = false

  (* ----------------------------------------------------------------------- *)
  (* --- Normal Assigns                                                  --- *)
  (* ----------------------------------------------------------------------- *)

  let assigns_goal _env _sloc _lto = unsupported "proof of assigns clauses"

  let assigns_supported = false

  (* ----------------------------------------------------------------------- *)
  (* --- Havoc                                                           --- *)
  (* ----------------------------------------------------------------------- *)

  (**
   * Compute the WP for the predicate [p] of something that modifies
   * only the assigned locations [out_locs].
   * The idea is to assigns arbitrary values "out" to the lvalues in [out_locs].
   * Because the modified locations are evaluated in the pre-state,
   * we don't have commutativity problems, ie. assigns X, T[X] can be processed
   * in any order because the X in mu (T, .(X) -> out) is \at(Lbefore, X).
   *)
  let subst_havoc env = function
    | D.Aloc(te,lval) ->
	let m = mem_at env Clabels.Here in
        let t = D.tau_of_object te in
        let v = D.fresh "out" (Mdata.Vmodel t) in
        let exp = F.var v in
        let x, _f_obj_n, f_mu = mk_mu m lval in
	let newterm (sigma : D.substitution) : Formula.abstract F.term =
          let obj_0 = D.apply sigma (F.var x) in
          f_mu obj_0 exp
	in
	[D.Fresh v ; D.Update(x,newterm)]
    | D.Arange(_,lval,rg) ->
	let m = mem_at env Clabels.Here in
	let upd_range rg t = F.wrap (D.set_range_index (F.unwrap t) rg) in
        let x, f_obj_n, f_mu = mk_mu m lval in
	let newterm (sigma : D.substitution) : Formula.abstract F.term =
          let obj_0 = D.apply sigma (F.var x) in
          let obj_n = f_obj_n obj_0 in
          let exp = upd_range rg obj_n in
          f_mu obj_0 exp
	in
	[D.Update(x,newterm)]

  (* ----------------------------------------------------------------------- *)
  (* --- User Predicate environment                                      --- *)
  (* ----------------------------------------------------------------------- *)

  type closure = Cil_types.varinfo * string

  let pp_closure fmt (vinfo,s) =
    Format.fprintf fmt "%a@@%s" !Ast_printer.d_var vinfo s

  let userdef_frame () =
    {
      kf = Kernel_function.dummy ();
      states = [];
      result = None;
      exit_status = None ;
      ret_type = None;
    }

  let collect_signature label varinfo fvar (variables,closures) =
    fvar :: variables , (varinfo,label) :: closures

  (* TODO filter with respect to tused and pused *)
  let userdef_signature frame _tused _pused =
    List.fold_left
      (fun s (clabel, mem) -> match clabel with
      | Clabels.LabelParam label ->
	Varinfo.Hashtbl.fold (collect_signature label) mem.variables s
      | _ ->
        Wp_parameters.fatal
          "Non-logic label in user-defined function or predicate")
      ([], [])
      frame.states

  let userdef_closure mem (varinfo,_lparam) =
    F.var (get_var ~create:true mem varinfo)

  let label_at_closure (_,label) = label

end



