(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

open Db_types
open Cil_types

type scope =
  | SC_Global
  | SC_Function_in    (* Just before the pre-state *)
  | SC_Function_frame (* Just after the introduction of formals *)
  | SC_Function_out   (* Post-state *)
  | SC_Block_in
  | SC_Block_out

type assigns_method =
  | NoAssigns
  | NormalAssigns
  | EffectAssigns
  | RegionAssigns

module type Export =
sig
  type pred
  val export_goal : Format.formatter -> string -> pred -> unit
  val export_decl : Format.formatter -> pred Formula.declaration -> unit
end

module type Splitter =
sig
  type pred
  val split : assigns_method -> pred -> pred Bag.t
end

(**
 * This is what is really needed to propagate something through the CFG.
 * Usually, the propagated thing should be a predicate,
 * but it can be more sophisticated like lists of predicates,
 * or maybe a structure to keep hypotheses and goals separated.
 * Moreover, proof obligations may also need to be handeled.
 **)
module type S = sig

  type t_prop
  val pretty : Format.formatter -> t_prop -> unit
  val merge : t_prop -> t_prop -> t_prop
  val empty : t_prop

  type t_env
  val new_env : kernel_function -> t_env

  val add_axiom : WpAnnot.prop_id -> string -> logic_label list -> predicate named -> unit
  val add_hyp  : t_env -> WpAnnot.prop_id -> predicate named -> t_prop -> t_prop
  val add_goal : t_env -> WpAnnot.prop_id -> predicate named -> t_prop -> t_prop

  (** [add_as_hyp env h p] add h as hypothesis in p (h => p) *)
  val add_as_hyp  : t_env -> t_prop -> t_prop -> t_prop

  (** [add_as_goal env g p] add g as goal to p (g /\ p) *)
  val add_as_goal : t_env -> t_prop -> t_prop -> t_prop

  val add_assigns : 
    t_env -> WpAnnot.prop_id -> 
    logic_label -> 
    WpAnnot.a_kind -> identified_term assigns list -> t_prop -> t_prop
  val use_assigns : 
    t_env -> WpAnnot.prop_id ->
    WpAnnot.a_kind -> identified_term assigns list -> t_prop -> t_prop

  val assigns_method : unit -> assigns_method

  val label :  t_env -> logic_label -> t_prop -> t_prop
  val assign : t_env -> lval -> Cil_types.exp -> t_prop -> t_prop
  val return : t_env -> exp option -> t_prop -> t_prop
  val test : t_env -> exp -> t_prop -> t_prop -> t_prop
  val switch : t_env -> exp -> (exp list * t_prop) list -> t_prop -> t_prop

  val call_normal : t_env ->
    get_pre:(kernel_function -> (WpAnnot.prop_id * predicate named) list) ->
    get_post:(kernel_function -> (WpAnnot.prop_id * predicate named) list) ->
    get_assigns:(kernel_function -> identified_term assigns list) ->
    stmt -> lval option -> exp -> exp list -> t_prop -> t_prop

  val call_exit : t_env -> 
    get_pre:(kernel_function -> (WpAnnot.prop_id * predicate named) list) ->
    get_exits:(kernel_function -> (WpAnnot.prop_id * predicate named) list) ->
    get_assigns:(kernel_function -> identified_term assigns list) ->
    stmt -> exp -> exp list -> t_prop -> t_prop

  val scope : t_env -> varinfo list -> scope -> t_prop -> t_prop

  (** build [p => alpha(p)] for functional dependencies verification. *)
  val build_prop_of_from : t_env -> WpAnnot.t_id_pred list -> t_prop -> t_prop

  val close : t_env -> t_prop -> t_prop

end
