(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* -------------------------------------------------------------------------- *)
(* --- Datatypes Interface                                                --- *)
(* -------------------------------------------------------------------------- *)

open Cil_types
open Formula
open Ctypes
      
(** {1 FOL Variable Types} *)

type vkind = 
  | Vmodel of tau 
  | Vacsl of logic_type
      
(** {1 Type of ACSL terms} *)

type kind =
  | Kcint of Ctypes.c_int
  | Kint
  | Kreal
  | Kbool
  | Kptr of Cil_types.typ
  | Kset of kind
  | Kstruct of compinfo
  | Karray of arrayinfo
  | Kadt of string * kind list

module type Addr =
sig
  module F : Formula.S
  type addr
  val equal : addr -> addr -> F.pred
  val closed : addr -> bool
  val fresh : F.pool -> F.var * addr
  val pp_addr : Format.formatter -> addr -> unit
end

module type S =
sig

  module F : Formula.S
  type format = Formula.m_format F.term
  type abstract = Formula.abstract F.term
  type integer = Formula.integer F.term
  type real = Formula.real F.term
  type boolean = Formula.boolean F.term
  type record =  m_array F.term
  type urecord = m_array F.term
  type array =  m_array F.term
  type set = m_set F.term
  type name = Formula.integer F.term
  type pointer = m_pointer F.term

  val pp_kind : Format.formatter -> kind -> unit
    
  type interval = {
    inf : integer option ;
    sup : integer option ;
  }

  type 'a assigned =
    | Aloc of c_object * 'a
    | Arange of c_object * 'a * interval

  val pp_interval : Format.formatter -> interval -> unit

  (** {2 FOL Helpers} *)

  val e_app0 : string -> 'a F.term
  val e_app1 : string -> 'a F.term -> 'b F.term
  val e_app2 : string -> 'a F.term -> 'b F.term -> 'c F.term
  val e_app3 : string -> 'a F.term -> 'b F.term -> 'c F.term -> 'd F.term
  val e_app4 : string -> 'a F.term -> 'b F.term -> 'c F.term -> 'd F.term -> 'e F.term
  val e_app5 : string -> 'a F.term -> 'b F.term -> 'c F.term -> 'd F.term -> 'e F.term -> 'f F.term

  val p_app0 : string -> F.pred
  val p_app1 : string -> 'a F.term -> F.pred
  val p_app2 : string -> 'a F.term -> 'b F.term -> F.pred
  val p_app3 : string -> 'a F.term -> 'b F.term -> 'c F.term -> F.pred
  val p_app4 : string -> 
      'a F.term -> 'b F.term -> 'c F.term -> 'd F.term -> F.pred
  val p_app5 : string -> 
      'a F.term -> 'b F.term -> 'c F.term -> 'd F.term -> 'e F.term -> F.pred

  (** {2 Representation of ACSL Values} *)

  val tau_of_logic_type : Cil_types.logic_type -> tau

  (** {2 Constrained Terms Construction} *)

  type context
  type bindings

  class type hint =
  object
    method capture : F.var list -> unit
    method has_var : F.var list -> bool
    method subst : F.var -> Formula.abstract F.term -> unit
    method pretty : Format.formatter -> unit
  end

  val closed : bindings
  val close : bindings -> F.pred -> F.pred
  val capture : hint list -> bindings -> unit

  val push : string -> F.pool -> bindings -> hint list -> context
  val pop : string -> context -> bindings
  val kill: string -> context -> unit
  val flush : string -> context -> F.pred -> F.pred (* pop and close *)

  val term_such_that : tau -> ('a F.term -> F.pred) -> 'a F.term
    
  val forall : F.var list -> F.pred -> F.pred
  val exists : F.var list -> F.pred -> F.pred
  val subst : F.var -> 'a F.term -> F.pred -> F.pred
  val fresh : string -> vkind -> F.var
  val alpha : F.var -> F.var option
  val pool : unit -> F.pool
  val vkind_of_var : F.var -> vkind
  val has_vars : F.var list -> F.pred -> bool

  (** {2 Generalized substitutions} *)

  type substitution
  val apply : substitution -> 'a F.term -> 'a F.term

  type havoc =
    | Fresh of F.var
    | Update of F.var * (substitution -> Formula.abstract F.term)

  val havoc_static : havoc list -> F.pred -> F.pred
  val havoc_inductive : havoc list -> F.pred -> F.pred

  (** {2 Global Declarations} *)

  val clear : unit -> unit
  val on_clear : (unit -> unit) -> unit
  val fresh_name : string -> string -> string
  val add_declaration : F.pred declaration -> unit
  val iter : section -> (F.pred declaration -> unit) -> unit
  val iter_all : (F.pred declaration -> unit) -> unit
  val dummy : unit -> F.pred

  (** {2 Functors and Types for Declaration} *)

  module type Identifiable =
  sig
    type t
    module H : Hashtbl.S
    val index : t -> H.key
    val prefix : string
    val basename : t -> string
    val location : t -> Log.source option
    val pp_title : Format.formatter -> t -> unit
    val pp_descr : Format.formatter -> t -> unit
  end

  module type Registry =
  sig
    type t
    val define : t -> unit
    val get_definition : t -> F.pred declaration
    val on_definition : (t -> F.pred declaration -> unit) -> unit
  end

  module type Declarator =
  sig
    include Identifiable
    val clear : unit -> unit
    val section : section
    val declare : t -> string -> F.pred item
  end

  module Register 
    (D : Declarator) : 
    (Registry with type t = D.t)

  (** {2 Build-int Identifiables and Registry} *)

  module Varinfo   : Identifiable with type t = varinfo
  module Varaddr   : Identifiable with type t = varinfo
  module Fieldinfo : Identifiable with type t = fieldinfo
  module Compinfo  : Identifiable with type t = compinfo
  module Arrayinfo : Identifiable with type t = arrayinfo
  module Logicvar  : Identifiable with type t = logic_var

  module type Indexed =
  sig
    include Registry
    val get_ind : t -> integer
  end

  module Dindex (I:Identifiable):
  sig include Indexed with type t = I.t end

  module Findex : Indexed with type t = fieldinfo
  module Xindex : Indexed with type t = varinfo
  module Aindex : Indexed with type t = varinfo
  module Tindex : Indexed with type t = compinfo

  (** {2 Integer Logic Cast} *)    

  val modulo : Ctypes.c_int -> integer -> integer
  val guard : Ctypes.c_int -> integer -> F.pred
  val i_convert: Ctypes.c_int -> Ctypes.c_int -> integer -> integer

  (** {2 Real Logic Cast} *)  

  val round : Ctypes.c_float -> real -> real
  val f_guard : Ctypes.c_float -> real -> F.pred
  val f_convert : Ctypes.c_float -> Ctypes.c_float -> real -> real

  val has_type: abstract -> logic_type -> F.pred

  (** {2 Sub-arrays} *) 

  val get_range_index : array -> interval -> abstract

  val set_range_index : array -> interval -> array

  (** {2 Set and Range as first Class Value} *) 
   
  val empty : set
    (**[empty()] returns the polymorphic empty set. *)
    
  val singleton : abstract -> set
    (** [singleton a] returns the singleton set containning [a]. *)
    
  val union : set -> set -> set
    (** [union s0 s1] returns the union set of [s0] and [s1]. *)
    
  val unions : set list -> set
      
  val inter : set -> set -> set
    (**[inter s0 s1] returns the intersection set of [s0] and [s1]. *)  
  
  val remove : set -> set -> set
    (** [remove s0 s1] returns a set [s'] such as \{b in s' | b in s0
	and !(b in s1) \}.*) 

  val set_of_list : abstract list -> set

  val add_set : set -> set -> set

  val mult_set : set -> set -> set

  val neg_set : set -> set

  val interval : interval -> set
    (** [range l h] returns the integer set [s] such as \{ i in s | l <=
	i <= h \}. *)

  (** {2 Pointer arithmetics} *)
    
  val equal_pointer_bool : pointer -> pointer -> boolean
    (** [equal_pointer_bool p q] is the boolean that is true when
	pointers [p] and [q], that points to elements of type [ty],
	are equal. Returns a term of type [bool]. *)
    
  val lt_pointer_bool : pointer -> pointer -> boolean
    (** [lt_pointer_bool p q] is the less than comparison
	of pointers [p] and [q], that points to elements of type [ty],
	ie. [p<q]. Returns a term of type [boolean]. 
	*)

  val le_pointer_bool : pointer -> pointer -> boolean
    (** [le_pointer_bool p q] is the less than or equal comparison
	of pointers [p] and [q], that points to elements of type [ty],
	ie. [p<=q]. Returns a term of type [boolean]. 
	*)
    
  val lt_pointer : pointer -> pointer -> F.pred
    (** [lt_pointer p q] is the less than comparison
	of pointers [p] and [q], that points to elements of type [ty],
	ie. [p<q]. 
	*)
    
  val le_pointer : pointer -> pointer -> F.pred
    (** [le_pointer p q] is the less than or equal comparison
	of pointers [p] and [q], that points to elements of type [ty],
	ie. [p<=q].
	*)  
        
  val minus_pointer : pointer -> pointer -> integer
    (** [minus_pointer ty p q] is the arithmetics difference of
	pointers [p] and [q], that points to elements of type [ty],
	ie. [p-q]. Returns a term of type [integer]. *)
    
  val is_null : pointer -> boolean
    (** Take a term representing an [address] and returns a
	term of type [bool]. *)
    
  (** {2 Special locations} *)
    
  val null : pointer
    (** [null] return the special location of the memory model designing 
      the null pointer. *)
 
  (** {2 Formats} *)

  val tau_of_object : c_object -> tau
  val tau_of_ctype : typ -> tau

  val int_format : format
  val real_format : format
  val pointer_format : format
  val record_format : format
  val urecord_format : format
  val array_format : format -> format
  val format_of_object : Ctypes.c_object -> format

  val decode : format -> abstract -> 'a F.term
  val encode : format -> 'a F.term -> abstract

  (** {2 Record as First Class Value}*)
    
  val acc_field : 
    record -> Cil_types.fieldinfo -> abstract
    (** Takes a record term of type (tau_of F.fcomp) and 
	returns the term of type (tau_of F) for its field 'F' *)
    
  val upd_field : 
    record -> Cil_types.fieldinfo -> abstract  -> record
    (** Takes a record term of type (tau_of F.fcomp) and 
	a new term of type (tau_of F) for its field 'F'
	and return the updated record. *)
    
  (** {2 Array as first Class Value} *) 
    
  val acc_index : array -> integer -> abstract
    (** Takes term of type ['a farray] and returns the ['a] at index [i]. *)
    
  val upd_index : array -> integer -> abstract -> array

  (** {2 ACSL Comparison} *)

  val equal : c_object -> abstract -> abstract -> F.pred
  val eq_array : arrayinfo -> array -> array -> F.pred
  val eq_record : compinfo -> record -> record -> F.pred

end

