(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* -------------------------------------------------------------------------- *)
(** Memory Model Signature                                                    *)
(* -------------------------------------------------------------------------- *)

open Ctypes
open Clabels
open Formula
open Cil_types
open Mdata

module type Model =
sig

  module D : Mdata.S
  module A : Mint.S with module D = D
  module R : Mfloat.S with module D = D

  type loc

  val loc_of_pointer : c_object -> D.pointer -> loc
  val pointer_of_loc : loc -> D.pointer

  (** 
   * [cast_loc_to_int t p c_int] :
   * cast [loc] if type [t*] into a term of type [c_int] 
   **)
 
  val cast_loc_to_int : Cil_types.typ -> loc -> Ctypes.c_int -> D.integer
  val cast_int_to_loc : Ctypes.c_int -> D.integer -> Cil_types.typ -> loc
 
  val pp_loc : Format.formatter -> loc -> unit

end

module type Values =
sig

  include Model
  
  (** The internal representation of an ACSL value *)
  type value = 
    | V_int of Ctypes.c_int * D.integer
    | V_float of Ctypes.c_float * D.real
    | V_pointer of Cil_types.typ * loc
    | V_record of compinfo * D.record
    | V_union of compinfo * D.urecord
    | V_array of arrayinfo * D.array
 
  val pp_value : Format.formatter -> value -> unit

  (** Conversion between internal representation of ACSL value and FOL term *)

  val logic_of_value : value -> D.abstract
  val value_of_logic : c_object -> D.abstract -> value

end

module type Data =
sig
  
  include Values
    
  type m_of_mem
  val tau_of_mem : tau
  val forall_loc : D.F.pool -> D.F.var list * loc

  val index : loc -> c_object -> D.integer -> loc
  val field : loc -> fieldinfo -> loc

  val load_mem  : m_of_mem D.F.term -> c_object -> loc -> value
  val store_mem : m_of_mem D.F.term -> c_object -> loc -> value -> m_of_mem D.F.term

end

module type S =
sig

  include Values

  (** {2 Memory, Field and Array access} *)

  type mem

  val cvar  : mem -> varinfo -> loc

  val shift : loc -> c_object -> D.integer -> loc
    (** [shift ptr tau k] computes the location
	of [ptr+k], where [ptr] is a pointer to a value of type [tau]. *)

  val index : loc -> c_object -> D.integer -> loc 
  (** [index tab tau k] computes the location
      of [tab[k]], where [tab] is an array with elements of type [tau]. *)

  val startof : loc -> c_object -> loc
  (** [startof] return a pointer to the first element of an array *)

  val field : loc -> fieldinfo -> loc

  val load  : mem -> c_object -> loc -> value

  val cast_loc_to_loc :  typ -> typ -> loc -> loc
  (** [cast_loc_to_loc t1 t2 l] returns the casted location of type [t2] 
            from the location [l] of type [t1] *)

end
