(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* ------------------------------------------------------------------------ *)
(* ---  Zone Memory Model                                               --- *)
(* ------------------------------------------------------------------------ *)

open Cil_types
open Ctypes
open Formula

module Create
  (D:Mdata.S)
  (A:Mdata.Addr with module F = D.F) =
struct

  type addr = A.addr
  module D = D
  module F = D.F

  (* ------------------------------------------------------------------------ *)
  (* --- Elementary Regions                                               --- *)
  (* ------------------------------------------------------------------------ *)

  type base =
    | Cvar of varinfo * A.addr
    | Root of int * A.addr
    | Base of A.addr

  type range =
    | Block (* offset = 0, length = +oo *)
    | Index of D.integer (* offset, length = 1 *)
    | Range of D.integer * D.integer (* offset, length *)

  let range_empty = function
    | Block | Index _ -> F.p_false
    | Range(_,n) -> F.p_icmp Cleq n (F.e_int 0)

  let range_subset p q =
    match p , q with
      | _ , Block -> F.p_true
      | Block , _ -> F.p_false
      | Index k1 , Index k2 -> F.p_icmp Ceq k1 k2
      | Index k1 , Range(k2,n2) -> 
	  F.p_and 
	    (F.p_icmp Cleq k2 k1)
	    (F.p_icmp Clt k2 (F.e_iop Iadd k2 n2))
      | Range(k1,n1) , Index k2 ->
	  F.p_and
	    (F.p_icmp Cleq k2 k1)
	    (F.p_icmp Ceq n1 (F.e_int 1))
      | Range(k1,n1) , Range(k2,n2) ->
	  F.p_and
	    (F.p_icmp Cleq k2 k1)
	    (F.p_icmp Cleq (F.e_iop Iadd k1 n1) (F.e_iop Iadd k2 n2))

  let range_disjoint p q =
    match p , q with
      | Block , _ | _ , Block -> F.p_false
      | Index k1 , Index k2 -> F.p_neq k1 k2
      | Index k1 , Range(k2,n2) | Range(k2,n2) , Index k1 ->
	  F.p_or
	    (F.p_icmp Clt k1 k2)
	    (F.p_icmp Cleq (F.e_iop Iadd k2 n2) k1)
      | Range(k1,n1) , Range(k2,n2) ->
	  F.p_or
	    (F.p_icmp Cleq (F.e_iop Iadd k1 n1) k2)
	    (F.p_icmp Cleq (F.e_iop Iadd k2 n2) k1)

  (* ------------------------------------------------------------------------ *)
  (* --- Path Management                                                  --- *)
  (* ------------------------------------------------------------------------ *)

  type path = D.integer * D.integer
  let e_sizeof te = F.e_int64 (Ctypes.sizeof_object te)
  let e_field f = F.e_int64 (Ctypes.field_offset f)

  let a_sizeof = function
    | C_array { arr_flat=Some f } -> F.e_int64 f.arr_size
    | _ -> F.e_int 1

  let sizeof ~atomic te = if atomic then a_sizeof te else e_sizeof te

  let rec is_atomic = function
    | C_int _ | C_float _ | C_pointer _ -> true
    | C_array { arr_element=te ; arr_flat=Some _ } -> 
	is_atomic (Ctypes.object_of te)
    | C_array { arr_flat=None } | C_comp _ -> false

  let p_block ~atomic te = F.e_int 0 , sizeof atomic te
  let p_field ~atomic (ofs,_) f =
    assert (not atomic) ;
    F.e_iop Iadd ofs (e_field f) , e_sizeof (Ctypes.object_of f.ftype)
  let p_index ~atomic (ofs,_) te ~offset = 
    let s = sizeof ~atomic te in
    F.e_iop Iadd ofs (F.e_iop Imul s offset) , s
  let p_range ~atomic (ofs,_) te ~offset ~length =
    let s = sizeof ~atomic te in
    F.e_iop Iadd ofs (F.e_iop Imul s offset) , (F.e_iop Imul s length)
  let p_interval ~atomic (ofs,sz) te rg =
    let s = sizeof ~atomic te in
    match rg with
      | { D.inf=None ; D.sup=None } -> ofs,sz
      | { D.inf=None ; D.sup=Some p } -> ofs,F.e_iop Imul s p
      | { D.inf=Some a ; D.sup=None } -> ofs,F.e_iop Isub sz (F.e_iop Imul s a)
      | { D.inf=Some a ; D.sup=Some b } ->
	  F.e_iop Iadd ofs (F.e_iop Imul s a) ,
	  F.e_iop Imul s (F.e_iop Iadd (F.e_int 1) (F.e_iop Isub b a))

  (* ------------------------------------------------------------------------ *)
  (* --- Map Merges                                                       --- *)
  (* ------------------------------------------------------------------------ *)

  module Vmap = Cil_datatype.Varinfo.Map
  module Imap = Datatype.Int.Map

  exception Trivial

  module Merge( M : Map.S ) =
  struct
    
    let union cup m1 m2 =
      M.fold
	(fun x e m ->
	   let e =
	     try cup e (M.find x m2)
	     with Not_found -> e
	   in M.add x e m
	) m1 m2

    let forall f m =
      try
	M.fold
	  (fun x e p ->
	     let q = f x e in
	     if F.is_false q then raise Trivial ;
	     F.p_and p q
	  ) m F.p_true
      with Trivial -> F.p_false

    let exists f m =
      try
	M.fold
	  (fun x e p ->
	     let q = f x e in
	     if F.is_true q then raise Trivial ;
	     F.p_or p q
	  ) m F.p_false
      with Trivial -> F.p_true

  end

  module Vmerge = Merge(Vmap)
  module Imerge = Merge(Imap)

  let rec forall f xs = 
    try
      List.fold_left 
	(fun p x -> 
	   let q = f x in
	   if F.is_false q then raise Trivial ;
	   F.p_and p q)
	F.p_true xs
    with Trivial -> F.p_false 

  let rec exists f xs =
    try
      List.fold_left 
	(fun p x -> 
	   let q = f x in
	   if F.is_true q then raise Trivial ;
	   F.p_or p q)
	F.p_false xs
    with Trivial -> F.p_true

  (* ------------------------------------------------------------------------ *)
  (* --- Complex Regions                                                  --- *)
  (* ------------------------------------------------------------------------ *)

  let roots = ref 0
  let () = D.on_clear (fun () -> roots := 0)

  type region = {
    cvars : (A.addr * range list) Vmap.t ;
    roots : (A.addr * range list) Imap.t ;
    bases : (A.addr * range) list ;
  }

  let pp_range fmt = function
    | Block -> Format.fprintf fmt "-"
    | Index k -> Format.fprintf fmt "[%a]" F.pp_term k
    | Range(a,b) -> Format.fprintf fmt "[%a:%a]" F.pp_term a F.pp_term b

  let pp_region fmt r =
    begin
      Vmap.iter
	(fun x (px,ds) ->
	   Format.fprintf fmt "@,@[<v 0>@[Variable %s %a@]:@[<hov 2>"  x.vname A.pp_addr px ;
	   List.iter (fun d -> Format.fprintf fmt "@ @[%a@]" pp_range d) ds ;
	   Format.fprintf fmt "@]@]"
	) r.cvars ;
      Imap.iter
	(fun k (px,ds) ->
	   Format.fprintf fmt "@,@[<v 0>@[Root %d %a@]:@[<hov 2>" k A.pp_addr px ;
	   List.iter (fun d -> Format.fprintf fmt "@ @[%a@]" pp_range d) ds ;
	   Format.fprintf fmt "@]@]"
	) r.roots ;
      List.iter
	(fun (b,d) ->
	   Format.fprintf fmt "@,@[<v 0>@[Base %a@]:@[<hov 2>@ %a@]@]@\n" A.pp_addr b pp_range d
	) r.bases ;
    end

  (* ------------------------------------------------------------------------ *)
  (* --- Oprations on Regions                                             --- *)
  (* ------------------------------------------------------------------------ *)

  let empty = {
    cvars = Vmap.empty ;
    roots = Imap.empty ;
    bases = [] ;
  }

  let union_range (p,xs) (_,ys) =
    match xs , ys with
      | [Block] , _ -> p,xs
      | _ , [Block] -> p,ys
      | [] , zs | zs , [] -> p,zs
      | _ -> p,List.append xs ys

  let union r1 r2 = 
    {
      cvars = Vmerge.union union_range r1.cvars r2.cvars ;
      roots = Imerge.union union_range r1.roots r2.roots ;
      bases = List.append r1.bases r2.bases ;
    }
	  
  let cvar p x = Cvar(x,p)
  let root p = incr roots ; Root( !roots , p )
  let base p = Base p

  let zone base range =
    match base with
      | Cvar(x,p) -> { empty with cvars = Vmap.add x (p,[range]) Vmap.empty }
      | Root(k,p) -> { empty with roots = Imap.add k (p,[range]) Imap.empty }
      | Base p -> { empty with bases = [p,range] }

  let block base = zone base Block
  let index base k = zone base (Index k)
  let range base k n = zone base (Range(k,n))
  let path base (k,n) = zone base (Range(k,n))

  let iter_others f r =
    begin
      let iter_map f _ (p,rs) = List.iter (fun r -> f p r) rs in
      Vmap.iter (iter_map f) r.cvars ;
      Imap.iter (iter_map f) r.roots ;
    end

  let iter_bases f r =
    List.iter (fun (p,r) -> f p r) r.bases 

  (* ------------------------------------------------------------------------ *)
  (* --- Emptyness of Regions                                             --- *)
  (* ------------------------------------------------------------------------ *)

  let is_empty r =
    F.p_conj [
      Vmerge.forall (fun _ (_,rs) -> forall range_empty rs) r.cvars ;
      Imerge.forall (fun _ (_,rs) -> forall range_empty rs) r.roots ;
      forall (fun (_,r) -> range_empty r) r.bases ;
    ]

  (* ------------------------------------------------------------------------ *)
  (* --- Inclusion of Regions                                             --- *)
  (* ------------------------------------------------------------------------ *)

  let included_ranges xs ys =
    forall (fun x -> exists (fun y -> range_subset x y) ys) xs

  let included_find x p rs find map bases =
    let rs' = try snd (find x map) with Not_found -> [] in
    forall
      (fun r ->
	 let found_in_map =
	   exists (fun r' -> range_subset r r') rs' in
	 if F.is_true found_in_map then found_in_map else
	   let found_in_bases =
	     exists
	       (fun (q,r') -> F.p_and (A.equal p q) (range_subset r r'))
	       bases in
	   F.p_or found_in_map found_in_bases
      ) rs

  let included_cvars r1 r2 =
    Vmerge.forall
      (fun x (p,rs) -> included_find x p rs Vmap.find r2.cvars r2.bases)
      r1.cvars

  let included_roots r1 r2 =
    Imerge.forall
      (fun k (p,rs) -> included_find k p rs Imap.find r2.roots r2.bases)
      r1.roots
    
  let included_zone_in_bases p r r2 =
    exists
      (fun (q,r') ->
	 F.p_and (A.equal p q) (range_subset r r')
      ) r2.bases

  let included_zone_in_cvars p r r2 =
    Vmerge.exists
      (fun _x (q,r2s) ->
	 F.p_and (A.equal p q) (exists (fun r' -> range_subset r r') r2s)
      ) r2.cvars

  let included_zone_in_roots p r r2 =
    Imerge.exists
      (fun _k (q,r2s) ->
	 F.p_and (A.equal p q) (exists (fun r' -> range_subset r r') r2s)
      ) r2.roots

  let included_bases r1 r2 =
    forall
      (fun (p,r) ->
	 exists 
	   (fun f -> f p r r2)
	   [ included_zone_in_bases ;
	     included_zone_in_cvars ;
	     included_zone_in_roots ]
      ) r1.bases

  let included r1 r2 =
    forall 
      (fun f -> f r1 r2) 
      [ included_cvars ; included_roots ; included_bases ]

  (* ------------------------------------------------------------------------ *)
  (* --- Separation of Regions                                            --- *)
  (* ------------------------------------------------------------------------ *)

  let separated_ranges rs rs' =
    forall (fun r -> forall (range_disjoint r) rs') rs

  let separated_find x rs find map =
    try
      let rs' = snd (find x map) in
      separated_ranges rs rs'
    with Not_found -> F.p_true

  let separated_cvars r1 r2 =
    Vmerge.forall
      (fun x (_,rs) -> separated_find x rs Vmap.find r2.cvars)
      r1.cvars

  let separated_roots r1 r2 = 
    Imerge.forall
      (fun k (_,rs) -> separated_find k rs Imap.find r2.roots)
      r1.roots

  let all_separated iter1 iter2 r1 r2 =
    try
      let w = ref F.p_true in
      iter1 
	(fun p r -> 
	   iter2 
	     (fun q r' ->
		let s = F.p_implies 
		  (A.equal p q)
		  (range_disjoint r r')
		in
		if F.is_false s then raise Trivial ;
		w := F.p_and !w s
	     ) r2
	) r1 ; !w
    with Trivial ->
      F.p_false

  let separated_bases = all_separated iter_bases iter_bases
  let separated_others_bases = all_separated iter_others iter_bases
  let separated_bases_others = all_separated iter_bases iter_others
    
  let separated r1 r2 =
    forall
      (fun f -> f r1 r2)
      [ separated_cvars ; separated_roots ; separated_bases ;
	separated_others_bases ; separated_bases_others ]

  (* ------------------------------------------------------------------------ *)
  (* --- Finger-print                                                     --- *)
  (* ------------------------------------------------------------------------ *)

  type env = {
    pool : F.pool ;
    mutable vars : F.var list ;
    mutable ctxt : F.pred ;
  }

  let fp_offset env k =
    if F.term_closed k then k else
      let x = F.fresh env.pool "k" (Model Integer) in
      let k' = F.var x in
      env.vars <- x :: env.vars ;
      env.ctxt <- F.p_and env.ctxt (F.p_eq k' k) ; k'

  let fp_size env s =
    if F.term_closed s then s else
      let x = F.fresh env.pool "n" (Model Integer) in
      let s' = F.var x in
      env.vars <- x :: env.vars ;
      env.ctxt <- F.p_and env.ctxt (F.p_eq s' s) ; s'
    
  let fp_addr env p =
    if A.closed p then p else
      let x,q = A.fresh env.pool in
      env.vars <- x :: env.vars ;
      env.ctxt <- F.p_and env.ctxt (A.equal q p) ; q

  let fp_range env = function
    | Block -> Block
    | Index x -> Index (fp_offset env x)
    | Range(x,n) -> Range (fp_offset env x,fp_size env n)

  let fp_base env (p,r) = fp_addr env p , fp_range env r

  let fp_map env (p,rs) =
    fp_addr env p , List.map (fp_range env) rs

  let fingerprint pool r =
    let env = { pool=pool ; vars=[] ; ctxt=F.p_true } in
    let r' = {
      cvars = Vmap.map (fp_map env) r.cvars ;
      roots = Imap.map (fp_map env) r.roots ;
      bases = List.map (fp_base env) r.bases ;
    } in
    env.vars , env.ctxt , r'

end
