(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* ------------------------------------------------------------------------ *)
(** {1 Zone Memory Model}                                                   *)
(* ------------------------------------------------------------------------ *)

open Cil_types
open Ctypes
open Formula

module Create
  (D:Mdata.S)
  (A:Mdata.Addr with module F = D.F) :
sig

  (** {2 Base Management} *)
    
  type base

  val cvar : A.addr -> varinfo -> base 
    (** Base of a C-variable. 
	[cvar p x] is the base of variable [x], with [p] its base-address *)

  val root : A.addr -> base
    (** Create a fresh root, separated from base address of all
	variables. [root p] is a new root equal to pointer [p]. *)

  val base : A.addr -> base
    (** An arbitrary base-address of any allocated C memory-block. *)

  (** {2 Path Management} *)

  type path
  val is_atomic : c_object -> bool    
    
  val p_block : atomic:bool -> c_object -> path
  val p_field : atomic:bool -> path -> fieldinfo -> path
  val p_index : atomic:bool -> path -> c_object -> offset:D.integer -> path
  val p_range : atomic:bool -> path -> c_object -> offset:D.integer -> length:D.integer -> path
  val p_interval : atomic:bool -> path -> c_object -> D.interval -> path

  (** {2 Region Management} *)

  type region
    
  val pp_region : Format.formatter -> region -> unit

  val empty : region
    (** The empty region *)

  val union : region -> region -> region
    (** Union of arbitrary regions *)

  val block : base -> region 
    (** The full block at given block *)

  val index : base -> D.integer -> region 
    (** [index b k] is the cell at index [k] in block [k].
	Internally, same as [range b k 1]. It is a safe abstraction
	of [range b (k * s) s] when all memory cells in the block
	have the same size in the underlying memory model. *)

  val range : base -> D.integer -> D.integer -> region
    (** [range b k n] builds a region of length [n] starting at offset [k] 
	in the block of base [b]. *)

  val path : base -> path -> region

  (** {2 Region Properties} *)

  val is_empty : region -> D.F.pred
  val included : region -> region -> D.F.pred
  val separated : region -> region -> D.F.pred

  (** {2 Substituion-free Regions} *)
    
  val fingerprint : D.F.pool -> region -> ( D.F.var list * D.F.pred * region )

  (** {2 Sizeof Abstraction } *)

  val a_sizeof : Ctypes.c_object -> D.integer

end

