(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

open Task
open Property
open Cil_types
open Db_types

(* -------------------------------------------------------------------------- *)
(* --- WP Models for VC generation                                        --- *)
(* -------------------------------------------------------------------------- *)

module W = Fol_why
module Q = Fol_coq
module E91 = Fol_ergo.E91
module E92 = Fol_ergo.E92

module F = Fol_formula
module D = Datalib.Create(F)
module A = Mint_natural.Create(D)
module R = Mfloat_natural.Create(D)

module Hoare = Hoare_mem.Create(D)(A)(R)
module Store = Store_mem.Create(D)(A)(R)
module Runtime = Runtime_mem.Create(D)(A)(R)

(*
module FQed = Fol_qed.F
module DQed = Datalib.Create(FQed)
module AQed = Mint_natural.Create(DQed)
module RQed = Mfloat_natural.Create(DQed)
module EQed = Fol_qed.E
*)

(* --------- WP Calculus Engines ------------------ *)

module WP_Hoare = 
  CfgProof.Create(Hoare)(W)(Q)(E91)(E92)(Fol_split)
    (struct
       let shared = "hoare"
       let context = "hoare"
       let updater = "Hoare"
       let name = "Hoare"
     end)

module WP_Caveat = 
  CfgProof.Create
    (Caveat_mem.Safe.WP)
    (Caveat_mem.Safe.Why)
    (Caveat_mem.Safe.Coq)
    (Caveat_mem.Safe.Ergo91)
    (Caveat_mem.Safe.Ergo92)
    (Caveat_mem.Safe.Splitter)
    (struct
       let shared = "caveat"
       let context = "caveat_safe"
       let updater = "Caveat-Safe"
       let name = "Caveat"
     end)

module WP_UnsafeCaveat = 
  CfgProof.Create
    (Caveat_mem.Unsafe.WP)
    (Caveat_mem.Unsafe.Why)
    (Caveat_mem.Unsafe.Coq)
    (Caveat_mem.Unsafe.Ergo91)
    (Caveat_mem.Unsafe.Ergo92)
    (Caveat_mem.Unsafe.Splitter)
    (struct
       let shared = "caveat"
       let context = "caveat"
       let updater = "Caveat-Unsafe"
       let name = "Caveat(without guards)"
     end)

module WP_Store = CfgProof.Create(Store)(W)(Q)(E91)(E92)
  (Fol_split)
  (struct
     let shared = "store"
     let context = "store_full"
     let updater = "Store-Full"
     let name = "Store(full memory)"
   end)

module WP_Storefun = CfgProof.Create(Funvar_mem.Create(Store))(W)(Q)(E91)(E92)
  (Fol_split)
  (struct
     let shared = "store"
     let context = "store"
     let updater = "Store-LogicVar"
     let name = "Store"
   end)

module WP_Runtime = CfgProof.Create(Runtime)(W)(Q)(E91)(E92)
  (Fol_split)
  (struct
     let shared = "runtime"
     let context = "runtime_full"
     let updater = "Runtime-Full"
     let name = "Runtime(full memory)"
   end)

module WP_Runtimefun = CfgProof.Create(Funvar_mem.Create(Runtime))(W)(Q)(E91)(E92)
  (Fol_split)
  (struct
     let shared = "runtime"
     let context = "runtime"
     let updater = "Runtime-LogicVar"
     let name = "Runtime"
   end)

(* --------- WP Dispatcher ------------------ *)

type feature = NA | Yes | No

type wp_model = {
  wp_name : string ;
  wp_qed : feature ;
  wp_logicvar : feature ;
  wp_method : unit -> Mcfg.assigns_method ;
  wp_computer : unit -> CfgProof.computer ;
  wp_altmodel : (unit -> CfgProof.computer) option ;
}

let wp_model name = {
  wp_name = name ;
  wp_qed = NA ;
  wp_logicvar = NA ;
  wp_method = (fun () -> Wp_parameters.fatal "no method implemented") ;
  wp_computer = (fun () -> Wp_parameters.fatal "computer not implemented") ;
  wp_altmodel = None ;
}

let option opt = function
  | NA -> true
  | Yes -> opt ()
  | No -> not (opt ())

type 'a computing_methods =
  | OneforBoth of 'a
  | NonAssigns of 'a
  | OneforEach of 'a * 'a (*non assigns, assigns only*)

let dispatch models =
  try
    let model = List.find
      (fun m ->
	 List.for_all
	   (fun (opt,f) ->
	      match f with
		| NA -> true
		| Yes -> opt ()
		| No -> not (opt ()) ) 
	   [ 
	     (fun () -> false) (*Wp_parameters.Qed.get*) , m.wp_qed ;
	     Wp_parameters.LogicVar.get , m.wp_logicvar ;
	   ]
      ) models in
    let computer = model.wp_computer () in
    match model.wp_method () , model.wp_altmodel with
      | Mcfg.NoAssigns , None -> NonAssigns computer
      | Mcfg.NoAssigns , Some alt -> OneforEach( computer , alt () )
      | _ -> OneforBoth computer
  with Not_found ->
    Wp_parameters.abort "No model found with provided criteria"

(* --------- WP Computer -------------------- *)

(*
  computer returns :
  - either one computer for both assigns and non-assigns
  - or a unique computer for only non-assigns
  - or two computers, one for assigns and one for non-assigns
*)

let computer = function

    | Wp_parameters.M_Hoare  ->
	NonAssigns (WP_Hoare.create ())

    | Wp_parameters.M_CaveatU ->
	NonAssigns (WP_UnsafeCaveat.create ())

    | Wp_parameters.M_CaveatG ->
	NonAssigns (WP_Caveat.create ())

    | Wp_parameters.M_Store ->
	dispatch [
	  { (wp_model "Store") with
	      wp_logicvar = Yes ; 
	      wp_method = WP_Storefun.assigns_method ;
	      wp_computer = WP_Storefun.create ;
	      wp_altmodel = Some WP_Store.create ;
	  } ;
	  { (wp_model "Store") with
	      wp_logicvar = No ;
	      wp_method = WP_Store.assigns_method ;
	      wp_computer = WP_Store.create ;
	  }
	]
	  
    | Wp_parameters.M_Runtime ->
	dispatch [
	  { (wp_model "Runtime") with
	      wp_logicvar = Yes ;
	      wp_method = WP_Runtimefun.assigns_method ;
	      wp_computer = WP_Runtimefun.create ;
	      wp_altmodel = Some WP_Runtime.create ;
	  } ;
	  { (wp_model "Runtime") with
	      wp_logicvar = No ;
	      wp_method = WP_Runtime.assigns_method ;
	      wp_computer = WP_Runtime.create ;
	  }
	]

(* ------------------------------------------------------------------------ *)
(* --- Iterators                                                        --- *)
(* ------------------------------------------------------------------------ *)

let on_definition phi kf =
  match kf.fundec with
    | Declaration _ ->
        Wp_parameters.warning ~current:false "Function %s has no body (skipped)"
	  (Kernel_function.get_name kf)
    | Definition _ -> phi kf

let on_all_functions = fun do_body ->
  Globals.Functions.iter (fun kf ->
                            !Db.progress ();
			    match kf.fundec with
			      | Declaration _ -> ()
			      | Definition _ ->
				  on_definition do_body kf)

let on_function_names fcts = fun do_body ->
  List.iter
    (fun fname ->
       try
	 let kf = Globals.Functions.find_by_name fname in
	 on_definition do_body kf
       with Not_found ->
	 Wp_parameters.error "Unknown function '%s' (skipped)" fname
    ) fcts

let on_function kf = fun do_body -> on_definition do_body kf

(* ------------------------------------------------------------------------ *)
(* --- Shared Functions for both GUI and command line                   --- *)
(* ------------------------------------------------------------------------ *)

let build_cfg ?(do_dot) kf =
  if Wp_parameters.RTE.get () 
  then WpAnnot.compute_rte_for kf ;
  let cfg, unreached = Cil2cfg.get kf in
  WpAnnot.process_unreached_annots cfg unreached ;
  let do_dot = match do_dot with
    | None -> Wp_parameters.Dot.get ()
    | Some b -> b
  in
  !Db.progress () ;
  let dot = if do_dot then Some (Cil2cfg.dot_cfg cfg) else None in
  cfg, dot

let dot_lannots cfg lannots =
  let do_dot annots =
    let pp_annots fmt e =
      WpAnnot.pp_annots fmt (annots.WpAnnot.get_annots e)
    in
    let bhv = annots.WpAnnot.behavior_name in
    ignore (Cil2cfg.dot_annots cfg bhv pp_annots)
  in List.iter do_dot lannots
       
type prop =
  | NamedProp of string
  | IdProp of Property.t

let do_compute
    (goals : Wpo.t Bag.t ref)
    (computer : CfgProof.computer)
    fun_iter assigns behaviors property
    =
  let do_kf kf =
    let cfg, _ = build_cfg kf in
    let annots =
      match behaviors, property with
	| None, None ->
            WpAnnot.get_function_strategies ~assigns cfg
	| Some bhvs, None ->
            WpAnnot.get_behavior_strategies ~assigns cfg bhvs
	| _, Some (IdProp p) ->
            WpAnnot.get_id_prop_strategies ~assigns cfg p
	| _, Some (NamedProp p) ->
            WpAnnot.get_prop_strategies ~assigns cfg (behaviors, p)
    in
      if Wp_parameters.Dot.get () then dot_lannots cfg annots;
      computer#add cfg annots
  in
  begin
    fun_iter do_kf ;
    goals := Bag.concat !goals (Bag.list computer#compute) ;
  end

(* ------------------------------------------------------------------------ *)
(* ---  Printing informations                                           --- *)
(* ------------------------------------------------------------------------ *)

let do_wp_print () =
  (* Printing *)
  if Wp_parameters.Print.get () then 
    try
      Wpo.iter ~on_goal:(fun _ -> raise Exit) () ;
      Wp_parameters.result "No proof obligations"
    with Exit ->
      Log.print_on_output "%t" 
	(fun fmt ->
	   Wpo.iter
	     ~on_environment:(Wpo.pp_environment fmt)
	     ~on_behavior:(Wpo.pp_function fmt)
	     ~on_goal:(Wpo.pp_goal_flow fmt) ())
	
let do_wp_print_for goals =
  if Wp_parameters.Print.get () then
    if Bag.is_empty goals 
    then Wp_parameters.result "No proof obligations"
    else Log.print_on_output "%t"
      (fun fmt -> Bag.iter (Wpo.pp_goal_flow fmt) goals)

(* ------------------------------------------------------------------------ *)
(* ---  Proving                                                         --- *)
(* ------------------------------------------------------------------------ *)

let do_wpo_feedback g prover result =
  Wp_parameters.feedback "[%a] Goal %s : %a"
    Wpo.pp_prover prover g.Wpo.po_gid Wpo.pp_result result

(*
  else Wp_parameters.feedback "[%a] Goal %a : %a"
    Wpo.pp_prover prover WpAnnot.pp_id g.Wpo.po_pid Wpo.pp_result result

let do_wp_feedback server kf bhv =
  begin
    Task.launch server ; (* synchronize previous goals *)
    match bhv with
      | None ->
	  Wp_parameters.feedback "Function %a"
	    Kernel_function.pretty_name kf
      | Some bhv ->
	  Wp_parameters.feedback "Function %a with behavior %s"
	    Kernel_function.pretty_name kf bhv
  end

*)

let do_wp_proof server interactive prover g =
  let already_valid (_,r) = r=Wpo.Valid in
  if not (List.exists already_valid (Wpo.get_results g))
  then Task.spawn server 
    (Prover.prove ~callout:do_wpo_feedback g ~interactive prover)
    
let do_wp_proofs () =
  let pname = Wp_parameters.Prover.get () in
  match Wpo.prover_of_name pname with
    | None -> ()
    | Some prover ->
	let interactive = Wpo.is_interactive pname in
	let server = Prover.server () in
	try
	  Wpo.iter ~on_goal:(do_wp_proof server interactive prover) () ;
	  Task.launch server ;
	with e ->
	  Task.cancel_all server ;
	  raise e

let do_wp_proofs_for goals =
  let pname = Wp_parameters.Prover.get () in
  match Wpo.prover_of_name pname with
    | None -> ()
    | Some prover ->
	let interactive = Wpo.is_interactive pname in
	let server = Prover.server () in
	try
	  Bag.iter (do_wp_proof server interactive prover) goals ;
	  Task.launch server ;
	with e ->
	  Task.cancel_all server ;
	  raise e

(* ------------------------------------------------------------------------ *)
(* ---  Type Checking prover's inputs                                   --- *)
(* ------------------------------------------------------------------------ *)

let do_check_feedback g lang result =
  Wp_parameters.feedback "[%a] Goal %s : %a"
    Wpo.pp_language lang g.Wpo.po_gid Wpo.pp_result result

let do_wp_check server lang g =
 Task.spawn server (Prover.check ~callout:do_check_feedback g lang)

let do_wp_checks () =
  match Wpo.language_of_name (Wp_parameters.Check.get ()) with
    | None -> ()
    | Some lang ->
	let server = Prover.server () in
	try
	  Wpo.iter ~on_goal:(do_wp_check server lang) () ;
	  Task.launch server ;
	with e ->
	  Task.cancel_all server ;
	  raise e

let do_wp_checks_for goals =
  match Wpo.language_of_name (Wp_parameters.Check.get ()) with
    | None -> ()
    | Some lang ->
	let server = Prover.server () in
	try
	  Bag.iter (do_wp_check server lang) goals ;
	  Task.launch server ;
	with e ->
	  Task.cancel_all server ;
	  raise e
  
(* ------------------------------------------------------------------------ *)
(* ---  Filtering WP passes                                             --- *)
(* ------------------------------------------------------------------------ *)

let do_wp_passes fun_iter behaviors property =
  let model = Wp_parameters.get_model () in
  let goals = ref Bag.empty in
  let wp_pass computer assigns =
    do_compute goals computer fun_iter assigns behaviors property
  in
  begin 
    match computer model with
      | NonAssigns c ->
	  Wp_parameters.warning
	    ~current:false ~once:true
	    "Ignored Assigns-Goals with '%s' model"
	    (Wp_parameters.Model.get ()) ;
	  wp_pass c WpAnnot.NoAssigns
      | OneforBoth c -> 
	  wp_pass c WpAnnot.WithAssigns
      | OneforEach(c1,c2) ->
	  match property with
	    | None | Some (NamedProp _ | IdProp(IPBlob _ | IPBehavior _)) ->
		begin
		  wp_pass c1 WpAnnot.NoAssigns ;
		  wp_pass c2 WpAnnot.OnlyAssigns ;
		end
	    | Some(IdProp( IPPredicate _ | IPAxiom _
		   | IPCodeAnnot _ | IPComplete _
		   | IPDisjoint _ | IPDecrease _ )) ->
		wp_pass c1 WpAnnot.NoAssigns
	    | Some (IdProp(IPAssigns _)) ->
		wp_pass c2 WpAnnot.OnlyAssigns
  end ; !goals

(* ------------------------------------------------------------------------ *)
(* ---  Secondary Entry Points                                          --- *)
(* ------------------------------------------------------------------------ *)

(* Registered entry point in Dynamic. *)

let wp_compute kf_opt behaviors property =
  let p = match property with None -> None | Some p -> Some (IdProp p) in
  let fun_iter =
    match kf_opt with
      | Some kf -> on_function kf
      | None -> on_all_functions
  in
  let goals = do_wp_passes fun_iter behaviors p in
  if not (Bag.is_empty goals) then do_wp_proofs_for goals

(* Build the CFG dot file, and return its name if everything is OK. *)

let wp_cfg_dot kf =
  let _cfg, dotfile = build_cfg ~do_dot:true kf in
    match dotfile with None -> assert false (* asked dot explicitly ! *)
      | Some dotfile -> dotfile

(* ------------------------------------------------------------------------ *)
(* ---  Command-line Entry Points                                       --- *)
(* ------------------------------------------------------------------------ *)

let cmdline_run () = 
  let wp_main kf_list =
    Ast.compute ();
    Axiomatics.compute ();
    let fun_iter = match kf_list with
      | [] -> on_all_functions
      | names -> on_function_names names
    in
    let bhvs = Wp_parameters.get_behaviors () in
    let property =
      match Wp_parameters.get_properties () with
        | None -> None
        | Some [p] -> Some (NamedProp p)
        | Some _ ->
            Wp_parameters.not_yet_implemented "several properties in -wp-prop"
    in
    do_wp_passes fun_iter bhvs property
  in
  match Wp_parameters.job () with

    | Wp_parameters.WP_None -> 
	do_wp_checks () ;
	do_wp_proofs () ;
	do_wp_print ()
	  
    | Wp_parameters.WP_All -> 
	ignore (wp_main []) ;
	do_wp_checks () ;
	do_wp_proofs () ;
	do_wp_print () 
	    
    | Wp_parameters.WP_Select fcts -> 
	let goals = wp_main fcts in
	do_wp_checks_for goals ;
	do_wp_proofs_for goals ;
	do_wp_print_for goals

(* ------------------------------------------------------------------------ *)
(* ---  Register external functions                                     --- *)
(* ------------------------------------------------------------------------ *)

let wp_compute = Dynamic.register ~plugin:"Wp" "wp_compute"
                   (Datatype.func3
                      (Datatype.option Kernel_datatype.Kernel_function.ty)
                      (Datatype.option (Datatype.list Datatype.string))
                      (Datatype.option Property.ty)
                      Datatype.unit
                   )
                   ~journalize:false
                   wp_compute

(** Use options to know what to do *)
let run = Dynamic.register ~plugin:"Wp" "run"
            (Datatype.func Datatype.unit Datatype.unit)
            ~journalize:true
            cmdline_run

(* ------------------------------------------------------------------------ *)
(* ---  Main Entry Point                                                --- *)
(* ------------------------------------------------------------------------ *)

let main _fmt =
  begin
    cmdline_run () ;
    Wp_parameters.reset () ;
  end

let () = Db.Main.extend main

(* ------------------------------------------------------------------------ *)

