(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* -------------------------------------------------------------------------- *)
(** Memory Model for Runtime: 
* more information about it in {{:../../wp/Notes/m3.html}this document}
* *)
(* -------------------------------------------------------------------------- *)

open Cil_types
open Cil_datatype

let unsupported = Wp_error.unsupported

type compute_int_mode = CIMterms | CIMvalues | CIMcompute

module Create 
  (D:Mdata.S)
  (A:Mint.S   with module D = D)
  (R:Mfloat.S with module D = D)
  = 
struct

  module F = D.F

  (*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*)
  (** The memory is composed of 2 parts: one that deals with the values,
  * and the other one that handle allocation information. More over, there
  * is a table in the [Pre] memory to handle logic variables that are used for
  * the initial value of the parameters.
  * *)

  type m_mbits
  let t_mbits : Formula.tau = Formula.ADT ("memory", [])
  type mem_bits = m_mbits F.term

  type m_alloc
  let t_alloc : Formula.tau = Formula.ADT ("memalloc", [])
  type mem_alloc = m_alloc F.term

  type m_mem
  let t_mem : Formula.tau = Formula.ADT ("memory", [])
  type memory = { vbits : F.var ; valloc : F.var }
  type mem = memory * ((F.var Varinfo.Hashtbl.t) option)

  let vbits_of_mem (m,_) = m.vbits
  let valloc_of_mem (m,_) = m.valloc
  let logic_of_mem (_,tbl) = tbl
  let mbits_of_mem (m,_) : mem_bits = F.var m.vbits
  let malloc_of_mem (m,_) : mem_alloc = F.var m.valloc

  (*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*)
  let tcomp_of_comp comp =  (* TODO: we shouldn't loose the TComp ! *)
    TComp (comp, {scache = Not_Computed}, [])

  let cil_field_info f = 
    let t = tcomp_of_comp (f.fcomp) in
      try
        let offset = Field (f, NoOffset) in
          Cil.bitsOffset t offset 
      with Cil.SizeOfError (msg,t) -> 
        unsupported "sizeof %a : %s for field '%s'"
	  !Ast_printer.d_type t msg f.fname
  let cil_field_offset f = fst (cil_field_info f)
  let cil_field_size f = snd (cil_field_info f)

  let name_of_var vi : D.name = D.Xindex.get_ind vi
  let name_of_field f = D.Findex.get_ind f

  (** ----- Formats : *)

  let format_of_c_int_type c_int : D.format = 
    let fname = Pretty_utils.sfprintf "%a_format" Ctypes.pp_int c_int in
      D.e_app0 fname

  let format_of_c_float_type c_float : D.format = 
    let fname = Pretty_utils.sfprintf "%a_format" Ctypes.pp_float c_float in
      D.e_app0 fname

  let format_of_addr _ty = 
    (* format_of_c_int_type (Ctypes.c_ptr()) *)
      D.e_app0 "rt_addr_format"


  (*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*)
  (** Size and offset are sometimes known constants. We can choose to 
 * represent them are simple [int] and compute values as far as possible,
 * or to keep them as named terms.
 * The first solution gives smaller - yet easier provable - goals,
 * but if if fails, the user has few information the understand the problem.
 * We will try to have both solutions in order to be able to test them.
 *)

  let compute_int_mode = CIMterms (* TODO: add an option ? *)

  (** integer terms with a phantom type to mark what kind of object it is. *)
  type 'a tint = D.integer

  module Aint : sig
    type 'a t

    val of_int : int -> 'a t
    val of_int64 : Int64.t -> 'a t
    val of_term : 'a tint -> 'a t

    val add : 'a t -> 'b t -> 'a t
    val sub : 'a t -> 'b t -> 'a t
    val cnst_mult :  Int64.t -> 'a t -> 'a t
    val mult : 'b t -> 'a t -> 'a t

    val to_term : 'a t -> 'a tint

    val eq_pred : 'a t -> 'a t -> F.pred

  end = struct

    let compute = (compute_int_mode = CIMcompute)

    type aint = 
      | AIcnst of Int64.t 
      | AIterm of D.integer
      | AImult of Int64.t * D.integer
      | AIadd of aint * Int64.t

    type 'a t = aint

    let term_of_cnst i =  F.e_icst (Int64.to_string i)

    let of_int64 i = if compute then AIcnst i else AIterm (term_of_cnst i)
    let of_int (i:int) = of_int64 (Int64.of_int i)
    let of_term t = AIterm t

    let term_of_add t1 t2 = F.e_iop Formula.Iadd t1 t2
    let term_of_sub t1 t2 = F.e_iop Formula.Isub t1 t2
    let term_of_mult t1 t2 = F.e_iop Formula.Imul t1 t2

    let rec to_term (ai: 'a t) : 'a tint = match ai with
      | AIcnst i -> term_of_cnst i
      | AImult (i, t) -> term_of_mult (term_of_cnst i) t
      | AIadd (t, i) -> term_of_add (to_term t) (term_of_cnst i)
      | AIterm t -> t

    let cnst_is_zero i = 0 = Int64.compare i (Int64.zero)

    let is_zero ai = match ai with
      | AIcnst i -> cnst_is_zero i
      | _ -> false

    let add_cnst ai i = match ai with
      | AIcnst i' when compute -> AIcnst (Int64.add i i')
      | AIadd (t, i')  when compute -> 
          AIadd (t, Int64.add i i')
      | _ when cnst_is_zero i && compute -> ai
      | _ -> AIadd (ai, i)

    let rec cnst_mult i ai = 
      if compute then match ai with
        | AIcnst i' -> AIcnst (Int64.mul i i')
        | AImult (i', t) -> AImult (Int64.mul i i', t)
        | _ when cnst_is_zero i -> AIcnst Int64.zero
        | AIadd (t, i') -> 
            add_cnst (cnst_mult i t) (Int64.mul i i')
        | AIterm t -> AImult (i, t)
      else of_term (term_of_mult (term_of_cnst i) (to_term ai))

    let mult ai1 ai2 = match ai1, ai2 with
      | AIcnst i1, _ -> cnst_mult i1 ai2
      | _, AIcnst i2 -> cnst_mult i2 ai1
      | AImult (i1, t1), AImult (i2, t2) when compute ->
          cnst_mult (Int64.mul i1 i2) (AIterm (term_of_mult t1 t2))
      | _, _ -> (* TODO: develop other cases ? *)
          let t1 = to_term ai1 in
          let t2 = to_term ai2 in
            AIterm (term_of_mult t1 t2)

    let rec add ai1 ai2 = match ai1, ai2 with
      | AIcnst i1, _ -> add_cnst ai2 i1
      | _, AIcnst i2 -> add_cnst ai1 i2
      | AIadd (t1, i1), AIadd (t2, i2) when compute ->
          add_cnst (add t1 t2) (Int64.add i1 i2)
      | _, _ -> (* TODO: develop other cases ? *)
          let t1 = to_term ai1 in
          let t2 = to_term ai2 in
            AIterm (term_of_add t1 t2)

    let sub ai1 ai2 = match ai1, ai2 with (* TODO: compute... *)
      | _, _ ->
          let t1 = to_term ai1 in
          let t2 = to_term ai2 in
            AIterm (term_of_sub t1 t2)

    let eq_cnst i1 i2 = 0 = Int64.compare i1 i2

    let eq_pred ai1 ai2 = match ai1, ai2 with
      | AIcnst i1, AIcnst i2 when compute ->
          if eq_cnst i1 i2 then F.p_true else F.p_false
      | _, _ ->  (* TODO: develop other cases ? *)
          F.p_eq (to_term ai1) (to_term ai2)
  end

  (** Phantom types to tag the terms *)
  type m_addr
  type m_offset
  type m_size
  type m_zone

  (** Specialized type of D.integer terms *)
  type t_addr = m_addr tint
  type t_offset = m_offset tint
  type t_size = m_size tint
  type t_zone = m_zone F.term

  module Tint : sig 

    type x_addr
    val xaddr_of_var : mem_alloc -> varinfo -> x_addr
    val xaddr_of_gvar : varinfo -> x_addr
    val xaddr_of_integer : D.integer -> x_addr
    val integer_of_xaddr : x_addr -> D.integer
    val pointer_of_xaddr : x_addr -> D.pointer
    val varinfo_of_xaddr : x_addr -> varinfo option
    val xaddr_of_pointer : D.pointer -> x_addr
    val pp_addr : Format.formatter -> x_addr -> unit
    val base : mem_alloc -> x_addr -> x_addr
    val term_of_xaddr : x_addr -> t_addr

    type x_size
    val size_of_int : int -> x_size
    val size_of_int64 : Int64.t -> x_size
    val cnst_mult_size : Int64.t -> x_size -> x_size
    val xsize_of_range : D.integer -> D.integer -> x_size -> x_size
    val term_of_xsize : x_size -> t_size

    val shift_n_elem : x_addr -> D.integer -> x_size -> x_addr
    val shift_field : x_addr -> fieldinfo -> x_addr

    val toffset_of_field : ?mode:compute_int_mode -> fieldinfo -> t_offset
    val tsize_of_field : ?mode:compute_int_mode -> fieldinfo -> t_size

    type x_zone
    val mk_xzone : x_addr -> x_size -> x_zone
    val xzone_of_var : mem_alloc -> varinfo -> x_size -> x_zone
    val pp_xzone : Format.formatter -> x_zone -> unit
    val term_of_xzone : x_zone -> m_zone F.term
    val eq_zone : x_zone -> x_zone -> F.pred
    val xzone_disj : x_zone -> x_zone -> F.pred

  end = struct

    (*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*)
    (**  {2 Extracted specification from runtime.why} 
    * see definitions and axioms in 
    * {{:../../../share/why/runtime.why}runtime.why}
    *)
    module L = struct
      let vaddr : mem_alloc -> D.name -> t_addr = D.e_app2 "rt_vaddr"
      let gaddr : D.name -> t_addr =  D.e_app1 "rt_gaddr"

      let zone : t_addr -> t_size -> t_zone =  D.e_app2 "rt_zone"
      let vzone : mem_alloc -> D.name-> t_zone =  D.e_app2 "rt_vzone"

      let foffset : D.name -> t_offset = D.e_app1 "rt_foffset"
      let fsize : D.name -> t_size = D.e_app1 "rt_fsize"

      let shift : t_addr -> t_offset -> t_addr = D.e_app2 "rt_shift"

      let disj : t_zone -> t_zone -> F.pred = D.p_app2 "rt_disj"
     
      let base : mem_alloc -> t_addr -> t_addr = D.e_app2 "rt_abase"

      let addr_of_pointer : D.pointer -> t_addr = D.e_app1 "rt_addr_of_pointer"
      let pointer_of_addr : t_addr -> D.pointer = D.e_app1 "rt_pointer_of_addr"

    end

    type x_addr = 
      | Lvaddr of (mem_alloc * varinfo)
      | Laddr of m_addr Aint.t 
      | Lptr of D.pointer

    type x_offset = m_offset Aint.t
    type x_size = m_size Aint.t
    type x_zone = 
      | Zterm of t_zone
      | Zpair of x_addr * x_size

    let xaddr_of_var ma v = Lvaddr (ma, v)
    let xaddr_of_gvar v = 
      Laddr (Aint.of_term (L.gaddr (name_of_var v)))

    let term_of_xaddr ai : t_addr = match ai with
      | Lvaddr (ma, v) -> L.vaddr ma (name_of_var v)
      | Laddr ai -> Aint.to_term ai
      | Lptr ptr -> L.addr_of_pointer ptr

    let aint_of_xaddr a : m_addr Aint.t = match a with
      | Laddr a -> a
      | _ -> Aint.of_term (term_of_xaddr a)

    let varinfo_of_xaddr a = match a with
      | Lvaddr (_, vi) -> Some vi
      | _ -> None

    (** Notice that [pointer_of_loc] and [loc_of_pointer] are abstract
    * operations to switch between generic part and model part.
    * They are not a real casts of any kind. *)
    let pointer_of_xaddr a : D.pointer = match a with
      | Lptr ptr -> ptr
      | _ -> L.pointer_of_addr (term_of_xaddr a)

    let xaddr_of_pointer p = Lptr p

    let pp_addr fmt a = match a with
      | Lptr t ->  F.pp_term fmt t
      | _ ->  F.pp_term fmt (term_of_xaddr a)

    let xaddr_of_integer ti = Laddr (Aint.of_term ti)
    let integer_of_xaddr xa = term_of_xaddr xa

    let base ma a =  Laddr (Aint.of_term (L.base ma (term_of_xaddr a)))

    let offset_of_int i : x_offset = Aint.of_int i
    let size_of_int sz : x_size = Aint.of_int sz
    let size_of_int64 sz : x_size = Aint.of_int64 sz
    let cnst_mult_size n (sz:x_size) : x_size = Aint.cnst_mult n sz

    let xsize_of_range min max sz =
      let nb = Aint.sub (Aint.of_term max) (Aint.of_term min) in
      let nb = Aint.add nb (Aint.of_int 1) in
         Aint.mult nb sz

    let term_of_xsize ai : t_size = Aint.to_term ai

    let xoffset_of_field ?(mode=compute_int_mode) f =  
      if mode = CIMterms then 
        Aint.of_term (L.foffset (name_of_field f))
      else offset_of_int (cil_field_offset f)

    let toffset_of_field ?(mode=compute_int_mode) f =
      Aint.to_term (xoffset_of_field ~mode f)

    let xsize_of_field ?(mode=compute_int_mode) f =  
      if mode = CIMterms then 
        Aint.of_term (L.fsize (name_of_field f))
      else size_of_int (cil_field_size f)

    let tsize_of_field ?(mode=compute_int_mode) f =
      Aint.to_term (xsize_of_field ~mode f)

    let shift (a:x_addr) (offset:x_offset) : x_addr = 
      if compute_int_mode = CIMcompute then
         Laddr (Aint.add (aint_of_xaddr a) offset)
      else
        Laddr (Aint.of_term 
                 (L.shift (term_of_xaddr a) (Aint.to_term offset)))

    let shift_n_elem a n sz = 
      let o : x_offset = Aint.of_term n in
      let o = Aint.mult sz o in
      shift a o

    let shift_field a f = shift a (xoffset_of_field f)

    let xzone_of_var ma v sz =
      if compute_int_mode = CIMterms then
        Zterm (L.vzone ma (name_of_var v))
      else 
        Zpair (xaddr_of_var ma v, sz)

    let mk_xzone a sz : x_zone = match a with
      | Lvaddr (ma, v) -> xzone_of_var ma v sz
      | _ -> Zpair (a, sz)

    let term_of_xzone xz = match xz with
      | Zterm t -> t
      | Zpair (a, sz) -> L.zone (term_of_xaddr a) (term_of_xsize sz)

    let pp_xzone fmt xz = F.pp_term fmt (term_of_xzone xz)

    let eq_zone z1 z2 = (* TODO: SMP *)
      F.p_eq (term_of_xzone z1) (term_of_xzone z2)

    let xzone_disj z1 z2 = (* TODO: SMP *)
      L.disj (term_of_xzone z1) (term_of_xzone z2)
  end

  (* open Tint *)
  (*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*)
  (** {3 about size and offset} *)

  let rec sizeof_c_object t : Tint.x_size = match t with
    | Ctypes.C_comp comp -> 
        begin
          try
            let t = tcomp_of_comp comp in
            let sz = Cil.bitsSizeOf t in
              Tint.size_of_int sz
          with Cil.SizeOfError (msg, t) -> 
            unsupported "sizeof %a : %s" !Ast_printer.d_type t msg
        end
    | Ctypes.C_array {Ctypes.arr_flat = Some flat} ->
        let nb = flat.Ctypes.arr_cell_nbr in
        let sz = sizeof_c_object (Ctypes.object_of flat.Ctypes.arr_cell) in
          Tint.cnst_mult_size nb sz
    | _ -> Tint.size_of_int64 (Ctypes.sizeof_object t)

  (*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*)
    (**  {2 Extracted specification from runtime.why} 
    * see definitions and axioms in 
    * {{:../../../share/why/runtime.why}runtime.why}
    *)

  type m_bits
  type t_bits = m_bits F.term

  type m_dzone
  type dzone = m_dzone F.term

  module RtLib = struct
    let rt_global : D.name -> F.pred = D.p_app1 "rt_global"
    let rt_gsize : D.name -> t_size = D.e_app1 "rt_gsize"

    let rt_vformat : D.name -> D.format = D.e_app1 "rt_vformat"
    let rt_fformat : D.name -> D.format = D.e_app1 "rt_fformat"

    let load : mem_bits -> t_zone -> t_bits = D.e_app2 "rt_load"
    let store : mem_bits -> t_addr -> t_bits -> mem_bits = D.e_app3 "rt_store"
    let havoc : mem_bits -> t_zone -> mem_bits = D.e_app2 "rt_havoc"

    let to_bits : D.format -> D.abstract -> t_bits = D.e_app2 "rt_to_bits"
    let from_bits : t_bits -> D.format -> D.abstract = D.e_app2 "rt_from_bits"

    let alloc : mem_alloc -> D.name -> t_size -> mem_alloc 
                                                   = D.e_app3 "rt_valloc"

    let block_length : mem_alloc -> t_addr -> t_size 
                                                = D.e_app2 "rt_block_length"
    let valid : mem_alloc -> t_zone -> F.pred =  D.p_app2 "rt_valid"

    let is_havoc : mem_alloc -> mem_bits -> dzone -> mem_bits -> F.pred = 
      D.p_app4 "rt_is_havoc"

    let free : mem_alloc -> D.name -> mem_alloc = D.e_app2 "rt_vfree"

    let zs_empty : dzone = D.e_app0 "zs_empty"
    let zs_singleton : t_zone -> dzone = D.e_app1 "zs_singleton"
    let zs_union : dzone -> dzone -> dzone = D.e_app2 "zs_union"
    let zs_incl : dzone -> dzone -> F.pred = D.p_app2 "zs_incl"
  end

  let z_from_bits b fmt : D.integer = F.unwrap (RtLib.from_bits b fmt)
  let real_from_bits b fmt : D.real = F.unwrap (RtLib.from_bits b fmt)

  (*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*)

  module Model =
  struct

    module D = D
    module A = A
    module R = R
    
    type loc = Tint.x_addr

    let pp_loc fmt l = Tint.pp_addr fmt l

    let pointer_of_loc loc : D.pointer = Tint.pointer_of_xaddr loc
    let loc_of_pointer _tp p : loc =  Tint.xaddr_of_pointer p

    let cast_loc_to_int _tp loc ti : D.integer =
      let int_term = Tint.term_of_xaddr loc in
      D.i_convert (Ctypes.c_ptr()) ti int_term
      

    let cast_int_to_loc ti (i:D.integer) _tp : loc =
      let i = D.i_convert ti (Ctypes.c_ptr()) i in
      (Tint.xaddr_of_integer i)

  end

  let startof loc _ = loc

  let cast_loc_to_loc _t1 _t2 p = p

  (*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*)
  (** {3 about zone} *)
  (** 
  * Mwp requires a type [dzone] defined as [m_zone F.term]
  * that represent why formula of the model why zone type
  * (which name is given by [tau_of_zone] below).
  *
  * Because we need to provide function such as [dzone_union],
  * this [dzone] is defined as a set of the [rt_zone] defined in WHY.
   *)

  let tau_of_dzone = Formula.ADT("zones",[])

(*
  let zone_of_loc t loc = Tint.mk_xzone loc (sizeof_c_object t)
  let tzone_of_loc te loc = Tint.term_of_xzone (Tint.mk_xzone loc (sizeof_c_object te)
    
  let zone_shift te addr rg = 
    match rg with
      | {D.inf = Some min; D.sup = Some max} ->
          let sz = sizeof_c_object te in
          let addr = Tint.shift_n_elem addr min sz in
          let size = Tint.xsize_of_range min max sz in
          Tint.mk_xzone addr size
      | _ -> unsupported "unbounded range"

  let zone_index = zone_shift

  let tzone_of_loc t loc = Tint.term_of_xzone (zone_of_loc t loc)

  let dzone_of_tzone tz = RtLib.zs_singleton tz
  let dzone_of_xzone xz = dzone_of_tzone (Tint.term_of_xzone xz)

  let dzone_of_zone _cobj xz = dzone_of_xzone xz

  let dzone_of_loc te loc = dzone_of_tzone (tzone_of_loc te loc)
*)

  let xzone_assigned = function
    | D.Aloc( te , loc ) ->
	Tint.mk_xzone loc (sizeof_c_object te)
    | D.Arange( te , loc , rg ) ->
	match rg with
	  | {D.inf = Some min; D.sup = Some max} ->
              let sz = sizeof_c_object te in
              let addr = Tint.shift_n_elem loc min sz in
              let size = Tint.xsize_of_range min max sz in
              Tint.mk_xzone addr size
	  | _ -> unsupported "unbounded range"

  let dzone_assigned a = RtLib.zs_singleton (Tint.term_of_xzone (xzone_assigned a))
  let dzone_empty () = RtLib.zs_empty

  let dzone_subset dz1 dz2 = RtLib.zs_incl dz1 dz2

  let dzone_union dz1 dz2 = RtLib.zs_union dz1 dz2

  let effect_supported = true

  (*---------------------------------- *)
  (** Regions *)

  type region
  let pp_region _ _ = ()
  let region_assigned _ = assert false
  let region_empty _ = assert false
  let region_union _ _ = assert false
  let region_included _ _ = assert false
  let region_separated _ _ = assert false
  let region_fingerprint _ _ = assert false

  let region_supported = false

  (*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*)

  let int_format_for_hyp t = match t with
    | Ctypes.C_int cint -> Some (format_of_c_int_type cint)
    | Ctypes.C_pointer ty -> Some (format_of_addr ty)
    | _ -> None

  let add_int_format_hyp vformat t h = match int_format_for_hyp t with
    | None -> h
    | Some fmt ->
        let h_format =  F.p_eq vformat fmt in
          F.p_and h_format h

  module VarDecl = D.Register
    (struct
       include D.Varinfo

     (** Global variable has a fixed zone in any allocation memory.
     * [forall ma, rt_vsize (ma, v) = sz /\ rt_vaddr (ma, v) = rt_gaddr v] *)
       let declare v _ =
         let t = Ctypes.object_of v.vtype in
         let h = 
           if v.vglob then
             let h_glob = RtLib.rt_global (name_of_var v) in
             let sz = sizeof_c_object t in
             let gsize = RtLib.rt_gsize (name_of_var v) in
             let h_size = F.p_eq gsize (Tint.term_of_xsize sz) in
               F.p_and h_glob h_size 
               else F.p_true
         in
         let vformat = RtLib.rt_vformat (name_of_var v) in
         let h = add_int_format_hyp vformat t h in
           (* TODO: size info might be redondant with format... *)
           (* TODO: format for other types... *)
           Formula.Axiom h

       let section = Formula.S_typedef
       let prefix = "Decl"
       let basename x = x.vname
       let clear () = ()
       let pp_descr fmt _x =
         Format.fprintf fmt "Global declaration"
     end)
  (*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*)
  module Fields = D.Register
    (struct
       include D.Fieldinfo

       let declare f _ =
         let foff = Tint.toffset_of_field ~mode:CIMterms f in
         let off = Tint.toffset_of_field ~mode:CIMvalues f in
         let hoff = F.p_eq foff off in

         let fsz = Tint.tsize_of_field ~mode:CIMterms f in
         let sz = Tint.tsize_of_field ~mode:CIMvalues f in
         let hsz = F.p_eq fsz sz in
         let h = F.p_and hoff hsz in

         let t = Ctypes.object_of  f.ftype in
         let fformat = RtLib.rt_fformat (name_of_field f) in
         let h = add_int_format_hyp fformat t h in

           Formula.Axiom h

       let section = Formula.S_typedef
       let prefix = "Finfo"
       let basename x = 
         let name = D.Compinfo.basename  x.fcomp in
           (name^"_"^x.fname)
       let clear () = ()
       let pp_descr fmt _x =
         Format.fprintf fmt "Field info"
     end)
  (*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*)
  module Data = struct 

    include Datalib.Cvalues(Model) 

    type m_of_mem = m_mbits

    let tau_of_mem = t_mbits

    let forall_loc pool =
      let p = F.fresh pool "p" (Formula.Model Formula.Integer) in
      [p] , (Tint.xaddr_of_integer (F.var p))
	
    let cvar m vi = 
      VarDecl.define vi ;
      Tint.xaddr_of_var (malloc_of_mem m) vi

    let shift (l:loc) t i : loc = Tint.shift_n_elem l i (sizeof_c_object t)

    let index = shift

    let field (l:loc) f = 
      Fields.define f;
      if f.fcomp.cstruct then Tint.shift_field l f else l

    let value_of_bits = ref (fun _ _ -> assert false)
    let bits_of_value = ref (fun _ _ -> assert false)

    (** Read a data of type [te] at [loc] and returns it as a logic value. *)
    let load_mem mb te loc =
      let xzone = Tint.mk_xzone loc (sizeof_c_object te) in
      let tzone = Tint.term_of_xzone xzone in
      let bits = RtLib.load mb tzone in
      (!value_of_bits te bits)

    let store_mem mb te loc v =
      let tzone = Tint.term_of_xaddr loc in
      RtLib.store mb tzone (!bits_of_value te v)

  end

  module DF = Data_mem.Create(Data)

  let load m te loc = 
    DF.loaded te; 
    match Tint.varinfo_of_xaddr loc, logic_of_mem m with
      | Some vi, Some tbl when vi.vformal -> 
          let v = 
            try Varinfo.Hashtbl.find tbl vi
            with Not_found ->
              let v = D.fresh vi.vname (Mdata.Vacsl (Ctype vi.vtype)) in
              (Varinfo.Hashtbl.add tbl vi v; v)
          in Data.value_of_logic te (F.var v)
      | _, _ -> 
          let mb = mbits_of_mem m in
          Data.load_mem mb te loc

  let store m loc t exp_val = 
    DF.stored t; 
    let mb = mbits_of_mem m in
    Data.store_mem mb t loc exp_val

  include Data
  (*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*)
  (** {3 about bits} 
  * [bits] type is used to represent the values that are stored in the memory.
  * We have functions that encode/decode bits to/from ACSL logic values
  * depending on the intermediate C type [t].
  * *)

  let format_of_compinfo comp =
    if comp.cstruct then DF.record_format comp
    else unsupported "format of union"

  let format_of_array arr = DF.array_format arr 

  let rt_format_of_ctype t = match t with
    | Ctypes.C_int c_int -> format_of_c_int_type c_int
    | Ctypes.C_float c_float ->  format_of_c_float_type c_float
    | Ctypes.C_comp comp -> format_of_compinfo comp
    | Ctypes.C_array arr -> format_of_array arr
    | Ctypes.C_pointer ty -> format_of_addr ty

  (** Compute the logic value from bits interpreted with type [t]. *)
  let rec value_of_bits t bits : Data.value = match t with
    | Ctypes.C_int c_int -> 
        let c_val = z_from_bits bits (format_of_c_int_type c_int) in
        V_int (c_int, c_val)
    | Ctypes.C_float c_float -> 
        let ft = format_of_c_float_type c_float in
        let c_val = real_from_bits bits ft in
        V_float (c_float, c_val)
    | Ctypes.C_comp comp -> 
        let ft = format_of_compinfo comp in
        let c_val =  RtLib.from_bits bits ft in
        V_record (comp, F.unwrap c_val)
    | Ctypes.C_array arr -> 
        let ft = format_of_array arr in
        let c_val =  RtLib.from_bits bits ft in
        let logic_val = (* D.encode ft*) (F.unwrap c_val) in
          V_array (arr, F.unwrap logic_val)
    | Ctypes.C_pointer ty ->
        let c_val = z_from_bits bits (format_of_addr ty) in
        let addr = Tint.xaddr_of_integer c_val in
        V_pointer (ty, addr)

  (* TODO: is it normal that we don't use t?
  * Maybe we should check that it is the same than in the value ??? *)
  let rec bits_of_value _t value : t_bits = match value with
    | V_int (c_int, i) -> 
        let ft = format_of_c_int_type c_int in
          RtLib.to_bits ft (F.wrap i)
    | V_float (c_float, f) ->
        let ft = format_of_c_float_type c_float in
          RtLib.to_bits ft (F.wrap f)
    | V_pointer (ty, loc) ->
        let ft = format_of_addr ty in
          RtLib.to_bits ft (F.wrap (Tint.integer_of_xaddr loc))
    | V_union _ -> unsupported "bits_of_value of union"
    | V_record (comp, r) ->
        let ft = format_of_compinfo comp in
        let e = (*D.decode ft*) (F.wrap r) in
          RtLib.to_bits ft e
    | V_array (arr, t) ->
        let ft = format_of_array arr in
        let e = (* D.decode ft*) (F.wrap t) in
          RtLib.to_bits ft e
    
  (** Horrible thing but no (known) way to avoid... *)
  let () = 
    begin
      Data.value_of_bits := value_of_bits ;
      Data.bits_of_value := bits_of_value ;
    end
  (*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*)

  (** {3 Frame Environment} *)

  type frame = {
    mutable memories : (Clabels.c_label * mem) list ;
    return : Cil_types.typ option ;
    mutable result : F.var option ;
    mutable exit_status : F.var option ;
  }


  let new_frame kf ~m_here ~m_pre ~m_post res_opt =
    let add mm l m = match m with None -> mm | Some m -> (l, m)::mm in
    let mm = [] in
    let mm = add mm Clabels.Post m_post in
    let mm = add mm Clabels.Pre  m_pre  in
    let mm = add mm Clabels.Here m_here in
    {
      memories = mm;
      return = Some(Kernel_function.get_return_type kf);
      result = res_opt;
      exit_status = None;
    }

  let new_mem () =
    let va = D.fresh "ma" (Mdata.Vmodel t_alloc) in
    let vb = D.fresh "mb" (Mdata.Vmodel t_mbits) in
      { vbits = vb; valloc = va }

  (** @raise Not_found when the label has no memory. *)
  let find_mem_at env l = List.assoc l env.memories

  let mem_at env l =
    try find_mem_at env l
    with Not_found ->
      let m = new_mem () in
      let params = 
        if l = Clabels.Pre then Some (Varinfo.Hashtbl.create 10) else None 
      in
      let m = (m, params) in
        env.memories <- (l,m) :: env.memories;
        m

  let result env =
    match env.result, env.return with
      | Some x, Some t  -> x, t
      | Some _,  _ -> Wp_parameters.fatal "Result type undefined"
      | None ,_ ->
          match env.return with
            | Some typ ->
                let x = D.fresh "result" (Mdata.Vacsl (Ctype typ)) in
                env.result <- Some x ; (x, typ)
            | None ->
                Wp_parameters.fatal "Result type undefined"

  let exit_status env =
    match env.exit_status with
      | Some x -> x
      | None ->
          let x = D.fresh "exit_status" (Mdata.Vmodel Formula.Integer) in
          env.exit_status <- Some x ; x

  let set_exit_status f x = { f with
    result = None ;
    exit_status = Some x ;
  }

  let get_exit_status env = env.exit_status

  (** {3 Validity }*)

  let valid m a = 
    let xzone = xzone_assigned a in
    RtLib.valid (malloc_of_mem m) (Tint.term_of_xzone xzone)

  let separated z1 z2 =
    Tint.xzone_disj (xzone_assigned z1) (xzone_assigned z2)

  let tbits_of_var a : t_bits = F.var a

  let subst_havoc env a =
    let xzone = xzone_assigned a in
    let m_here = mem_at env  Clabels.Here in
    let vmh = vbits_of_mem m_here in
    (* let v = D.fresh "v" (Mdata.Vmodel(Formula.ADT("bits",[]))) in
    let bits = tbits_of_var v in *)
    let new_vmh sigma =
      let mb = D.apply sigma (F.var vmh) in
        F.wrap (RtLib.havoc mb (Tint.term_of_xzone xzone))
        (* missing hyp about [bits] size :
        * F.wrap (RtLib.store mb (Tint.term_of_xzone xzone) bits) *)
    in
    [(* D.Fresh v; *) D.Update(vmh, new_vmh)]

  let assigns_goal env from_label region to_label =
    let mfrom = mem_at env from_label in
    let mto = mem_at env to_label in
    let zones =
      match region with
	| [] -> dzone_empty ()
	| [a] -> dzone_assigned a
	| a::others -> 
	    List.fold_left
	      (fun dz a ->
		 dzone_union dz (dzone_assigned a)
	      ) (dzone_assigned a) others
    in
    let ma = malloc_of_mem mfrom in
    RtLib.is_havoc ma (mbits_of_mem mfrom) zones (mbits_of_mem mto)
      
  let assigns_supported = true
      
  (** {3  Special locations} *)

  let base_address m loc = Tint.base (malloc_of_mem m) loc

  let block_length m p = 
    RtLib.block_length (malloc_of_mem m) (Tint.term_of_xaddr p)

  (*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*)
  (** {3 User-defined Predicates} *)

  type closure =
    | Mem of string  (* The logic labels used *)
    | Alloc of string  (* The logic labels used *)

  let pp_closure fmt  = function
    | Mem s ->Format.fprintf fmt "@@%s" s
    | Alloc s ->Format.fprintf fmt "@@%s" s

  let userdef_frame () = {
    memories = [] ;
    result = None ;
    exit_status = None;
    return = None ;
  }

  (* TODO filter with respect to tused and pused *)
  let userdef_signature frame _tused _pused =
    let add (acc_vmem, acc_lab) (label, m) =
      let acc_lab = match label with
        | Clabels.LabelParam l -> Alloc l :: Mem l :: acc_lab
        | _ -> Wp_parameters.fatal
                 "Unexpected statement label in user-defined symbol"
      in 
      let vma = valloc_of_mem m in
      let vmb = vbits_of_mem m in
      vma::vmb::acc_vmem, acc_lab
    in List.fold_left add ([], []) frame.memories
    
  let userdef_closure mem cl = match cl with
    | Mem _ -> F.wrap (mbits_of_mem mem)
    | Alloc _ -> F.wrap (malloc_of_mem mem)

  let label_at_closure = function
    | Mem l -> l
    | Alloc l -> l

  (*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*)
  (** {2 Mwp.S requirements} *)

  let update_at_label env label p = 
    if label = Clabels.Here then p
    else try
      let mem_here = mem_at env Clabels.Here in
      let mem_label = find_mem_at env label in
      let vma_label = valloc_of_mem mem_label in
      let vmb_label = vbits_of_mem mem_label in
      let p = D.subst vmb_label (mbits_of_mem mem_here) p in
      let p = D.subst vma_label (malloc_of_mem mem_here) p in
        p
    with Not_found -> p

  let quantify_at_label env label p = 
    try
      let m = find_mem_at env label in
      let l = match logic_of_mem m with
      | None -> []
      | Some tbl -> Varinfo.Hashtbl.fold (fun _ v l -> v::l) tbl [] 
      in
      let ma = valloc_of_mem m in
      let mb = vbits_of_mem m in
      let l = ma::mb::l in
      D.forall l p 
    with Not_found -> p

  let subst_lval env t ptr exp p = 
    let mem_here = mem_at env Clabels.Here in
    let new_mem = store mem_here ptr t exp in
    let vmb = vbits_of_mem mem_here in
    D.subst vmb (F.wrap new_mem) p

  let subst_result env v_exp p = 
    match env.result with
      | None -> p
      | Some xres -> 
          match v_exp with 
            | Some v_exp -> D.subst xres (logic_of_value v_exp) p
            | None -> D.forall [xres] p

  (* ------------------------------------------------------------------------ *)
  (* ---  Local Scope                                                     --- *)
  (* ------------------------------------------------------------------------ *)

  let local_scope env vars scope_kind p =
    let m_here = mem_at env Clabels.Here in
    let vmh = valloc_of_mem m_here in
    let mh = F.var vmh in
    let do_var p v =
      let t = Ctypes.object_of v.vtype in
      let alloc v p =
        Wp_parameters.debug ~level:3 "[local_scope] alloc %s@." v.vname;
        let sz = sizeof_c_object t in
        let sz = Tint.term_of_xsize sz in
        let ma = RtLib.alloc mh (name_of_var v) sz in
        let p = D.subst vmh ma p in
        p
      in
      let init v p = (* initialize parameter : v@Here = v@Pre *)
        let m_pre = mem_at env Clabels.Pre in
        let v_loc = cvar m_here v in
        let v_pre = load m_pre t v_loc in
        let p = subst_lval env t v_loc v_pre p in
        p
      in
        let p = match scope_kind with
        | Mcfg.SC_Function_in -> 
            (* let p = alloc v p in p  *)
            let p = match logic_of_mem (mem_at env Clabels.Pre) with
              | None -> p
              | Some tbl ->
                try let lv = Varinfo.Hashtbl.find tbl v in
                  D.forall [lv] p
                with Not_found -> p
            in p
        | Mcfg.SC_Function_frame ->
            let p = init v p in 
            let p = alloc v p in
            p
        | Mcfg.SC_Block_in ->
            let p = alloc v p in
            (*let h = add_int_format_hyp (name_of_var v) t F.p_true in
             F.p_implies h p 
             *) p
        | Mcfg.SC_Block_out | Mcfg.SC_Function_out ->
            Wp_parameters.debug ~level:3 "[local_scope] free %s@." v.vname;
            D.subst vmh (RtLib.free mh (name_of_var v)) p 
	| Mcfg.SC_Global -> (* nothing to do *) p
      in p
    in List.fold_left do_var p vars

  (* ------------------------------------------------------------------------ *)
  (* ---  Export                                                          --- *)
  (* ------------------------------------------------------------------------ *)

  type goal = F.pred
  let make_goal _env p = p

end 
