(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* -------------------------------------------------------------------------- *)
(** Memory Model with separation                                              *)
(* -------------------------------------------------------------------------- *)

module WpLog = Wp_parameters
open Cil_types
open Formula
open Ctypes
open Clabels
open Mdata

module Create
  (D:Mdata.S)
  (A:Mint.S   with module D = D)
  (R:Mfloat.S with module D = D)
  =

struct

  let unsupported = Wp_error.unsupported

  module F=D.F

  (* ------------------------------------------------------------------------ *)
  (* --- Term Types                                                       --- *)
  (* ------------------------------------------------------------------------ *)

  type m_memory
  let t_memory : tau = Formula.ADT("farray",[Formula.ADT("data",[])])
  type store = m_memory F.term

  type m_alloc
  let t_alloc : tau = Formula.ADT("farray",[Formula.Integer])
  type tbl_alloc = m_alloc F.term

  type mem = { store : store  ; talloc : tbl_alloc}

  type m_dzone
  type dzone = m_dzone F.term

  let addr_format : D.format = D.e_app0 "addr_format"

  (* ----------------------------------------------------------------------- *)
  (* --- Extracted specification from store.why                          --- *)
  (* ----------------------------------------------------------------------- *)

  (* Address *)

  let model_addr : D.integer -> D.integer -> D.integer = D.e_app2 "addr"
  let model_base : D.integer -> D.integer = D.e_app1 "base"
  let model_offset : D.integer -> D.integer = D.e_app1 "offset"

  let model_addr_shift : D.integer -> D.integer -> D.integer =
    D.e_app2 "addr_shift"

  let model_addr_of_pointer : D.pointer -> D.integer =
    D.e_app1 "addr_of_pointer"

  let model_offset_of_pointer : D.pointer -> D.integer =
    D.e_app1 "offset_of_pointer"

  let model_pointer_of_addr : D.integer -> D.pointer =
    D.e_app1 "pointer_of_addr"

  (* Allocation *)

  let model_valid :
      tbl_alloc -> D.integer -> D.integer -> F.pred =
    D.p_app3 "valid"

  let model_isfresh : store -> tbl_alloc -> D.integer -> F.pred =
    D.p_app3 "is_fresh"

  let model_alloc
      (talloc:tbl_alloc) (p:D.integer) (sz:D.integer) : tbl_alloc =
    D.e_app3 "update" talloc p sz

  let model_free (talloc : tbl_alloc) (p: D.integer) : tbl_alloc =
    D.e_app3 "update" talloc p (F.e_int 0)

  let model_block (talloc : tbl_alloc) (p: D.integer) : D.integer =
    D.e_app2 "access" talloc p

  (* Memory manipulation *)

  let model_access : store -> D.integer -> D.abstract =
    D.e_app2 "access"

  let model_update: store -> D.integer -> D.abstract -> store =
    D.e_app3 "update"

  (* Zone *)

  let model_zempty : dzone = D.e_app0 "zempty"
  let model_zunion : dzone -> dzone -> dzone = D.e_app2 "zunion"
  let model_included : dzone -> dzone -> F.pred = D.p_app2 "included"

  let model_zrange : D.integer -> D.integer -> D.integer -> dzone = 
    D.e_app3 "zrange"

  let model_zrange_of_addr_range : 
      D.integer -> D.integer -> D.integer -> dzone = 
    D.e_app3 "zrange_of_addr_range"

  let model_separated : dzone -> dzone -> F.pred = D.p_app2 "separated"

  (* Record, Array acces, update  *)

  let model_update_range: store -> dzone -> D.abstract -> store = 
    D.e_app3 "update_range"
  let model_access_range: store -> dzone  -> D.abstract = 
    D.e_app2 "access_range"

      
  let model_update_havoc: store -> dzone -> D.abstract -> store = 
    D.e_app3 "update_havoc"
  let model_ishavoc : tbl_alloc -> store -> dzone -> store -> F.pred =
    D.p_app4 "is_havoc"

  (* ----------------------------------------------------------------------- *)
  (* --- Region instanciation                                            --- *)
  (* ----------------------------------------------------------------------- *)


  module Z = Region.Create(D)
    (struct
       module F = D.F
       type addr = D.integer
       let pp_addr fmt a = Format.fprintf fmt "Addr %a" F.pp_term a
       let closed = F.term_closed
       let equal = F.p_eq
       let fresh pool =
	 let x = F.fresh pool "A" (Model Integer) in
	 x , F.var x
     end)


  (* ----------------------------------------------------------------------- *)
  (* --- Instanciation of MVALUE and MLOGIC : Store implemantation       --- *)
  (* ----------------------------------------------------------------------- *)

  let zero = F.e_int 0
  let one = F.e_int 1
  let sub = F.e_iop Isub
  let add = F.e_iop Iadd
  let mul = F.e_iop Imul

  let rec sizeof = function
 | C_comp cinfo ->
     List.fold_left
       (fun sz f -> add sz  (sizeof (object_of f.ftype))) zero cinfo.cfields
 | C_array ainfo ->
     begin
       match ainfo.arr_flat with
	 | Some a -> mul
	     (sizeof (object_of a.arr_cell))  (F.e_int (Int64.to_int a.arr_cell_nbr))
	 | None -> WpLog.not_yet_implemented "Sizeof unknown-size array"
     end
 | _ -> one 
    
  let n_size n te =  mul (sizeof te) n
  let add_offset d te k = add d (n_size k te)
  let cardinal a b = add one (sub b a)

  let offset_of_field f =
    let rec acc sz l f =
      match l with
	| [] -> assert false
	| fi::m ->
	    if Cil_datatype.Fieldinfo.equal f fi
	    then sz
	    else  acc (add sz  (sizeof (object_of fi.ftype))) m f
    in
    acc zero f.fcomp.cfields f


  module Model =
  struct

    module A=A
    module R=R
    module D=D

    type loc =
      | Loc of varinfo * D.integer (*base,offset*)
      | Addr of D.integer

    let addr = function
      | Loc(x,k) -> model_addr (D.Xindex.get_ind x) k 
      | Addr p -> p

    let base = function
      | Loc(x,_) -> D.Xindex.get_ind x
      | Addr p -> model_base p

    let offset = function
      | Loc(_,k) -> k
      | Addr p -> model_offset p


    let loc_of_pointer _ ptr = Addr(model_addr_of_pointer ptr)
    let pointer_of_loc loc   = model_pointer_of_addr (addr loc)

    let cast_loc_to_int _tp loc _ti = addr loc

    let cast_int_to_loc _ti i _tp = Addr i

    let pp_loc fmt = function
      | Loc(x,k) ->
	  Format.fprintf fmt "addr(%a,%a)" !Ast_printer.d_var x F.pp_term k
      | Addr(p) ->  F.pp_term fmt p

  end

  module Globals = D.Register
    (struct
       include D.Varinfo

       let declare x _ =
	 let pool = F.pool () in
	 let xa = F.fresh pool "ta" (Model t_alloc) in
	 let sx = sizeof (Ctypes.object_of x.vtype) in
	 let xk = D.Xindex.get_ind x in
	 let sa = D.e_app2 "access" (F.var xa) xk in
	 let gta = D.p_app1 "global" (F.var xa) in
	 Axiom (F.p_forall [xa] (F.p_implies gta (F.p_eq sa sx)))

       let section = S_typedef
       let prefix = "Alloc"
       let basename x = x.vname
       let clear () = ()
       let pp_descr fmt _x =
	 Format.fprintf fmt "Global allocation table"
     end)

  module Data =
  struct

    include Datalib.Cvalues(Model)

    type m_of_mem = m_memory
    let tau_of_mem = t_memory

    let forall_loc pool =
      let p = F.fresh pool "p" (Model Integer) in
      [p] , Model.Addr (F.var p)

    let load_rec = ref (fun _ _ _ -> assert false)
    let store_rec = ref (fun _ _ _ _ -> assert false)

    let define_vinfo x = if x.vglob then Globals.define x

    let cvar_of_var vinfo =
      define_vinfo vinfo; Model.Loc(vinfo, zero)

    let cvar _mem vinfo = cvar_of_var vinfo

    let offset loc te n =
      match loc with
	| Model.Loc (b,d) -> Model.Loc(b, (add_offset d te n))
	| Model.Addr p -> Model.Addr(model_addr_shift p (n_size n te))

    let shift = offset
    let index = offset

    let field loc f =
      if f.fcomp.cstruct then
	let pos = offset_of_field f in
	match loc with
	  | Model.Loc(b,d) -> Model.Loc(b,(add d pos))
	  | Model.Addr p -> Model.Addr (model_addr_shift p pos)
      else
	unsupported "union field"

    let load_mem m t l = !load_rec m t l
    let store_mem m t l v = !store_rec m t l v

  end

  module DF = Data_mem.Create(Data)

  include Data

  let startof l _cv =
    match l with
      | Model.Loc(b,ofs) -> Model.Loc(b,ofs)
      | Model.Addr p -> Model.Addr (model_addr_shift p zero)


  let base_address _mem = function
    | Model.Loc(b,_) -> Model.Loc(b,zero)
    | Model.Addr p -> 	Model.Addr(model_addr p zero)
        
  let block_length mem loc = model_block mem.talloc (Model.addr loc)

  let cast_loc_to_loc _ty1 _ty2 _l = unsupported "pointer cast"


 let zrange loc n = 
      match loc with 
	| Model.Loc(x,k) -> model_zrange (D.Xindex.get_ind x) k n
	| Model.Addr p -> model_zrange_of_addr_range p zero n  

  (* ------------------------------------------------------------------------ *)
  (* --- Load                                                             --- *)
  (* ------------------------------------------------------------------------ *)

  let load_with fmt mem loc =
    D.decode fmt (model_access mem (Model.addr loc))

  let load_mem mem te loc =
    match te with
      | C_pointer ty -> V_pointer(ty,Model.Addr(load_with addr_format mem loc))
      | C_int i -> V_int(i,load_with D.int_format mem loc)
      | C_float f -> V_float(f,load_with D.real_format mem loc)
      | C_comp comp ->
	  if comp.cstruct then
	    let fmt = DF.record_format comp in
	    let z = zrange loc  (F.e_int (List.length comp.cfields))in
	    let d = model_access_range mem z in
	    V_record(comp,D.decode fmt d)
          else Wp_parameters.not_yet_implemented "load of union"
      | C_array arr ->
	  let fmt = DF.array_format arr in
	  let z = zrange loc  (sizeof te) in
	  let d = model_access_range mem z in
	  V_array(arr,D.decode fmt d)

  let store_with mem loc fmt v =
    model_update mem (Model.addr loc) (D.encode fmt v)

  let store_mem mem te loc v =
    match v with
      | V_int(_,t) -> store_with mem loc D.int_format t
      | V_float(_,t) -> store_with mem loc D.real_format t
      | V_pointer(_,lv) -> store_with mem loc addr_format (Model.addr lv)
      | V_record (comp,r) ->
	  let fmt = DF.record_format comp in
	  let dr = D.encode fmt r in
	  let zp = zrange loc (F.e_int (List.length comp.cfields)) in
	  model_update_range mem zp dr
      | V_union _ -> unsupported "union"
      | V_array (arr,r) ->
	  let fmt = DF.array_format arr in
	  let dr = D.encode fmt r in
	  let zp = zrange loc  (sizeof te) in
	  model_update_range mem zp dr

  let () =
    begin
      Data.load_rec := load_mem ;
      Data.store_rec := store_mem ;
    end

  let load m te loc =
    DF.loaded te ; load_mem m.store te loc

  (* ------------------------------------------------------------------------ *)
  (* --- Zone                                                             --- *)
  (* ------------------------------------------------------------------------ *)

  (* --- Effect Assigns Method --- *)

  let tau_of_dzone = Formula.ADT("zone",[])

  (* Elementary Zones *)
  type assignable =
    | Avar   of varinfo * D.integer 
    | Aaddr  of D.integer
    | Arange of D.integer * D.integer * D.integer

  let addr_of_assignable = function
    | Avar(x,ofs) -> model_addr (D.Xindex.get_ind x) ofs
    | Aaddr p -> p
    | Arange(p,ofs,_sz) -> model_addr_shift p ofs

  let zone_of_assignable = function
    | Avar(x,ofs) -> model_zrange (D.Xindex.get_ind x) ofs one
    | Aaddr p -> model_zrange_of_addr_range p zero one
    | Arange(p,ofs,sz) -> model_zrange_of_addr_range p ofs sz

  let assignable_loc _te loc =
    match loc with
      | Model.Loc(x,d) -> Avar(x,d)
      | Model.Addr p -> Aaddr p
	
  let assignable_range te loc rg =
    match loc with
      | Model.Loc(x,d) ->
	  let b = D.Xindex.get_ind x in 
	  begin
	    match rg with
	      | {D.inf=None;D.sup=Some h} -> 
		  Arange(b,d,n_size (add h one) te)
	      | {D.inf=Some l;D.sup=Some h} ->
		  Arange(b,add_offset d te l,n_size (cardinal l h) te)
	      | _ -> unsupported "infinite range for array"
	  end
      | Model.Addr p ->
	  begin
	    match rg with
	      | {D.inf=None;D.sup=Some h} ->
		  Arange(p,zero,n_size (add h one) te)
	      | {D.inf=Some l;D.sup=Some h} ->
		  Arange(p,n_size l te,n_size (cardinal l h) te)
	      | _ -> unsupported "infinite range for array"
	  end

  let assignable = function
    | D.Aloc( te , loc ) -> assignable_loc te loc
    | D.Arange( te , loc , rg ) -> assignable_range te loc rg

  let dzone_assigned a = zone_of_assignable (assignable a)
  let dzone_subset = model_included
  let dzone_union = model_zunion
  let dzone_empty () = model_zempty

  let effect_supported = true

  let assignable_sizeof = function 
    | D.Aloc( te , _ ) -> sizeof te
    | D.Arange( te , _ , {D.inf=Some l;D.sup=Some h} ) ->  
	n_size (cardinal l h) te
    | D.Arange _ -> unsupported "infinite range for array"

  let valid mem a = 
    model_valid mem.talloc 
      (addr_of_assignable (assignable a))  
      (assignable_sizeof a)

  let separated a1 a2 =
    model_separated (dzone_assigned a1) (dzone_assigned a2)
   

  (* --- Region Assigns Method --- *)



  type region = Z.region
  let pp_region = Z.pp_region

  let region_of_loc te = function
    | Model.Loc(x,ofs) ->
	let p = D.Xindex.get_ind x in
	Z.range (Z.cvar p x) ofs one
    | Model.Addr p ->
	Z.range (Z.base (model_base p)) (model_offset p) (sizeof te)

  let region_range te loc rg =
    match loc with
      | Model.Loc(x,d) ->
	  begin
	    let base = Z.cvar (D.Xindex.get_ind x) x in
	    match rg with
	      | {D.inf=None;D.sup=None} ->
		  Z.block base
	      | {D.inf=None;D.sup=Some h} -> 
		  Z.range base d (n_size (add h one) te)
	      | {D.inf=Some l;D.sup=Some h} ->
		  Z.range base (add_offset d te l) (n_size (cardinal l h) te)
	      | _ -> unsupported "infinite range for array"
	  end
      | Model.Addr p ->
	  begin
	    let base = Z.base (model_base p) in
	    let ofs = model_offset p in
	    match rg with
	      | {D.inf=None;D.sup=None} ->
		  Z.block base
	      | {D.inf=None;D.sup=Some h} ->
		  Z.range base ofs (n_size (add h one) te)
	      | {D.inf=Some l;D.sup=Some h} ->
		  Z.range base (add_offset ofs te l) (n_size (cardinal l h) te)
	      | _ -> unsupported "infinite range for array"
	  end

  let region_assigned = function
    | D.Aloc(te,loc) -> region_of_loc te loc
    | D.Arange(te,loc,rg) -> region_range te loc rg
	    
  let region_empty () = Z.empty
  let region_union = Z.union
  let region_included = Z.included
  let region_separated = Z.separated
  let region_fingerprint = Z.fingerprint

  let region_supported = true

  (* --- Translation Environment --- *)

  type frame = {
    mutable states : (c_label * (F.var*F.var) ) list ;
    mutable memories : (c_label * mem) list ;
    mutable result : F.var option ;
    mutable exit_status : F.var option ;
    return : Cil_types.typ option ;
  }

  let new_frame kf ~m_here ~m_pre ~m_post ropt =
    let may_cons acc lab x =
      match x with
	| None -> acc
	| Some x -> (lab,x)::acc
    in
    let memories =
      (may_cons
	 (may_cons
	    (may_cons [] Clabels.Post m_post)
	    Clabels.Pre m_pre)
	 Clabels.Here m_here)
    in
    {
      states = [];
      memories = memories;
      result = ropt;
      exit_status = None;
      return = Some(Kernel_function.get_return_type kf);
    }


  let get_result env =
    match env.result,env.return with
      | Some x,Some t -> x, t
	  | Some _, _ -> Wp_parameters.fatal "Result type undefined"
      | None,_ ->
	  match env.return with
	    | Some typ ->
		let x = D.fresh "result" (Mdata.Vacsl(Ctype typ)) in
		env.result <- Some x ; (x, typ)
	    | None ->
		Wp_parameters.fatal "Result type undefined"

  let get_exit env =
    match env.exit_status with
      | Some x -> x
      | None ->
	  let x = D.fresh "exit_status" (Mdata.Vmodel Formula.Integer) in
          env.exit_status <- Some x ; x

  let set_exit_status f x =
    {
      states = f.states ;
      memories = f.memories ;
      result = None ;
      exit_status = Some x ;
      return = f.return ;
    }

  let varmem_at env l =
    try (List.assoc l env.states)
    with Not_found ->
      let x = D.fresh "m" (Mdata.Vmodel t_memory) in
      let ta = D.fresh "ta" (Mdata.Vmodel t_alloc) in
      env.states <- (l,(x,ta)) :: env.states ;
      (x,ta)

  let mem_at env lab =
    try List.assoc lab env.memories
    with Not_found ->
      let m,ta = varmem_at env lab in
      {store = F.var m ; talloc = F.var ta }

  let result env = get_result env

  let exit_status env = get_exit env
  let get_exit_status env = env.exit_status

  (* ----------------------------------------------------------------------- *)
  (* --- Instanciation of MWP : WP Calculus                              --- *)
  (* ----------------------------------------------------------------------- *)

  let update_at_label frame lab p =
    let m_h,a_h = varmem_at frame Here in
    try
      let (m_l,a_l) = List.assoc lab frame.states in
      D.subst m_l (F.var m_h) (D.subst a_l (F.var a_h) p)
    with Not_found -> p

  let quantify_at_label frame lab p =
    try
      let x_l,a_l = List.assoc lab frame.states in
      D.forall [x_l;a_l] p
    with Not_found -> p

  let subst_lval frame te loc v p =
    DF.stored te ;
    let xm,_ = varmem_at frame Here in
    let tm = store_mem (F.var xm) te loc v in
    D.subst xm tm p

  let subst_result frame v p =
    match frame.result with
      | None -> p
      | Some xres -> match v with
          | Some v -> D.subst xres (logic_of_value v) p
          | None -> D.forall [xres] p

  let alloc_vars frame xs p =
    let _,xta = varmem_at frame Here in
    List.fold_left
      (fun p x ->
	 let ta0 = F.var xta in
	 let v_x = D.Xindex.get_ind x in
	 let sz_x = sizeof (object_of x.vtype) in
	 let ta1 = model_alloc ta0 v_x sz_x in
	 D.subst xta ta1 p
      ) p xs

  let free_vars frame xs p =
    let _,xta = varmem_at frame Here in
    List.fold_left
      (fun p x ->
	 let ta0 = F.var xta in
	 let v_x = D.Xindex.get_ind x in
	 let ta1 = model_free ta0 v_x in
	 D.subst xta ta1 p
      ) p xs

  let fresh_vars frame xs p =
    let m,xta = varmem_at frame Here in
    List.fold_left
      (fun p x ->
	 let v_x = D.Xindex.get_ind x in
	 let q = model_isfresh (F.var m) (F.var xta) v_x in
	 F.p_implies q p
      ) p xs

  let notexists_vars frame xs p =
    let _,xta = varmem_at frame Here in
    List.fold_left
      (fun p x ->
	 let ta0 = F.var xta in
	 let v_x = D.Xindex.get_ind x in
	 let q = F.p_eq (D.e_app2 "access" ta0 v_x) (F.e_int 0) in
	 F.p_implies q p
      ) p xs

  let local_scope frame vars scope p =
    match scope with
      | Mcfg.SC_Block_in | Mcfg.SC_Function_in ->
	  notexists_vars frame vars 
	    (alloc_vars frame vars (fresh_vars frame vars p))
      | Mcfg.SC_Function_frame-> (* nothing to do *) p
      | Mcfg.SC_Block_out | Mcfg.SC_Function_out  ->
	  fresh_vars frame vars (free_vars frame vars p)
      | Mcfg.SC_Global ->
	  let (_,a_h) = varmem_at frame Here in
	  if D.has_vars [a_h] p
	  then F.p_implies (D.p_app1 "global" (F.var a_h)) p
	  else p
	    

  let subst_havoc frame a =
    let addr = zone_of_assignable (assignable a) in
    let m,_ = varmem_at frame Here in
    let v = D.fresh "v" (Vmodel(ADT("data",[]))) in
    let m' sigma =
      let m0 = D.apply sigma (F.var m) in
      F.wrap(model_update_havoc m0 addr (F.var v))
    in
    [D.Fresh v;D.Update(m,m')]

  let rec region = function
    | [] -> model_zempty
    | [a] -> dzone_assigned a
    | a :: others -> model_zunion (dzone_assigned a) (region others)

  let assigns_goal frame from_label assigned to_label =
    let mfrom = mem_at frame from_label in
    let mto = mem_at frame to_label in
    model_ishavoc mfrom.talloc mfrom.store (region assigned) mto.store

  let assigns_supported = true

  (* ----------------------------------------------------------------------- *)
  (* --- User Predicate environment                                      --- *)
  (* ----------------------------------------------------------------------- *)

  type closure =
    | Mem of string  (* The logic labels used *)
    | Alloc of string  (* The logic labels used *)

  let pp_closure fmt  = function
    | Mem s ->Format.fprintf fmt "@@%s" s
    | Alloc s ->Format.fprintf fmt "@@%s" s

  let userdef_frame () = {
    states = [] ;
    memories = [] ;
    result = None ;
    exit_status = None;
    return = None ;
  }

  (* TODO filter with respect to tused and pused -> (ZD) ??*)
  let userdef_signature frame _tused _pused =
    let vars =
      List.fold_left
	(fun xs (_label,(xm,xa)) -> xa::xm::xs)
	[] frame.states
    in
    let closures =
      List.fold_left
	(fun cs (label,_) ->
	   let l = match label with
	     | LabelParam l -> l
	     | _ ->
		 Wp_parameters.fatal
		   "Unexpected statement label in user-defined symbol"
	   in Alloc l :: Mem l :: cs)
	[] frame.states
    in
    vars , closures

  let userdef_closure mem = function
    | Mem _ -> F.wrap mem.store
    | Alloc _ -> F.wrap mem.talloc

  let label_at_closure = function
    | Mem l -> l
    | Alloc l -> l

end
