(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* ------------------------------------------------------------------------ *)
(* ---  Translation of Term and Predicats                               --- *)
(* ------------------------------------------------------------------------ *)

module WpLog = Wp_parameters

open Ctypes
open Clabels
open Formula
open Cil_types
open Cil_datatype

module Create
  (M : Mlogic.S)
  =

struct

  module F = M.D.F
  module D = M.D

  (* ----------------------------------------------------------------------- *)
  (* --- Registration of User-defined Predicate and Functions            --- *)
  (* ----------------------------------------------------------------------- *)

  type userdef = {
    d_info : logic_info ;
    d_callname : string ;
    d_implicit : F.var list ;
    d_closure : M.closure list ;
    d_formals : F.var list;
  }

  type axiomdef = {
    a_name : string ;
    a_labels : string list ;
    a_hyp_name : string ;
    a_def_name : string ;
    a_implicit : F.var list ;
    a_closure : M.closure list ;
    a_property : F.pred ;
  }

  module Hdef = Logic_var.Hashtbl

  let user_axioms = Hashtbl.create 131
  let user_axiomdefs = Hashtbl.create 131
  let user_variables = Hdef.create 131
  let user_definitions = Hdef.create 131

  module UserVariable = D.Register
    (struct
       include D.Logicvar
       let pp_title fmt x = Format.fprintf fmt "User-defined constant %s"
	 x.lv_name
       let pp_descr fmt x = Format.fprintf fmt "logic %s : %a" x.lv_name
	 !Ast_printer.d_logic_type x.lv_type
       let section = S_userdef
       let clear () = Hdef.clear user_variables
       let prefix = "D"
       let declare x _ =
	 Formula.Function( [] , D.tau_of_logic_type x.lv_type )
     end)

  let get_user_variable x =
    try Hdef.find user_variables x
    with Not_found ->
      let d = UserVariable.get_definition x in
      let t = F.e_call d.d_name [] in
      Hdef.add user_variables x t ; t

  let rec pp_closures fmt xs cs =
    match xs , cs with
      | [] , [] -> ()
      | x::xs , c::cs ->
	  Format.fprintf fmt "{%s:%a=%a}"
	    (F.name_of_var x)
	    Datalib.Tau.pp_tau (F.tau_of_var x)
	    M.pp_closure c ;
	  pp_closures fmt xs cs
      | x::xs , [] ->
	  Format.fprintf fmt "{%s:%a=?}"
	    (F.name_of_var x)
	    Datalib.Tau.pp_tau (F.tau_of_var x) ;
	  pp_closures fmt xs []
      | [] , c::cs ->
	  Format.fprintf fmt "{?=%a}"
	    M.pp_closure c ;
	  pp_closures fmt [] cs

  let pp_formals fmt xs =
    List.iter
      (fun x ->
	 Format.fprintf fmt "(%s:%a)"
	   (F.name_of_var x)
	   Datalib.Tau.pp_tau (F.tau_of_var x) ;
      ) xs

  module UserAxiom = D.Register
    (struct
       type t = string
       module H = Hashtbl.Make
	 (struct
	    type t = string
	    let hash = Hashtbl.hash
	    let equal: string -> string -> bool = (=)
	  end)
       let section = S_userdef
       let source = None
       let prefix = "Def"
       let clear () = Hashtbl.clear user_axioms
       let index x = x
       let basename x = x
       let location _x = None
       let declare x _ =
	 try Formula.Axiom( Hashtbl.find user_axioms x )
	 with Not_found -> Wp_parameters.fatal "Uncompiled axiom (%s)" x
       let pp_title fmt x = Format.fprintf fmt "User-defined axiom %s" x
       let pp_descr fmt _x = Format.fprintf fmt "No parameters."
     end)

  module UserDefinition =
  struct
    let index = Hdef.create 131
    let () = D.on_clear (fun () -> Hdef.clear index)

    let pp_userdef_title fmt d =
      let x = d.d_info.l_var_info.lv_name in
      match d.d_info.l_type with
	| Some _ -> Format.fprintf fmt "User-defined function %s" x
	| None -> Format.fprintf fmt "User-defined predicate %s" x

    let pp_userdef_descr fmt d =
      begin
	if d.d_implicit <> [] then
	  begin
	    Format.fprintf fmt "Locations:" ;
	    pp_closures fmt d.d_implicit d.d_closure ;
	  end ;
	if d.d_implicit <> [] && d.d_formals <> [] then
	  Format.pp_print_newline fmt () ;
	if d.d_formals <> [] then
	  begin
	    Format.fprintf fmt "Parameters:" ;
	    pp_formals fmt d.d_formals
	  end ;
      end

    let define duser items =
      let k = duser.d_info.l_var_info in
      if not (Hdef.mem index k) then
	begin
	  List.iter
	    (fun (name,item) ->
	       D.add_declaration {
		 d_name = name ;
		 d_section = S_userdef ;
		 d_title = (fun fmt -> pp_userdef_title fmt duser) ;
		 d_descr = (fun fmt -> pp_userdef_descr fmt duser) ;
		 d_source = None ;
		 d_item = item ;
	       })
	    items ;
	  Hdef.add index k ()
	end
  end

  module UserAxiomDefs =
  struct
    let index = Hashtbl.create 131
    let () = D.on_clear (fun () -> Hashtbl.clear index)

    let pp_axiomdef_title fmt a =
      Format.fprintf fmt "User-defined axiom %s" a.a_name

    let pp_axiomdef_descr fmt a =
      Format.fprintf fmt "Axiom %s%a@\n" a.a_name
	Axiomatics.pp_labels a.a_labels ;
      Format.fprintf fmt "Locations:" ;
      pp_closures fmt a.a_implicit a.a_closure

    let define axdef =
      if not (Hashtbl.mem index axdef.a_name) then
	begin
	  let signature = List.map F.tau_of_var axdef.a_implicit in
	  D.add_declaration {
	    d_name = axdef.a_hyp_name ;
	    d_section = S_userdef ;
	    d_title = (fun fmt -> pp_axiomdef_title fmt axdef) ;
	    d_descr = (fun fmt -> pp_axiomdef_descr fmt axdef) ;
	    d_source = None ;
	    d_item = Formula.Predicate signature ;
	  } ;
	  D.add_declaration {
	    d_name = axdef.a_def_name ;
	    d_section = S_userdef ;
	    d_title = (fun fmt -> pp_axiomdef_title fmt axdef) ;
	    d_descr = (fun fmt -> pp_axiomdef_descr fmt axdef) ;
	    d_source = None ;
	    d_item = Formula.Axiom axdef.a_property ;
	  } ;
	  Hashtbl.add index axdef.a_name () ;
	end

  end

  (* ----------------------------------------------------------------------- *)
  (* --- Translation Environment                                         --- *)
  (* ----------------------------------------------------------------------- *)

  type xvar =
    | PrmVar of M.value
    | MemVar

  type env = {
    formals_in_pre : bool ;
    frame : M.frame ;
    label : c_label ;
    lvars : abstract F.term Logic_var.Map.t ;
    xvars : M.value Varinfo.Map.t ;
  }

  let env_at e label = {
    formals_in_pre = e.formals_in_pre ;
    frame = e.frame ;
    label = label ;
    lvars = e.lvars ;
    xvars = e.xvars ;
  }

  let get_frame e = e.frame

  let mem_at_env env = M.mem_at env.frame env.label

  let quant_env e assoc = {
    formals_in_pre = e.formals_in_pre ;
    frame = e.frame ;
    label = e.label ;
    lvars =
      List.fold_right
	(fun (x,v) t -> Logic_var.Map.add x v t ) assoc e.lvars ;
    xvars = e.xvars ;
  }

  let env_body frame = {
    formals_in_pre = false ;
    frame = frame ;
    label = Here ;
    lvars = Logic_var.Map.empty ;
    xvars = Varinfo.Map.empty ;
  }

  let call_post called_kf vs m_pre m_post xopt =
    let frame =
      M.new_frame called_kf
	~m_here:(Some m_post)
	~m_pre:(Some m_pre)
	~m_post:(Some m_post) xopt in
    let xenv1 =
      try
	List.map2
	  (fun x v -> x ,  v)
	  (Kernel_function.get_formals called_kf) vs
      with Invalid_argument _ -> []
	(* implicit prototype of callee*)
    in
    let xenv =
      List.fold_right
	(fun (x,v) t -> Varinfo.Map.add x v t) xenv1 Varinfo.Map.empty
    in
    {
      formals_in_pre = true ;
      frame = frame ;
      label = Here ;
      lvars = Logic_var.Map.empty ;
      xvars = xenv ;
    }

  let call_pre called_kf vs m_pre =
    let frame =
      M.new_frame called_kf
	~m_here:(Some m_pre)
	~m_pre:(Some m_pre)
	~m_post:None
	None
    in
    let xenv1 =
      try
	List.map2
	  (fun x v -> x , v)
	  (Kernel_function.get_formals called_kf) vs
      with Invalid_argument _ -> []
	(* implicit prototype of callee*)
    in
    let xenv =
      List.fold_right
	(fun (x,v) t -> Varinfo.Map.add x v t) xenv1 Varinfo.Map.empty
    in
    {
      formals_in_pre = false ;
      frame = frame ;
      label = Here ;
      lvars = Logic_var.Map.empty ;
      xvars = xenv ;
    }

  let call_exit env x =
    {
      formals_in_pre = false ;
      frame = M.set_exit_status env.frame x ;
      label = env.label ;
      lvars = env.lvars ;
      xvars = env.xvars ;
     }

  let assigns_clause frame =
    {
      formals_in_pre = false ;
      frame = frame ;
      label = Here ; (* actually, HERE is PRE when proving the clause *)
      lvars = Logic_var.Map.empty ;
      xvars = Varinfo.Map.empty ;
    }

  (* Consulting the logic variable environment.*)
  let lvar lmap x =
    try Logic_var.Map.find x lmap
    with Not_found -> get_user_variable x

  let add_lvar e x t =
   {
    formals_in_pre = e.formals_in_pre ;
    frame = e.frame ;
    label = e.label ;
    lvars = Logic_var.Map.add x t e.lvars ;
    xvars = e.xvars ;
  }

  let xvar_of_varinfo env vinfo =
    try PrmVar (Varinfo.Map.find vinfo env.xvars)
    with Not_found -> MemVar

  (* ----------------------------------------------------------------------- *)
  (* --- Translation Values                                              --- *)
  (* ----------------------------------------------------------------------- *)

  (* kinds are logic types *)
  type kind =
    | Kcint of Ctypes.c_int
    | Kint
    | Kreal
    | Kbool
    | Kptr of Cil_types.typ
    | Kset of kind
    | Kstruct of compinfo
    | Karray of arrayinfo
    | Kadt of string * kind list

   let pp_kind fmt = function
     | Kcint i -> Ctypes.pp_int fmt i
     | Kint -> Format.pp_print_string fmt "int"
     | Kreal -> Format.pp_print_string fmt "real"
     | Kptr _-> Format.pp_print_string fmt "pointer"
     | Kset _-> Format.pp_print_string fmt "set"
     | Kstruct c -> Format.pp_print_string fmt c.cname
     | Karray a -> Ctypes.pretty fmt (C_array a)
     | Kadt (s,_) -> Format.pp_print_string fmt s
     | Kbool ->  Format.pp_print_string fmt "bool"

  type data =
    | Data     of abstract F.term (* Singleton *)
    | Loc      of M.loc
    | Value    of M.value
    | Interval of D.interval
    | Range    of c_object * M.loc * D.interval
    | List     of data list
    | Set      of D.set

  let pp_data fmt = function
    | Data d -> Format.fprintf fmt "Data=%a" F.pp_term d
    | Loc l -> Format.fprintf fmt "Loc=%a" M.pp_loc l
    | Value v -> Format.fprintf fmt "Value=%a" M.pp_value v
    | _ -> Format.fprintf fmt "Blob"

  let data_of_integer (x : Formula.integer F.term) : data = Data (F.wrap x)
  let data_of_real (x : Formula.real F.term) : data = Data(F.wrap x)
  let data_of_boolean (x:Formula.boolean F.term) : data = Data (F.wrap x)

  let integer_of_value = function
    | M.V_int(_,t) -> t
    | v -> Wp_parameters.fatal "integer_of_value %a" M.pp_value v

  let real_of_value = function
    | M.V_float(_,t) -> t
    | M.V_int(_,t) -> F.real_of_integer  t
    | _ -> assert false




  let extract_from_data = function
    | Data d -> F.unwrap d
    | Value v -> F.unwrap (M.logic_of_value v)
    | _ -> assert false

  let boolean_of_data d : boolean F.term = extract_from_data d
  let array_of_data d : D.array = extract_from_data d
  let record_of_data d : D.record = extract_from_data d
  let urecord_of_data d : D.urecord = extract_from_data d

  let loc_of_data obj d =
    match d with
    | Value(M.V_pointer(_,l)) -> l
    | Value v -> M.loc_of_pointer obj (F.unwrap (M.logic_of_value v))
    | Data l -> M.loc_of_pointer obj (F.unwrap l)
    | Loc l ->  l
    | _ ->
	Wp_parameters.fatal ~current:true "not a loc (%a)" pp_data d

 let integer_of_data k d : integer F.term =
    match d with
    | Data d ->
	begin
	  match k with
	    | Kint | Kcint _ -> F.unwrap d
	    | k -> WpLog.fatal ~current:true "not an integer (%a : %a)" F.pp_term d pp_kind k
	end
    | Value v -> integer_of_value v
    | _ -> WpLog.fatal ~current:true "not an integer (%a)" pp_data d

 let real_of_data k d =
    match k with
      | Kint | Kcint _-> F.real_of_integer (integer_of_data k d)
      | Kreal ->
	  begin
	    match d with
	      | Data d -> F.unwrap d
	      | Value v -> real_of_value v
	      | _ -> WpLog.fatal ~current:true "not a real (%a)" pp_data d
	  end
      | _ -> WpLog.fatal ~current:true "not a real (%a)" pp_data d


  let rec set_of  = function
    | Value v -> D.singleton (M.logic_of_value v)
    | Data d -> D.singleton d
    | Loc l ->  D.singleton (F.wrap (M.pointer_of_loc l))
    | List ds -> D.unions (List.map set_of ds)
    | Interval i -> D.interval i
    | Set s -> s
    | Range _ -> Wp_parameters.not_yet_implemented "set of zone"

  let list_of = function
    | (Value _|Data _|Loc _|Interval _|Range _) as d -> [d]
    | List xs -> xs
    | Set _ -> assert false

  let union_data a b =
    match a,b with
      | Set a , b | b, Set a -> Set (D.union a (set_of b))
      | List _ , _ | _ , List _ -> List (list_of a @ list_of b)
      | ( (Value _|Data _|Loc _|Range _|Interval _ ) ,
	  (Value _|Data _|Loc _|Range _|Interval _ ) ) -> List[a;b]

  let union_map f = function
    | [] -> List []
    | d::ds ->
	List.fold_left
	  (fun w x -> union_data w (f x))
	  (f d) ds

  let term_of_data = function
    | Data t  -> t
    | Value v -> M.logic_of_value v
    | Loc l   -> F.wrap (M.pointer_of_loc l)
    | Set s   -> F.wrap s
    | List dl -> F.wrap (D.unions (List.map (set_of) dl))
    | Range _ -> Wp_parameters.not_yet_implemented "set of zone"
    | Interval i  -> F.wrap (D.interval i )

  let neg_interval r =
    match r.D.inf,r.D.sup with
      | Some j, None -> {D.inf = None ; D.sup = Some (F.e_ineg j)}
      | None , Some j -> {D.inf = Some (F.e_ineg j) ; D.sup = None}
      | Some j, Some k -> {D.inf = Some (F.e_ineg k) ; D.sup = Some (F.e_ineg j)}
      | None, None -> r

  (* ----------------------------------------------------------------------- *)
  (* --- Logic Types                                                     --- *)
  (* ----------------------------------------------------------------------- *)

  let rec object_of_pointed = function
    | Kptr te | Kset(Kptr te) ->  Ctypes.object_of te
    | _ -> WpLog.fatal "Dereferencing a non-pointer value"

  let kind_of_typ c=
      match object_of c with
	| C_int i -> Kcint i
	| C_float _ -> Kreal
	| C_pointer te -> Kptr te
	| C_comp comp -> Kstruct comp
	| C_array arr -> Karray arr

  let kind_of_data ty = function
    | (Data _ | Loc _ | Value _) -> kind_of_typ ty
    | (Interval _ | Range _ | List _ | Set _) -> Kset (kind_of_typ ty)

  let rec kind_of = function
    | Ctype c -> kind_of_typ c
    | Linteger -> Kint
    | Lreal -> Kreal
    | Ltype ({lt_name="𝔹"}, _) -> Kbool
    | Ltype( {lt_name="set"} , [elt] ) -> Kset(kind_of elt)
    | Ltype( {lt_name=adt} , args ) -> Kadt(adt,List.map kind_of args)
    | Lvar _ -> WpLog.not_yet_implemented "logic type variables"
    | Larrow _ -> WpLog.not_yet_implemented "type of logic function"

  let typ_of_elements = function
    | Ctype c ->
	let o = object_of c in
	begin
	  match o with
	    | C_pointer te -> te
	    | C_array arr -> arr.arr_element
	    | _ ->
                WpLog.fatal "elements of non-pointer type %a" Ctypes.pp_object o
	end
    | t -> WpLog.fatal "elements of non-pointer type: %a"
        !Ast_printer.d_logic_type t

  (* ----------------------------------------------------------------------- *)
  (* --- Global Recursions                                               --- *)
  (* ----------------------------------------------------------------------- *)

  let data_rec : (env -> term -> data) ref = ref (fun _ _ -> assert false)

  let rec_apply_function
      : (env -> logic_info ->
	   (logic_label * logic_label) list ->
	     term list -> abstract F.term) ref
      = ref (fun _ _ _ _ -> assert false)

  (* ---------------------------------------------------------------------- *)
  (* --- Data memory predicate                                          --- *)
  (* ---------------------------------------------------------------------- *)

  let rec assigned_of_data te acc d =
    match d with
      | (Data _|Value _|Loc _) ->
	  D.Aloc(te,loc_of_data te d)::acc
      | Range(te,loc,range) -> 
	  D.Arange(te,loc,range)::acc
      | List ds -> List.fold_left (assigned_of_data te) acc ds
      | _ -> Wp_parameters.not_yet_implemented "arbitrary zone"

  let data_valid m k d =
    let t =
      match k with
	| Kptr te -> te
	| Kset (Kptr te) -> te
	| _ -> WpLog.fatal "unexpected type for valid predicate"
    in
    let te = Ctypes.object_of t in
    F.p_conj (List.map (M.valid m) (assigned_of_data te [] d))
    
  let data_separated (t1, d1) (t2, d2) =
    let r1 = assigned_of_data t1 [] d1 in
    let r2 = assigned_of_data t2 [] d2 in
    let p = ref F.p_true in
    List.iter
      (fun a ->
	 List.iter
	   (fun b ->
	      p := F.p_and !p (M.separated a b)
	   ) r2
      ) r1 ;
    !p

  (* ---------------------------------------------------------------------- *)
  (* --- Data memory operation                                          --- *)
  (* ---------------------------------------------------------------------- *)

  let rec data_shift_range tobj loc kbi dindex ~is_pos =
    match dindex with
      | Data _ | Value _ | Loc _  as d ->
	  let idx =integer_of_data kbi d in
	  let idx = if is_pos then idx else F.e_ineg idx in
	  Loc (M.shift loc tobj idx)
      | List il -> union_map (data_shift_range tobj loc kbi ~is_pos) il
      | Interval r ->
	  let r = if is_pos then r else neg_interval r in Range(tobj,loc,r)
      | Set _ | Range _ ->
	  Wp_parameters.not_yet_implemented "pointer shift over arbitrary sets"

  let rec data_index_range tobj loc kbi dindex =
    match dindex with
      | Data _ | Value _ | Loc _  as d ->
	  Loc (M.index loc tobj (integer_of_data kbi d))
      | List il -> union_map (data_index_range tobj loc kbi ) il
      | Interval r -> Range(tobj,loc,r)
      | Set _ | Range _ ->
	  Wp_parameters.not_yet_implemented "array access over arbitrary sets"


  let data_shift ka ga kb gb ~is_pos =
    match ka,kb with
      | (Karray _ | Kptr _ ) , (Kint | Kcint _) ->
	  let obj = object_of_pointed ka in
	  let gb = integer_of_data kb gb in
	  let gb = if is_pos then gb else (F.e_ineg gb) in
	  Loc (M.shift
		 (loc_of_data obj ga)
		  obj gb)
      | (Kptr _ | Karray _) , (Kset ((Kint | Kcint _) as kbi)) ->
	  let obj = object_of_pointed ka in
	  data_shift_range obj (loc_of_data obj ga) kbi gb ~is_pos
      | _ -> WpLog.not_yet_implemented "shift over arbitrary sets of pointers"

 let data_index ta ka ga kb gb =
   match ka,kb with
     | (Karray _ | Kptr _) , (Kint | Kcint _) ->
	 Loc (M.index (loc_of_data ta ga) ta (integer_of_data kb gb))
     | (Karray _ | Kptr _) , (Kset ((Kint |Kcint _) as kbi)) ->
	 data_index_range ta (loc_of_data ta ga) kbi gb
     | _ ->  WpLog.not_yet_implemented "shift over arbitrary sets of pointers"

  let data_field ka ga f =
    match ka with
      | Kset _ ->
	  (* data_field_set env ga f *)
	  Wp_parameters.not_yet_implemented "field access over sets"
      | _ ->
	  let obj = object_of f.ftype in
	  Loc (M.field (loc_of_data obj ga) f)

  let rec data_startof_set ta ga =
    match ga with
      | Value _ | Data _ | Loc _ ->
	  Loc (M.startof (loc_of_data ta ga) ta)
      | Set _ | Range _ -> Wp_parameters.not_yet_implemented "start-of over sets"
      | List pl ->
	  union_map (fun p -> data_startof_set ta p ) pl
      | Interval _ -> WpLog.fatal "unexpected argument for [startof]"

  let data_startof ta ka ga =
    match ka with
      | Kset _ -> data_startof_set ta ga
      | _ -> Loc( M.startof (loc_of_data ta ga) ta )

  let rec data_load env ty  = function
    | Loc _ | Data _ | Value _  as d ->
	let obj = object_of ty in
	Value (M.load (mem_at_env env) obj (loc_of_data obj d))
    | Range _ ->
	Wp_parameters.not_yet_implemented "load of arbitrary region"
	(* Set (M.load_zone (mem_at_env env) te z) *)
    | Set _ ->
	Wp_parameters.not_yet_implemented "load of arbitrary sets"
    | List pl -> union_map (data_load env ty) pl
    | _ -> WpLog.fatal "unexpected argument for [load]"


  (* ---------------------------------------------------------------------- *)
  (* --- Offsets inside functional memory model                         --- *)
  (* ---------------------------------------------------------------------- *)

  let rec logic_offset env a = function
    | TNoOffset -> Data a
    | TField(f,off) ->
	let fieldvalue = D.acc_field (F.unwrap a) f in
	logic_offset env fieldvalue off
    | TIndex (t,off) ->
	let i = integer_of_data (kind_of t.term_type) (!data_rec env t) in
	logic_offset env (D.acc_index (F.unwrap a) i) off

  let rec loc_offset env loc ty = function
    | TNoOffset -> loc
    | TField(f,off) ->
	loc_offset env (M.field loc f) f.ftype off
    | TIndex(t,off) ->
	let k = integer_of_data (kind_of t.term_type) (!data_rec env t) in
	let ty_elt = Cil.typeOf_array_elem ty in
	let te = Ctypes.object_of ty_elt in
	loc_offset env (M.index loc te k) ty_elt off

  (* ---------------------------------------------------------------------- *)
  (* --- Offsets in the C-memory model                                  --- *)
  (* ---------------------------------------------------------------------- *)

  let rec memory_offset env ty dp = function
    | TNoOffset -> ty,dp
    | TIndex(t,off) ->
	let kp = kind_of_data ty dp in
	let ki = kind_of t.term_type in
	let di = !data_rec env t in
	let te = Cil.typeOf_array_elem ty in
	let ta = object_of te in
	let dq = data_index ta kp dp ki di in
	memory_offset env te dq off
    | TField(f,off) as offset ->
	(match dp with
	  | Range _ | Set _ | List _ ->
	      Datalib.Collector.add_warning
		~reason:"field access over set of l-values"
		"Ignored offset '%a' in assign clause"
		!Ast_printer.d_term_offset offset ;
	      ty, dp
	  | dp ->
	      let dq = data_field (kind_of_typ ty) dp f in
	      memory_offset env f.ftype dq off )

  let gaddress_of_cvar tenv x off =
    let tenv =
      if tenv.formals_in_pre && x.vformal
      then (env_at tenv Pre) else tenv
    in
     memory_offset tenv x.vtype (Loc (M.cvar (mem_at_env tenv) x)) off

  let gaddress_of_mem tenv e off =
    let g = !data_rec tenv e in
    let te =
      match kind_of e.term_type with
	| Kptr telt -> telt
	| Kset(Kptr telt) -> telt
	| _ -> WpLog.fatal "expected pointer"
    in
    memory_offset tenv te g off

  let gstartof (ty,g) =
    let te = Cil.typeOf_array_elem ty in
    let ta = object_of te in
    data_startof ta (kind_of_typ ty) g

  let gstartof_cvar env x off =
    gstartof (gaddress_of_cvar env x off)

  let gstartof_mem env e off =
    gstartof (gaddress_of_mem env e off)

  let gstartof_value env ty v off =
    gstartof (memory_offset env ty (Value v) off)

  (* ---------------------------------------------------------------------- *)
  (* --- Arithmetics Cast                                               --- *)
  (* ---------------------------------------------------------------------- *)

  let cast v ty_from ty_to =
    if ty_from = ty_to then v
    else
      match ty_from with
	| Kcint i ->
	    let vi = integer_of_data ty_from v in
	    begin match ty_to with
	      | Kcint j -> 
		  if Ctypes.sub_c_int i j then v
		  else 
		    Value (M.V_int(j,D.modulo j vi))
	      | Kint -> Data(F.wrap vi)
	      | Kreal -> Data(F.wrap (F.real_of_integer vi))
	      | Kptr te -> Loc (M.cast_int_to_loc i vi te)
	      | k -> WpLog.not_yet_implemented "logic cast from %a to %a"
		  Ctypes.pp_int i pp_kind k
	    end
	| Kint ->
	    let vi = integer_of_data ty_from v in
	    begin match ty_to with
	      | Kcint j -> Value (M.V_int(j,D.modulo j vi))
	      | Kint -> Data(F.wrap vi)
	      | Kptr te -> Loc (M.cast_int_to_loc (Ctypes.c_ptr()) vi te)
	      | Kreal -> Data(F.wrap (F.real_of_integer vi))
	      | k -> WpLog.not_yet_implemented
		  "logic cast from integer to %a" pp_kind k
	    end
	| Kreal ->
	    let vr = real_of_data ty_from v in
	    begin match ty_to with
	      | Kcint j -> Value(M.V_int(j,D.modulo j
					       (F.integer_of_real vr)))
	      | Kint -> Data(F.wrap (F.integer_of_real vr))
	      | Kreal -> Data(F.wrap vr)
	      | k ->  WpLog.not_yet_implemented
		  "logic cast from real to %a" pp_kind k
	    end
	| Kptr tfrom ->
	    let loc = loc_of_data (object_of tfrom) v in
	    begin match ty_to with
	      | Kcint j -> Value (M.V_int(j, M.cast_loc_to_int tfrom loc j))
	      | Kptr tto -> Loc (M.cast_loc_to_loc tfrom tto loc)
	      | k -> WpLog.not_yet_implemented
		  "logic cast from pointer over %a to %a"
		    !Ast_printer.d_type tfrom pp_kind k
	    end
	| k -> WpLog.not_yet_implemented "logic cast from %a" pp_kind k


  (* ---------------------------------------------------------------------- *)
  (* --- Binary Operators                                               --- *)
  (* ---------------------------------------------------------------------- *)

  let int_op = function
    | PlusA   -> F.e_iop Formula.Iadd
    | MinusA  -> F.e_iop Formula.Isub
    | Mult    -> F.e_iop Formula.Imul
    | Div     -> F.e_iop Formula.Idiv
    | Mod     -> F.e_iop Formula.Imod
    | BAnd    -> F.e_band
    | BXor    -> F.e_bxor
    | BOr     -> F.e_bor
    | Shiftlt -> F.e_lshift
    | Shiftrt -> F.e_rshift
    | _ -> assert false

  let real_op = function
    | PlusA  -> F.e_rop Formula.Radd
    | MinusA -> F.e_rop Formula.Rsub
    | Mult   -> F.e_rop Formula.Rmul
    | Div    -> F.e_rop Formula.Rdiv
    | _ -> assert false

  let rel_op = function
    | Rlt  -> Lt
    | Rgt  -> Gt
    | Rle  -> Le
    | Rge  -> Ge
    | Req  -> Eq
    | Rneq -> Ne

  let real_cmp op r1 r2 =
    match op with
      | Lt -> F.e_rcmp Formula.Clt r1 r2
      | Gt -> F.e_rcmp Formula.Clt r2 r1
      | Le -> F.e_rcmp Formula.Cleq r1 r2
      | Ge -> F.e_rcmp Formula.Cleq r2 r1
      | Eq -> F.e_rcmp Formula.Ceq r1 r2
      | Ne -> F.e_rcmp Formula.Cneq r1 r2
      | _ -> assert false

  let int_cmp op i1 i2 =
    match op with
      | Lt -> F.e_icmp Formula.Clt i1 i2
      | Gt -> F.e_icmp Formula.Clt i2 i1
      | Le -> F.e_icmp Formula.Cleq i1 i2
      | Ge -> F.e_icmp Formula.Cleq i2 i1
      | Eq -> F.e_icmp Formula.Ceq i1 i2
      | Ne -> F.e_icmp Formula.Cneq i1 i2
      | _ -> assert false

  let preal_cmp op r1 r2 =
    match op with
      | Lt -> F.p_rcmp Formula.Clt r1 r2
      | Gt -> F.p_rcmp Formula.Clt r2 r1
      | Le -> F.p_rcmp Formula.Cleq r1 r2
      | Ge -> F.p_rcmp Formula.Cleq r2 r1
      | Eq -> F.p_rcmp Formula.Ceq r1 r2
      | Ne -> F.p_rcmp Formula.Cneq r1 r2
      | _ -> assert false

  let pint_cmp op i1 i2 =
    match op with
      | Lt -> F.p_icmp Formula.Clt i1 i2
      | Gt -> F.p_icmp Formula.Clt i2 i1
      | Le -> F.p_icmp Formula.Cleq i1 i2
      | Ge -> F.p_icmp Formula.Cleq i2 i1
      | Eq -> F.p_icmp Formula.Ceq i1 i2
      | Ne -> F.p_icmp Formula.Cneq i1 i2
      | _ -> assert false

  let ptr_rel op l1 l2 =
    match op with
      | Lt -> D.lt_pointer l1 l2
      | Gt -> D.lt_pointer l2 l1
      | Le -> D.le_pointer l1 l2
      | Ge -> D.le_pointer l2 l1
      | Eq -> F.p_eq l1 l2
      | Ne -> F.p_neq l1 l2
      | _ -> assert false

  let ptr_cmp op l1 l2 =
    match op with
      | Lt -> D.lt_pointer_bool l1 l2
      | Gt -> D.lt_pointer_bool l2 l1
      | Le -> D.le_pointer_bool l1 l2
      | Ge -> D.le_pointer_bool l2 l1
      | Eq -> D.equal_pointer_bool l1 l2
      | Ne -> F.e_not (D.equal_pointer_bool l1 l2)
      | _ -> assert false

  let plus i j  = F.e_iop Formula.Iadd i j

  let plus_interval r i =
    match r.D.inf , r.D.sup with
      | None , None -> r
      | Some j, None -> {D.inf = Some (plus i j ); D.sup = None}
      | None, Some k -> {D.inf = None ; D.sup = Some (plus i k)}
      | Some j, Some k -> {D.inf = Some (plus i j );D.sup = Some (plus i k)}

  let plus_interval_interval a b =
    match a , b with
      | ({ D.sup=None ; D.inf=None } as top) , _
      | _ , ({ D.sup=None ; D.inf=None } as top) -> Interval top
      | _ -> Set (D.add_set (D.interval a) (D.interval b))


  let rec add_integer  ka ga kb gb =
    match ga,gb with
      | Set _ , _ | _ , Set _ -> Set (D.add_set (set_of ga) (set_of gb))
      | (Value _ | Data _), (Value _ | Data _) ->
	  data_of_integer (plus (integer_of_data ka ga) (integer_of_data kb gb))
      | Interval r1 , Interval r2 ->plus_interval_interval r1 r2
      | (Value _ | Data _ ),Interval r ->
	  Interval (plus_interval r (integer_of_data ka ga))
      | Interval r,(Value _ | Data _ ) ->
	  Interval (plus_interval r (integer_of_data kb gb))
      | List _ , List _ -> Set (D.add_set (set_of ga) (set_of gb))
      | List ds , b ->  union_map (add_integer kb b ka) ds
      | b , List ds -> union_map (add_integer ka b kb) ds
      | _ -> WpLog.fatal "unsuitable arguments for integer addition"

  let data_cmp binop ka ga kb gb =
    match ka,kb with
      |( Kint  | Kcint _) , (Kint| Kcint _) ->
	int_cmp binop (integer_of_data ka ga) (integer_of_data kb gb)
      | (Kreal|Kint| Kcint _) , (Kreal|Kint| Kcint _) ->
	 real_cmp binop (real_of_data ka ga) (real_of_data kb gb)
      | Kptr ty , Kptr _ ->
	  let obj = object_of ty in
          ptr_cmp  binop  (M.pointer_of_loc (loc_of_data obj ga))
	   (M.pointer_of_loc (loc_of_data obj gb))
      | _ ->
	  Wp_error.not_yet_implemented
	    "boolean comparison between %a and %a"
	    pp_kind ka pp_kind kb

  let data_binop kr binop ka ga kb gb =
    match binop with

      (* pointer arithmetics *)

      | (IndexPI | PlusPI) -> data_shift ka ga kb gb ~is_pos:true
      | MinusPI -> data_shift ka ga kb gb ~is_pos:false
      | MinusPP ->
	  begin
	    match ka with
	      | Kptr te ->
		  let obj = object_of te in
		 data_of_integer
		   (D.minus_pointer
		      (M.pointer_of_loc (loc_of_data obj ga))
		      (M.pointer_of_loc (loc_of_data obj gb)))
	      | _ -> WpLog.fatal "wrong parameters for pointer arithmetics"
	  end

      (* scalar arithmetics *)

      | PlusA | MinusA | Mult | Div | Mod
      | BAnd | BXor | BOr | Shiftlt | Shiftrt ->
	  begin
	    match kr with
	     | (Kint  | Kcint _)  -> data_of_integer
		   (int_op binop
		      (integer_of_data ka ga) (integer_of_data kb gb))
	      | Kreal -> data_of_real
		  (real_op binop (real_of_data ka ga) (real_of_data kb gb))
	      | (Kset Kint |Kset (Kcint _)) when binop = PlusA ->
		  add_integer ka ga kb gb
	      | _ -> WpLog.fatal "wrong parameters for scalar arithmetics"
	  end

      (* comparisons *)
      | Eq | Ne | Gt | Ge | Lt | Le ->
	  data_of_boolean (data_cmp binop ka ga kb gb)

      (*logic or *)
      | LOr ->
	  data_of_boolean (F.e_or (boolean_of_data ga) (boolean_of_data gb))

      (* logic and *)
      | LAnd ->
	  data_of_boolean (F.e_and (boolean_of_data ga) (boolean_of_data gb))

  (* ------------------------------------------------------------------------ *)
  (* --- Unary Operators                                                  --- *)
  (* ------------------------------------------------------------------------ *)

  let data_unop kr unop ka ga =
    match unop with
      | Neg ->
	  begin
	    match kr with
	      | (Kint  | Kcint _)  ->
		  data_of_integer (F.e_ineg (integer_of_data ka ga))
	      | Kreal -> data_of_real (F.e_rneg (real_of_data ka ga))
	      | _ -> WpLog.fatal
		  "wrong type of parameter for negation operator"
	  end
      | LNot ->
	  let b =
	    match kr with
	      |(Kint | Kcint _)->
		 F.e_icmp Ceq (integer_of_data ka ga) (F.e_icst "0")
	      |Kreal -> F.e_rcmp Ceq (real_of_data ka ga) (F.e_rcst "0.0")
	      |Kptr ty ->
		 let obj = object_of ty in
		 D.is_null (M.pointer_of_loc (loc_of_data obj ga))
	      | Kbool ->
		  F.e_not (boolean_of_data ga)
	      | k -> WpLog.fatal
		  "%a : wrong type of parameter for logic not operator" pp_kind k
	  in data_of_boolean b
      | BNot ->
	  begin
	    match kr with
	     | (Kint  | Kcint _) ->
		 data_of_integer (F.e_bnot (integer_of_data ka ga))
	     | _ -> WpLog.fatal
		 "wrong type of parameter for bitwise not operator"
	  end

  (* ------------------------------------------------------------------------ *)
  (* --- Constants                                                        --- *)
  (* ------------------------------------------------------------------------ *)

  let data_const = function
    | CInt64(k,_ik,_) ->
	data_of_integer (F.e_icst (Int64.to_string k))

    | CChr c         ->
	data_of_integer (F.e_icst (string_of_int (Char.code c)))

    | CReal(f,_fk,_) ->
	data_of_real (F.e_rcst (string_of_float f))

    | CEnum e ->
	let machdep = true in
	let e' = Cil.constFold machdep e.eival in
	begin
	  match e'.enode with
	    | Const CInt64(k,_,_) ->
		data_of_integer (F.e_icst (Int64.to_string k))
	    | Const CChr c ->
		data_of_integer (F.e_icst (string_of_int (Char.code c)))
	    | _ -> WpLog.fatal "unrecognized sizeof/alignof "
	end
    | CWStr _        ->
        WpLog.not_yet_implemented "wide character string constant"
    | CStr s         ->
        WpLog.not_yet_implemented "character string constant (%S)" s

  (* ------------------------------------------------------------------------ *)
  (* --- Terms                                                            --- *)
  (* ------------------------------------------------------------------------ *)

  let rec data_of_term env term =
    match term.term_node with

      (* Constants *)
      | TConst c -> data_const c

      (* Operators *)

      | TUnOp (unop,a) ->
	  data_unop (kind_of term.term_type) unop
	    (kind_of a.term_type) (data_of_term env a)

      | TBinOp(binop,a,b) ->
	  data_binop
	    (kind_of term.term_type) binop
	    (kind_of a.term_type) (data_of_term env a)
	    (kind_of b.term_type) (data_of_term env b)

      (* L-Values *)

      | TLval(TResult _ ,off) ->
	  logic_offset env (F.var(fst (M.result env.frame))) off

      | TLval (TVar{lv_name = "\\exit_status"},_) ->
	  Data (F.var (M.exit_status env.frame))

      | TLval (TVar({lv_origin=None} as x),off) ->
	  logic_offset env (lvar env.lvars x) off

      | TLval(TVar{lv_origin=Some x},off) ->
	  begin
	    match xvar_of_varinfo env x with
	      | MemVar ->
		  let tr,gloc = gaddress_of_cvar env x off in
		  data_load env tr gloc

	      | PrmVar(M.V_pointer(ty,loc)) ->
		  Loc (loc_offset env loc ty off)

	      | PrmVar v ->
		  let r = M.logic_of_value v in
		  logic_offset env r off
	  end

      | TLval(TMem e,off) ->
	  let tr,gloc = gaddress_of_mem env e off in
	  data_load env tr gloc

      | TAddrOf(TVar{lv_origin=Some x},off) ->
	   begin
	    match xvar_of_varinfo env x with
	      | MemVar -> snd (gaddress_of_cvar env x off)
	      | PrmVar v -> snd (memory_offset env x.vtype (Value v) off)
	  end

      | TStartOf(TVar{lv_origin=Some x},off) ->
	  begin
	    match xvar_of_varinfo env x with
	      | MemVar -> gstartof_cvar env x off
	      | PrmVar v -> gstartof_value env x.vtype v off
	  end

      | TAddrOf(TMem e,off) ->
	  snd (gaddress_of_mem env e off)
      | TStartOf(TMem e,off) ->
	  gstartof_mem env e off

      | TAddrOf(TResult _,_)
      | TStartOf(TResult _,_) -> WpLog.not_yet_implemented "&\\result"

      | TAddrOf(TVar {lv_origin=None},_) ->
          WpLog.fatal "taking address of a logic variable"

      | TStartOf(TVar {lv_origin=None},_) ->
          WpLog.not_yet_implemented "reference to a logic array"

      (* At *)

      | Tat(t,label) -> data_of_term (env_at env (c_label label)) t
      | Told t ->  data_of_term (env_at env Pre)  t

      (* Sizeof and alignment *)

      | TSizeOf _
      | TSizeOfE _
      | TSizeOfStr _
      | TAlignOf _
      | TAlignOfE _ ->
	  let machdep = true in
	  let e' = Cil.constFoldTerm machdep term in
	  begin
	    match e'.term_node with
	    | TConst _ -> data_of_term env e'
	    | _ -> WpLog.fatal "unrecognized sizeof/alignof (%a)"
		!Ast_printer.d_term term
	  end

      (* Conditional *)

      | Tif (b, t, f) ->
	  Data
	  (F.e_cond (boolean_of_data (data_of_term env b))
	    (term_of_data (data_of_term env t))
	    (term_of_data (data_of_term env f)))

      (* Memory call *)

      | Tbase_addr t ->
	  let obj = match t.term_type with
	    | Ctype ty -> object_of ty
	    | _ -> WpLog.fatal "Base-address of logic type object"
	  in
	  Loc (M.base_address
		 (mem_at_env env)
		 (loc_of_data  obj (data_of_term env t)))

      | Tblock_length t ->
	  let obj = match t.term_type with
	    | Ctype ty -> object_of ty
	    | _ -> WpLog.fatal "Block-length of logic type object"
	  in
	  data_of_integer
	    (M.block_length
	       (mem_at_env env)
	       (loc_of_data obj (data_of_term env t)))

      (* Range *)

      | Trange (ti,tj ) ->
	  let option_int env  = function
	    | None -> None
	    | Some x -> Some (integer_of_data (kind_of x.term_type)
				 (data_of_term env x))
	  in
	  let r = {D.inf =(option_int env ti);
		   D.sup =(option_int env tj)} in
	  Interval r

      | Tempty_set -> List  []

      | Tunion xs -> union_map (data_of_term env) xs


      | Tinter(a::b) ->
	  Set (List.fold_left
		 (fun s1 s2 ->
		    D.inter s1 (set_of (data_of_term env s2)))
		 (set_of (data_of_term env a)) b)

      | Tinter [] -> WpLog.fatal "empty intersection"

      | Tcomprehension (_, _, _) ->
	  WpLog.not_yet_implemented "Set comprehension"

      (* Conversions *)

      | Tnull -> Data (F.wrap D.null)

      | TCastE (ty,t) ->
	  cast (data_of_term env t) (kind_of t.term_type) (kind_of_typ ty)

      (* Logic ADT *)

      | TUpdate (_,TNoOffset,tv) ->
	  data_of_term env tv

      | TUpdate (r,TField (f, TNoOffset),tv) ->
	  begin
            match kind_of (r.term_type) with
	      | Kstruct _ ->
		  let record = record_of_data (data_of_term env r) in
		  let v = term_of_data (data_of_term env tv) in
		  let r = D.upd_field record f v in
		  Data (F.wrap r)

	      | _ ->
		  WpLog.fatal "Functional update of a non-record value"
	  end

      | TUpdate (r,TIndex(k, TNoOffset),tv) ->
	  begin
            match kind_of (r.term_type) with
	      | Karray _ ->
		  let array = array_of_data (data_of_term env r) in
		  let idx = integer_of_data (kind_of k.term_type)
		    (data_of_term env k) in
		  let v = term_of_data (data_of_term env tv) in
		  let r = D.upd_index array idx v in
		  Data (F.wrap r)

	      | _ ->
		  WpLog.fatal "Functional update of a non-array value"
	  end

      | TUpdate (_,_,_) ->
	  WpLog.not_yet_implemented "ACSL extension for functional update"

      | TDataCons({ctor_name="\\true"},[]) -> Data(F.wrap F.e_true)
      | TDataCons({ctor_name="\\false"},[]) -> Data(F.wrap F.e_false)

      | TDataCons (c,_)  ->
	  WpLog.not_yet_implemented "Constructor (%s)" c.ctor_name

      (* Jessie *)
      | TCoerce (_,_)
      | TCoerceE (_,_) ->
	  assert false (* Only produced by Jessie plugin *)

      (* Type *)

      | Ttypeof _
      | Ttype _ ->
	  WpLog.not_yet_implemented "Type Tag"

      (* Let binding  *)

      | Tlet (({l_var_info =x;
	       l_labels=[];l_tparams=[];
	       l_profile =[];l_body=(LBterm t1); l_type=Some _ } as linfo),t2) ->
	   if Logic_env.Logic_builtin_used.mem linfo then
	    Wp_parameters.not_yet_implemented "Built-ins symbols"
	   else
	     let var = D.fresh x.lv_name (Mdata.Vacsl x.lv_type ) in
	     let env2 = add_lvar env x (F.var var) in
	     let t1' = term_of_data (data_of_term env t1) in
	     let t2' = term_of_data (data_of_term env2 t2) in
	     Data (F.e_subst D.alpha var t1' t2')
      | Tlet _ ->
          WpLog.not_yet_implemented
	    "Complex Let-binding"

      (* Logic-Function Call *)

      | Tapp (lfun,labels,args) ->
	   if Logic_env.Logic_builtin_used.mem lfun then
	    Wp_parameters.not_yet_implemented "Built-ins symbols"
	  else
	  Data (!rec_apply_function env lfun labels args)

      (* Higer order function *)

      | Tlambda (_,_) ->
	  WpLog.not_yet_implemented "Higher order functions"

  let () = data_rec := data_of_term

  let term env t = term_of_data (data_of_term env t)

  (* ------------------------------------------------------------------------ *)
  (* --- Assignable                                                       --- *)
  (* ------------------------------------------------------------------------ *)

  let rec data_of_assignable env t=
    match t.term_node with

    | TLval (TVar{lv_origin=Some x}, off) ->
	begin
	  match xvar_of_varinfo env x with
	    | MemVar -> snd (gaddress_of_cvar env  x off)
	    | PrmVar v -> Value v
	end
    | TLval (TMem e, off) ->
	snd (gaddress_of_mem env e off)
    | TAddrOf _lv -> WpLog.fatal "Unexpected address value"

    | TStartOf (TVar{lv_origin=Some x}, off)  ->
	begin
	  match xvar_of_varinfo env x with
	    | MemVar -> gstartof_cvar env x off
	    | PrmVar v -> gstartof_value env x.vtype v off
	end

    | TStartOf (TMem e, off) -> gstartof_mem env e off

    | Tempty_set  | TLval (TResult _, _ )-> List []

    | Tat(t ,lab) -> data_of_assignable (env_at env (c_label lab)) t
    | Told t -> data_of_assignable (env_at env Pre) t

    | TBinOp((IndexPI | PlusPI),a,b) ->
	data_shift
	  (kind_of a.term_type) (data_of_term env a)
	  (kind_of b.term_type) (data_of_term env b)
	  ~is_pos:true

    | TBinOp (MinusPI,a,b) ->
	data_shift
	  (kind_of a.term_type) (data_of_term env a)
	  (kind_of b.term_type) (data_of_term env b)
	  ~is_pos:false

    | Tunion ts -> union_map (data_of_assignable env) ts
    | Tinter (t::ts) ->
	let f t = set_of (data_of_assignable env t) in
	Set(List.fold_left
	      (fun s t -> D.inter s (f t)) (f t) ts)
    | Tinter [] -> WpLog.fatal "empty intersection"

    | TStartOf (TResult _, _ )
    | Tlet (_, _) | Tcomprehension (_, _, _)
    | Tif (_, _, _) | Tapp (_, _, _) | TCastE (_, _)

	-> WpLog.not_yet_implemented "assignation of (%a)"
	!Ast_printer.d_term t

    | TLval (TVar {lv_origin=None}, _)
    | TStartOf (TVar {lv_origin=None}, _)
    | Trange (_, _)|Ttype _|Ttypeof _|TCoerceE (_, _)|TCoerce (_, _)
    | TUpdate (_, _, _)|Tblock_length _
    | TDataCons (_, _)|Tlambda (_, _)
    | TUnOp (_, _)|TAlignOfE _|TAlignOf _|TSizeOfStr _|TSizeOfE _|TSizeOf _
    | TConst _ | Tnull | Tbase_addr _
    | TBinOp(
	(LOr|LAnd|BOr|BXor|BAnd|Ne|Eq|Ge|Le|Gt|Lt|
	     Shiftrt|Shiftlt|Mod|Div|Mult|
		 MinusPP|MinusA|PlusA),_,_)
      -> WpLog.fatal "not assignable terms"

  let assigned env t =
    (* (ZD) DO preserve this filter, as \empty is polymorphic! *)
    match t.term_node with
      |  Tempty_set  | TLval (TResult _, _ ) -> []
      | _ ->
	  let data = data_of_assignable env t in
	  let te =
	    match t.term_type with
	      | Ctype te -> te
	      | Ltype( {lt_name="set"} , [Ctype elt] ) -> elt
 	      | _ -> WpLog.fatal
		  "unexpected logic-type for assignable term %a"
		    !Ast_printer.d_logic_type t.term_type
	  in
	  assigned_of_data (object_of te) [] data

  (* ------------------------------------------------------------------------ *)
  (* --- Properties                                                       --- *)
  (* ------------------------------------------------------------------------ *)

  let rec_apply_predicate
      : (env -> logic_info ->
	   (logic_label * logic_label) list -> term list -> F.pred) ref
      = ref (fun _ _ _ _ -> assert false)

  let pred_cmp binop ka ga kb gb =
    match ka,kb with
      | (Kint  | Kcint _) , (Kint  | Kcint _) ->
	  pint_cmp binop (integer_of_data ka ga) (integer_of_data kb gb)
      | (Kreal|Kint| Kcint _) , (Kreal|Kint| Kcint _) ->
	  preal_cmp binop (real_of_data ka ga) (real_of_data kb gb)
      |  Kptr _ty , Kptr _ ->
	   ptr_rel binop
	     (F.unwrap (term_of_data ga))
	     (F.unwrap (term_of_data gb))
      | Kstruct s, Kstruct s' when (s.Cil_types.ckey=s'.Cil_types.ckey) ->
	  D.eq_record s (record_of_data ga) (record_of_data gb)
      | Karray arr, Karray arr' when Ctypes.AinfoComparable.equal arr arr' ->
	  D.eq_array arr (array_of_data ga) (array_of_data gb)
      | _ ->
	  begin
            match binop with
	      | Eq -> F.p_eq (term_of_data ga) (term_of_data gb)
	      | Ne -> F.p_not (F.p_eq (term_of_data ga) (term_of_data gb))
	      | _  -> WpLog.fatal ~current:true "Unexpected comparison"
	  end

  let rec prop env p =
    List.fold_right F.p_named p.name (prop_body env p)

  and prop_body env p =
    match p.content with
      | Pfalse -> F.p_false

      | Ptrue  -> F.p_true

      | Pand(p1,p2) ->
	  F.p_and (prop env p1)(prop env p2)

      | Por(p1,p2)  ->
	  F.p_or (prop env p1)(prop env p2)

      | Pxor(p1,p2) ->
	  (*TODO : no primitive translation for XOR in WHY *)
	  (*Infact in bool.why there is bool_xor *)
	  F.p_xor(prop env p1)(prop env p2)

      | Pimplies(p1,p2) ->
	  F.p_implies(prop env p1)(prop env p2)

      | Piff(p1,p2)->
	  F.p_iff(prop env p1)(prop env p2)

      | Pnot p -> F.p_not (prop env p)

      | Pif(c,pt,pf) ->
	  F.p_cond (F.unwrap (term env c))
	    (prop env pt)
	    (prop env pf)

      | Pold p -> prop (env_at env Pre) p

      | Pat (p,l) ->  prop (env_at env (c_label l))  p

      | Prel ( rel ,t1,t2) ->
	  let ct1 = kind_of t1.term_type in
	  let ct2 = kind_of t2.term_type in
	  let m1 = data_of_term env t1 in
	  let m2 = data_of_term env t2 in
	  pred_cmp (rel_op rel) ct1 m1 ct2 m2

      | Pvalid t ->
	  let k = kind_of t.term_type in
	  let d  = data_of_term env t in
	  data_valid (mem_at_env env) k d

      | Pvalid_index(tp,ti) ->
	  let ty = match tp.term_type with
	    | Ctype te -> te
	    | _ -> WpLog.fatal "expected a non logic type"
	  in
	  let kp = kind_of_typ ty in
	  let ki = kind_of ti.term_type in
	  let dp = data_of_term env tp in
	  let di = data_of_term env ti in
	  if Cil.isArrayType ty then
	    let te = Cil.typeOf_array_elem ty in
	    let ta = object_of te in
	    let d = data_index ta kp dp ki di in
	    data_valid (mem_at_env env) kp d
	  else
	    if Cil.isPointerType ty then
	      let d = data_shift kp dp ki di ~is_pos:true in
	      data_valid (mem_at_env env) kp d
	    else WpLog.fatal "unexepected type for valid index"

      | Pvalid_range(b,l,h) ->
	  let tb = b.term_type in
	  let k = kind_of tb in
	  begin
	    match k with
	      | Kptr _ ->
		  let ty = typ_of_elements tb in
		  let obj = object_of ty in
		  let loc = loc_of_data obj (data_of_term env b) in
		  let rg = {
		    D.inf = Some (integer_of_data (kind_of l.term_type)
				    (data_of_term env l));
		    D.sup = Some (integer_of_data (kind_of h.term_type)
				    (data_of_term env h));
		  } in
		  M.valid (mem_at_env env) (D.Arange(obj,loc,rg))
	      | _ -> WpLog.fatal "unsuitable argument for [valid_range]"
	  end

      | Pfresh _t -> WpLog.not_yet_implemented "fresh"

      | Psubtype (_t1,_t2) ->  WpLog.not_yet_implemented "subtype"

      | Plet(def, p) ->
          begin
            let lv = def.l_var_info in
            match def.l_body, def.l_profile with
		LBterm t, [] ->
	          let x = D.fresh lv.lv_name (Mdata.Vacsl lv.lv_type) in
	          D.subst x (term env t)
		    (prop (quant_env env [(lv,F.var x)]) p)
              | _ -> WpLog.not_yet_implemented "local binding"
          end

      | Pforall (xs,p) ->
	  let freshes =
	    List.map
	      (fun x -> x ,
		 D.fresh x.lv_name (Mdata.Vacsl x.lv_type)) xs in
	  let quantified = List.map snd freshes in
	  let assoc = List.map (fun (x,v) -> x,F.var v) freshes in
	  D.forall quantified (prop (quant_env env assoc) p)

      | Pexists (xs,p) ->
	  let freshes =
	    List.map
	      (fun x -> x , D.fresh x.lv_name (Mdata.Vacsl x.lv_type)) xs
	  in
	  let quantified = List.map snd freshes in
	  let assoc = List.map (fun (x,v) -> x,F.var v) freshes in

 	  D.exists quantified
	    (prop (quant_env env assoc) p)

      | Pseparated tl ->
	  let gs =
	    List.map
	      (fun t ->
		 let te =
		   match kind_of t.term_type with
		     | Kptr te -> te
		     | Kset (Kptr te) -> te
		     | k -> WpLog.fatal "separated on non pointer type : %a " pp_kind k
		 in
		 Ctypes.object_of te, data_of_term env t) tl in
	  let ags = Array.of_list gs in
	  let p = ref F.p_true in
	  for i=0 to Array.length ags - 2 do
	    for j=i+1 to Array.length ags - 1 do
	      p := F.p_and !p (data_separated ags.(i) ags.(j))
	    done
	  done ;
	  !p

      | Papp (predicate,labels,args) ->
	  !rec_apply_predicate env predicate labels args

  (* ------------------------------------------------------------------------ *)
  (* --- Applying Hints                                                   --- *)
  (* ------------------------------------------------------------------------ *)

  open Axiomatics

  let get_axiom name =
    try Hashtbl.find user_axiomdefs name
    with Not_found ->
      Wp_parameters.failure "Non-compiled axiom '%s'" name ;
      None

  type partial =
    | Svalue of abstract F.term
    | Sclosure of M.closure

  let hints = ref 0
  let () = D.on_clear (fun () -> hints := 0)

  let pp_partial fmt = function
    | Svalue e -> Format.fprintf fmt "(%a)"F.pp_term e
    | Sclosure c -> Format.fprintf fmt "{@@%s}" (M.label_at_closure c)

  let apply_sigma (env:env) sigma closure =
    try
      let k_label = M.label_at_closure closure in
      let c_label = List.assoc k_label sigma in
      Svalue( M.userdef_closure (M.mem_at env.frame c_label) closure )
    with Not_found ->
      Sclosure closure

  (* Hints *)

  class instance axdef partials =
  object(self)

    val axdef = axdef
    val mutable id = 0
    val mutable partials = partials
    val mutable captured = false

    method id = id

    method pretty fmt =
      if captured then Format.pp_print_string fmt "[captured]" ;
      Format.pp_print_string fmt axdef.a_name ;
      List.iter (fun s -> pp_partial fmt s) partials

    initializer
      id <- (incr hints ; !hints)

    (* Mdata.hint interface *)

    method has_var xs =
      not captured && xs <> [] &&
	List.exists
	(function
	   | Svalue t -> F.term_has_var xs t
	   | Sclosure _ -> false)
	partials

    method capture xs =
      if self#has_var xs then captured <- true

    method subst x (v : Formula.abstract F.term) =
      if not captured then
	let modified = ref false in
	let partials' = List.map
	  (function
	     | Svalue t as s ->
		 if F.term_has_var [x] t then
		   ( modified := true ; Svalue (F.e_subst D.alpha x v t) )
		 else s
	     | Sclosure _ as s -> s)
	  partials
	in
	if !modified then
	  begin
	    id <- (incr hints ; !hints) ;
	    partials <- partials'
	  end

    (* Saturation *)

    method extend_with_hint env hint =
      if not captured && axdef.a_name = hint.h_name then
	let better = ref false in
	let partials' = List.map
	  (function
	     | Svalue _ as s -> s
	     | Sclosure c ->
		 let s = apply_sigma env hint.h_sigma c in
		 (match s with Svalue _ -> better := true | _ -> ()) ; s)
	  partials in
	if !better then Some(new instance axdef partials') else None
      else None

    (* Axiom instanciation *)

    method instantiate =
      try
	let params =
	  List.map
	    (function
	       | Svalue v -> v
	       | Sclosure _ -> raise Not_found)
	    partials
	in
	UserAxiomDefs.define axdef ;
	Some (F.p_call axdef.a_hyp_name params)
      with Not_found ->	None

    method alive = not captured

  end

  let instance_of_hint env hint =
    match get_axiom hint.h_name with
      | Some axdef ->
	  let partials = List.map
	    (apply_sigma env hint.h_sigma)
	    axdef.a_closure
	  in Some( new instance axdef partials )
      | None -> None

  let saturate instances env hints =
    begin
      let acc = ref instances in
      List.iter
	(fun h ->
	   match instance_of_hint env h with
	     | None -> ()
	     | Some a -> 
		 Wp_parameters.debug ~level:2 "New axiom hint %t@." a#pretty ;
		 acc := a :: !acc)
	hints ;
      List.iter
	(fun a ->
	   List.iter
	     (fun h ->
		match a#extend_with_hint env h with
		  | None -> ()
		  | Some a' -> 
		      Wp_parameters.debug ~level:2 "Extended axiom hint %t@." a#pretty ;
		      acc := a' :: !acc)
	     hints)
	instances ; (* only the ones previously known *)
      !acc
    end

  let hint_context = ref None

  let push_instances where instances =
    match !hint_context with
      | None -> hint_context := Some instances
      | Some _ ->
	  Wp_parameters.fatal "Hint context mismatch (push %s)" where

  let pop_instances where =
    match !hint_context with
      | Some instances ->
	  hint_context := None ;
	  List.filter (fun h -> h#alive) instances
      | None ->
	  Wp_parameters.fatal "Hint context mismatch (pop %s)" where

  let apply_hints env def labels =
    match !hint_context with
      | None -> ()
      | Some instances ->
	  try
	    let hints = Axiomatics.apply def.l_var_info labels in
	    hint_context := Some(saturate instances env hints)
	  with err -> (*TODO nettoyer, ne pas eliminer
                             des exceptions trop vite*)
	    Wp_parameters.warning
	      "Error during hints computation (%s)"
	      (Printexc.to_string err)

  let rec do_instantiate hs p = function
    | [] -> hs , p
    | h::others ->
	match h#instantiate with
	  | Some hyp ->
	      (*TODO: add pid to collector *)
	      do_instantiate hs (F.p_implies hyp p) others
	  | None ->
	      do_instantiate
		(if h#alive then h::hs else hs)
		p others

  let instantiate hs p = do_instantiate [] p hs

  let as_hint (h : instance) = (h :> D.hint)

  let merge hs1 hs2 =
    let cmp h1 h2 = Datatype.Int.compare h1#id h2#id in
    let norm hs = List.sort cmp (List.filter (fun h -> h#alive) hs) in
    let rec do_merge hs1 hs2 =
      match hs1 , hs2 with
	| h1::hst1 , h2 :: hst2 ->
	    let k1 = h1#id in
	    let k2 = h2#id in
	    if k1 < k2 then h1 :: do_merge hst1 hs2 else
	      if k1 > k2 then h2 :: do_merge hs1 hst2 else
		h1 :: do_merge hst1 hst2
	| hs , [] | [] , hs -> hs
    in do_merge (norm hs1) (norm hs2)

  (* ------------------------------------------------------------------------ *)
  (* --- Accessing User-Definitions                                       --- *)
  (* ------------------------------------------------------------------------ *)

  let get_definition cc fdef =
    let xdef = fdef.l_var_info in
    let udef , items =
      try Hdef.find user_definitions xdef
      with Not_found ->
	let cdef = cc fdef in
	Hdef.add user_definitions xdef cdef ; cdef
    in
    UserDefinition.define udef items ; udef

  (* ----------------------------------------------------------------------- *)
  (* --- Compilation of User-defined Predicate and Functions             --- *)
  (* ----------------------------------------------------------------------- *)

  let push_context where =
    let hints = !hint_context in
    hint_context := None ;
    hints , D.push where (F.pool()) D.closed []

  let flush_context where (hints,context) p =
    hint_context := hints ; D.flush where context p

  let kill_context where (hints,context) =
    hint_context := hints ; D.kill where context

  let user_env pdef =
    let context = push_context "user" in
    try
      let frame = M.userdef_frame () in
      let formals = List.map
	(fun lv ->
	   let x = D.fresh lv.lv_name (Mdata.Vacsl lv.lv_type) in
	   lv , x
	) pdef.l_profile
      in
      let here =
	match pdef.l_labels with
	  | [] -> Clabels.LabelParam "WP_nowhere" (* Non-expected ! *)
	  | LogicLabel (None, first) :: _ -> Clabels.LabelParam first
	  | LogicLabel (Some _, _) :: _ -> 
	      Wp_parameters.fatal
		"Unexpected redefined labels in user-defined predicates"
	  | StmtLabel _ :: _ ->
	      Wp_parameters.fatal
		"Unexpected stmt-labels in user-defined predicates"
      in
      context , formals,
      {
	formals_in_pre = false ;
	frame = frame ;
	label = here ;
	lvars =
	  List.fold_right
	    (fun (lv,x) t -> Logic_var.Map.add lv (F.var x) t) formals Logic_var.Map.empty;
	xvars = Varinfo.Map.empty ;
      }
    with err ->
      kill_context "user env" context; raise err

  let compile_predicate pdef =
    let o_name = Datalib.identifier (pdef.l_var_info.lv_name) in
    let p_name = "D_" ^ o_name in
    let d_name = "Def_" ^ o_name in
    let context, formals, env = user_env pdef in
    try
      let body,tused,pused =
	match pdef.l_body with
	  | LBpred named ->
	      let v= prop env named in (Some v, [],[v])
	  | LBterm def  ->
	      let v = term env def in (Some (F.p_bool (F.unwrap v)), [v],[])
          | LBnone ->
	      None,[],[]
	  | LBreads xs ->
	      (*TODO: Below is an incorrect translation (LC+BM) (Idem for "\from")
		(LC) Reason : xs are to be interpreted as left-values.
		(Use case : f reads t[0..n])
		let vs = List.map (fun x -> term env x.it_content) xs in
		None,vs,[]
	      *)
	      Wp_parameters.warning ~once:true "Interpreting reads-definition as expressions rather than tsets" ;
	      let vs = List.map (fun x -> term env x.it_content) xs in
	      None,vs,[]
	  | LBinductive _ ->
	      Wp_parameters.not_yet_implemented "Inductive predicates"
      in
      let implicits , closures = M.userdef_signature env.frame tused pused in
      let p_formals = List.map snd formals @ implicits in
      let declaration =
	p_name ,
	Formula.Predicate(List.map F.tau_of_var p_formals) in
      let definitions =
	match body with
	  | None -> kill_context "compile" context ; []
	  | Some body ->
	      let p_axiom =
		D.forall p_formals
		  (F.p_iff
		     (F.p_call p_name (List.map F.var p_formals))
		     (flush_context "compile" context body))
	      in
              [ d_name , Formula.Axiom p_axiom ]
      in
      {
	d_info = pdef ;
	d_callname = p_name ;
	d_implicit = implicits ;
	d_closure = closures ;
	d_formals = List.map snd formals ;
      } ,
      declaration :: definitions
    with err ->
      kill_context "compile" context ;
      raise err

  let compile_function fdef =
    let o_name = Datalib.identifier (fdef.l_var_info.lv_name) in
    let f_name = "D_" ^ o_name in
    let d_name = "Def_" ^ o_name in
    let context, formals, env = user_env fdef in
    try
      let body, used =
	match fdef.l_body with
	  | LBterm def  ->
	      let v = term env def in (Some v, [v])
	  | LBreads xs ->
	      (*TODO: Incorrect translation (Cf. predicates)
	      None,List.map (fun x -> term env x.it_content) xs
	      *)
	      Wp_parameters.warning ~once:true "Interpreting reads-definition as expressions rather than tsets" ;
	      let vs = List.map (fun x -> term env x.it_content) xs in
	      None,vs
	  | LBnone ->
	      None,[]
	  | LBinductive _ ->
	      Wp_parameters.fatal "Inductive function"
	  | LBpred _ ->
	      Wp_parameters.fatal "Function defined by a predicate"
      in
      let t_result =
	match fdef.l_type with
	  | Some ltyp -> D.tau_of_logic_type ltyp
	  | None -> Wp_parameters.fatal "Function defined with not result type"
      in
      let implicits , closures = M.userdef_signature env.frame used [] in
      let f_formals = List.map snd formals @ implicits in
      let declaration =
	f_name ,
	Formula.Function(List.map F.tau_of_var f_formals,t_result)
      in
      let definitions =
	match body with
	  | None -> kill_context "compile" context ; []
	  | Some def ->
	      let f_axiom = F.p_forall f_formals
		(flush_context "compile" context
		   (F.p_eq (F.e_call f_name (List.map F.var f_formals)) def))
	      in
	      [ d_name , Formula.Axiom f_axiom ]
      in
      {
	d_info = fdef ;
	d_callname = f_name ;
	d_implicit = implicits ;
	d_closure = closures ;
	d_formals = List.map snd formals ;
      } ,
      declaration :: definitions
    with err ->
      kill_context "compile" context ;
      raise err

  (* ----------------------------------------------------------------------- *)
  (* --- Compilation of User-defined Axiom                               --- *)
  (* ----------------------------------------------------------------------- *)

  let axiom_env here =
    let frame = M.userdef_frame () in
    let here = Clabels.LabelParam here (* Non-expected ! *) in
    {
      formals_in_pre = false ;
      frame = frame ;
      label = here ;
      lvars = Logic_var.Map.empty ;
      xvars = Varinfo.Map.empty ;
    }

  let compile_user_axiom name predicate =
    if not (Hashtbl.mem user_axioms name) then
      let context = push_context "axiom" in
      try
	let env = axiom_env "WP_nowhere" in
	let def = prop env predicate in
	Hashtbl.add user_axioms name
	  (flush_context "axiom" context def)
      with err ->
	kill_context "axiom" context ;
	raise err

  let compile_user_hypothesis name labels predicate =
    let h_name = "H_" ^ name in
    let d_name = "Def_" ^ name in
    let context = push_context "compile-axiom" in
    try
      let here =
	try Axiomatics.label_name (List.hd labels)
	with _ -> Wp_parameters.fatal "No default label for Axiom '%s'" name
      in
      let env = axiom_env here in
      let body = prop env predicate in
      let implicits , closures = M.userdef_signature env.frame [] [body] in
      let property =
	D.forall implicits
	  (F.p_iff
	     (F.p_call h_name (List.map F.var implicits))
	     (flush_context "compile-axiom" context body)) in
      Hashtbl.add user_axiomdefs name
	(Some {
	   a_name = name ;
	   a_labels = List.map Axiomatics.label_name labels ;
	   a_hyp_name = h_name ;
	   a_def_name = d_name ;
	   a_implicit = implicits ;
	   a_closure = closures ;
	   a_property = property ;
	 }) ;
    with err ->
      Hashtbl.add user_axiomdefs name None ;
      kill_context "compile-axiom" context ;
      raise err

  let add_axiom name labels predicate =
    if labels = [] then
      ( compile_user_axiom name predicate ;
	UserAxiom.define name )
    else
      if Wp_parameters.Axioms.get () then
	compile_user_hypothesis name labels predicate
      else
	Wp_parameters.warning ~once:true
	  "Ignoring axiom %s (because of labels)" name

  (* ------------------------------------------------------------------------ *)
  (* --- Applying Definitions                                             --- *)
  (* ------------------------------------------------------------------------ *)

  let apply_closure env labels closure =
    let label = Clabels.lookup labels (M.label_at_closure closure) in
    let mem = M.mem_at env.frame label in
    M.userdef_closure mem closure

  let apply_predicate env def labels args =
    apply_hints env def labels ;
    let definition = get_definition compile_predicate def in
    let params = List.map (term env) args in
    let implicits = List.map (apply_closure env labels) definition.d_closure in
    F.p_call definition.d_callname (params @ implicits)

  let apply_function env def labels args =
    apply_hints env def labels ;
    let definition = get_definition compile_function def in
    let params = List.map (term env) args in
    let implicits = List.map (apply_closure env labels) definition.d_closure in
    F.e_call definition.d_callname (params @ implicits)

  let () =
    begin
      rec_apply_predicate := apply_predicate ;
      rec_apply_function  := apply_function ;
    end

end
