(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(** Every access to annotations have to go through here,
  * so this is the place where we decide what the computation
  * is allowed to use. *)

open Db_types
open Cil_types

(** Kind of PO: normal, establishment, preservation, ...*)
type prop_id

(** returns the annotation which lead to the given PO. *)
val property_of_id : prop_id -> Property.t
val compare_prop_id : prop_id -> prop_id -> int

(** splits a prop_id goals into prop_id parts for each sub-goals *)

val split : ( prop_id -> 'a -> unit ) -> prop_id -> 'a Bag.t -> unit

(** A proof accumulator for a set of related prop_id *)
type proof

val create_proof : prop_id -> proof
  (** to be used only once for one of the related prop_id *)

val add_proof : proof -> prop_id -> prop_id list -> unit
  (** accumulate int the proof the partial proof for this prop_id *)

val is_composed : proof -> bool
  (** whether a proof needs several lemma to be complete *)

val is_proved : proof -> bool
  (** wether all partial proofs have been accumulated or not *)

val consolidate_loop : proof -> unit

val target : proof -> Property.t
val dependencies : proof -> Property.t list
val pid_for_function : kernel_function -> model:string -> prop_id
val missing_rte : kernel_function -> string list
val compute_rte_for : kernel_function -> unit

(** generates an unique name for a given PO. *)
val prop_txt : prop_id -> string

val pp_id : Format.formatter -> prop_id -> unit

(* TODO: this one should be in Properties_status. *)
val id_prop_txt : Property.t -> string

(** Short description of the kind of PO *)
val label_of_prop_id: prop_id -> string

(* module Hprops : Hashtbl.S with type key = prop_id
* see Properties_status.Htbl *)
(*
type prop_body = (* Cil_types.predicate named *)
  | Ppred of Cil_types.predicate named
  | Passigns of Cil_types.logic_label (* label to know where to stop *)
              * Cil_types.identified_term Cil_types.assigns list

val pp_prop_body : Format.formatter -> prop_body -> unit
*)
type t_pred = Cil_types.predicate named
type a_fun = Assigns_FctOut | Assigns_FctExit | Assigns_Stmt | Assigns_Loop
type a_kind = LoopAssigns | StmtAssigns
type t_assigns = {
  a_label : Cil_types.logic_label ;
  a_fun : a_fun ;
  a_kind : a_kind ;
  a_assigns : Cil_types.identified_term Cil_types.assigns list ;
}
type t_axiom = string * Cil_types.logic_label list * Cil_types.predicate named

type t_id_pred = prop_id * t_pred
type t_id_assigns = prop_id * t_assigns
type t_id_axiom = prop_id * t_axiom

type t_annots

val get_hyp_only : t_annots -> t_id_pred list
val get_goal_only : t_annots -> t_id_pred list
val get_hyp_goal : t_annots -> t_id_pred list
(** the [bool] in [get_cut] results says if the property has to be
* considered as a both goal and hyp ([goal=true], or hyp only ([goal=false]) *)
val get_cut : t_annots -> (bool * t_id_pred) list
(** [get_inv] only gives the general invariants (not loop invariant) *)
val get_inv : t_annots -> t_id_pred list
val get_asgn_hyp : t_annots -> t_id_assigns option
val get_asgn_goal : t_annots -> t_id_assigns option

val body_of_id_pred : t_id_pred ->  t_pred
val id_of_id_pred : t_id_pred -> prop_id
val body_of_id_assigns : t_id_assigns -> t_assigns
val id_of_id_assigns : t_id_assigns -> prop_id
val is_assigns : prop_id -> bool

val pp_annots : Format.formatter -> t_annots -> unit

type strategy_info
val new_loop_computation : strategy_info -> bool
val pp_strategy_info : Format.formatter -> strategy_info -> unit

type strategy_for_froms = {
  get_pre : unit -> t_annots
}
type strategy_kind =
  | SKannots (** normal mode for annotations *)
  | SKfroms of strategy_for_froms

type strategy = {
  behavior_name : string option ;
  info : strategy_info;

  (** Tell if the stategy includes some assigns goals *)
  has_asgn_goal : bool;
  (** Tell if the stategy includes some other goals *)
  has_prop_goal : bool;
  (** Tell if the stategy includes some arbitrary invariant properties *)
  has_inv_prop : Cil2cfg.edge list;

  (** Hypothesis and goals to be considered on an edge *)
  get_annots : Cil2cfg.edge -> t_annots ;

  global_axioms : t_id_axiom list ;

  (** Needed to process calls : *)
  called_preconditions : Kernel_function.t -> t_id_pred list ;
  called_postconditions : Kernel_function.t -> t_id_pred list ;
  called_exit_postconditions : Kernel_function.t -> t_id_pred list ;
  called_assigns : Kernel_function.t ->
    Cil_types.identified_term Cil_types.assigns list;

  strategy_kind : strategy_kind;

}

(* val default_behavior : string : rather use Cil.default_behavior_name now *)

type assigns = NoAssigns | OnlyAssigns | WithAssigns

type asked_prop =
  | AllProps
  | NamedProp of string
  | IdProp of Property.t


(**
 * Defines how annotations are applied on the Cfg during the WP.
 * - [~behaviors:bhv] : only annotations related
 *   to [b] in [bhv] are taken into account.
 *   Otherwise all behaviors are included.
 *   The default behavior is one behavior among the others.
 * - [~property:pid] : only the mentioned property id is turned into a goal.
 *   Others are only used in hypothesis.
 *   Otherwise, annotations are generally turned into goals unless
 *   they already have a valid current status.
 * - [~assigns] : filter in or out the assigns clauses.
 *
 **)
(*
val get_strategies : Cil2cfg.t
  -> ?behaviors:string list
  -> ?property:asked_prop
  -> ?assigns:assigns
  -> unit -> strategy list
  *)

(** Get the not yet proved properties of unreachable CFG nodes. *)
val process_unreached_annots :
  Cil2cfg.t -> Cil2cfg.node_type list -> unit

(** Compute the strategies to prove the selected property.  *)
val get_id_prop_strategies :
    ?assigns:assigns -> Cil2cfg.t ->
    Property.t -> strategy list

(** Similar to [get_id_prop_strategies] but with a named property.
* (useful for command line option).
* The behavior list has to be the behaviors of the property.
* TODO: it should be removed when we will be able to compute it. *)
val get_prop_strategies : ?assigns:assigns -> Cil2cfg.t ->
  (string list option * string) -> strategy list

(** Compute the strategy to prove all the properties of the behavior.
* Notice that is a property is related to several behaviors,
* it might not be fully proved with this strategy.
* Can return more strategies than names in the input list
* because a name can be used for several (disjoint) statement spec.
* *)
val get_behavior_strategies :
    ?assigns:assigns -> Cil2cfg.t -> string list -> strategy list

(** Compute the strategies to prove all the properties of the selected function.
* *)
val get_function_strategies : ?assigns:assigns -> Cil2cfg.t -> strategy list

(*
Local Variables:
compile-command: "make -C ../.."
End:
*)
