(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(** This file groups functions needed to check the fonctional dependencies *)

open Cil_types
open Db_types

(* -------------------------------------------------------------------------- *)
(** Build the logic type of the function that takes parameters of [in_types]
* and return an [out_type] result ([None] for a predicate) *)
let mk_linfo_type (out_type, in_types) =
  let lvar_out_type = match out_type with Some t -> t
    | None -> (* TODO: ugly ! but see in Logic_typing.logic_decl *)
        Ctype Cil.voidType
  in
  let ltype = match in_types with [] -> lvar_out_type
    | _ -> Larrow (in_types, lvar_out_type) 
  in ltype

(** Build a [logic_info] with [fname] and the signature given by 
* [(out_type, in_types)] (see {!mk_linfo_type})
TODO: should be Cil_const.make_logic_info when it be finished. *)
let make_logic_info fname (out_type, in_types) =
  let ltype = mk_linfo_type (out_type, in_types) in
  let lvar = Cil_const.make_logic_var fname ltype in
  let mk_in_lvar t = Cil_const.make_logic_var "x" t in
  let in_vars = List.map mk_in_lvar in_types in
  let linfo = {
    l_var_info = lvar;
    l_labels = [];
    l_tparams = [];
    l_type = out_type;
    l_profile = in_vars;
    l_body = LBnone;
  } in linfo

(** Find the [logic_info] for the given name and signature.
* Build and register it if it doesn't exist yet. 
* Notice that the returned [logic_info] can have a different name than 
* the asked one because we cannot create several function with same input
* signature and different output type. 
* TODO: the detection to know if we have to build a new name or not could be
* enhanced.
* *)
let get_linfo_cpt = ref 0
let get_linfo name sgn =
  let ptype = mk_linfo_type sgn in
  let linfos = Logic_env.find_all_logic_functions name in
  let info_ok info = Logic_utils.is_same_type info.l_var_info.lv_type ptype in
  let rec find l = match l with [] -> None
    | info::l -> if info_ok info then Some info else find l
  in match find linfos with 
    | Some info -> info
    | None ->
        let name = match linfos with [] -> name 
          | _ -> incr get_linfo_cpt; (name ^ (string_of_int !get_linfo_cpt))
        in
        let linfo = make_logic_info name sgn in
          Logic_utils.add_logic_function linfo;
          linfo

(** Build the implicit function for the nth assign clause of behavior b
* in the ki element of function kf. *)
let get_implicit_fun kf ki b n (out_type, inputs_type) =
  let ki_info = match ki with Kglobal -> "" 
    |  Kstmt s -> ("_stmt"^(string_of_int s.sid))
  in
  let bhv_name = 
    if b.b_name =  Cil.default_behavior_name then "" else ("_"^b.b_name) 
  in
  let name = Pretty_utils.sfprintf "FI_%a%s%s_%d"
               Kernel_function.pretty_name  kf
               bhv_name
               ki_info
               n
  in let linfos = Logic_env.find_all_logic_functions name in
  let f = match linfos with
    | f::l -> assert (l == []); (*TODO: error message *) f
    | [] ->
        let linfo = make_logic_info name (Some out_type, inputs_type) in
          Logic_utils.add_logic_function linfo;
          linfo
  in f

let get_pred_linfo t = get_linfo "Pred_to_check_fun_deps" (None, [t])

let get_init_linfo t = get_linfo "Init_to_check_fun_deps" (Some t, [Linteger])

(** Build [P(output)] *)
let mk_assign_post termination_kind (output, _) = 
  let out = match output with
    | Nothing -> Wp_parameters.fatal "Nothing as output in an assigns clause ?" 
    | Location {Cil_types.it_content=lv} -> lv
  in
    match termination_kind, out.term_node with
      | Exits, TLval (TResult _, _ )  -> None
      | Normal, TLval (TVar{lv_name = "\\exit_status"},_) -> None
      | _, _ ->
          let linfos = get_pred_linfo out.term_type in
          let p = Logic_const.papp (linfos, [], [out]) in
            Some (Logic_const.new_predicate p)

(** Build [ xi = Init (i) /\ ...] forall inputs. *)
let mk_assign_pre (_, inputs) =
  let get_init lv n =
    let linfo = get_init_linfo lv.term_type in
      Logic_const.term (Tapp (linfo, [], [Logic_const.tinteger n])) lv.term_type
  in
  let add_in (n, acc) input = match input with
    | Nothing -> n, acc
    | Location {it_content=lv} ->
        let _name = lv.term_name in (* TODO process name *)
        let init = get_init lv n in
        let pre = Logic_const.prel (Req, lv, init) in
          n+1, pre::acc
  in
  let _, pres = List.fold_left add_in (1, []) inputs in
    Logic_const.new_predicate (Logic_const.pands pres)

let build_bhv_assumes b =
  let assumes = List.map Logic_const.pred_of_id_pred b.b_assumes in
  let assumes = Logic_const.pands assumes in
    assumes

(** Build out = f_n (inputs) *)
let build_fimpl_call kf ki bhv n output inputs =
  let out_type = output.term_type in
  let input_term i acc = match i with Nothing -> acc
    | Location {it_content=lv} -> lv::acc
  in
  let inputs = List.fold_right input_term inputs [] in
  let fimpl_sig = (out_type, List.map (fun i -> i.term_type) inputs) in
  let fun_impl = get_implicit_fun kf ki bhv n fimpl_sig in
  let inputs = 
    List.map (fun x -> Logic_const.term (Told x) x.term_type) inputs 
  in
  let fun_impl = Logic_const.term (Tapp (fun_impl, [], inputs)) out_type in
  let p_eq = Logic_const.prel (Req, output, fun_impl) in
    p_eq

let post_of_assigns_in_spec kf ki spec termination_kind =
  let add_behav acc bhv = 
    (* post for behavior is [\old(assumes) => out = f(in)]*)
    let add_assign (n, acc) (output, inputs) = 
      let output = match output with 
        | Nothing -> 
            Wp_parameters.fatal "Nothing as output in an assigns clause ?"
        | Location {Cil_types.it_content=lv} -> lv
      in 
      let acc = match termination_kind, output.term_node with
        | Exits, TLval (TResult _, _ )  -> acc
        | Normal, TLval (TVar{lv_name = "\\exit_status"},_) -> acc
        | _, _ ->
            let p_eq = build_fimpl_call kf ki bhv n output inputs in
              p_eq::acc
      in
        n+1, acc
    in
    let _, bacc = List.fold_left add_assign (1, []) bhv.b_assigns in
    let p = Logic_const.pands bacc in
    let assumes = Logic_const.pold (build_bhv_assumes bhv) in
    let p = Logic_const.pimplies (assumes, p) in
    let p = Logic_const.new_predicate  p in
      (bhv,p)::acc
  in let acc = List.fold_left add_behav [] spec.spec_behavior in
    acc

