(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

open WpAnnot
open Cil_types
open Db_types
open Viewgraph

(* ------------------------------------------------------------------------ *)
(* ---  RUN WP                                                          --- *)
(* ------------------------------------------------------------------------ *)

exception Stop

let get_annot_opt localizable =
  match localizable with
    | Pretty_source.PIP ip ->
        Property.get_kf ip,
        Extlib.may_map ~dft:None (fun x -> Some [x.b_name])
          (Property.get_behavior ip),
        Some ip
    | Pretty_source.PStmt _ | Pretty_source.PLval _ | Pretty_source.PTermLval _
    | Pretty_source.PVDecl _ | Pretty_source.PGlobal _
      -> Pretty_source.kf_of_localizable localizable, None, None

let run_and_prove (main_ui:Design.main_window_extension_points) localizable =
  try
    let kf_opt, bhv, annot_id = get_annot_opt localizable in
    let annot_id = match annot_id with
      | Some (Property.IPBehavior _) -> None (* this is not really a property *)
      | _ -> annot_id
    in
    let fstr = match kf_opt with None -> "all functions"
      | Some kf -> ("function '"^( Kernel_function.get_name kf)^"'")
    in
    let pstr = match annot_id with None -> "all properties"
      | Some id -> (WpAnnot.id_prop_txt id)
    in
    let bstr = match bhv with None -> "all behaviors"
      | Some b -> Pretty_utils.sfprintf "%a behavior%s"
                    (Pretty_utils.pp_list ~sep:"_" Format.pp_print_string) b
                    (if List.length b > 1 then "s" else "")
    in
    let mstr = Wp_parameters.Model.get () in
    Wp_parameters.log "compututing WP on %s, for %s, in %s with model %s"
      fstr pstr bstr mstr;
    Register.wp_compute kf_opt bhv annot_id;
    main_ui#rehighlight ()
  with Stop -> ()

(* ------------------------------------------------------------------------ *)
(* ---  CFG Viewer                                                      --- *)
(* ------------------------------------------------------------------------ *)


module V = ViewGraph_select.VG (ViewGraph_core.EmptyCb)

let show_cfg (main_ui:Design.main_window_extension_points) kf =
  let kf_name = Kernel_function.get_name kf in
    main_ui#annot_window#buffer#insert ("[cfg] compute for "^kf_name^"\n");
  let select_init_env = ViewGraph_utils.create_graph_win (kf_name ^ " CFG") in
  let filename = Register.wp_cfg_dot kf in
  try ignore (V.open_dot_file () select_init_env filename)
  with ViewGraph_core.DotError _ ->
    GToolbox.message_box "Error"
      (Printf.sprintf "Dot failed on graph '%s'" filename)

(* ------------------------------------------------------------------------ *)
(* ---  Source Callback                                                 --- *)
(* ------------------------------------------------------------------------ *)

let add_wp_menu
    (popup_factory:GMenu.menu GMenu.factory)
    (main_ui:Design.main_window_extension_points)
    localizable =
  let wp_callback () = run_and_prove main_ui localizable in
  ignore (popup_factory#add_item "WP" ~callback:wp_callback)

let add_cfg_menu
    (popup_factory:GMenu.menu GMenu.factory)
    (main_ui:Design.main_window_extension_points)
    localizable =
  match Pretty_source.kf_of_localizable localizable with
    | None -> ()
    | Some kf ->
	let cfg_callback () = show_cfg main_ui kf in
	ignore (popup_factory#add_item "CFG" ~callback:cfg_callback)

let selector 
    (popup_factory:GMenu.menu GMenu.factory)
    (main_ui:Design.main_window_extension_points)
    ~button localizable =
  match button with
    | 3 -> (* Popup Menu: *)
	begin
	  (* WP  *) add_wp_menu  popup_factory main_ui localizable ;
	  (* CFG *) add_cfg_menu popup_factory main_ui localizable ;
	end
    | _ -> (* Other buttons... *) ()

(* ------------------------------------------------------------------------ *)
(* ---  WP Panel                                                        --- *)
(* ------------------------------------------------------------------------ *)

let wp_model_menu = [
  "Hoare" , "Hoare" ;
  "Store" , "Store" ;
  "Runtime" , "Runtime" ;
]

type p_select =
  | NoProver
  | AltErgo of bool
  | Prover of string

let wp_prover_menu = [
  "None" , NoProver ;
  "Alt-Ergo" , AltErgo true ;
  "Alt-Ergo (no array)" , AltErgo false ;
  "Coq" , Prover "coqide" ;
  "Z3" , Prover "z3" ; 
  "Simplify" , Prover "simplify" ;
  "CVC3" , Prover "cvc3" ;
  "Yices" , Prover "yices" ;
  "Zenon" , Prover "zenon" ;
]

let wp_prover_get () =
  match Wp_parameters.Prover.get () with
    | "alt-ergo" -> AltErgo (Wp_parameters.Arrays.get ())
    | "coqide" -> Prover "coqide"
    | "coq" -> Wp_parameters.Prover.set "coqide" ; Prover "coqide"
    | "simplify" -> Prover "simplify"
    | "z3" -> Prover "z3"
    | "cvc3" -> Prover "cvc3"
    | "yices" -> Prover "yices"
    | "zenon" -> Prover "zenon"
    | _ -> NoProver

let wp_dir = ref
  (try
     let home = Sys.getenv "HOME" in
     if Sys.file_exists home && Sys.is_directory home
     then home else raise Not_found
   with Not_found -> Sys.getcwd ())
     
let wp_script () =
  let file = GToolbox.select_file 
    ~title:"Script File for Coq proofs" 
    ~dir:wp_dir ~filename:"wp.script" ()
  in
  match file with
    | Some f -> Wp_parameters.Script.set f
    | None -> ()

let wp_prover_set = function
  | AltErgo a -> 
      Wp_parameters.Prover.set "alt-ergo" ; 
      Wp_parameters.Arrays.set a
  | Prover p -> 
      if p="coqide" && Wp_parameters.Script.get() = "" 
      then wp_script () ;
      Wp_parameters.Prover.set p
  | NoProver -> 
      Wp_parameters.Prover.set "none"

let wp_panel (_main_ui:Design.main_window_extension_points) =
  let vbox = GPack.vbox () in
  let demon = Gtk_form.demon () in
  let packing = vbox#pack in
  let form = new Gtk_form.form ~packing in
  
  form#label "Model" ;
  Gtk_form.menu wp_model_menu
    ~tooltip:"Memory model selection" ~packing:form#item
    Wp_parameters.Model.get Wp_parameters.Model.set demon ;

  form#label "Prover" ;
  Gtk_form.menu wp_prover_menu
    ~tooltip:"Prover selection"
    ~packing:form#item
    wp_prover_get wp_prover_set demon ;

  let options = GPack.hbox ~spacing:8 ~packing () in

  Gtk_form.check ~label:"RTE" 
    ~tooltip:"Generates RTE guards for WP"
    ~packing:options#pack
    Wp_parameters.RTE.get Wp_parameters.RTE.set demon ;

  Gtk_form.check ~label:"Split" 
    ~tooltip:"Split cunjunctions into sub-goals"
    ~packing:options#pack
    Wp_parameters.Split.get Wp_parameters.Split.set demon ;
  
  Gtk_form.check ~label:"Trace"
    ~tooltip:"Keep user labels in proof obligations"
    ~packing:options#pack
    Wp_parameters.Trace.get Wp_parameters.Trace.set demon ;

  Gtk_form.check ~label:"Axioms"
    ~tooltip:"Instanciation of memory-dependent axioms (Experimental)"
    ~packing:options#pack
    Wp_parameters.Axioms.get Wp_parameters.Axioms.set demon ;

  let control = GPack.hbox ~packing () in

  Gtk_form.button ~label:"Scripts"
    ~tooltip:"Script file for saving Coq proofs"
    ~callback:wp_script ~packing:control#pack () ;

  Gtk_form.label ~text:"Timeout" ~packing:control#pack () ;
  Gtk_form.spinner ~lower:0 ~upper:100000 
    ~tooltip:"Timeout for proving one proof obligation"
    ~packing:control#pack
    Wp_parameters.Timeout.get Wp_parameters.Timeout.set demon ;

  Gtk_form.label ~text:"Process" ~packing:control#pack () ;
  Gtk_form.spinner ~lower:1 ~upper:32
    ~tooltip:"Maximum number of parallel running provers"
    ~packing:control#pack
    Wp_parameters.Procs.get Wp_parameters.Procs.set demon ;

  let progress = GRange.progress_bar ~packing ~show:false () in
  let server = Prover.server () in
  let state = ref 0 in
  Task.on_server_activity server 
    (fun () ->
       let n = Task.load server in
       if n=0 then
	 ( progress#misc#hide () ; state := 0 )
       else
	 begin
	   if !state = 0 then progress#misc#show () ;
	   if n > !state then state := n ;
	   let r = !state - n in
	   progress#set_text (Printf.sprintf "%d / %d" r !state) ;
	   progress#set_fraction ( float r /. float !state ) ;
	 end) ;

  if Wp_parameters.Prover.get () = "none" 
  then Wp_parameters.Prover.set "alt-ergo" ;

  "WP" , vbox#coerce , Some (Gtk_form.refresh demon)

(* ------------------------------------------------------------------------ *)
(* --- Registering WP GUI                                               --- *)
(* ------------------------------------------------------------------------ *)

let main main_ui =
  main_ui#register_source_selector selector ;
  main_ui#register_panel wp_panel

let () = Design.register_extension main

(* ------------------------------------------------------------------------ *)

(*
  Local Variables:
  compile-command: "make -C ../.."
  End:
*)
