(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

include Plugin.Register
  (struct
     let name = "wp"
     let shortname = "wp"
     let help = "Weakest Preconditions Calculus."
   end)

(* localize all warnings inside WP *)

let warning ?current =
  match current with
    | None -> warning ~current:true
    | Some b -> warning ~current:b

let resetdemon = ref []
let on_reset f = resetdemon := f :: !resetdemon
let reset () = List.iter (fun f -> f ()) !resetdemon

(* ------------------------------------------------------------------------ *)
(* ---  WP Generation                                                   --- *)
(* ------------------------------------------------------------------------ *)

let wp_generation = add_group "Goal Selection"

let () = Plugin.set_group wp_generation
module WP =
  Action(struct
	   let option_name = "-wp"
	   let help = "Computes wp on all functions."
           let kind = `Tuning
	 end)
let () = on_reset WP.clear

let () = Plugin.set_group wp_generation
module Functions =
  StringSet(struct
              let option_name = "-wp-fct"
              let arg_name = "f,..."
              let help = "Computes wp only for the selected functions."
              let kind = `Tuning
            end)
let () = on_reset Functions.clear

let () = Plugin.set_group wp_generation
module Behaviors =
  StringSet(struct
                let option_name = "-wp-bhv"
		let arg_name = "b,..."
                let help = "Computes wp only for the selected behaviors."
                let kind = `Tuning
              end)
let () = on_reset Behaviors.clear

let () = Plugin.set_group wp_generation
module Properties =
  StringSet(struct
                let option_name = "-wp-prop"
		let arg_name = "p,..."
                let help = "Computes wp only for the selected properties.\n\
                            Type 'assigns' for all assigns clauses"
                let kind = `Tuning
              end)
let () = on_reset Properties.clear

type job =
  | WP_None
  | WP_All
  | WP_Select of string list

let job () =
  if Functions.is_set () 
  then WP_Select(Datatype.String.Set.elements (Functions.get()))
  else if WP.get () then WP_All 
  else if Properties.is_set () then WP_Select []
  else WP_None
	
let get_behaviors () =
  if Behaviors.is_set ()
  then Some(Datatype.String.Set.elements (Behaviors.get()))
  else None

let get_properties () =
  if Properties.is_set ()
  then Some(Datatype.String.Set.elements (Properties.get()))
  else None

(* ------------------------------------------------------------------------ *)
(* --- Froms                                                            --- *)
(* ------------------------------------------------------------------------ *)

(*
let () = Plugin.set_group wp_generation
module Froms =
  False(struct
          let option_name = "-wp-froms"
          let help = "Undocumented (dot not use)."
          let kind = `Tuning
        end)
*)

(* ------------------------------------------------------------------------ *)
(* ---  Memory Models                                                   --- *)
(* ------------------------------------------------------------------------ *)

let wp_model = add_group "Model Selection"

type model_kind =
  | M_Hoare
  | M_CaveatU
  | M_CaveatG
  | M_Store
  | M_Runtime

let model_names =
  [ "Hoare" ; "Caveat"; "UnsafeCaveat"; "Store"; "Runtime" ]

let model_of_name = function
  | "Runtime" -> M_Runtime
  | "Store" -> M_Store
  | "UnsafeCaveat" -> M_CaveatU
  | "Caveat" -> M_CaveatG
  | "Hoare" -> M_Hoare
  | _ -> raise Not_found

let () = Plugin.set_group wp_model
module Model =
  String(struct
	   let option_name = "-wp-model"
	   let arg_name = "model"
	   let help =
             "Memory model selection:\n \
              - 'Hoare':  no indirect access to the memory\n \
              - 'Store':  no heterogeneous casts (default)\n \
              - 'Runtime': low-level model"
              (* "- 'Runtime': low-level model\n \
              - 'UnsafeCaveat': no alias\n \
              - 'Caveat': no alias with guards" *)
	   let default = "Store"
           let kind = `Tuning
	 end)
let () = Model.set_possible_values model_names

let get_model () =
  let m = Model.get () in
  try model_of_name m
  with Not_found -> abort "Unknown model '%s'" m

let get_models () = model_names

let () = Plugin.set_group wp_model
module LogicVar =
  True(struct
	 let option_name = "-wp-logicvar"
	 let help = "Apply Hoare model for variables when possible."
	 let kind = `Tuning
       end)
(* pas de reset -> GUI pb *)

let () = Plugin.set_group wp_model
module Assigns =
  String(struct
	   let option_name = "-wp-assigns"
	   let arg_name = "mth"
	   let help = "Method for proving assigns clauses:\n\
                       - 'effect' one sub-goal per assignment (default)\n\
                       - 'region' optimization for effect (Store model only)\n\
                       - 'memory' strong proof (incompatible with Hoare)\n\
                       - 'none'   skip assigns clause (default with Hoare)"
	   let default = "effect"
	   let kind = `Tuning
	 end)
let () = Assigns.set_possible_values [ "effect" ; "region" ; "memory" ; "none" ]

let get_assigns_method () =
  match Assigns.get () with
    | "effect" -> Mcfg.EffectAssigns
    | "region" -> Mcfg.RegionAssigns
    | "memory" -> Mcfg.NormalAssigns
    | "none"   -> Mcfg.NoAssigns
    | m -> abort "Unknown assigns method '%s'" m

(* ------------------------------------------------------------------------ *)
(* ---  WP Strategy                                                     --- *)
(* ------------------------------------------------------------------------ *)

let wp_strategy = add_group "Computation Strategies"

let () = Plugin.set_group wp_strategy
module RTE =
  False(struct
	  let option_name = "-wp-rte"
	  let help = "Generates RTE guards before WP"
	  let kind = `Tuning
	end)

let () = Plugin.set_group wp_strategy
module Simpl = 
  True(struct
	let option_name = "-wp-simpl"
	let help = "Simplify constant terms and predicates."
	let kind = `Tuning
       end)

let () = Plugin.set_group wp_strategy
module Split =
  False(struct
	  let option_name = "-wp-split"
	  let help = "Split conjunctions into sub-goals."
	  let kind = `Tuning
	end)

let () = Plugin.set_group wp_strategy
module Norm =
  String(struct
	   let option_name = "-wp-norm"
	   let arg_name = "norm"
	   let help =
             "Predicate normalization for Coq and Alt-Ergo provers:\n\
              - Exp: let-expansion (Ergo default)\n\
              - Let: preserve let-bindings (Coq default).\n\
              - Eqs: replace let-bindings by logic equalities.\n\
              - Cc:  generates local functions by closure-conversion"
	   let default = "Let"
           let kind = `Correctness
	 end)
let () = Norm.set_possible_values
           [ "Let";"Exp";"Cc";"Eqs" ]

type norm = Let | Exp | Eqs | Cc

let get_norm () =
  match Norm.get () with
  | "Let" -> Let
  | "Exp" -> Exp
  | "Eqs" -> Eqs
  | "Cc" -> Cc
  | m -> abort "Unknown normalization '%s'" m

let () = Plugin.set_group wp_strategy
module Huge =
  Int(struct
	let option_name = "-wp-huge"
        let default = 30
        let arg_name = "s"
        let help = 
	  Printf.sprintf
	    "Limits the size for generated proof obligation.\n\
             Proof terms of size exceeding 2^s are not generated.\n\
             (default: 2^%d)" 
	    default
	let kind = `Tuning
      end)

let () = Plugin.set_group wp_strategy
module Axioms =
  False(struct
          let option_name = "-wp-axioms"
          let help = "Instanciate memory-dependent axioms (Experimental)."
          let kind = `Tuning
        end)

(* ------------------------------------------------------------------------ *)
(* ---  Prover Interface                                                --- *)
(* ------------------------------------------------------------------------ *)

let wp_prover = add_group "Prover Interface"

let () = Plugin.set_group wp_prover
module Prover =
  String(struct
           let option_name = "-wp-proof"
           let default = "none"
           let arg_name = "dp"
           let help =
	     "Submit proof obligations to external prover:\n \
               * 'none' to skip proofs (default)\n \
              Directly supported provers:\n \
               * 'alt-ergo' (see also -wp-arrays)\n \
               * 'coq', 'coqide' (see also -wp-script)\n \
              Supported provers via Why:\n \
               * 'simplify', 'yices', 'cvc3', 'z3', 'zenon'"
           let kind = `Tuning
         end)
(* no reset *)

let prover_names = 
  [ "none" ;
    "coq" ; "coqide" ;
    "alt-ergo";
    "simplify";
    "yices";
    "cvc3";
    "z3" ]

let () = Prover.set_possible_values prover_names
let get_provers () = prover_names

let () = Plugin.set_group wp_prover

module Check =
  String(struct
           let option_name = "-wp-check"
           let default =  "none"
           let arg_name = "dp"
           let help =
	     "Typecheck proof obligations for external prover:\n \
               * 'none' to skip checks (default)\n \
               * 'alt-ergo' (see also -wp-arrays)\n \
               * 'coq'\n \
               * 'why' "
           let kind = `Tuning
         end)
(* no reset *)

let check_names = 
  [ "none" ;
    "coq" ;
    "alt-ergo";
    "why" ]

let () = Plugin.set_group wp_prover
module Arrays = 
  True(struct
	 let option_name = "-wp-arrays"
	 let help = "Use build-in array theory of alt-ergo."
	 let kind = `Tuning
       end)

let () = Plugin.set_group wp_prover
module Script = 
  String(struct
	   let option_name = "-wp-script"
	   let arg_name = "f.script"
	   let default = "" 
	   let help = "Set user's file for saving Coq proofs."
	   let kind = `Tuning
	 end)

let () = Plugin.set_group wp_prover
module Timeout =
  Int(struct
        let option_name = "-wp-timeout"
        let default = 10
        let arg_name = "n"
        let help = 
	  Printf.sprintf 
	    "Set the timeout (in seconds) for provers (default: %d)." default
       let kind = `Tuning
      end)

let () = Plugin.set_group wp_prover
module Procs =
  Int(struct
	let option_name = "-wp-par"
        let arg_name = "p"
        let default = 4
        let help = 
	  Printf.sprintf
	    "Number of parallel proof process (default %d)" default
	let kind = `Tuning
      end)

let () = Plugin.set_group wp_prover
module Trace =
  False(struct
	  let option_name = "-wp-trace"
	  let help = "Keep labels in proof obligations (default: no)."
          let kind = `Tuning
	end)


module ShareDir = 
   String(struct
	   let option_name = "-wp-share"
	   let arg_name = "dir"
	   let default = ""
	   let help = "Set working directory for compiling specifications.\n\
                       Defaults: directory where architecture independent \n\
                       files are."
           let kind = `Tuning
	 end)
(* no reset *)


(* ------------------------------------------------------------------------ *)
(* --- PO Management                                                    --- *)
(* ------------------------------------------------------------------------ *)

let wp_po = add_group "Proof Obligations"

let () = Plugin.set_group wp_po
module Print =
  False(struct
	  let option_name = "-wp-print"
	  let help = "Pretty-prints proof obligations on standard output."
          let kind = `Tuning
	end)
let () = on_reset Print.clear

let () = Plugin.set_group wp_po
module Dot =
  False(struct
          let option_name = "-wp-dot"
          let help = "Generates dot files for wp computations."
          let kind = `Tuning
        end)
let () = on_reset Dot.clear

let () = Plugin.set_group wp_po
module OutputDir =
  String(struct
	   let option_name = "-wp-out"
	   let arg_name = "dir"
	   let default = ""
	   let help = "Set working directory for generated files.\n\
                       Defaults to some temporary directory."
           let kind = `Tuning
	 end)
(* no reset *)




(* -------------------------------------------------------------------------- *)
(* --- OS environment variables                                           --- *)
(* -------------------------------------------------------------------------- *)

let get_env ?default var =
  try
    let varval = Sys.getenv var in
    debug "ENV %s=%S" var varval ; varval
  with Not_found ->
    debug "ENV %s not set." var ;
    match default with
      | Some varval ->
	  debug "ENV %s default(%S)" var varval ; varval
      | None ->
	  debug "ENV %s undefined." var ;
	  raise Not_found

(*TODO: Projectifier cette reference*)
let tmp_dir = ref None

let get_output () =
  match OutputDir.get () with
    | "" -> 
	begin
	  match !tmp_dir with
	    | Some dir -> dir
	    | None ->
		let dir = Extlib.temp_dir_cleanup_at_exit "wp" in
		tmp_dir := Some dir ; dir
	end
    | dir -> 
	if Sys.file_exists dir then
	  begin
	    if not (Sys.is_directory dir) then 
	      abort "File '%s' is not a directory (WP aborted)" dir ;
	  end
	else
	  begin
	    try 
	      Unix.mkdir dir 0o770 ;
	      debug "Created output directory '%s'" dir
	    with e -> 
	      debug "System error '%s'" (Printexc.to_string e) ;
	      abort "Can not create output directory '%s'" dir 
	  end ;
	dir

let get_share () = 
  match ShareDir.get() with 
    | "" -> Config.datadir^"/wp" 
    | dir -> 
	if Sys.file_exists dir && Sys.is_directory dir then dir
	  else abort "'%s': no such directory" dir 
