(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2010                                               *)
(*    CEA   (Commissariat  l'nergie atomique et aux nergies            *)
(*           alternatives)                                                *)
(*    INRIA (Institut National de Recherche en Informatique et en         *)
(*           Automatique)                                                 *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version v2.1                *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

module type Actor =
sig
  val name : string
  val kind : State.kind
  val dependencies : State.t list
end

module Updater(A : Actor) =
struct

  open Cil_types

  module S = State_builder.Ref
    (Datatype.Unit)
    (struct
       let name = A.name
       let kind = A.kind
       let dependencies = A.dependencies
       let default () = ()
     end)

  module U = Properties_status.Make_updater
    (struct
       let name = A.name
       let emitter = S.self
     end)

  let set_checked ?(descr=A.name) ip ips valid = 
    U.set ip ips (Checked{emitter=descr;valid=valid})

  let set_unknown ip ips =
    U.set ip ips Unknown

end

module WP = State_builder.Ref
  (Datatype.Unit)
  (struct
     let name = "WP"
     let kind = `Internal
     let dependencies = [Ast.self]
     let default () = ()
   end)

module type Info =
sig
  type key
  val size : int
  val kind : State.kind
  val dependencies: State.t list
  val state_name : key -> string
  val state_proxy : key -> Property.t -> (string * Property.t list) option
end

module type Indexed =
sig
  type key
  val state : key -> State.t (** Memoized state for key *)
  val property : key -> Property.t (** IPBlob of state for key *)
end

let kindex = ref 0

module Indexed
  (Key:Datatype.S_with_collections)
  (Info:Info with type key = Key.t) =
struct

  open Cil_types

  type key = Key.t

  module Dash = State_builder.Dashtbl
    (Dashtbl.Default_key_marshaler(Key))
    (Dashtbl.Default_data_marshaler(Datatype.Unit))
    (struct
       let size = Info.size
       let dependencies = Info.dependencies
       let name = incr kindex ; Printf.sprintf "Wprop.Indexed.%d" !kindex
       let kind = `Internal (* global map *)
       let internal_kind = Info.kind (* each generated state *)
     end)

  module Proxy = 
    Properties_status.Make_updater
    (struct
       let name = "WP"
       let emitter = WP.self
     end)

  let set_proxy key state =
    let ip = Property.IPBlob state in
    match Info.state_proxy key ip with
      | None -> ()
      | Some ( descr , ips ) ->
	  let st = Checked{emitter=descr;valid=True} in
	  Proxy.set ip ips st

  let state (key:key) =
    try Dash.find_state key WP.self
    with Not_found ->
      let name = Info.state_name key in
      Dash.add name key [WP.self] () ;
      let state = Dash.find_state key WP.self in
      set_proxy key state ; state

  let property (key:key) = Property.IPBlob (state key)

end

module type Indexed2 =
sig
  type key1
  type key2
  val state : key1 -> key2 -> State.t
  val property : key1 -> key2 -> Property.t
end

module Indexed2
  (Key1:Datatype.S_with_collections)
  (Key2:Datatype.S_with_collections)
  (Info:Info with type key = Key1.t * Key2.t) =
struct

  module P = Datatype.Pair_with_collections(Key1)(Key2)
    (struct
       let module_name = "WpropIndexed2" 
     end)
  module I = Indexed(P)(Info)

  type key1 = Key1.t
  type key2 = Key2.t      

  let state a b = I.state (a,b)
  let property a b = I.property (a,b)

end

